#include "twodel.h"

void CTwoDelIndex::insert(const char *p)
{
    string str(p);

    int sz = str.size();
    if (sz < 2) { return; }

    char origChar1 = str[0];

    str.erase(str.begin());

    for (int i = 1; i < sz; ++i) {
        char origChar2 = str[i - 1];

        insertInternal(str.c_str(), i, origChar1, 0, 0);

        str.erase(str.begin() + i - 1);

        //cout << "Transform: '" << p << "' -> '" << str << "'" << endl;


        for (int k = i + 1; k <= sz; ++k) {
            insertInternal(str.c_str(), i, origChar1, k - 1, origChar2);
            if (k < sz) {
                swap(origChar2, str[k - 2]);
            }
            //cout << "Transform: '" << p << "' -> '" << str << "'" << endl;
        }
        str.insert(str.end(), origChar2);
        if (i < sz - 1) {
            swap(origChar1, str[i - 1]);
        }
    }
    str.insert(str.end() - 1, origChar1);
#ifndef NDEBUG
    if (str != p) {
        cout << "Transform: '" << p << "' -> '" << str << "'" << endl;
    }
#endif
    assert(str == p);
    RecalcIterators();
}

void CTwoDelIndex::search(string &str, bool hasTransp, int *numComp, StringList *resList)
{
    int sz = str.size();

    if (!sz) return;

    if (sz == 1) {
        searchInternal("", false, 1, str[0], 0, 0, numComp, resList);
        return;
    }
    
#ifndef NDEBUG
    string str1 = str;
#endif

    char origChar1 = str[0];

    str.erase(str.begin());

    for (int i = 1; i < sz; ++i) {
        searchInternal(str.c_str(), hasTransp, i, origChar1, 0, 0, numComp, resList);

        char origChar2 = str[i - 1];
        str.erase(str.begin() + i - 1);

        for (int k = i + 1; k <= sz; ++k) {
            searchInternal(str.c_str(), hasTransp, i, origChar1, k - 1, origChar2, numComp, resList);
            if (k < sz) {
                swap(origChar2, str[k - 2]);
            }
        }
        str.insert(str.end(), origChar2);
        if (i < sz - 1) {
            swap(origChar1, str[i - 1]);
        }
    }
    str.insert(str.end() - 1, origChar1);
    assert(str == str1);

    return;
}

void CTwoDelIndex::searchInternal(const char *p, 
                                  bool hasTransp,
                                  unsigned char pos1, unsigned char sym1,
                                  unsigned char pos2, unsigned char sym2,
                                  int *numComp, StringList *resList)
{
    hmapci::iterator it = mIndex.find(p);
    if (numComp) (*numComp)++;
    if (it != mIndex.end()) {
        const Record &rec = it->second;
        unsigned int cnt = 0;
        const unsigned char *pPos1 = NULL;
        const unsigned char *pSym1 = NULL;
        const unsigned char *pPos2 = NULL;
        const unsigned char *pSym2 = NULL;
        if (rec.mIsExt) {
            unsigned int pos = rec.GetPos();
            assert(pos < mExtPos1.size());
            assert(mExtPos2.size() == mExtPos1.size());
            assert(mExtSym2.size() == mExtSym1.size());
            assert(mExtSym2.size() == mExtPos1.size());
            const vuc &  arrPos1 = mExtPos1[pos];
            const vuc &  arrSym1 = mExtSym1[pos];
            const vuc &  arrPos2 = mExtPos2[pos];
            const vuc &  arrSym2 = mExtSym2[pos];
            cnt = arrPos1.size();
            pPos1 = &arrPos1[0];
            pSym1 = &arrSym1[0];
            pPos2 = &arrPos2[0];
            pSym2 = &arrSym2[0];
        } else {
            pPos1 = rec.mPos1;
            pSym1 = rec.mSym1;
            pPos2 = rec.mPos2;
            pSym2 = rec.mSym2;
            cnt  = rec.mCnt;
        }
        if (numComp) (*numComp) += cnt;
        for (unsigned int i = 0; i < cnt; ++i) {
            assert(pos1 && pPos1[i]); // It is so by construction

            bool bOk = (  pos1 == pPos1[i] && (pos2 == pPos2[i] || !pPos2[i] || !pos2) );


            if (!bOk && hasTransp) {
                unsigned char pos1Plus = pos1 + 1;
                unsigned char pos2Plus = pos2 + 1;

                // There 4 possible cases
                if (pPos1[i] == pos1Plus && pSym1[i] == sym1 && 
                    (pos2 == pPos2[i] || !pos2 || !pPos2[i])) {
                    bOk = true;
                } else if (pPos2[i] == pos1Plus && pSym2[i] == sym1 
                            && pos1Plus /* check for overflow */ 
                            && !pos2) {
                    bOk = true;
                } else if (pPos2[i] == pos2Plus && pSym2[i] == sym2 && pos2 
                            && pos2Plus /* check for overflow */ 
                            && pos1 == pPos1[i]) {
                    bOk = true;
                } else if (pPos1[i] == pos2Plus && pSym1[i] == sym2 && pos2 && !pPos2[i]) {
                    bOk = true;
                }
                           
            }
            
            if (bOk) {
                if (resList) {
                    resList->insert(resList->begin(), p);
                    string &str = *resList->begin();
                    if (pPos2[i]) {
                        str.insert(str.begin() + pPos2[i] - 1, pSym2[i]);
                    }
                    str.insert(str.begin() + pPos1[i] - 1, pSym1[i]);
                }
            }
        }
    }
}

void CTwoDelIndex::insertInternal(const char *p, 
                                unsigned char pos1, unsigned char sym1,
                                unsigned char pos2, unsigned char sym2
                                )
{
    hmapci::iterator it = mIndex.find(p);
    if (it == mIndex.end()) {
        mIndex.insert(make_pair(strDup(p), Record()));
        it = mIndex.find(p);
        assert(it != mIndex.end());
    }
    Record &rec = it->second;
    if (!rec.mIsExt) {
        if (rec.mCnt < INTERN_CNT2) {
            rec.mPos1[rec.mCnt] = pos1;
            rec.mSym1[rec.mCnt] = sym1;
            rec.mPos2[rec.mCnt] = pos2;
            rec.mSym2[rec.mCnt] = sym2;
            rec.mCnt++;
        } else {
            size_t extSize = mExtPos1.size();
            if (extSize % REALLOC_QTY2 == 0) {
                mExtPos1.reserve(extSize + REALLOC_QTY2);
                mExtSym1.reserve(extSize + REALLOC_QTY2);
                mExtPos2.reserve(extSize + REALLOC_QTY2);
                mExtSym2.reserve(extSize + REALLOC_QTY2);
            }

            mExtPos1.resize(extSize + 1);
            mExtSym1.resize(extSize + 1);
            mExtPos2.resize(extSize + 1);
            mExtSym2.resize(extSize + 1);

            vuc &tmpPos1 = mExtPos1[extSize];
            vuc &tmpSym1 = mExtSym1[extSize];
            vuc &tmpPos2 = mExtPos2[extSize];
            vuc &tmpSym2 = mExtSym2[extSize];

            tmpPos1.reserve(INTERN_CNT2 + 1); // More CPU but less memory
            tmpSym1.reserve(INTERN_CNT2 + 1); // More CPU but less memory
            tmpPos2.reserve(INTERN_CNT2 + 1); // More CPU but less memory
            tmpSym2.reserve(INTERN_CNT2 + 1); // More CPU but less memory

            for (int i = 0; i < INTERN_CNT2; ++i) {
                tmpPos1.push_back(rec.mPos1[i]);
                tmpSym1.push_back(rec.mSym1[i]);
                tmpPos2.push_back(rec.mPos2[i]);
                tmpSym2.push_back(rec.mSym2[i]);
            }

            tmpPos1.push_back(pos1);
            tmpSym1.push_back(sym1);
            tmpPos2.push_back(pos2);
            tmpSym2.push_back(sym2);

            rec.SetPos(extSize);
            assert(rec.mIsExt); // changes mIsExt to 1
        }
    } else {
        assert(mExtPos2.size() == mExtPos1.size());
        assert(mExtSym2.size() == mExtSym1.size());
        assert(mExtSym2.size() == mExtPos1.size());
        unsigned int extPos = rec.GetPos();

        unsigned sz = mExtPos1[extPos].size();
        mExtPos1[extPos].reserve(sz + 1);
        mExtSym1[extPos].reserve(sz + 1);
        mExtPos2[extPos].reserve(sz + 1);
        mExtSym2[extPos].reserve(sz + 1);

        mExtPos1[extPos].push_back(pos1);
        mExtSym1[extPos].push_back(sym1);
        mExtPos2[extPos].push_back(pos2);
        mExtSym2[extPos].push_back(sym2);
    }
}
