#include "index.h"
#include "hash.h"


static long long seekVal = 0;

CSpellIndex::CSpellIndex (bool hasTransp, EAlphabet AlphaType)
              : CInterface(hasTransp, AlphaType), 
              openedAll(false), opened1(false), opened2(false)
{
    const char * _ABC = GetAlphabet(AlphaType);

    int size = strlen (_ABC);
    ABC = new char[size + 1];
    memcpy (ABC, _ABC, size + 1);
}


CSpellIndex::~CSpellIndex ()
{
    delete [] ABC;
}

bool CSpellIndex::Insert(const char *p)
{
    mSpellIndex1.insert(p);
    mSpellIndex2.insert(p);

    return true;
}

void CSpellIndex::ExpandAndSearch(int start,
                        int &numComp, 
                        string & str, 
                        int kerr, 
                        StringList *resList)
{
    if (1) {
#if 0
        if (GetEnvSwitch("SHOW_VARIANTS")) {
            std::cout << "Variant: " << str << std::endl;
        }
#endif
        mSpellIndex1.search(str, mHasTransp, &numComp, resList);
    }

    int size = str.size ();
    if (!kerr || start > size) return;

    int i = 0;

    if (mHasTransp) {
    // Transposition
        for (i = start; i < size - 1; i++) {
            swap(str[i], str[i+1]);
            ExpandAndSearch(i + 2, numComp,  str, kerr - 1, resList);
            swap(str[i], str[i+1]);
        }
    }
    //mSpellIndex2.search(str, mHasTransp, &numComp, resList);

}


bool CSpellIndex::Search(const char *p, int kerr, int *numComp,
                         StringList * resList, bool silent)
{
    string str(p);
    if (!kerr) {
        mSpellIndex1.search(str, mHasTransp, numComp, resList, true);
    } if (kerr < 2) {
        mSpellIndex1.search(str, mHasTransp, numComp, resList);
    } else if (kerr < 3) {
        int numComp1 = 0;
        /*
        if (mHasTransp) {
            ExpandAndSearch(0, numComp1, str, 1, resList);
        }
        if (numComp) *numComp = numComp1;
        */
        numComp1 = 0;
        mSpellIndex2.search(str, mHasTransp, &numComp1, resList);
        if (numComp) (*numComp) += numComp1;
    }
    return true;
}

bool CSpellIndex::OpenDB1(const char *base_name, bool create)
{
    string    headerName = string(base_name) + "1.hdr";
    string    hashName = string(base_name) + "1.hash";
    string    dictName = string(base_name) + "1.dict";

    int flags = O_RDWR;
    int access = 0;

    pDict1 = NULL;
    pHash1 = NULL;
    pParams1 = NULL;

    if (create)  {
        flags = O_RDWR | O_TRUNC | O_CREAT;
        access = defaultAccess;
    }
    fdParams1 = open(headerName.c_str(), flags, access);
    if (fdParams1 < 0) {
        Perror("Cannot open header file (1)");
        return false;
    }
    fdHash1 = open(hashName.c_str(), flags, access);
    if (fdHash1 < 0) {
        Perror("Cannot open hash file (1)");
        close(fdParams1);
        return false;
    }
    fdDict1 = open(dictName.c_str(), flags, access);
    if (fdDict1 < 0) {
        Perror("Cannot open dict file (1)");
        close(fdHash1);
        close(fdParams1);
        return false;
    }

    if (!create) {
        opened1 = true;
        if (!mapData1()) {
            CloseDB1();
            return false;
        }
    }
    return true;    
}

bool CSpellIndex::OpenDB2(const char *base_name, bool create)
{
    string    headerName = string(base_name) + "2.hdr";
    string    hashName = string(base_name) + "2.hash";
    string    dictName = string(base_name) + "2.dict";

    int flags = O_RDWR;
    int access = 0;

    pDict2 = NULL;
    pHash2 = NULL;
    pParams2 = NULL;

    if (create)  {
        flags = O_RDWR | O_TRUNC | O_CREAT;
        access = defaultAccess;
    }
    fdParams2 = open(headerName.c_str(), flags, access);
    if (fdParams2 < 0) {
        Perror("Cannot open header file (2)");
        return false;
    }
    fdHash2 = open(hashName.c_str(), flags, access);
    if (fdHash2 < 0) {
        Perror("Cannot open hash file (2)");
        close(fdParams2);
        return false;
    }
    fdDict2 = open(dictName.c_str(), flags, access);
    if (fdDict2 < 0) {
        Perror("Cannot open dict file (2)");
        close(fdHash2);
        close(fdParams2);
        return false;
    }

    if (!create) {
        opened2 = true;
        if (!mapData2()) {
            CloseDB2();
            return false;
        }
    }
    return true;    
}

bool CSpellIndex::OpenDB(const char *base_name, bool create)
{
    if (openedAll)
        CloseDB ();
    if (!OpenDB1(base_name, create) || !OpenDB2(base_name, create)) {
        return false;
    }
    return openedAll = true;
}


bool CSpellIndex::DumpDB1(const char *base_name)
{
    size_t numWords = mSpellIndex1.GetDictSize();
    vector<pair<size_t, COneDelIndex::ConstIndexIterator> >   pos2it;
    
    size_t dictSize = 0;
    vuc arrPos, arrSym;
    seekVal = 0;

    int cntIt = 0;
    int cntRec = 0, cntRecMax = 0;
    // The first set iteration - write 0 and calc size
    for (COneDelIndex::ConstIndexIterator it = mSpellIndex1.begin(); it != mSpellIndex1.end(); ++it, ++cntIt) {
        const string & str = it.GetWord();
        size_t         strSize = str.size() + 1;

        int cnt;
        it.GetPosSym(arrPos, arrSym, cnt);

        cntRec += cnt;
        cntRecMax = max(cnt, cntRecMax);

        // if the word length is less than 255, then 
        // no more than 255 words can be converted into a single
        // word by the deletion of a single character
        if (cnt > 255) {
           CloseDB1();
           cerr << Format("Count %d for word '%s' is too large (1)\n", cnt, str.c_str());
           return false;
        }
        size_t addSize = strSize + 1 + 2 * cnt; // 1 counter 
        
        char           zeroBuf[addSize];

        memset(zeroBuf, 0, addSize);
        if (write(fdDict1, zeroBuf, addSize) != addSize) {
            Perror("Cannot write to dict file (1)");
            CloseDB1();
            return false;
        }
        dictSize += addSize;
        pos2it.push_back(make_pair(0U, it));
    }

    if (GetEnvSwitch("PRINT_ADD_STAT")) {
        cout << "Average list len (k = 1): " << double(cntRec) / cntIt << " max: " << cntRecMax << endl;
    }

    int fillRatio = defaultHashRatio;

    GetEnv("FILL_RATIO", fillRatio);
    if (!fillRatio) fillRatio = defaultHashRatio;

    if (fillRatio < minHashRatio) fillRatio = minHashRatio;
    if (fillRatio > maxHashRatio) fillRatio = maxHashRatio;

    size_t hashSize = from__stl_next_prime((size_t) 100LL * numWords / fillRatio); 
    if (!hashSize) hashSize = 1;


    size_t tmp(0);
    if (lseek(fdHash1, getPageAlignedValue((hashSize - 1) * sizeof(size_t)), SEEK_SET) < 0 ||
         write(fdHash1, &tmp, sizeof(size_t)) != sizeof(size_t)) {
        Perror("Cannot lseek or write");
        CloseDB1();
        return false;
    }
    CHashParams prm(fillRatio, hashSize, dictSize);

    if (write(fdParams1, &prm, sizeof prm) != sizeof prm) {
        Perror("Cannot write params");
        CloseDB1();
        return false;
    }

    if (!mapData1()) {
        CloseDB1();
        return false;
    }

    memset(pHash1, 0, pParams1->hashSize * sizeof(size_t));

    size_t n = 0;
    cntIt = 0;
    // Find out which hash table elements to fill
    for (COneDelIndex::ConstIndexIterator it = mSpellIndex1.begin(); it != mSpellIndex1.end(); ++it, ++cntIt) {
        const string & str = it.GetWord();
        size_t         buckNum;
        if (!insertHashDB(str, buckNum, *pParams1, pHash1, &seekVal)) {
            cerr << "Failed to insert string internally (1)" << endl;
            CloseDB1();
            return false;
        }
#if 0
        if (cntIt % 100000 == 0) {
            cout << cntIt <<" of " << numWords << " " << 100 * cntIt / numWords << "% (1) seekval" << seekVal / max(1, cntIt) << endl;
        }
#endif
        pos2it[n++].first = buckNum;
    }
    // Finally store strings in the order they are in hash table
    // and fill hash positions
    sort(pos2it.begin(), pos2it.end()); // will sort by position

    size_t pos = 0;
    for (n = 0; n < pos2it.size(); ++n) {
        const COneDelIndex::ConstIndexIterator &it = pos2it[n].second;

        pHash1[pos2it[n].first] = pos + 1; // do not store zero value

        const string & str = it.GetWord();
        size_t         strSize = str.size() + 1;

        memcpy(&pDict1[pos], str.data(), strSize);
        pos += strSize;

        int cnt;
        it.GetPosSym(arrPos, arrSym, cnt);

        // if the word length is less than 255, then 
        // no more than 255 words can be converted into a single
        // word by the deletion of a single character
        
        if (cnt > 255) {
           CloseDB1();
           cerr << Format("Count %d for word '%s' is too large (1)\n", cnt, str.c_str());
           return false;
        }

        unsigned char uCnt = cnt;
        memcpy(&pDict1[pos], &uCnt, sizeof uCnt);
        pos += 1;

        memcpy(&pDict1[pos], &arrPos[0], cnt);
        pos += cnt;

        memcpy(&pDict1[pos], &arrSym[0], cnt);
        pos += cnt;

    }
    return true;
}

bool CSpellIndex::DumpDB2(const char *base_name)
{
    size_t numWords = mSpellIndex2.GetDictSize();
    vector<pair<size_t, CTwoDelIndex::ConstIndexIterator> >   pos2it;
    
    size_t dictSize = 0;
    vuc arrPos1, arrSym1;
    vuc arrPos2, arrSym2;
    // The first set iteration - write 0 and calc size
    for (CTwoDelIndex::ConstIndexIterator it = mSpellIndex2.begin(); it != mSpellIndex2.end(); ++it) {
        const string & str = it.GetWord();
        size_t         strSize = str.size() + 1;

        int cnt;
        it.GetPosSym(arrPos1, arrSym1, arrPos2, arrSym2, cnt);

        // if the word length is less than 255, then 
        // no more than 65535 words can be converted into a single
        // word by the deletion of two characters
        
        if (cnt > 65535) {
           CloseDB2();
           cerr << Format("Count %d for word '%s' is too large (2)\n", cnt, str.c_str());
           return false;
        }
        size_t addSize = strSize + 2 + 4 * cnt; // 1 two-byte counter 
        
        char           zeroBuf[addSize];

        memset(zeroBuf, 0, addSize);
        if (write(fdDict2, zeroBuf, addSize) != addSize) {
            Perror("Cannot write to dict file (2)");
            CloseDB2();
            return false;
        }
        dictSize += addSize;
        pos2it.push_back(make_pair(0U, it));
    }

    int fillRatio = defaultHashRatio;

    GetEnv("FILL_RATIO", fillRatio);
    if (!fillRatio) fillRatio = defaultHashRatio;

    if (fillRatio < minHashRatio) fillRatio = minHashRatio;
    if (fillRatio > maxHashRatio) fillRatio = maxHashRatio;

    size_t hashSize = from__stl_next_prime((size_t) 100LL * numWords / fillRatio); 
    if (!hashSize) hashSize = 1;


    size_t tmp(0);
    if (lseek(fdHash2, getPageAlignedValue((hashSize - 1) * sizeof(size_t)), SEEK_SET) < 0 ||
         write(fdHash2, &tmp, sizeof(size_t)) != sizeof(size_t)) {
        Perror("Cannot lseek or write (2)");
        CloseDB2();
        return false;
    }
    CHashParams prm(fillRatio, hashSize, dictSize);

    if (write(fdParams2, &prm, sizeof prm) != sizeof prm) {
        Perror("Cannot write params (2)");
        CloseDB2();
        return false;
    }

    if (!mapData2()) {
        CloseDB2();
        return false;
    }

    memset(pHash2, 0, pParams2->hashSize * sizeof(size_t));

    size_t n = 0;
    seekVal = 0;
    // Find out which hash table elements to fill
    int cntIt = 0;
    for (CTwoDelIndex::ConstIndexIterator it = mSpellIndex2.begin(); it != mSpellIndex2.end(); ++it, ++cntIt) {
        const string & str = it.GetWord();
        size_t         buckNum;
        if (!insertHashDB(str, buckNum, *pParams2, pHash2, &seekVal)) {
            cerr << "Failed to insert string internally (2)" << endl;
            CloseDB2();
            return false;
        }
#if 0
        if (cntIt % 100000 == 0) {
            cout << cntIt <<" of " << numWords << " " << 100 * cntIt / numWords << "% (2) seekval" << seekVal / max(1, cntIt) << endl;
        }
#endif
        pos2it[n++].first = buckNum;
    }
    // Finally store strings in the order they are in hash table
    // and fill hash positions
    sort(pos2it.begin(), pos2it.end()); // will sort by position

    size_t pos = 0;
    for (n = 0; n < pos2it.size(); ++n) {
        const CTwoDelIndex::ConstIndexIterator &it = pos2it[n].second;

        pHash2[pos2it[n].first] = pos + 1; // do not store zero value

        const string & str = it.GetWord();
        size_t         strSize = str.size() + 1;

        memcpy(&pDict2[pos], str.data(), strSize);
        pos += strSize;

        int cnt;
        it.GetPosSym(arrPos1, arrSym1, arrPos2, arrSym2, cnt);

        // if the word length is less than 255, then 
        // no more than 65535 words can be converted into a single
        // word by the deletion of a two characters
        
        if (cnt > 65535) {
           CloseDB2();
           cerr << Format("Count %d for word '%s' is too large (2)\n", cnt, str.c_str());
           return false;
        }

        u_int16_t uCnt = cnt;
        memcpy(&pDict2[pos], &uCnt, sizeof uCnt);
        pos += 2;

        memcpy(&pDict2[pos], &arrPos1[0], cnt);
        pos += cnt;

        memcpy(&pDict2[pos], &arrSym1[0], cnt);
        pos += cnt;

        memcpy(&pDict2[pos], &arrPos2[0], cnt);
        pos += cnt;

        memcpy(&pDict2[pos], &arrSym2[0], cnt);
        pos += cnt;

    }
    return true;
}

bool CSpellIndex::DumpDB(const char *base_name)
{
    if (openedAll)
        CloseDB ();
    if (!OpenDB (base_name, true))
        return false;           // Create database

    if (!DumpDB2(base_name) || !DumpDB1(base_name)) {
        CloseDB();
        return false;
    }

    return CloseDB();
}

bool CSpellIndex::mapData1()
{
    pParams1 = reinterpret_cast<CHashParams *>(mmap(NULL, getPageAlignedValue(sizeof *pParams1), 
                                        PROT_WRITE | PROT_READ, MAP_SHARED, fdParams1, 0));
    if (!pParams1 || pParams1 == MAP_FAILED) {
        Perror("Cannot map params (1)");
        pParams1 = NULL;
        return false;
    }
    pHash1 = reinterpret_cast<size_t *>(mmap(NULL, getPageAlignedValue(pParams1->hashSize * sizeof(size_t)), 
                                        PROT_WRITE | PROT_READ, MAP_SHARED, fdHash1, 0));
    if (!pHash1 || pHash1 == MAP_FAILED) {
        Perror("Cannot map hash (1)");
        pHash1 = NULL;
        unmapData1();
        return false;
    }

    pDict1 = reinterpret_cast<char *>(mmap(NULL, getPageAlignedValue(pParams1->dictSize), 
                                        PROT_WRITE | PROT_READ, MAP_SHARED, fdDict1, 0));
    if (!pDict1 || pDict1 == MAP_FAILED) {
        Perror("Cannot map dict (1)");
        pDict1 = NULL;
        unmapData1();
        return false;
    }
    return true;
}

bool CSpellIndex::mapData2()
{
    pParams2 = reinterpret_cast<CHashParams *>(mmap(NULL, getPageAlignedValue(sizeof *pParams2), 
                                        PROT_WRITE | PROT_READ, MAP_SHARED, fdParams2, 0));
    if (!pParams2 || pParams2 == MAP_FAILED) {
        Perror("Cannot map params (2)");
        pParams2 = NULL;
        return false;
    }
    pHash2 = reinterpret_cast<size_t *>(mmap(NULL, getPageAlignedValue(pParams2->hashSize * sizeof(size_t)), 
                                        PROT_WRITE | PROT_READ, MAP_SHARED, fdHash2, 0));
    if (!pHash2 || pHash2 == MAP_FAILED) {
        Perror("Cannot map hash (2)");
        pHash2 = NULL;
        unmapData2();
        return false;
    }

    pDict2 = reinterpret_cast<char *>(mmap(NULL, getPageAlignedValue(pParams2->dictSize), 
                                        PROT_WRITE | PROT_READ, MAP_SHARED, fdDict2, 0));
    if (!pDict2 || pDict2 == MAP_FAILED) {
        Perror("Cannot map dict (2)");
        pDict2 = NULL;
        unmapData2();
        return false;
    }
    return true;
}

bool CSpellIndex::mapData()
{
    if (!openedAll) {
        cerr << "DB is not open, cannot map" << endl;
        return false;
    }
    if (!mapData1() || !mapData2()) {
        return false;
    }
    return true;
}

void CSpellIndex::unmapData1(){
    if (pParams1) {
        if (pHash1 && munmap(pHash1, getPageAlignedValue(pParams1->hashSize * sizeof(size_t)))) {
            Perror("Cannot unmap hash");
        }
        pHash1 = NULL;
        if (pDict1 && munmap(pDict1, getPageAlignedValue(pParams1->dictSize))) {
            Perror("Cannot unmap dict");
        }
        pDict1 = NULL;
        if (munmap(pParams1, getPageAlignedValue(sizeof *pParams1))) {
            Perror("Cannot unmap params");
        }
        pParams1 = NULL;
    }
}

void CSpellIndex::unmapData2(){
    if (pParams2) {
        if (pHash2 && munmap(pHash2, getPageAlignedValue(pParams2->hashSize * sizeof(size_t)))) {
            Perror("Cannot unmap hash");
        }
        pHash2 = NULL;
        if (pDict2 && munmap(pDict2, getPageAlignedValue(pParams2->dictSize))) {
            Perror("Cannot unmap dict");
        }
        pDict2 = NULL;
        if (munmap(pParams2, getPageAlignedValue(sizeof *pParams2))) {
            Perror("Cannot unmap params");
        }
        pParams2 = NULL;
    }
}

bool CSpellIndex::CloseDB1()
{
    if (!opened1) {
        return true;
    }
    unmapData1();
    if (close(fdDict1)) {
        Perror("Cannot close dict file");
        return false;
    }
    if (close(fdHash1)) {
        Perror("Cannot close hash file");
        return false;
    }
    if (close(fdParams1)) {
        Perror("Cannot close param file");
        return false;
    }
    opened1 = false;

    return true;
}

bool CSpellIndex::CloseDB2()
{
    if (!opened2) {
        return true;
    }
    unmapData2();
    if (close(fdDict2)) {
        Perror("Cannot close dict file");
        return false;
    }
    if (close(fdHash2)) {
        Perror("Cannot close hash file");
        return false;
    }
    if (close(fdParams2)) {
        Perror("Cannot close param file");
        return false;
    }
    opened2 = false;

    return true;
}

bool CSpellIndex::CloseDB()
{
    if (!openedAll)
        return true;
    if (!CloseDB1() || !CloseDB1()) {
        return false;
    }
    openedAll = false;
    return true;
}
