#include "onedel.h"

void COneDelIndex::insert(const char *p)
{
    string str(p);

    int sz = str.size();
    if (!sz) return ;

    char origChar = str[0];

    insertInternal(p, 0, 0); // the string itself

    str.erase(str.begin());
    for (int i = 1; i <= sz; ++i) {
        insertInternal(str.c_str(), i, origChar);
        if (i < sz) {
            swap(origChar, str[i - 1]);
        }
    }
    RecalcIterators();
}

void COneDelIndex::search(string &str, bool hasTransp, int *numComp, StringList *resList, bool exOnly)
{
    int sz = str.size();
    if (!sz) return ;

    char origChar = str[0];

    searchInternal(str.c_str(), false, 0, 0, numComp, resList, exOnly); // the string itself
    if (exOnly) return;

    str.erase(str.begin());
    for (int i = 1; i <= sz; ++i) {
        searchInternal(str.c_str(), hasTransp, i, origChar, numComp, resList);
        if (i < sz) {
            swap(origChar, str[i - 1]);
        }
    }
    str.insert(str.end(), origChar);

    return;
}

void COneDelIndex::searchInternal(const char* p, bool hasTransp, unsigned char pos, unsigned char sym, int *numComp, StringList *resList, bool exOnly)
{
    hmapci::iterator it = mIndex.find(p);
    if (numComp) (*numComp)++;
    if (it != mIndex.end()) {
        const Record &rec = it->second;
        unsigned int cnt = 0;
        const unsigned char *pPos = NULL;
        const unsigned char *pSym = NULL;
        if (rec.mIsExt) {
            unsigned int pos = rec.GetPos();
            assert(pos < mExtPos.size());
            const vuc &  arrPos = mExtPos[pos];
            const vuc &  arrSym = mExtSym[pos];
            cnt = arrPos.size();
            pPos = &arrPos[0];
            pSym = &arrSym[0];
        } else {
            pPos = rec.mPos;
            pSym = rec.mSym;
            cnt  = rec.mCnt;
        }
        if (numComp) (*numComp) += cnt;
        if (exOnly) {
            for (unsigned int i = 0; i < cnt; ++i) 
            if (!pPos[i]) {
                if (resList) {
                    resList->insert(resList->begin(), p);
                    string &str = *resList->begin();
                    if (pPos[i]) {
                        str.insert(str.begin() + pPos[i] - 1, pSym[i]);
                    }
                }
            }
            return;
        }
        for (unsigned int i = 0; i < cnt; ++i) {
            unsigned char Pos1 = pPos[i];
            if (!pos || !Pos1 || Pos1 == pos || 
                (hasTransp && pos && Pos1 == pos + 1 && pSym[i] == sym)) {
                if (resList) {
                    resList->insert(resList->begin(), p);
                    string &str = *resList->begin();
                    if (pPos[i]) {
                        str.insert(str.begin() + pPos[i] - 1, pSym[i]);
                    }
                }
            }
        }
    }
}

void COneDelIndex::insertInternal(const char *p, unsigned char pos, unsigned char sym)
{
    hmapci::iterator it = mIndex.find(p);
    if (it == mIndex.end()) {
        mIndex.insert(make_pair(strDup(p), Record()));
        it = mIndex.find(p);
        assert(it != mIndex.end());
    }
    Record &rec = it->second;
    if (!rec.mIsExt) {
        if (rec.mCnt < INTERN_CNT) {
            rec.mPos[rec.mCnt] = pos;
            rec.mSym[rec.mCnt++] = sym;
        } else {
            size_t extSize = mExtPos.size();
            if (extSize % REALLOC_QTY1 == 0) {
                mExtPos.reserve(extSize + REALLOC_QTY1);
                mExtSym.reserve(extSize + REALLOC_QTY1);
            }

            mExtPos.resize(extSize + 1);
            mExtSym.resize(extSize + 1);

            vuc &tmpPos = mExtPos[extSize];
            vuc &tmpSym = mExtSym[extSize];

            tmpPos.reserve(INTERN_CNT + 1); // More CPU but less memory
            tmpSym.reserve(INTERN_CNT + 1); // More CPU but less memory

            for (int i = 0; i < INTERN_CNT; ++i) {
                tmpPos.push_back(rec.mPos[i]);
                tmpSym.push_back(rec.mSym[i]);
            }

            tmpPos.push_back(pos);
            tmpSym.push_back(sym);

            rec.SetPos(extSize); 
            assert(rec.mIsExt); // changes mIsExt to 1
            //mExtPos.push_back(tmpPos);
            //mExtSym.push_back(tmpSym);
        }
    } else {
        size_t extPos = rec.GetPos();
        assert(extPos < mExtPos.size());
        mExtPos[extPos].push_back(pos);
        mExtSym[extPos].push_back(sym);
    }
}

