#ifndef TWODEL_H
#define TWODEL_H

#include <vector>
#include <set>
#include <list>
#include "interface.h"

using namespace std;

typedef vector<unsigned char> vuc;

const unsigned int DefaultSize2 = 1000000U;

#define INTERN_CNT2 3 // Should be more than 0
#define REALLOC_QTY2 100000

class CTwoDelIndex {
public:
    struct Record {
        unsigned char mPos1[INTERN_CNT2];
        unsigned char mSym1[INTERN_CNT2];
        unsigned char mPos2[INTERN_CNT2];
        unsigned char mSym2[INTERN_CNT2];
        unsigned char mIsExt ;
        unsigned char mCnt   ;
        Record() { memset(this, 0, sizeof *this); }
        unsigned int GetPos() const { return *((u_int32_t *)&mPos1[0]);}
        unsigned int SetPos(u_int32_t val) { 
            *((u_int32_t *)&mPos1[0]) = val;
            mIsExt = 1;
        }
    };
    typedef hash_map<const char *, Record, MyHash, Equal >  hmapci;
protected:
    struct MemBlock {
        char *mpBeg;
        int  mSize;
        int  mFreePos;
        MemBlock(int size) : mSize(size), mpBeg(new char[size]), mFreePos(0) {}
        ~MemBlock(){delete [] mpBeg;}
    };
public:
    class ConstIndexIterator {
        friend class CTwoDelIndex;
        hmapci::const_iterator   mIt;
        const CTwoDelIndex       *mpIndex;
        ConstIndexIterator() :mpIndex(NULL) {}
        ConstIndexIterator(const CTwoDelIndex *pIndex, const hmapci::const_iterator &it) : mIt(it), mpIndex(pIndex) {}
    public:
        ConstIndexIterator(const ConstIndexIterator &oth) {
            mpIndex = oth.mpIndex;
            mIt = oth.mIt;
        }
        const ConstIndexIterator &operator=(const ConstIndexIterator &oth) {
            mpIndex = oth.mpIndex;
            mIt = oth.mIt;
        }
        ConstIndexIterator &operator++() {++mIt;return *this;} // prefix
        ConstIndexIterator operator++(int k) { // postfix
            hmapci::const_iterator oldIt(mIt);
            ++mIt;
            return ConstIndexIterator(mpIndex, oldIt);
        }
        bool operator<(const ConstIndexIterator & oth) const {
            return mpIndex == oth.mpIndex && (*mIt).first < (*oth.mIt).first;
        }
        bool operator==(const ConstIndexIterator & oth) const {
            return mpIndex == oth.mpIndex && mIt == oth.mIt;
        }
        bool operator!=(const ConstIndexIterator & oth) const {return !this->operator==(oth);}
        const string GetWord() const {
            return mIt->first;
        }
        void GetPosSym(vuc &arrPos1, vuc &arrSym1, 
                       vuc &arrPos2, vuc &arrSym2,
                        int &cnt) const {
            const Record &rec = mIt->second;
            if (rec.mIsExt) {
                unsigned int pos = rec.GetPos();
                assert(pos < mpIndex->mExtPos1.size());
                arrPos1 =  mpIndex->mExtPos1[pos];
                arrSym1 =  mpIndex->mExtSym1[pos];
                arrPos2 =  mpIndex->mExtPos2[pos];
                arrSym2 =  mpIndex->mExtSym2[pos];
                cnt = arrPos1.size();
            } else {
                cnt  = rec.mCnt;
                arrPos1.resize(cnt);
                arrSym1.resize(cnt);
                arrPos2.resize(cnt);
                arrSym2.resize(cnt);
                for (int i = 0; i < cnt; ++i) {
                   arrPos1[i] = rec.mPos1[i]; 
                   arrSym1[i] = rec.mSym1[i];
                   arrPos2[i] = rec.mPos2[i]; 
                   arrSym2[i] = rec.mSym2[i];
                }
            }
        }
    };
    CTwoDelIndex() {allocMemBlock(DefaultSize2);RecalcIterators();}
    ~CTwoDelIndex() {
        for(list<MemBlock *>::iterator it = mMemBlocks.begin(); it != mMemBlocks.end(); ++it) {
            delete *it;
        }
    }
    void insert(const char *p);
    void search(string &str, bool hasTransp, int *numCmp, StringList *resList);
    int  GetDictSize() {return mIndex.size();}
protected:
    ConstIndexIterator mBegin, mEnd;
    friend class ConstIterIndex;
    hmapci             mIndex;
    vector<vuc>        mExtPos1;
    vector<vuc>        mExtSym1;
    vector<vuc>        mExtPos2;
    vector<vuc>        mExtSym2;
    list<MemBlock *>   mMemBlocks;
    MemBlock           *mpLastMemBlock;
    void allocMemBlock(int sz) {
        mpLastMemBlock = new MemBlock(sz);
        mMemBlocks.insert(mMemBlocks.begin(), mpLastMemBlock);
    }
    char *strDup(const string &str) {
        assert(mpLastMemBlock); 
        unsigned int sz = str.size() + 1;
        int freeSz = mpLastMemBlock->mSize - mpLastMemBlock->mFreePos;
        if (freeSz < sz) {
            allocMemBlock(max(DefaultSize2, sz));
        }
        char *p = mpLastMemBlock->mpBeg + mpLastMemBlock->mFreePos;
        strcpy(p, str.c_str());
        mpLastMemBlock->mFreePos += sz;
        return p;
    }
    void insertInternal(const char *p, unsigned char pos1, unsigned char sym1, unsigned char pos2, unsigned char sym2);
    void searchInternal(const char *p, bool hasTransp, 
                        unsigned char pos1, unsigned char sym1, unsigned char pos2, unsigned char sym2, 
                        int *numCmp, StringList *resList);
    void RecalcIterators() {
        mBegin = ConstIndexIterator(this, mIndex.begin());
        mEnd = ConstIndexIterator(this, mIndex.end());
    }
public:
    const ConstIndexIterator & begin() const {return mBegin;}
    const ConstIndexIterator & end()   const {return mEnd;}
};


#endif
