#ifndef ALTERN_MORFRANK_PERFHASH_H
#define ALTERN_MORFRANK_PERFHASH_H

#include "interface.h"
#include <fcntl.h>
#include <iostream>
#include <vector>
#include <string>

#include <cmph.h>

#include "../Common/gendel.h"
#include "../Common/comptable.h"

#define SHARD_QTY 100

typedef hash_map<string, StringList> String2StringList;
typedef hash_map<unsigned, StringList> Unsigned2StringList;

class CAlternGenMorFrankPerfHash : public CInterface {
public:
    CAlternGenMorFrankPerfHash(bool hasTransp, EAlphabet AlphaType);
    ~CAlternGenMorFrankPerfHash() {
        for (unsigned kerr = 0; kerr < MAX_KERR; ++kerr) {
            for (unsigned ShardId = 0; ShardId < SHARD_QTY; ++ShardId) {
                delete [] maapHashContent[kerr][ShardId];
                delete maapCompactHashTable[kerr][ShardId];
            }
            ClearPerfectHash(kerr);
        }
        if (mpDictPerfectHash) {
            cmph_destroy(mpDictPerfectHash);
        }
    }

    void LoadPerfectHash(unsigned kerr);
    void LoadDictPerfectHash();
    void ClearPerfectHash(unsigned kerr) {
        for (unsigned ShardId = 0; ShardId < SHARD_QTY; ++ShardId) {
            if (maapPerfectHash[kerr][ShardId]) {
                cmph_destroy(maapPerfectHash[kerr][ShardId]);
                maapPerfectHash[kerr][ShardId] = NULL;
            }
        }
    }
    bool CreateInMemIndex(const char *inputFile,
                             int insertCount, int *dcnt, 
                             bool silent = true);
    //=========================================================================
    bool Insert(const char *str);
    bool Search(const char *str, int kerr, int *numComp,
                 StringList * resList, bool silent = true);
    //=========================================================================
    bool SearchInDB(const char *str, int kerr, int *numComp,
                    StringList * resList, bool silent = true) {
        cerr << "SearchInDB is not supported" << endl;
        return false;
    }
    //=========================================================================
    bool DumpDB (const char *base_name);
    bool OpenDB (const char *base_name, bool create = false);
    bool CloseDB ();
    void PrintAddStat(SearchType type, int testNum, int kerr, ostream &);
#ifdef _DEBUG
    void Print ();
#endif

    bool FindInDict(const char* pStr, size_t sz) const {
        size_t Pos = cmph_search(mpDictPerfectHash, pStr, sz) % mvDictStr.size();

        return !strcmp(pStr, mvDictStr[Pos].c_str());
    }

    static string PerfectHashName(unsigned kerr, unsigned ShardId) {
        stringstream fn;

        fn << "PerfectHashK=" << kerr << "_ShardId=" << ShardId << ".mph";

        return fn.str();
    }

    static string PerfectHashDictName() {
        return "PerfectHash.mph";
    }

    static string CompactHashTableName(unsigned kerr, unsigned ShardId) {
        stringstream fn;

        fn << "CompactHashTableK=" << kerr << "_ShardId=" << ShardId;

        return fn.str();
    };

    static string HashContentName(unsigned kerr, unsigned ShardId) {
        stringstream fn;

        fn << "HashContentK=" << kerr << "_ShardId=" << ShardId;

        return fn.str();
    }
private:
    vector<string>      mvDictStr;
    bool                mbOpened;

    string TmpDictShardName(unsigned kerr, unsigned ShardId) {
        stringstream fn;

        fn << "TmpDict_K=" << kerr << "_ShardId=" << ShardId;

        return fn.str();
    }

    string TmpSortedDictShardName(unsigned kerr, unsigned ShardId) {
        stringstream fn;

        fn << "TmpSortedDict_K=" << kerr << "_ShardId=" << ShardId;

        return fn.str();
    }

    string TmpChunkName(unsigned kerr, unsigned ShardId) {
        stringstream fn;

        fn << "TmpChunk_K=" << kerr << "_ShardId=" << ShardId;

        return fn.str();
    }

    unsigned char*  maapHashContent[MAX_KERR + 1][SHARD_QTY];
    size_t          maaTableSize[MAX_KERR + 1][SHARD_QTY];
    CCompactTable*  maapCompactHashTable[MAX_KERR + 1][SHARD_QTY];
    cmph_t*         maapPerfectHash[MAX_KERR + 1][SHARD_QTY];
    cmph_t*         mpDictPerfectHash;
    bool            maaHasString[MAX_KERR + 1][SHARD_QTY];
};

template <>
inline bool SearchDbNotSupported<CAlternGenMorFrankPerfHash>() {
    return true;
}


#endif
