#ifndef PACK_H
#define PACK_H

#define MAX_ALPHA_SIZE_LOG          6
#define MAX_STRING_SIZE_LOG         6
#define MAX_SUPPORTED_ALPHA_SIZE    (1<<MAX_ALPHA_SIZE_LOG) 
#define MAX_STRING_LEN              ((1<<MAX_STRING_SIZE_LOG) - 2)

// Exactly 3 bytes!!! (instead of 4, if each value were stored using a byte)
struct CCharPos2 {
    unsigned mCharIndex1:6;
    unsigned mPos1:6;
    unsigned mCharIndex2:6;
    unsigned mPos2:6;
}
#ifdef __GNUC__
 __attribute__ ((aligned(1)))
 __attribute__ ((packed))
#endif
;

// 4.5 -> 5 bytes
struct CCharPos3_1 {
    unsigned mCharIndex1:6;
    unsigned mPos1:6;
    unsigned mCharIndex2:6;
    unsigned mPos2:6;
    unsigned mCharIndex3:6;
    unsigned mPos3:6;
}
#ifdef __GNUC__
 __attribute__ ((aligned(1)))
 __attribute__ ((packed))
#endif
;

// 9 bytes (otherwise would be 12)
struct CCharPos3_2 {
    unsigned mCharIndex1_1:6;
    unsigned mPos1_1:6;
    unsigned mCharIndex1_2:6;
    unsigned mPos1_2:6;
    unsigned mCharIndex1_3:6;
    unsigned mPos1_3:6;

    unsigned mCharIndex2_1:6;
    unsigned mPos2_1:6;
    unsigned mCharIndex2_2:6;
    unsigned mPos2_2:6;
    unsigned mCharIndex2_3:6;
    unsigned mPos2_3:6;
}
#ifdef __GNUC__
 __attribute__ ((aligned(1)))
 __attribute__ ((packed))
#endif
;

// Exactly 6 bytes (as opposed to 8, if each value were stored using a byte)
struct CCharPos4 {
    unsigned mCharIndex1:6;
    unsigned mPos1:6;
    unsigned mCharIndex2:6;
    unsigned mPos2:6;
    unsigned mCharIndex3:6;
    unsigned mPos3:6;
    unsigned mCharIndex4:6;
    unsigned mPos4:6;
}
#ifdef __GNUC__
 __attribute__ ((aligned(1)))
 __attribute__ ((packed))
#endif
;

/* 
   To ensure that there is always enough space 
   in the destination buffer, use the function PackedSize
   below.

   pCharPosArr contains 2 * k * Qty entries in the form

    char1_1 pos1_1 char1_2 pos1_2 ... char1_k pos1_k
    char2_1 pos2_1 char2_2 pos2_2 ... char2_k pos2_k
                        * * *
    charN_1 posN_1 charN_2 posN_2 ... charN_k posN_k,

    where N = Qty

    RETURN VALUE: The size of packed data 
*/
size_t PackCharPos(EAlphabet kAlphaType, 
                   const unsigned char pCharPosArr[], 
                   size_t SrcQty, unsigned k,
                   char* pDst, size_t DstSize);

// RETURN VALUE: the number of bytes required 
size_t PackedSize(size_t Qty, unsigned k);


/*
    To ensure that there is always enough space
    in the destination buffer,
    DstSize should be >= SrcSize (each packed element requires at least 1 byte 2 bits)
    Note that each destination buffer should have at least DstSize bytes!!

    RETURN VALUE: The number of elements in each of the target buffers: pDstchar, pDstPos
 */
size_t UnpackCharPos(const char* pABC,
                    const unsigned char* pSrc, size_t SrcSize, unsigned k,
                    unsigned char* pDstChar, unsigned char* pDstPos, size_t DstSize);
                    


#endif
