#include "index.h"

static long long seekVal = 0;

// Doesn't clear vRes!!! This is an important property.
static void 
SearchHammingConstLenUnrolledLoop(unsigned Qty, unsigned kerr, const char* pStr, const char* pPatBeg, unsigned PatLen, StringList *pResList) {
    if (!Qty || !PatLen) {
        return;
    }
    const char *pText = pStr;
    const char *pPat = pPatBeg;
    const char *pNext = pStr + PatLen;
    const char *pStrEnd = pStr + PatLen * Qty;

    unsigned ErrQty = 0;

    #define STEP {ErrQty += (*pPat++ != *pText++); \
                  if (ErrQty > kerr) { pText = pNext; goto exit;}}
    //#define STEP ErrQty += (*pPat++ != *pText++);

    while (true) {
        switch (PatLen % 8) {
            case 0: while (pText < pNext) {
                        STEP
            case 7:     STEP
            case 6:     STEP
            case 5:     STEP
            case 4:     STEP
            case 3:     STEP
            case 2:     STEP
            case 1:     STEP
                    }
        }
        if (ErrQty <= kerr) {
            if (pResList) {
                pResList->push_back(string(pNext - PatLen, PatLen));
            }
       }
exit:
       if (pNext < pStrEnd) { pNext += PatLen; } 
       else break;
       ErrQty = 0;
       pPat = pPatBeg;
   }
}

// Doesn't clear vRes!!! This is an important property.
static void 
SearchHammingConstLen(unsigned Qty, unsigned kerr, const char* pStr, const char* pPatBeg, unsigned PatLen, StringList *pResList) {
    const char *pText = pStr;
    const char *pPat = pPatBeg;
    const char *pNext = pStr + PatLen;
    const char *pStrEnd = pStr + PatLen * Qty;

    for (unsigned ErrQty = 0; pText <= pStrEnd;) {
        if (pText < pNext) {
            ErrQty += (*pPat++ != *pText++);
            if (ErrQty > kerr) {
                pPat = pPatBeg;
                pText = pNext;
                pNext += PatLen;
                ErrQty = 0;
           }
       } else {
           if (ErrQty <= kerr) {
                if (pResList) {
                    pResList->push_back(string(pNext - PatLen, PatLen));
                }
           }
           pPat = pPatBeg;
           if (pNext < pStrEnd) { pNext += PatLen; } 
           else break;
           ErrQty = 0;
       }
   }
}

CAlternGenRedAlpha::CAlternGenRedAlpha(bool hasTransp, EAlphabet AlphaType) : CInterface(hasTransp, AlphaType), mbOpened(false), mpABC(GetAlphabet(AlphaType)), mRedAlphaSize(GetRedAlphaSize())
{
    cout << "Reduced alphabet size: " << mRedAlphaSize << endl;
    if (!mpABC) {
        cerr << "Fatal internal error, cannot retrieve alphabet information, AlphaType:" << AlphaType << endl;
        abort();
    }
    mAlphaSize = strlen(mpABC);

    if ((unsigned char)REDALPHA_FIRST_CHAR + (unsigned)MAX_REDALPHA_SIZE > 256) {  
        cerr << "Fatal error: invalid combination of REDALPHA_FIRST_CHAR & MAX_REDALPHA_SIZE" << endl;
        abort();
    }

    if (mRedAlphaSize > mAlphaSize) {
        cerr << "Fatal error: The size of the reduced alphabet " << mRedAlphaSize << ") should be smaller than the size of the original alphabet ("  << mAlphaSize << ")" << endl;
        abort();
    }

    if (mRedAlphaSize > MAX_REDALPHA_SIZE) {
        cerr << "Fatal error: The size of the reduced alphabet " << mRedAlphaSize << ") should be smaller than MAX_REDALPHA_SIZE ("  << MAX_REDALPHA_SIZE << ")" << endl;
        abort();
    }

    for (unsigned i = 0; i < mRedAlphaSize; ++i) {
        maRedABC[i] = REDALPHA_FIRST_CHAR + i;
    }
    maRedABC[mRedAlphaSize] = 0;
}

int 
CAlternGenRedAlpha::GetRedAlphaSize()
{
    int res = ALPHABET_SIZE;

    GetEnv("ALPHABET_SIZE", res);

    return res;
}

bool 
CAlternGenRedAlpha::Insert (const char *str)
{
    cerr << "Insert unsupported" << endl;
    return false;
}

bool 
CAlternGenRedAlpha::CreateInMemIndex(const char *inputFile,
                                         int insertCount, int *dcnt,
                                         bool silent)
{
    bool bOptim = false;
#ifdef OPTIMIZE_BY_STAT
    bOptim = true;
#endif

    const char *p = getenv("OPTIMIZE_BY_STAT");

    if (p) {
        bOptim = atol(p) != 0;
    }

    if (bOptim) {
        CollectStat st;

        st.CollectFileStat (inputFile);
        if (!CalcHMap (mRedAlphaSize, st.stat (), mAlphaMap, getenv("PRINT_ADD_STAT"))) {
            std::cerr << "Hash generation failed!" << std::endl;
            return false;
        }
    } else {
        for (const char *pSym = mpABC; char c = *pSym; ++pSym) {
            int v = (pSym - mpABC) % mRedAlphaSize + 1;

            mAlphaMap.setValue(tolower_eng_rus(c), v);
            mAlphaMap.setValue(toupper_eng_rus(c), v);
        }
    }

    cout << "Optimized by stat? " << bOptim << endl;
    cout << "The size of reduced alphabet: " << mRedAlphaSize << endl;

    ifstream    InputFile(inputFile);
    string      s;

    for ((*dcnt) = 0; getline(InputFile, s); (*dcnt)++) {
        toUpper (s);
        InsertInternal(s);
    }

    return true;
}

void
CAlternGenRedAlpha::InsertInternal(const string& sOrig)
{
    string sRed = sOrig;

    ToRedStr(sRed);
    THashData::iterator it = mRedHash.find(sRed.c_str());

    if (it == mRedHash.end()) {
        mRedHash.insert(make_pair(newstrdup(sRed.c_str()), CHashElem()));
        it = mRedHash.find(sRed.c_str());
        if (it == mRedHash.end()) {
            THROW_AND_PRINT_RUNTIME_ERROR("Internal error: element is not found after insertion");
        }
    }
    CHashElem&  elem = it->second;

    elem.mQty++;
    elem.mStrings.append(sOrig);

}

bool 
CAlternGenRedAlpha::Search (const char *str, int kerr, int *numComp, StringList * pResList, bool silent)
{
    string      sOrigPat(str);
    string      sRedPat(str);
    int         nc = 0;

    ToRedStr(sRedPat);
#if 0
    if (GetEnvSwitch("SHOW_VARIANTS")) {
        cout << "\"" << sOrigPat << "\" -> \"" << sRedPat << "\" kerr = " << kerr << endl;
    }
#endif

    ExpandAndSearch (0, nc, sRedPat, sOrigPat, kerr, kerr, pResList);

    if (numComp) *numComp = nc;

    return true;
}

void 
CAlternGenRedAlpha::SearchInBucket(const char* pOrigPat,
                                   const char* pRedPat,
                                   int kErrOrig,
                                   StringList* pResList,
                                   int& NumComp) 
{
    THashData::const_iterator it = mRedHash.find(pRedPat);

    if (it != mRedHash.end()) {
        const CHashElem&    elem = it->second;

        size_t OrigPatSize = strlen(pOrigPat);
        size_t ExpSize = OrigPatSize * elem.mQty;

        if (elem.mStrings.size() != ExpSize) {
            THROW_AND_PRINT_RUNTIME_ERROR("Fatal error: search bucket size and pattern size mismatch! String size: " << elem.mStrings.size() << " qty: " << elem.mQty << " Pattern size: " << OrigPatSize);
        }

        NumComp += elem.mQty;

        SearchHammingConstLenUnrolledLoop(elem.mQty, kErrOrig, elem.mStrings.data(), pOrigPat, OrigPatSize, pResList);
        //SearchHammingConstLen(elem.mQty, kErrOrig, elem.mStrings.data(), sOrigPat.data(), sOrigPat.size(), pResList);
    }
}

void 
CAlternGenRedAlpha::ExpandAndSearch(unsigned start, int& NumComp, 
                          string& sRedPat, 
                          string& sOrigPat, 
                          unsigned kErr, 
                          unsigned kErrOrig,
                          StringList* pResList)
{
    int size = sRedPat.size ();
    if (size != sOrigPat.size()) {
        THROW_AND_PRINT_RUNTIME_ERROR("Fatal error: orig (size:" << sOrigPat.size() << ") and reduced pattern (size:" << sRedPat << ") have different size!");
    }

    if (1) {
#if 0
        if (GetEnvSwitch("SHOW_VARIANTS")) {
            std::cout << "Variant: \"" << sRedPat << "\" and \"" << sOrigPat <<  "\"" << std::endl;
        }
#endif
        /*
         * Find strings in the bucket. Store them in pResList
         */
        SearchInBucket(sOrigPat.c_str(), sRedPat.c_str(), kErrOrig, pResList, NumComp);
    }
    if (!kErr || start > size) return;

    int i = 0;
    // Substitution: don't have to modify the original pattern
    for (i = start; i < size; i++) {
        char origChar = sRedPat[i];
        for (const unsigned char *ptr = maRedABC; *ptr; ptr++) {
            if (*ptr != sRedPat[i]) {
                sRedPat[i] = *ptr;
                ExpandAndSearch(i + 1, NumComp, sRedPat, sOrigPat, kErr - 1, kErrOrig, pResList);
            }
        }
        sRedPat[i] = origChar;
    }
    // We well try to make as few insertions and erases as possible
    // It is enough to make 1 insertion, 1 deletion per insertion case
    // and 1 insertion, 1 deletion per deletion case
    // Insertion
    sRedPat.insert(sRedPat.begin() + start, ' ');
    sOrigPat.insert(sOrigPat.begin() + start, ' ');

    for (i = start; i <= size; i++) {
        for (const unsigned char *ptr = maRedABC; *ptr; ptr++) {
            sRedPat[i] = *ptr; 
            ExpandAndSearch(i + 1, NumComp, sRedPat, sOrigPat, kErr - 1, kErrOrig, pResList);
        }
        if (i < size) {
            swap(sRedPat[i], sRedPat[i + 1]);
            swap(sOrigPat[i], sOrigPat[i + 1]);
        }
    }
    sRedPat.erase(sRedPat.end() - 1);
    sOrigPat.erase(sOrigPat.end() - 1);

    if (mHasTransp) {
    // Transposition
        for (i = start; i < size - 1; i++) {
            swap(sRedPat[i],  sRedPat[i+1]);
            swap(sOrigPat[i], sOrigPat[i+1]);

            ExpandAndSearch(i + 2, NumComp, sRedPat, sOrigPat, kErr - 1, kErrOrig, pResList);
            swap(sRedPat[i],  sRedPat[i+1]);
            swap(sOrigPat[i], sOrigPat[i+1]);
        }
    }
    // Deletion
    if (start >= size) return;
    char FirstCharRed  = sRedPat[start];
    char FirstCharOrig = sOrigPat[start];

    sRedPat.erase(sRedPat.begin() + start);
    sOrigPat.erase(sOrigPat.begin() + start);

    for (i = start; i < size; i++) {
        ExpandAndSearch(i, NumComp, sRedPat, sOrigPat, kErr - 1, kErrOrig, pResList);
        if (i < size - 1) {
           swap(sRedPat[i],  FirstCharRed);
           swap(sOrigPat[i], FirstCharOrig);
        }
    }
    sRedPat.insert(sRedPat.end(), FirstCharRed);
    sOrigPat.insert(sOrigPat.end(), FirstCharOrig);
}

bool
CAlternGenRedAlpha::DumpDB(const char* base_name)
{
    if (mbOpened)
        CloseDB ();
    if (!OpenDB (base_name, true))
        return false;           // Create database


    size_t numWords = mRedHash.size();
    vector<pair<size_t, THashDataIteratorWrap> >   pos2word;
    
    size_t dictSize = 0;
    // The first set iteration - write 0 and calc size
    //
    vector<char>     vZeroBuf;

    for (THashData::iterator it = mRedHash.begin (); it != mRedHash.end (); ++it) {
        const CHashElem&    Elem = it->second;
        size_t              ElemSize = strlen(it->first) + 1 + sizeof(Elem.mQty) + Elem.mStrings.size() + 1;
        vZeroBuf.resize(ElemSize);

        if (write(mFdDict, &vZeroBuf[0], ElemSize) != ElemSize) {
            Perror("Cannot write to dict file");
            CloseDB();
            return false;
        }

        dictSize += ElemSize;
        pos2word.push_back(make_pair(0U, it));
    }

    int fillRatio = defaultHashRatio;

    GetEnv("FILL_RATIO", fillRatio);
    if (!fillRatio) fillRatio = defaultHashRatio;

    if (fillRatio < minHashRatio) fillRatio = minHashRatio;
    if (fillRatio > maxHashRatio) fillRatio = maxHashRatio;

    size_t hashSize = from__stl_next_prime((size_t) 100LL * numWords / fillRatio); 
    if (!hashSize) hashSize = 1;

    size_t tmp(0);
    if (lseek(mFdHash, getPageAlignedValue((hashSize - 1) * sizeof(size_t)), SEEK_SET) < 0 ||
         write(mFdHash, &tmp, sizeof(size_t)) != sizeof(size_t)) {
        Perror("Cannot lseek or write");
        CloseDB();
        return false;
    }
    CHashParams prm(fillRatio, hashSize, dictSize);

    if (write(mFdParams, &prm, sizeof prm) != sizeof prm) {
        Perror("Cannot write params");
        CloseDB();
        return false;
    }

    if (!mapData()) {
        CloseDB();
        return false;
    }

    memset(mpHash, 0, mpParams->hashSize * sizeof(size_t));

    seekVal = 0;

    size_t n = 0;
    int    cntIt = 0;
    
    // Find out which hash table elements to fill
    for (THashData::iterator it = mRedHash.begin (); it != mRedHash.end (); ++it) {
        const CHashElem&    Elem = it->second;
        size_t              ElemSize = strlen(it->first) + 1 + sizeof(Elem.mQty) + Elem.mStrings.size() + 1;
        size_t              buckNum;

        if (!insertHashDB(it->first, buckNum, *mpParams, mpHash, &seekVal)) {
            cerr << "Failed to insert string internally" << endl;
            CloseDB();
            return false;
        }
#if 0
        if (cntIt % 10000 == 0) {
            cout << 100 * cntIt / numWords << "% seekVal " << double(seekVal) / max(1, cntIt) << endl;
        }
#endif
        pos2word[n++].first = buckNum;
    }
    // Finally store strings in the order they are in hash table
    // and fill hash positions
    sort(pos2word.begin(), pos2word.end()); // will sort by position

    size_t pos = 0;
    for (n = 0; n < pos2word.size(); ++n) {
        THashData::iterator&    it = pos2word[n].second.mIt;
        const CHashElem&        Elem = it->second;
        size_t                  ElemSize = strlen(it->first) + 1 + sizeof(Elem.mQty) + Elem.mStrings.size() + 1;

        size_t pos1 = pos;
        memcpy(&mpDict[pos], it->first, strlen(it->first) + 1);
        pos += strlen(it->first) + 1;
        memcpy(&mpDict[pos], &Elem.mQty, sizeof Elem.mQty);
        pos += sizeof Elem.mQty;
        memcpy(&mpDict[pos], Elem.mStrings.c_str(), Elem.mStrings.size() + 1);
        pos += Elem.mStrings.size() + 1;
        mpHash[pos2word[n].first] = pos1 + 1; // do not store zero value

        if (pos - pos1 != ElemSize) {
            THROW_AND_PRINT_RUNTIME_ERROR("Internal error: position diff mismatch");
        }
    }

    return CloseDB ();
}


bool 
CAlternGenRedAlpha::OpenDB (const char *base_name, bool create)
{
    if (mbOpened)
        CloseDB ();

    string    headerName = HeaderFileName(base_name);
    string    hashName = HashFileName(base_name);
    string    dictName = DictFileName(base_name);

    int flags = O_RDWR;
    int access = 0;

    mpDict = NULL;
    mpHash = NULL;
    mpParams = NULL;

    if (create)  {
        flags = O_RDWR | O_TRUNC | O_CREAT;
        access = defaultAccess;
    }
    mFdParams = open(headerName.c_str(), flags, access);
    if (mFdParams < 0) {
        Perror("Cannot open header file");
        return false;
    }
    mFdHash = open(hashName.c_str(), flags, access);
    if (mFdHash < 0) {
        Perror("Cannot open hash file");
        close(mFdParams);
        return false;
    }
    mFdDict = open(dictName.c_str(), flags, access);
    if (mFdDict < 0) {
        Perror("Cannot dict file");
        close(mFdHash);
        close(mFdParams);
        return false;
    }

    if (!create) {
        mbOpened = true;
        if (!mapData()) {
            CloseDB();
            return false;
        }
    }
    
    return mbOpened = true;
}

bool
CAlternGenRedAlpha::CloseDB ()
{
    if (!mbOpened)
        return true;
    unmapData();
    if (close(mFdDict)) {
        Perror("Cannot close dict file");
        return false;
    }
    if (close(mFdHash)) {
        Perror("Cannot close hash file");
        return false;
    }
    if (close(mFdParams)) {
        Perror("Cannot close param file");
        return false;
    }
    mbOpened = false;
    return true;
}

bool 
CAlternGenRedAlpha::mapData() 
{
    if (!mbOpened) {
        cerr << "DB is not open, cannot map" << endl;
        return false;
    }
    mpParams = reinterpret_cast<CHashParams *>(mmap(NULL, getPageAlignedValue(sizeof *mpParams), 
                                        PROT_WRITE | PROT_READ, MAP_SHARED, mFdParams, 0));
    if (!mpParams || mpParams == MAP_FAILED) {
        Perror("Cannot map params");
        mpParams = NULL;
        return false;
    }
    mpHash = reinterpret_cast<size_t *>(mmap(NULL, getPageAlignedValue(mpParams->hashSize * sizeof(size_t)), 
                                        PROT_WRITE | PROT_READ, MAP_SHARED, mFdHash, 0));
    if (!mpHash || mpHash == MAP_FAILED) {
        Perror("Cannot map hash");
        mpHash = NULL;
        unmapData();
        return false;
    }

    mpDict = reinterpret_cast<char *>(mmap(NULL, getPageAlignedValue(mpParams->dictSize), 
                                        PROT_WRITE | PROT_READ, MAP_SHARED, mFdDict, 0));
    if (!mpDict || mpDict == MAP_FAILED) {
        Perror("Cannot map dict");
        mpDict = NULL;
        unmapData();
        return false;
    }
    return true;
}

void CAlternGenRedAlpha::unmapData() 
{
    if (mpParams) {
        if (mpHash && munmap(mpHash, getPageAlignedValue(mpParams->hashSize * sizeof(size_t)))) {
            Perror("Cannot unmap hash");
        }
        mpHash = NULL;
        if (mpDict && munmap(mpDict, getPageAlignedValue(mpParams->dictSize))) {
            Perror("Cannot unmap dict");
        }
        mpDict = NULL;
        if (munmap(mpParams, getPageAlignedValue(sizeof *mpParams))) {
            Perror("Cannot unmap params");
        }
        mpParams = NULL;
    }
}
