#include <fstream>
#include <limits>
#include <iomanip>

#include "index.h"
#include "../Common/hash.h"

unsigned    aNumCheck[MAX_KERR + 1];
unsigned    aNumSearchInBucket[MAX_KERR + 1];

#define SKIP_LOG   5

CAlternGenRedAlphaMorFrankPerfHash::CAlternGenRedAlphaMorFrankPerfHash(bool hasTransp, EAlphabet AlphaType) : 
                                                        CInterface(hasTransp, AlphaType), 
                                                        mRedHash(hasTransp, AlphaType), 
                                                        mbOpened(false)
{
    memset(maapHashContent,  0, sizeof maapHashContent);
    memset(maapCompactHashTable, 0, sizeof maapCompactHashTable);
    memset(maapPerfectHash,  0, sizeof maapPerfectHash);
    memset(maaHasString, 0, sizeof maaHasString);
}

struct CUniqueStrCollectorShard {
    ofstream*           mpTmpDictShard;
    bool*               mpHasString;

    CUniqueStrCollectorShard(ofstream* pTmpDictShard, bool* pHasString) : 
                            mpTmpDictShard(pTmpDictShard), 
                            mpHasString(pHasString) {}

    void operator()(const string& s, unsigned char* pPos, char* pCurrChar, unsigned k) {
        unsigned iShardId = hash_func(s) % SHARD_QTY;
        mpHasString[iShardId] = true;

        mpTmpDictShard[iShardId] << s << endl;
    }
};

struct CInsertStr {
    cmph_t**    mppPerfectHash; 
    ofstream*   mpTmpChunk;
    unsigned    maMaxHashVal[SHARD_QTY];

    CInsertStr(cmph_t** ppPerfectHash, ofstream* pTmpChunk) : mppPerfectHash(ppPerfectHash), mpTmpChunk(pTmpChunk) 
                                                            {
        fill(maMaxHashVal, maMaxHashVal + SHARD_QTY, 0);
    }

    void operator()(const string& s, unsigned char* pPos, char* pCurrChar, unsigned k) {
        unsigned    NonPerfHashVal = hash_func(s);
        size_t      iShardId = NonPerfHashVal % SHARD_QTY;
        unsigned    HashVal = cmph_search(mppPerfectHash[iShardId], s.data(), s.size());

        maMaxHashVal[iShardId] = max(maMaxHashVal[iShardId], HashVal);

        size_t sz = s.size();

        mpTmpChunk[iShardId].write((const char*)&HashVal, sizeof(HashVal));
        mpTmpChunk[iShardId].write((const char*)pPos, k);
        mpTmpChunk[iShardId].write((const char*)pCurrChar, k);
    }
};

struct CSearchStr {
    CAlternGenRedAlpha&     mRedHash;
    cmph_t**                mppPerfectHash; 
    size_t*           mpTableSize;
    CCompactTable**   mppCompactHashTable;
    unsigned char**   mppHashContent;
    string                  mOrigPat;
    unsigned                mKerr;
    StringList*             mpResList;
    int                     mNumComp;
    unsigned                mNumCheck;
    unsigned                mNumSearchInBucket;

    CSearchStr(CAlternGenRedAlpha& RedHash,
               cmph_t** ppPerfectHash, 
               size_t* pTableSize, CCompactTable** ppCompactHashTable, unsigned char** ppHashContent, 
               const string& sOrigPat, unsigned kErr, StringList*   pResList) :
                   mRedHash(RedHash),
                   mppPerfectHash(ppPerfectHash), mpTableSize(pTableSize),
                   mppCompactHashTable(ppCompactHashTable), mppHashContent(ppHashContent), 
                   mOrigPat(sOrigPat), mKerr(kErr), mpResList(pResList) {
                        mNumCheck = 0;
                        mNumSearchInBucket = 0;
                        mNumComp = 0;
                   }

    void operator()(const string& s, unsigned char* pPatternPos, char* pCurrChar, unsigned kErrPattern) {
/*
            i1 i2 ... ik   IndexWord 
            j1 j2 ... jk   QueryWord (represented by pPos)
            ____________________________________________________________
            Transposition-unaware case.
            The comparing function produces indices:
            i_a1 ... i_al (in the index word, but not in the query word)
            j_b1 ... j_bm (in the query word, but not in the index word)
            k_1  ... k_n  (common indices)
            ____________________________________________________________
            
*/
        while (kErrPattern && pPatternPos[kErrPattern - 1] == 255) {
            --kErrPattern;
        }

        if (!mKerr) {
            string        sPat = mOrigPat;
            const string& sRedPat = s;

            // Apply the same deletions to the original pattern as to the reduced pattern
            for (unsigned i = 0; i < kErrPattern; ++i) {
                sPat.erase(pPatternPos[i], 1);
            }

            if (sPat.size() != sRedPat.size()) {
                THROW_AND_PRINT_RUNTIME_ERROR("Internal error reduced and original patterns have different length (after deletions are applied to the original pattern)")        ;
            }
        
            ++mNumCheck;
            ++mNumSearchInBucket;

            unsigned tmp = mNumComp;

            mRedHash.SearchInBucket(sPat.c_str(), sRedPat.c_str(), mKerr, mpResList, mNumComp);

            return;
        }
        unsigned ShardId = hash_func(s) % SHARD_QTY;
        if (!mppPerfectHash[ShardId]) return; // Shard is empty
        size_t HashVal = cmph_search(mppPerfectHash[ShardId], s.data(), s.size()) % mpTableSize[ShardId];

        size_t              DataSize = mKerr * 2;
        unsigned            RedPatSize = s.size();

        for (const unsigned char*   pIndexPos = mppHashContent[ShardId] + DataSize * mppCompactHashTable[ShardId]->GetVal(HashVal),
                                *   pIndexPosEnd = mppHashContent[ShardId] + DataSize * mppCompactHashTable[ShardId]->GetVal(HashVal + 1);
                                    pIndexPos < pIndexPosEnd;
                                    pIndexPos += DataSize) {
            int NumIndex = 0, NumPattern = 0; 
            unsigned kEstim = 0;

            ++mNumCheck;

            unsigned kErrIndex = mKerr;

            while (kErrIndex && pIndexPos[kErrIndex - 1] == 255) {
                --kErrIndex;
            }

            if (kErrIndex && pIndexPos[kErrIndex - 1] > RedPatSize) {
                continue;
            }

            unsigned kDel = 0;

            while (NumIndex < kErrIndex && NumPattern < kErrPattern) {
                unsigned char IndexPos = pIndexPos[NumIndex];
                unsigned char PatternPos = pPatternPos[NumPattern];

                if (IndexPos < PatternPos) {
                    ++NumIndex;
                } else if (IndexPos > PatternPos) {
                    ++NumPattern;
                } else {
                    ++NumIndex;
                    ++NumPattern;
                }
                ++kEstim;
            }
            kEstim += kErrIndex - NumIndex;
            kEstim += kErrPattern - NumPattern;

            //continue;

            if (kEstim <= mKerr) {
                ++mNumSearchInBucket;
                // Generate updated reduced-alphabet pattern as well as updated original pattern
                // Update includes deletion of certain characters as well as insertion of wildcards.
                const char *  pIndexChar = reinterpret_cast<const char *>(pIndexPos) + mKerr;


                //continue;

                const char* pOrigPat = mOrigPat.data();
                const char* pOrigRedPat  = s.data();

                unsigned MaxPossibSize = mOrigPat.size() + kErrIndex;
                char aPat[MaxPossibSize + 1];
                char aRedPat[MaxPossibSize + 1];

                unsigned NumIndex = 0;
                unsigned NumPattern = 0;

                unsigned PrevPatPos = 0;
                unsigned PatAccountForDel = 0;
                unsigned PatAccountForIns = 0;

                while (NumIndex < kErrIndex && NumPattern < kErrPattern ) {
                    unsigned char IndexPos = pIndexPos[NumIndex];
                    unsigned char PatternPos = pPatternPos[NumPattern];

                    if (IndexPos < PatternPos) {
                        // Extra letter in the index => 
                        // 1. insert a wildcard in the original pattern
                        // 2. insert a character from an indexed string

                        unsigned qty = PatAccountForIns + IndexPos - PrevPatPos;
                        
                        memcpy(aPat    + PrevPatPos, pOrigPat    + PrevPatPos - PatAccountForIns + PatAccountForDel, qty);
                        memcpy(aRedPat + PrevPatPos, pOrigRedPat + PrevPatPos - PatAccountForIns, qty);

                        PrevPatPos += qty;

                        aRedPat[PrevPatPos] = pIndexChar[NumIndex];
                        aPat[PrevPatPos]    = REDALPHA_WILDCARD_CHAR;

                        ++PrevPatPos;
                        ++PatAccountForIns;
                        ++NumIndex;
                    } else if (IndexPos > PatternPos) {
                        // Skip extra letters in the pattern
                        unsigned qty = PatAccountForIns + PatternPos - PrevPatPos;

                        memcpy(aPat    + PrevPatPos, pOrigPat    + PrevPatPos - PatAccountForIns + PatAccountForDel, qty);
                        memcpy(aRedPat + PrevPatPos, pOrigRedPat + PrevPatPos - PatAccountForIns, qty);

                        PrevPatPos += qty;
                        ++PatAccountForDel;
                        
                        ++NumPattern;
                    } else {
                        // Substitution: 
                        // 1. In the original pattern: will be handled through a Hamming-distance computation (insert a wildcard)
                        // 2. In the reduced pattern replace a character with a character from an indexed string
                        unsigned qty = PatAccountForIns + PatternPos - PrevPatPos;

                        memcpy(aPat    + PrevPatPos, pOrigPat    + PrevPatPos - PatAccountForIns + PatAccountForDel, qty);
                        memcpy(aRedPat + PrevPatPos, pOrigRedPat + PrevPatPos - PatAccountForIns, qty);

                        PrevPatPos += qty; 

                        aRedPat[PrevPatPos] = pIndexChar[NumIndex];
                        aPat[PrevPatPos]    = REDALPHA_WILDCARD_CHAR;

                        ++PrevPatPos;

                        ++PatAccountForDel;
                        ++PatAccountForIns;
                        
                        ++NumPattern;
                        ++NumIndex;
                    }
                }

                while (NumIndex < kErrIndex) {
                    unsigned char IndexPos = pIndexPos[NumIndex];
 
                    unsigned qty = PatAccountForIns + IndexPos - PrevPatPos;
                        
                    memcpy(aPat    + PrevPatPos, pOrigPat    + PrevPatPos - PatAccountForIns + PatAccountForDel, qty);
                    memcpy(aRedPat + PrevPatPos, pOrigRedPat + PrevPatPos - PatAccountForIns, qty);

                    PrevPatPos += qty;

                    aRedPat[PrevPatPos] = pIndexChar[NumIndex];
                    aPat[PrevPatPos]    = REDALPHA_WILDCARD_CHAR;

                    ++PrevPatPos;
                    ++PatAccountForIns;
                    ++NumIndex;
                }

                while (NumPattern < kErrPattern) {
                    unsigned char PatternPos = pPatternPos[NumPattern];
                    unsigned qty = PatAccountForIns + PatternPos - PrevPatPos;

                    memcpy(aPat    + PrevPatPos, pOrigPat    + PrevPatPos - PatAccountForIns + PatAccountForDel, qty);
                    memcpy(aRedPat + PrevPatPos, pOrigRedPat + PrevPatPos - PatAccountForIns, qty);

                    PrevPatPos += qty;
                    ++PatAccountForDel;
                        
                    ++NumPattern;
                }

                unsigned qty = mOrigPat.size() - PatAccountForDel + PatAccountForIns - PrevPatPos;

                memcpy(aPat    + PrevPatPos, pOrigPat    + PrevPatPos - PatAccountForIns + PatAccountForDel, qty);
                memcpy(aRedPat + PrevPatPos, pOrigRedPat + PrevPatPos - PatAccountForIns, qty);

                PrevPatPos += qty;

                if (PrevPatPos > MaxPossibSize) {
                    THROW_AND_PRINT_RUNTIME_ERROR("Internal error, PrevPatPos is too big: " << PrevPatPos);
                }

                aPat[PrevPatPos] = 0;
                aRedPat[PrevPatPos] = 0;

                unsigned tmp = mNumComp;

                mRedHash.SearchInBucket(aPat, aRedPat, mKerr, mpResList, mNumComp);

                if (tmp == mNumComp) {
                    return; // The indexed string is different than a string obtained from the pattern by deletions
                }
            }
        }
    }
};

void
CAlternGenRedAlphaMorFrankPerfHash::LoadPerfectHash(unsigned kerr)
{
    if (kerr > MAX_KERR) {
        THROW_AND_PRINT_RUNTIME_ERROR("Kerr is too big: " << kerr);
    }

    for (unsigned ShardId = 0; ShardId < SHARD_QTY; ++ShardId) {
        if (!maaHasString[kerr][ShardId]) {
            continue;
        }
        if (maapPerfectHash[kerr][ShardId]) {
            THROW_AND_PRINT_RUNTIME_ERROR("The hash is already loaded, kerr: " << kerr);
        }

        const string&   sHashName = PerfectHashName(kerr, ShardId);

        FILE* fdHash = fopen(sHashName.c_str(), "r");

        if (!fdHash) {
            THROW_AND_PRINT_RUNTIME_ERROR("Cannot open file: '" << sHashName << "'");
        }

        if (!(maapPerfectHash[kerr][ShardId] = cmph_load(fdHash))) {
            THROW_AND_PRINT_RUNTIME_ERROR("Cannot load hash for kerr = " << kerr);
        }
    }
}

void 
CAlternGenRedAlphaMorFrankPerfHash::PrintAddStat(SearchType type, int testNum, int kerr, ostream &out)
{
    out << " NumCheck:" << aNumCheck[kerr]/float(testNum) << " NumSearchInBucket: " << aNumSearchInBucket[kerr]/float(testNum) << endl; 
}

bool
CAlternGenRedAlphaMorFrankPerfHash::CreateInMemIndex(const char *inputFile,
                 int insertCount, int *dcnt, 
                 bool silent)
{

    bool bRetStat = false;

    try {
        // First create a regular reduced-alphabet index
        if (!mRedHash.CreateInMemIndex(inputFile, insertCount, dcnt, silent)) {
            return false;
        }

        // Then create an additional hash over reduced-alphabet strings in several steps
        for (unsigned k = 1; k <= MaxTestKerr<CAlternGenRedAlphaMorFrankPerfHash>(); ++k) {
            // Write unsorted data into several files/chunks/shards.
            
            const THashData& HashData = mRedHash.GetHashData();

            {
                ofstream         aTmpDict[SHARD_QTY];

                for (unsigned ShardId = 0; ShardId < SHARD_QTY; ++ShardId) {
                    string fn = TmpDictShardName(k, ShardId);

                    aTmpDict[ShardId].open(fn.c_str(), ios_base::out | ios_base::trunc);

                    if (!aTmpDict[ShardId]) {
                        Perror("Cannot open file: " + fn);
                        return false; 
                    }
                    aTmpDict[ShardId].exceptions(ios::failbit | ios::badbit | ios::eofbit);
                }

           
                cout << "K = " << k << " Preparing strings to generate perfect hash" << endl;

                CUniqueStrCollectorShard    ShardWriter(aTmpDict, maaHasString[k]);

                for (THashData::const_iterator it = HashData.begin(); it != HashData.end(); ++it) {
                    const string& s = it->first;
                    GenAllDeletions(s, 0, k, ShardWriter);
                }

                for (unsigned ShardId = 0; ShardId < SHARD_QTY; ++ShardId) {
                    aTmpDict[ShardId].close();

                    if (!maaHasString[k][ShardId]) continue;

                    string sTmpDictShardName = TmpDictShardName(k, ShardId);
                    string sTmpSortedDictShardName = TmpSortedDictShardName(k, ShardId);

                    stringstream CmdCAT;

                    // Sort the file and ignore all repeating values (there is no overlap among files)
                    CmdCAT << "/bin/bash -c \"sort -u '" << sTmpDictShardName << "' > '" << sTmpSortedDictShardName << "'\"" << endl;  

                    if (system(CmdCAT.str().c_str())) {
                        Perror("Cannot execute command: " + CmdCAT.str());
                        return false;
                    }

                    if (unlink(sTmpDictShardName.c_str())) {
                        Perror("Cannot delete file: '" + sTmpDictShardName + "'");
                        return false;
                    }

                    stringstream CmdCMPH;
                    string       sType = "chd"; // One of the most compact methods

                    CmdCMPH << "cmph -g -a " << sType << "  -m '" <<  PerfectHashName(k, ShardId) << "' '" << sTmpSortedDictShardName << "'"; 

                    // Create a perfect hash using an external utility
                    if (system(CmdCMPH.str().c_str())) {
                        Perror("Cannot execute command: " + CmdCMPH.str());
                        return false;
                    }

                    cout << "Created a perfect hash for shard # " << ShardId << endl;
                }

                LoadPerfectHash(k);
            }

            ofstream         aTmpChunk[SHARD_QTY];

            for (unsigned ShardId = 0; ShardId < SHARD_QTY; ++ShardId) {
                if (!maaHasString[k][ShardId]) continue;

                string fn = TmpChunkName(k, ShardId);

                aTmpChunk[ShardId].open(fn.c_str(), ios_base::out | ios_base::trunc);

                if (!aTmpChunk[ShardId]) {
                    Perror("Cannot open file: " + fn);
                    return false; 
                }
                aTmpChunk[ShardId].exceptions(ios::failbit | ios::badbit | ios::eofbit);
            }

            

            cout << "K = " << k << " The number of unique dictionary entries in the reduced alphabet: " << HashData.size() << endl;

            CInsertStr      Inserter(maapPerfectHash[k], aTmpChunk);

            // 2. Produce strings obtained by up to k deletions, save this data to the disk (in chunks)

            for (THashData::const_iterator it = HashData.begin(); it != HashData.end(); ++it) {
                const string& s = it->first;
#if 0
                cout << it->first << " : " << HashVal << endl;
#endif
                
                GenAllDeletions(s, 0, k, Inserter);
            }

            cout << "Closing temp chunks" << endl;

            for (unsigned ShardId = 0; ShardId < SHARD_QTY; ++ShardId) {
                if (!maaHasString[k][ShardId]) continue;
                aTmpChunk[ShardId].close();
            }


            long long TotalQty = 0;
            long long TotalTableDataSize = 0;
            long long TotalTableSize = 0;

            // 3. Read all the chunks, sort the data, and remove duplicates.
            cout << "Reading temp chunks." << endl;

            char                aData[2*k];
            size_t              ElemSize = sizeof(aData);

            for (unsigned ShardId = 0; ShardId < SHARD_QTY; ++ShardId) {
                if (!maaHasString[k][ShardId]) continue;

                size_t TableSize = maaTableSize[k][ShardId] = Inserter.maMaxHashVal[ShardId] + 1;
                cout << "K = " << k << " ShardId: " << ShardId << " Table size: " << TableSize << endl;


                vector<size_t>      vHashTable(TableSize + 1);
                size_t*             pTmpHashTable =  &vHashTable[0];

                maapCompactHashTable[k][ShardId] = new CCompactTable(TableSize + 1, SKIP_LOG);

                CCompactTable&      CompactHashTable = *maapCompactHashTable[k][ShardId]; 

                fill(pTmpHashTable, pTmpHashTable + TableSize + 1, 0);

                ifstream    inp;
                string      fn = TmpChunkName(k, ShardId);

                inp.open(fn.c_str());

                if (!inp) {
                    Perror("Cannot open file '" + fn + "' for reading");
                    return false;
                }

                inp.exceptions(ios::badbit);

                Unsigned2StringList hTmp;

                char c;
                inp.read(&c, 0);

                while (!inp.eof()) {
                    size_t          HashVal;

                    if (!inp.read((char*)&HashVal, sizeof HashVal) && inp.eof()) break;
                    inp.read((char*)aData, ElemSize);

                    hTmp[HashVal].push_back(string(aData, ElemSize));
                };

                inp.close();

                size_t ShardQty = 0;
        
                for (Unsigned2StringList::iterator it = hTmp.begin(); it != hTmp.end(); ++it) {
                    size_t HashVal = it->first;

                    if (HashVal >= TableSize) {
                        cerr << "Internal error: a HashVal = " << HashVal << " is >= than TableSize = " << TableSize << " Shard id: " << ShardId << " k = " << k << endl;
                        return false;
                    }
                    StringList& lst = it->second;

                    sort(lst.begin(), lst.end());
                    unsigned qty = unique(lst.begin(), lst.end()) - lst.begin();
                    lst.erase(lst.begin() + qty, lst.end());
                    ShardQty += qty;
                    pTmpHashTable[HashVal + 1] += qty;
                }


                TotalQty += ShardQty;

                size_t ShardSize = ShardQty * ElemSize;
                maapHashContent[k][ShardId] = new unsigned char [ShardSize];
                unsigned char* pHashContent = maapHashContent[k][ShardId];

                fill(pHashContent, pHashContent + ShardSize, 0);

                unsigned Skip = CompactHashTable.GetSkip();

                for (size_t i = 0; i <= TableSize; i += Skip) {
                    size_t iNext = min(TableSize + 1, i + Skip);
                    size_t aBlock[Skip];
                    unsigned BlockQty = iNext - i;
                    for (unsigned j = 0; j < BlockQty; ++j) {
                        aBlock[j] = pTmpHashTable[i + j];
                    }
                    CompactHashTable.AddBlock(aBlock, BlockQty);
                }

                if (CompactHashTable.GetVal(0)) {
                    THROW_AND_PRINT_RUNTIME_ERROR("Internal error: the first hash table element is non-zero and is equal to: " << CompactHashTable.GetVal(0));
                }

                size_t cumul = 0;

                // Verify the correctness
                for (size_t i = 0; i <= TableSize; i++) {
                    size_t val = CompactHashTable.GetVal(i);

                    cumul += pTmpHashTable[i];

                    if (cumul != val ) {
                        THROW_AND_PRINT_RUNTIME_ERROR("Internal error, table verification failed for element index: " << i
                                                  << " expected value: " << cumul * ElemSize << " obtained value: " << val
                                                  << " TableSize: " << TableSize);
                    }
                }

                TotalTableDataSize += CompactHashTable.GetDataSize();
                TotalTableSize += TableSize + 1;


                for (Unsigned2StringList::iterator it = hTmp.begin(); it != hTmp.end(); ++it) {
                    size_t HashVal = it->first;

                    if (HashVal >= TableSize) {
                        cerr << "Internal error: a HashVal = " << HashVal << " is > than TableSize = " << TableSize << endl;
                        return false;
                    }
                    StringList& lst = it->second;
                    unsigned qty = lst.size();

                    size_t RealSize = qty * ElemSize;
                    size_t ExpSize =  ElemSize * (CompactHashTable.GetVal(HashVal + 1) - CompactHashTable.GetVal(HashVal));

                    if (RealSize != ExpSize) {
                        THROW_AND_PRINT_RUNTIME_ERROR("Internal error, expected size: " << ExpSize << 
                                                      " Obtained: " << RealSize << 
                                                      " The number of elements: " << qty << 
                                                      " Element size: " << ElemSize << 
                                                      " HashVal: " << HashVal);
                    }
                    size_t Start = CompactHashTable.GetVal(HashVal) * ElemSize;

                    for (unsigned i = 0; i < qty; ++i) {
                        if (lst[i].size() != ElemSize) {
                            THROW_AND_PRINT_RUNTIME_ERROR("Internal error, element size: " << lst[i].size() << " is not equal to the expected value: " << ElemSize << " HashVal = " << HashVal << " k = " << k << " i = " << i);
                        }
                        memcpy(&pHashContent[Start], lst[i].data(), ElemSize);
                        Start += ElemSize;
                    }

                    if (Start != ElemSize * CompactHashTable.GetVal(HashVal + 1)) {
                        THROW_AND_PRINT_RUNTIME_ERROR("Internal error: Start = " << Start << " does not match the next offset: " << (ElemSize * CompactHashTable.GetVal(HashVal + 1)) << " HashVal: " << HashVal);
                    }
                }

                string TmpName = TmpChunkName(k, ShardId);

                if (unlink(TmpName.c_str())) {
                    Perror("Cannot delete file: '" + TmpName + "'");
                    return false;
                }
            }

            long long TotalSize = ((long long)TotalQty) * ElemSize;
            cout << "Total # of entries: " << TotalQty  << " kerr = " << k << endl
             << " Total # of bytes for the hash content: " << TotalSize  << endl
             << " Total # of bytes for the hash content + hash table: " << (TotalSize + TotalTableDataSize) << endl
             << " The average size of table entry: " << TotalTableDataSize / float(TotalTableSize) << endl;
        
        }

        for (unsigned k = 1; k <= MaxTestKerr<CAlternGenRedAlphaMorFrankPerfHash>(); ++k) {
            ClearPerfectHash(k);
        }
        for (unsigned k = 1; k <= MaxTestKerr<CAlternGenRedAlphaMorFrankPerfHash>(); ++k) {
            LoadPerfectHash(k);
        }

        bRetStat = true;
    } catch (const exception& e) {
        cerr << "Exception during indexing: " << e.what() << endl;
    }

    return bRetStat;
}


bool
CAlternGenRedAlphaMorFrankPerfHash::Insert(const char *str)
{
    return true;
}

bool
CAlternGenRedAlphaMorFrankPerfHash::Search(const char *str, int kerr, int *numComp,
       StringList * resList, bool silent)
{
    string sOrig(str);
    string sReduced(str);

    mRedHash.ToRedStr(sReduced);

    CSearchStr Searcher(mRedHash,
                        maapPerfectHash[kerr], 
                        maaTableSize[kerr], 
                        maapCompactHashTable[kerr], 
                        maapHashContent[kerr], 
                        sOrig, kerr, resList);

    GenAllDeletions(sReduced, 0, kerr, Searcher);

    if (numComp) *numComp += Searcher.mNumComp;
    aNumCheck[kerr] += Searcher.mNumCheck;
    aNumSearchInBucket[kerr] += Searcher.mNumSearchInBucket;

    return true;
}

bool
CAlternGenRedAlphaMorFrankPerfHash::DumpDB(const char *base_name)
{
    return true;
}

bool
CAlternGenRedAlphaMorFrankPerfHash::OpenDB(const char *base_name, bool create)
{
    return true;
}

bool
CAlternGenRedAlphaMorFrankPerfHash::CloseDB()
{
    return true;
}

#ifdef _DEBUG
void 
CAlternGenRedAlphaMorFrankPerfHash::Print()
{
}
#endif
