#include "index.h"
#include "myagrep.h"
#ifdef GEN_USING_DP
#include "gensupercond.h"
#else
#include "gensupercond_nfa.h"
#endif

static long long seekVal = 0;

struct CChecker {
    CChecker(const CAlternGenSuperCond& Indx, const char* pPat, unsigned kerr, bool bHasTransp, int* pNumComp, StringList* pResList, bool bSilent) :
        mIndx(Indx), mvDict(Indx.GetDict()), mpNumComp(pNumComp), mpResList(pResList), mbSilent(bSilent), 
        mKerr(kerr), mAgrep(pPat, kerr, bHasTransp) {
    }

    void operator()(const char* p, unsigned sz) {
        const CAlternGenSuperCond::vu*       pvIds = mIndx.GetIdList(string(p, sz), mKerr);

        if (!pvIds) {
            return;
        }

        const CAlternGenSuperCond::vu&      vIds(*pvIds);

        unsigned qty = vIds.size();

        if (mpNumComp) {
            *mpNumComp += qty; 
        }

        //return;

        for (unsigned i = 0; i < qty; ++i) {
            const string&   w = mvDict[vIds[i]];

            if (mAgrep.checkRTR(w.c_str(), w.c_str() + w.size())) {
                if (mpResList) {
                    mpResList->push_back(w);
                    if (!mbSilent) {
                        cout << w << endl;
                    }
                }
            }
        }
    };

    const CAlternGenSuperCond&      mIndx;
    const CAlternGenSuperCond::vs&  mvDict;  
    int*                            mpNumComp;
    StringList*                     mpResList;
    bool                            mbSilent;
    unsigned                        mKerr;
    myagrep                         mAgrep;

};

bool 
CAlternGenSuperCond::Search(const char *str, int kerr, int *numComp, StringList * resList, bool silent) {

    CChecker    chk(*this, str, kerr, mHasTransp, numComp, resList, silent);

    if (mHasTransp) {
        cerr << "Transpositions are not supported yet!" << endl;
        throw string("Wrong setup: no transpositions are supported!");
    }

#ifdef GEN_USING_DP
    GenSuperCondNG(chk, str, GetAlphabet(mAlphaType), kerr);
#else
    GenSuperCondNG_NFA(chk, str, GetAlphabet(mAlphaType), kerr);
#endif

    return true;
}

bool 
CAlternGenSuperCond::CreateInMemIndex(const char *inputFile,
                                         int insertCount, int *dcnt,
                                         bool silent)
{
    ifstream    InputFile(inputFile);
    string      s;

    for ((*dcnt) = 0; getline(InputFile, s); (*dcnt)++) {
        toUpper(s);
        mvDict.push_back(s);
    }

    for (unsigned kerr = 0; kerr <= MaxTestKerr<CAlternGenSuperCond>(); ++kerr) {
#if 1
        cout << "Generating for kerr = " << kerr << endl;
#endif
        for (unsigned WordId = 0; WordId < mvDict.size(); ++WordId) {
            string w = mvDict[WordId];

            // We don't have to generate all substrings, because 
            // a condensed NG substring cannot differ from an original string by more than 2kerr deletions
            for (unsigned k1 = 0; k1 <= 2 * kerr; ++k1) {
                string w1 = w;

                for (unsigned k2 = 0; k2 <= 2 * kerr - k1; ++k2) {
                    if (w1.empty()) {
                        break;
                    }
#if 0
                    cout << mvDict[WordId] << " -> '" << w1 << "' kerr = " << kerr << " k1 = " << k1  << " k2 = " << k2 << endl;
#endif
                    vu& vList = mahSubstrLists[kerr][w1];
                    if (!vList.size() || vList.back() != WordId) {
                        vList.push_back(WordId);
                    }
                    w1.erase(w1.size() - 1, 1);
                }
                if (w.empty()) {
                    break;
                }
                w.erase(0, 1);
            }

        }
    }

    return true;
}

typedef pair<size_t, CAlternGenSuperCond::hvu::const_iterator>   Pos2WordType;

struct CSorter {
    inline bool operator()(const Pos2WordType& This, const Pos2WordType& That) const {
        if (This.first != That.first) {
         return This.first < That.first;
        }

        return This.second->first < That.second->first;
    }
};

bool 
CAlternGenSuperCond::DumpDB(const char *base_name)
{
    if (!OpenDB(base_name, true /* do create */)) {
        return false;
    }

    unsigned            MainDictQty = mvDict.size();
    vector<unsigned>    vMainDictOffset(MainDictQty);

    mMainDictSize = 0;

    for (unsigned i = 0; i < mvDict.size(); ++i) {
        unsigned sz = mvDict[i].size() + 1;

        if (write(mFdDict, mvDict[i].c_str(),  sz) != sz) {
            Perror("Cannot write string to main dict");
            return false;
        }

        vMainDictOffset[i] = mMainDictSize;
        mMainDictSize += sz;
    }


    vector<Pos2WordType>                    avPos2Word[MAX_KERR_SUPERCOND + 1];

    vector<unsigned>                        vDictSize(MAX_KERR_SUPERCOND + 1);
    vector<char>                            vZeroBuf;

    int fillRatio = defaultHashRatio;

    GetEnv("FILL_RATIO", fillRatio);
    if (!fillRatio) fillRatio = defaultHashRatio;

    if (fillRatio < minHashRatio) fillRatio = minHashRatio;
    if (fillRatio > maxHashRatio) fillRatio = maxHashRatio;

    for (unsigned kerr = 0; kerr <= MaxTestKerr<CAlternGenSuperCond>(); ++kerr) {
        for (hvu::const_iterator    it = mahSubstrLists[kerr].begin(); it != mahSubstrLists[kerr].end(); ++it) {
            unsigned ElemSize = it->first.size() + 1 + sizeof(unsigned) + sizeof(unsigned) * it->second.size(); 

            vZeroBuf.resize(ElemSize);
            vDictSize[kerr] += ElemSize;
            if (write(mFdDict, &vZeroBuf[0], ElemSize) != ElemSize) {
                Perror(Format("Cannot write to dict file, kerr = %u, string = '%s', rec size = %u", kerr, it->first.c_str(), ElemSize));
                return false;
            }
            avPos2Word[kerr].push_back(make_pair(0U, it));
        }

        unsigned numWords = mahSubstrLists[kerr].size();
        size_t hashSize = from__stl_next_prime((size_t) 100LL * numWords / fillRatio); 

        if (!hashSize) hashSize = 1;

        size_t TmpUint(0);
        if (lseek(m_aFdHash[kerr], getPageAlignedValue((hashSize - 1)* sizeof(size_t)), SEEK_SET) < 0 ||
            write(m_aFdHash[kerr], &TmpUint, sizeof(TmpUint)) != sizeof(TmpUint)) {
            Perror(Format("Cannot lseek or write (hash), kerr = %u", kerr));
            return false;
        }
        char   TmpChar(0);
        if (vDictSize[kerr]) {
            if (lseek(m_aFdDictSubstr[kerr], getPageAlignedValue(vDictSize[kerr]) - 1, SEEK_SET) < 0 ||
                write(m_aFdDictSubstr[kerr], &TmpChar, sizeof(TmpChar)) != sizeof(TmpChar)) {
                Perror(Format("Cannot lseek or write (substr dict), kerr = %u", kerr));
                return false;
            }
        }
        CHashParams prm(fillRatio, hashSize, vDictSize[kerr]);

        if (write(m_aFdParams[kerr], &prm, sizeof prm) != sizeof prm) {
            Perror(Format("Cannot write params, kerr = %u", kerr));
            return false;
        }

    }

    if (!mapData()) {
        Perror("Cannot map data");
        return false;
    }

    for (unsigned kerr = 0; kerr <=  MaxTestKerr<CAlternGenSuperCond>(); ++kerr) {
        memset(m_apHash[kerr], 0, m_apParams[kerr]->hashSize * sizeof(size_t));
        memset(m_apDictSubstr[kerr], 0, m_apParams[kerr]->dictSize);

        unsigned n = 0;

        for (hvu::const_iterator    it = mahSubstrLists[kerr].begin(); it != mahSubstrLists[kerr].end(); ++it) {
            unsigned ElemSize = it->first.size() + 1 + sizeof(unsigned) + sizeof(unsigned) * it->second.size(); 

            size_t  buckNum;

            if (!insertHashDB(it->first, buckNum, *m_apParams[kerr], m_apHash[kerr], &seekVal)) {
                cerr << "Failed to insert string internally" << endl;
                CloseDB();
                return false;
            }
            avPos2Word[kerr][n++].first = buckNum;
        }
        // Finally store strings in the order they are in hash table
        // and fill hash positions
        CSorter sorter;

        sort(avPos2Word[kerr].begin(), avPos2Word[kerr].end(), sorter); // will sort by position

        size_t pos = 0;
        for (n = 0; n < avPos2Word[kerr].size(); ++n) {
            const hvu::const_iterator&  it = avPos2Word[kerr][n].second;
            const vector<unsigned>&     vInvList = it->second;

            unsigned ElemSize = it->first.size() + 1 + sizeof(unsigned) + sizeof(unsigned) * vInvList.size(); 
            size_t pos1 = pos;

            unsigned InvQty = vInvList.size();

            memcpy(&m_apDictSubstr[kerr][pos], it->first.c_str(), it->first.size() + 1);
            pos += it->first.size() + 1;
        
            memcpy(&m_apDictSubstr[kerr][pos], &InvQty, sizeof InvQty);
            pos += sizeof InvQty;

            unsigned* pList =  reinterpret_cast<unsigned *>(&m_apDictSubstr[kerr][pos]);

            for (unsigned IdNum = 0; IdNum < InvQty; ++IdNum) {
                *pList++ = vInvList[IdNum];
                pos += sizeof *pList;
            }

            m_apHash[kerr][avPos2Word[kerr][n].first] = pos1 + 1; // do not store zero value

            if (pos - pos1 != ElemSize) {
                THROW_AND_PRINT_RUNTIME_ERROR("Internal error: position diff mismatch");
            }
        }
    }

    return true;
}

bool
CAlternGenSuperCond::OpenHashKerr(const char* base_name, unsigned kerr, bool create)
{
    string    headerName = HeaderFileName(base_name, kerr);
    string    hashName   = HashFileName(base_name, kerr);
    string    dictName   = DictSubstrFileName(base_name, kerr);

    int flags = O_RDWR;
    int access = 0;

    if (create)  {
        flags = O_RDWR | O_TRUNC | O_CREAT;
        access = defaultAccess;
    }

    m_aFdParams[kerr] = open(headerName.c_str(), flags, access);

    if (m_aFdParams[kerr] < 0) {
        Perror(Format("Cannot open header file, kerr = %u", kerr));
        return false;
    }

    m_aFdHash[kerr] = open(hashName.c_str(), flags, access);

    if (m_aFdHash[kerr] < 0) {
        Perror(Format("Cannot open hash file, kerr = %u", kerr));
        return false;
    }
 
    m_aFdDictSubstr[kerr] = open(dictName.c_str(), flags, access);

    if (m_aFdDictSubstr[kerr] < 0) {
        Perror(Format("Cannot open substr dict file, kerr = %u", kerr));
        return false;
    }
      
    return true;
}

bool 
CAlternGenSuperCond::OpenDB(const char *base_name, bool create)
{
    if (mbOpened)
        CloseDB ();

    string    dictName   = DictFileName(base_name);

    int flags = O_RDWR;
    int access = 0;

    mpDict   = NULL;

    memset(m_apHash,   0, sizeof m_apHash);
    memset(m_apParams, 0, sizeof m_apParams);
    memset(m_apDictSubstr, 0, sizeof m_apDictSubstr);

    if (create)  {
        flags = O_RDWR | O_TRUNC | O_CREAT;
        access = defaultAccess;
    }

    for (unsigned kerr = 0; kerr <=  MaxTestKerr<CAlternGenSuperCond>(); ++kerr) {
        if (!OpenHashKerr(base_name, kerr, create)) {
            return false;
        }
    }

    mFdDict = open(dictName.c_str(), flags, access);

    if (mFdDict < 0) {
        Perror("Cannot main dict file");
        return false;
    }

    if (!create) {
        mbOpened = true;
        if (!mapData()) {
            return false;
        }
    }
    
    return mbOpened = true;
}

bool 
CAlternGenSuperCond::CloseDB()
{
    if (!unmapData()) {
        return false;
    }

    if (close(mFdDict)) {
        Perror("Cannot close main dict");
    }

    for (unsigned kerr = 0; kerr <=  MaxTestKerr<CAlternGenSuperCond>(); ++kerr) {
        if (close(m_aFdParams[kerr])) {
            Perror(Format("Cannot close params, %u", kerr));
        }
        if (close(m_aFdHash[kerr])) {
            Perror(Format("Cannot close hash, %u", kerr));
        }
        if (close(m_aFdDictSubstr[kerr])) {
            Perror(Format("Cannot close substr dict, %u", kerr));
        }
    }

    mbOpened = false;
    return true;
}

bool 
CAlternGenSuperCond::mapDataHashKerr(unsigned kerr)
{
    m_apParams[kerr] = reinterpret_cast<CHashParams *>(mmap(NULL, getPageAlignedValue(sizeof **m_apParams), 
                                                            PROT_WRITE | PROT_READ, MAP_SHARED, m_aFdParams[kerr], 0));

    if (!m_apParams[kerr] || m_apParams[kerr] == MAP_FAILED) {
        Perror(Format("Cannot map params, kerr = %u", kerr));
        return false;
    }

    m_apHash[kerr] = reinterpret_cast<size_t *>(mmap(NULL, getPageAlignedValue(m_apParams[kerr]->hashSize * sizeof(size_t)), 
                                                    PROT_WRITE | PROT_READ, MAP_SHARED, m_aFdHash[kerr], 0));

    if (!m_apHash[kerr] || m_apHash[kerr] == MAP_FAILED) {
        Perror(Format("Cannot map hash, kerr = %u", kerr));
        return false;
    }

    m_apDictSubstr[kerr] = reinterpret_cast<char *>(mmap(NULL, getPageAlignedValue(m_apParams[kerr]->dictSize), 
                                        PROT_WRITE | PROT_READ, MAP_SHARED, m_aFdDictSubstr[kerr], 0));

    if (!m_apDictSubstr[kerr] || m_apDictSubstr[kerr] == MAP_FAILED) {
        Perror(Format("Cannot map substr dict, kerr = %u", kerr));
        return false;
    }

    return true;
}

bool 
CAlternGenSuperCond::mapData() 
{
    if (!mbOpened) {
        cerr << "DB is not open, cannot map" << endl;
        return false;
    }

    mpDict = reinterpret_cast<char *>(mmap(NULL, getPageAlignedValue(mMainDictSize), 
                                        PROT_WRITE | PROT_READ, MAP_SHARED, mFdDict, 0));

    if (!mpDict || mpDict == MAP_FAILED) {
        Perror("Cannot map main dict");
        return false;
    }

    for (unsigned kerr = 0; kerr <=  MaxTestKerr<CAlternGenSuperCond>(); ++kerr) {
        if (!mapDataHashKerr(kerr)) {
            Perror(Format("Cannot map hash dict, %u", kerr));
            return false;
        }
    }

    return true;
}

bool 
CAlternGenSuperCond::unmapHashKerr(unsigned kerr)
{
    if (m_apParams[kerr]) {
        if (m_apHash[kerr] && munmap(m_apHash[kerr], getPageAlignedValue(m_apParams[kerr]->hashSize * sizeof(size_t)))) {
            Perror(Format("Cannot unmap hash, kerr %u", kerr));
            return false;
        }
                
        if (m_apDictSubstr[kerr] && munmap(m_apDictSubstr[kerr], getPageAlignedValue(m_apParams[kerr]->dictSize))) {
            Perror(Format("Cannot unmap dict, kerr %u", kerr));
            return false;
        }

        if (munmap(m_apParams[kerr], getPageAlignedValue(sizeof **m_apParams))) {
            Perror(Format("Cannot unmap params, kerr %u", kerr));
            return false;
        }
    }

    return true;
}

bool
CAlternGenSuperCond::unmapData() 
{
    for (unsigned kerr = 0; kerr <=  MaxTestKerr<CAlternGenSuperCond>(); ++kerr) {
        if (!unmapHashKerr(kerr)) {
            return false;
        }
    }

    if (mpDict && munmap(mpDict, getPageAlignedValue(mMainDictSize))) {
        Perror("Cannot unmap dict");
        return false;
    }

    return true;
}
