#ifndef ALTERN_SUPERCOND_INDEX_H
#define ALTERN_SUPERCOND_INDEX_H

#include "interface.h"
#include <fcntl.h>

#include <iostream>
#include <vector>
#include <string>

#include "langcase.h"
#include "alphabet.h"
#include "myexcept.h"
#include "../Common/hash.h"

using namespace std;

#define MAX_KERR_SUPERCOND  3

// Super-condensed neighborhood generation 

class CAlternGenSuperCond : public CInterface {
  public:
     CAlternGenSuperCond(bool hasTransp, EAlphabet AlphaType) : CInterface(hasTransp, AlphaType), mbOpened(false) {}
    ~CAlternGenSuperCond() {}

    bool CreateInMemIndex(const char *inputFile,
                             int insertCount, int *dcnt, 
                             bool silent = true);
    //=========================================================================
    bool Insert(const char *str) { 
        cerr << "Insert is not suppported!" << endl;
        return false;
    }
    bool Search(const char *str, int kerr, int *numComp, StringList * resList, bool silent = true);
    //=========================================================================
    bool SearchInDB(const char *str, int kerr, int *numComp,
                    StringList * resList, bool silent = true) {
        cerr << "SearchInDB is not supported" << endl;
        return false;
    }
    //=========================================================================
    bool DumpDB (const char *base_name);
    bool OpenDB (const char *base_name, bool create = false);
    bool CloseDB ();

    typedef vector<unsigned>        vu;
    typedef vector<string>          vs;
    const vs& GetDict() const { return mvDict; }
    const vu* GetIdList(const string& w, unsigned kerr) const {
        assert(kerr <= MAX_KERR_SUPERCOND);
        hvu::const_iterator it = mahSubstrLists[kerr].find(w);
        return (it == mahSubstrLists[kerr].end()) ? NULL : &it->second;
    }

    static string HeaderFileName(const string& BaseName, unsigned kerr) { return BaseName + Format("%u", kerr) + ".hdr";  }
    static string HashFileName(const string& BaseName, unsigned kerr)   { return BaseName + Format("%u", kerr) + ".hash"; }
    static string DictSubstrFileName(const string& BaseName, unsigned kerr)  { return BaseName + Format("%u", kerr) + ".substrdict"; }
    static string DictFileName(const string& BaseName)   { return BaseName + ".dict"; }
 
    typedef hash_map<string,vu>     hvu;
  protected:

    vs                              mvDict;
    hvu                             mahSubstrLists[MAX_KERR_SUPERCOND + 1];

    bool            mbOpened;
    int             m_aFdHash[MAX_KERR_SUPERCOND + 1];
    int             m_aFdParams[MAX_KERR_SUPERCOND + 1];
    int             m_aFdDictSubstr[MAX_KERR_SUPERCOND + 1];
    int             mFdDict;

    size_t*         m_apHash[MAX_KERR_SUPERCOND + 1]; 
    CHashParams*    m_apParams[MAX_KERR_SUPERCOND + 1];
    char*           m_apDictSubstr[MAX_KERR_SUPERCOND + 1];
    char*           mpDict;

    size_t          mMainDictSize;

    bool mapData();
    bool mapDataHashKerr(unsigned kerr);
    bool unmapData();
    bool OpenHashKerr(const char* base_name, unsigned kerr, bool create);
    bool mapHashKerr(unsigned kerr);
    bool unmapHashKerr(unsigned kerr);
};

template <>
inline int MaxTestKerr<CAlternGenSuperCond>() {
    int kerr = MAX_KERR_SUPERCOND;
    GetEnv(ENV_KERR, kerr);
    return min(kerr, MAX_KERR_SUPERCOND);
}

#endif
