#ifndef TRIEEX_H
#define TRIEEX_H

#include <vector>
#include <list>
#include <iostream>

#include "interface.h"

#define NUM_SHORT_NODE 2

using namespace std;

class CTrieExact {
public:
    static unsigned GetUsedMem();
    struct CTrieNodeBase {
        unsigned char          mSym;
        bool                   mIsFull;
    };
    const static char EOS = 1;
    CTrieExact(const char *pABC);
    ~CTrieExact();
    string Name() {  return "Trie"; }
    void Insert(string s) {
        s += EOS;
        InsertString(&m_pRoot, s, 0);
    }
    bool Check(const string& s) {
        unsigned    sz = s.size();

        CTrieNodeBase* pNode = m_pRoot;

        for (unsigned i = 0; i < sz; ++i) {
            CTrieNodeBase** ppTmp = GetPtrOfChildPtr(pNode, s[i]);
            //cout << pNode << " " << s[i] << " " << int(s[i]) << " " << ppTmp << endl;

            if (!ppTmp) return false;
            pNode = *ppTmp;
        }
        assert(pNode);

        CTrieNodeBase** ppTmp = GetPtrOfChildPtr(pNode, EOS);
        if (!ppTmp) return false;
        assert((*ppTmp)->mSym == EOS);
        return true;
    }
private:
    friend class CTrieNodeFull;
    // CTrieNodeShort aligned
    char           m_ABC[256];
    unsigned int   m_abcMap[256];
    unsigned int   m_abcSize;
    unsigned int   m_nodeSizeShort;
    unsigned int   m_nodeSizeFull;

    struct CTrieNodeShort : public CTrieNodeBase {
        unsigned char  mChildSym1; 
    #if NUM_SHORT_NODE > 1
        unsigned char  mChildSym2;
    #endif
    #if NUM_SHORT_NODE > 2
        unsigned char  mChildSym3;
    #endif
    #if NUM_SHORT_NODE > 3
        unsigned char  mChildSym4;
    #endif
    #if NUM_SHORT_NODE > 4
        unsigned char  mChildSym5; 
    #endif
    #if NUM_SHORT_NODE > 5
        unsigned char  mChildSym6;
    #endif
    #if NUM_SHORT_NODE > 6
        unsigned char  mChildSym7;
    #endif
    #if NUM_SHORT_NODE > 7
        unsigned char  mChildSym8;
    #endif
    #if NUM_SHORT_NODE > 8
        #error "NUM_SHORT_NODE > 4 is not supported!"
    #endif
        CTrieNodeBase*  mpChild1;
    #if NUM_SHORT_NODE > 1
        CTrieNodeBase*  mpChild2;
    #endif
    #if NUM_SHORT_NODE > 2
        CTrieNodeBase*  mpChild3; 
    #endif
    #if NUM_SHORT_NODE > 3
        CTrieNodeBase*  mpChild4;
    #endif
    #if NUM_SHORT_NODE > 4
        CTrieNodeBase*  mpChild5;
    #endif
    #if NUM_SHORT_NODE > 5
        CTrieNodeBase*  mpChild6;
    #endif
    #if NUM_SHORT_NODE > 6
        CTrieNodeBase*  mpChild7;
    #endif
    #if NUM_SHORT_NODE > 7
        CTrieNodeBase*  mpChild8;
    #endif
    };
    struct CTrieNodeFull : public CTrieNodeBase {
        CTrieNodeBase   *mpChildren[0]; // Memory will be allocated if needed
    };

    CTrieNodeBase **GetPtrOfChildPtr(CTrieNodeBase *pNode, unsigned char c) {
        if (!pNode->mIsFull) {
            CTrieNodeShort *p = (CTrieNodeShort *)pNode;
            if (p->mChildSym1 == c) return &p->mpChild1;
    #if NUM_SHORT_NODE > 1
            if (p->mChildSym2 == c) return &p->mpChild2;
    #endif
    #if NUM_SHORT_NODE > 2
            if (p->mChildSym3 == c) return &p->mpChild3;
    #endif
    #if NUM_SHORT_NODE > 3
            if (p->mChildSym4 == c) return &p->mpChild4;
    #endif
    #if NUM_SHORT_NODE > 4
            if (p->mChildSym5 == c) return &p->mpChild5;
    #endif
    #if NUM_SHORT_NODE > 5
            if (p->mChildSym6 == c) return &p->mpChild6;
    #endif
    #if NUM_SHORT_NODE > 6
            if (p->mChildSym7 == c) return &p->mpChild7;
    #endif
    #if NUM_SHORT_NODE > 7
            if (p->mChildSym8 == c) return &p->mpChild8;
    #endif
            return NULL;
        } else {
            CTrieNodeFull *p = (CTrieNodeFull *)pNode;
            int pos = GetPos(c);
            return (p->mpChildren[pos]) ? &(p->mpChildren[pos]):NULL;
        }
    }
    // might create new node if short is not enough
    CTrieNodeBase *insertChild(CTrieNodeBase *pNode, unsigned char c, CTrieNodeBase ***pppChild)
    {
        CTrieNodeShort *pNewNode = getNewNodeShort(c);

        CTrieNodeBase *pNodeToIns = pNode;

        if (!pNode->mIsFull) {
            CTrieNodeShort *pNodeShort = (CTrieNodeShort *)pNode;
            if (!pNodeShort->mpChild1) {
                *pppChild = &(pNodeShort->mpChild1);
                **pppChild = pNewNode;
                pNodeShort->mChildSym1 = c;
                return pNode;
            } 
    #if NUM_SHORT_NODE > 1
            else if (!pNodeShort->mpChild2) {
                *pppChild = &(pNodeShort->mpChild2);
                **pppChild = pNewNode;
                pNodeShort->mChildSym2 = c;
                return pNode;
            } 
    #endif
    #if NUM_SHORT_NODE > 2
            else if (!pNodeShort->mpChild3) {
                *pppChild = &(pNodeShort->mpChild3);
                **pppChild = pNewNode;
                pNodeShort->mChildSym3 = c;
                return pNode;
            } 
    #endif
    #if NUM_SHORT_NODE > 3
            else if (!pNodeShort->mpChild4) {
                *pppChild = &(pNodeShort->mpChild4);
                **pppChild = pNewNode;
                pNodeShort->mChildSym4 = c;
                return pNode;
            } 
     #endif
     #if NUM_SHORT_NODE > 4
            else if (!pNodeShort->mpChild5) {
                *pppChild = &(pNodeShort->mpChild5);
                **pppChild = pNewNode;
                pNodeShort->mChildSym5 = c;
                return pNode;
            } 
     #endif
     #if NUM_SHORT_NODE > 5
            else if (!pNodeShort->mpChild6) {
                *pppChild = &(pNodeShort->mpChild6);
                **pppChild = pNewNode;
                pNodeShort->mChildSym6 = c;
                return pNode;
            } 
     #endif
     #if NUM_SHORT_NODE > 6
            else if (!pNodeShort->mpChild7) {
                *pppChild = &(pNodeShort->mpChild7);
                **pppChild = pNewNode;
                pNodeShort->mChildSym7 = c;
                return pNode;
            } 
     #endif
     #if NUM_SHORT_NODE > 7
            else if (!pNodeShort->mpChild8) {
                *pppChild = &(pNodeShort->mpChild8);
                **pppChild = pNewNode;
                pNodeShort->mChildSym8 = c;
                return pNode;
            } 
     #endif
            else {
                // Not enough cells in short node, create full one
                pNodeToIns = createFull(pNodeShort); 
            }
        }
        CTrieNodeFull *p = (CTrieNodeFull *)pNodeToIns;
        int pos = GetPos(c); 
        *pppChild = &(p->mpChildren[pos]);
        p->mpChildren[pos] = pNewNode;
        return pNodeToIns;
    }
    CTrieNodeFull  *createFull(const CTrieNodeShort *pShort) {
        CTrieNodeFull  *pFull = getNewNodeFull(pShort->mSym);
        if (pShort->mpChild1) {
            int pos = GetPos(pShort->mChildSym1);
            pFull->mpChildren[pos] = pShort->mpChild1;
        }
    #if NUM_SHORT_NODE > 1
        if (pShort->mpChild2) {
            int pos = GetPos(pShort->mChildSym2);
            pFull->mpChildren[pos] = pShort->mpChild2;
        }
    #endif
    #if NUM_SHORT_NODE > 2
        if (pShort->mpChild3) {
            int pos = GetPos(pShort->mChildSym3);
            pFull->mpChildren[pos] = pShort->mpChild3;
        }
    #endif
    #if NUM_SHORT_NODE > 3
        if (pShort->mpChild4) {
            int pos = GetPos(pShort->mChildSym4);
            pFull->mpChildren[pos] = pShort->mpChild4;
        }
    #endif
    #if NUM_SHORT_NODE > 4
        if (pShort->mpChild5) {
            int pos = GetPos(pShort->mChildSym5);
            pFull->mpChildren[pos] = pShort->mpChild5;
        }
    #endif
    #if NUM_SHORT_NODE > 5
        if (pShort->mpChild6) {
            int pos = GetPos(pShort->mChildSym6);
            pFull->mpChildren[pos] = pShort->mpChild6;
        }
    #endif
    #if NUM_SHORT_NODE > 6
        if (pShort->mpChild7) {
            int pos = GetPos(pShort->mChildSym7);
            pFull->mpChildren[pos] = pShort->mpChild7;
        }
    #endif
    #if NUM_SHORT_NODE > 7
        if (pShort->mpChild8) {
            int pos = GetPos(pShort->mChildSym8);
            pFull->mpChildren[pos] = pShort->mpChild8;
        }
    #endif
        return pFull;
    }
    CTrieNodeBase *m_pRoot;

    list<CTrieNodeFull *>   m_freeNodesFull;
    list<CTrieNodeShort *>  m_freeNodesShort;
    list<void *>            m_memBlocks;

    CTrieNodeFull *getNewNodeFull(unsigned char c);
    CTrieNodeShort *getNewNodeShort(unsigned char c);

    void freeShortNode(CTrieNodeShort *p) {
        memset(p, 0, sizeof(CTrieNodeShort));
        m_freeNodesShort.insert(m_freeNodesShort.end(), p);
    }
    void allocateNodesFull(int N);
    void allocateNodesShort(int N);
    inline unsigned int GetPos(unsigned char c) {
        return m_abcMap[c];
    }
    inline void InsertString(CTrieNodeBase **ppCurrNode, const string & str, int start) {
        int sz = str.size();
        while (start < sz) {
            unsigned char c = str[start];
            CTrieNodeBase **ppTmp = GetPtrOfChildPtr(*ppCurrNode, c);
            if (!ppTmp) {
                CTrieNodeBase *pModNode = insertChild(*ppCurrNode, c, &ppTmp);
                assert(*ppTmp);
                
                if (pModNode != *ppCurrNode) {
                    freeShortNode((CTrieNodeShort *)*ppCurrNode); // free old node
                    *ppCurrNode = pModNode; // modify pointer in the parent node
                }
            }
            ppCurrNode = ppTmp;
            ++start;
        }
    }
};

#endif
