#ifndef ERRATA_TREE_H
#define ERRATA_TREE_H

#include "comptrie.h"
#include "addtrieinfo.h"
#include "nodealloc.h"
#include "addcombketsalloc.h"

#include <vector>

#define NODE_ALLOCATOR          CNodeAllocator
#define KET_PRIME_ALLOCATOR     CAddCombinedKetsAllocator

/*
 *
 *  THERE CAN BE ONLY ONE instance of CErrataTree at a time.
 * 
 */

class CErrataTree {
public:
    typedef CCompTrie<NODE_ALLOCATOR>           TTrie;
    typedef std::vector<CAddNodeInfo>           TNodeInfo;
    typedef std::vector<CAddCombinedKets*>      TCombinedKetsInfo; 
    typedef std::vector<std::string>            vs;

    void Create(const vs& StrList, int kerr, bool bHasTransp);
    size_t  GetTotalMemUsed() const { 
    /* 
     * This is for debug purposes only: it returns correct figures only if there is a single 
     * instance of CErrataTree
     */
        return NODE_ALLOCATOR::GetTotalMemUsed() +  KET_PRIME_ALLOCATOR::GetTotalMemUsed() +
                mNodeInfo.size() * sizeof(TNodeInfo) + 
                mCombinedKetsInfo.size() * sizeof(TCombinedKetsInfo);
    }
    void Search(const std::string& Pattern, int kerr, vs* pvResList) const {
        if (mTrie.Empty()) { return; }
        if (TTrie::IsLongString(Pattern)) {
            THROW_RUNTIME_ERROR("Pattern '" << Pattern << "' is too long");
        }
        char         aPrefix[MAX_SMALL_COUNTER + 1];
        SmallCounter aPatternStart[MAX_SMALL_COUNTER];
        SmallCounter aPrefixStart [MAX_SMALL_COUNTER];
        NodeIndex    aVisitedNodes[MAX_SMALL_COUNTER];

        char pPattern[Pattern.size() + 1];

        strncpy(pPattern, Pattern.c_str(), Pattern.size());

        return SearchErrataTreesStartWith(0, aVisitedNodes, 0, 
                                          mTrie.GetRoot(), 0,
                                          aPrefix, 0, aPrefixStart,
                                          pPattern, 0, aPatternStart, 
                                          kerr, pvResList);
    }
    void PrintErrataTrie(const TTrie &trie, int kerr) const;
    void PrintMainTrie() const { PrintErrataTrie(mTrie, mKerr); }
    
    virtual ~CErrataTree() {
    #if 0
        for (size_t i = 0; i < mCombinedKetsInfo.size(); ++i) {
            KET_PRIME_ALLOCATOR::FreeKets(mCombinedKetsInfo[i]);
        }
        NODE_ALLOCATOR::PurgeAllMem();
    #endif
    }
    bool DumpErrataTree(int FDTrieOffsets, int FDTrie,
                  int FDAddNodeInfo,
                  int FDCombKetOffsets, int FDCombKets);
protected:
    void CreateErrataTrees(TTrie trie, int kerr, bool bHasTransp);
    void CreateErrataTreesStartWith(SmallCounter Level, NodeIndex* pVisitedNodes, SmallCounter StartCentroidLevel, 
                                    NodeIndex StartIndex, int kerr, bool bHasTransp, bool bReuseExistingNodes);
    void SearchErrataTreesStartWith(SmallCounter Level, NodeIndex*   pVisitedNodes, SmallCounter StartCentroidLevel, 
                                    NodeIndex StartIndex, SmallCounter LabelStartPoint,
                                    char*       pPrefix,  SmallCounter PrefixStartPoint,  SmallCounter* pPrefixStart,
                                    char* pPattern, SmallCounter PatternStartPoint, SmallCounter* pPatternStart,
                                    int kerr, vs* pvResList) const;

    void DoKetPrimeForInterval(NodeIndex* pLeafQty, NodeIndex* pVisitedNodes, SmallCounter StartCentroidLevel, SmallCounter LevelQty, 
                                int kerr, bool bHasTransp, bool bReuseExistingNodes);
    void ResizeAddInfo() {
        size_t          OldNodeInfoSize = mNodeInfo.size();
        CAddNodeInfo    EmptyNodeInfo;

        mNodeInfo.resize(NODE_ALLOCATOR::GetMaxNodeId() + 1);

        for (size_t i = OldNodeInfoSize; i < mNodeInfo.size(); ++i) {
            mNodeInfo[i] = EmptyNodeInfo;
        }


        mCombinedKetsInfo.resize(NODE_ALLOCATOR::GetMaxNodeId() + 1);
    }

    void FillChildQty(NodeIndex StartIndex);
    void RecordCentroidPaths(NodeIndex StartIndex);
    void DumpTrie(int fd, NodeIndex StartIndex, off_t* pOffsets);

    TTrie               mTrie;
    int                 mKerr;
    TNodeInfo           mNodeInfo;
    TCombinedKetsInfo   mCombinedKetsInfo;
    bool                mHasTransp;
};


#endif
