#include "index.h"

CColeEtAlIndex::CColeEtAlIndex (bool hasTransp, EAlphabet AlphaType)
              : CInterface(hasTransp, AlphaType), 
              mbOpened(false)
{
}


CColeEtAlIndex::~CColeEtAlIndex ()
{
}

bool CColeEtAlIndex::Insert(const char *p)
{
    return true;
}

bool CColeEtAlIndex::Search(const char *p, int kerr, int *numComp,
                         StringList * resList, bool silent)
{
    mTrie.Search(p, kerr, resList);
    return true;
}

bool CColeEtAlIndex::OpenDB(const char *base_name, bool create)
{
    if (mbOpened)
        CloseDB ();
    /* 
     *  The following files are used during searching:
     *  1) Mapping info (node indices -> node offsets).
     *  2) Trie nodes themselves.  
     *  3) File for CAddNodeInfo
     *  4) 2 Files for CAddCombinedKets (which are variable-size records)
     *  4.1) One file encodes offsets 
     *  4.2) Another file stores the data itself.
     */

    errno = 0;

    if (!create) {
        cerr << "Readonly mode is not implemented!" << endl;
        return false;
    } else {
        mFDTrieOffsets = open(TrieOffsetsName(base_name).c_str(), O_RDWR | O_TRUNC | O_CREAT, defaultAccess);
        if (mFDTrieOffsets < 0) {
            Perror("Cannot open trie offset file");
            return false;
        }
        mFDTrie = open(TrieName(base_name).c_str(), O_RDWR | O_TRUNC | O_CREAT, defaultAccess);
        if (mFDTrie < 0) {
            Perror("Cannot open trie file");
            return false;
        }
        mFDAddNodeInfo = open(AddNodeInfoName(base_name).c_str(), O_RDWR | O_TRUNC | O_CREAT, defaultAccess);
        if (mFDAddNodeInfo < 0) {
            Perror("Cannot open add node info file");
            return false;
        }
        mFDCombKetOffsets = open(CombKetOffsetsName(base_name).c_str(), O_RDWR | O_TRUNC | O_CREAT, defaultAccess);
        if (mFDCombKetOffsets < 0) {
            Perror("Cannot open  comb ket offset file");
            return false;
        }
        mFDCombKets= open(CombKetName(base_name).c_str(), O_RDWR | O_TRUNC | O_CREAT, defaultAccess);
        if (mFDCombKets < 0) {
            Perror("Cannot open comb ket file");
            return false;
        }
    }

    return mbOpened = true;
}


bool CColeEtAlIndex::DumpDB(const char *base_name)
{
    if (mbOpened)
        CloseDB ();
    if (!OpenDB (base_name, true)) {
        return false;           // Create database
    }

    return mTrie.DumpErrataTree(mFDTrieOffsets, mFDTrie, mFDAddNodeInfo, mFDCombKetOffsets, mFDCombKets) && CloseDB();
}

bool CColeEtAlIndex::CloseDB()
{
    if (!mbOpened)
        return true;
        
    if (close(mFDTrieOffsets)) {
        Perror("Cannot close trie offset file");
        return false;
    }
    if (close(mFDTrie)) {
        Perror("Cannot close trie file");
        return false;
    }
    if (close(mFDAddNodeInfo)) {
        Perror("Cannot close add node info file");
        return false;
    }
    if (close(mFDCombKetOffsets)) {
        Perror("Cannot close comb ket offset file");
        return false;
    }
    if (close(mFDCombKets)) {
        Perror("Cannot close comb ket file");
        return false;
    }

    mbOpened = false;
    return true;
}

bool 
CColeEtAlIndex::CreateInMemIndex(const char *inputFile, int insertCount, int *dcnt, bool silent)
{
    *dcnt = 0;
    string s;

    ifstream f (inputFile);
    if (!f) {
        string err = string("Cannot open '") + inputFile + string("'");
        Perror(err.c_str());
        return false;
    }

    vector<string> lst;
    int lstCnt = 0;
    while (f >> s) {
        toUpper (s);

        lst.push_back (s);
        if (dcnt)
            (*dcnt)++;
        if (!silent && !((*dcnt) % 10000))
            cout << (*dcnt) << "\n";
    }

    int kerr = MaxTestKerr<CColeEtAlIndex>();
    cout << "Creating index for kerr = " << kerr << endl;

    mTrie.Create(lst, kerr, mHasTransp);

    
    return true;
}

