#ifndef NODE_HELPER_H
#define NODE_HELPER_H

#include <cstddef>
#include <sys/types.h>

// The data is stored in the following order:
// 1. Label length. 
// 2. The number of child nodes.
// 2. Variable length (non-zero terminated) label
// 4. If the number of child nodes is non-zero, a variable length array of child indices.

typedef u_int32_t                           NodeIndex;
typedef u_int8_t                            SmallCounter;

#define NILL_NODE_INDEX     u_int32_t(~NodeIndex(0))
#define MAX_NODE_INDEX      u_int32_t(NILL_NODE_INDEX-1)
#define MAX_SMALL_COUNTER   u_int8_t(~SmallCounter(0))

struct CNode {
    NodeIndex*        GetChildData() { 
        if (!mChildQty) { return NULL; }
        return reinterpret_cast<NodeIndex *>(&this->mpLabel[mLabelLen]);
    }
    const NodeIndex*        GetChildData() const { 
        if (!mChildQty) { return NULL; }
        return reinterpret_cast<const NodeIndex *>(&this->mpLabel[mLabelLen]);
    }
    static unsigned int  NodeSize(SmallCounter LabelLen, SmallCounter ChildQty) {
        return sizeof(CNode) + LabelLen + ChildQty * sizeof(NodeIndex);
    }

    SmallCounter      mLabelLen; 
    SmallCounter      mChildQty;
    char              mpLabel[0];

}
#ifdef __GNUC__ 
__attribute__ ((packed))
#endif
;

#endif

