#include <cassert>
#include <fstream>

#include "comptable.h"


using namespace std;

void 
CCompactTable::AddBlock(size_t aDiffVals[], unsigned DiffQty) {
    if (mCurrBlockId >= mBlockQty) {
        THROW_AND_PRINT_RUNTIME_ERROR("The table is already filled!");
    }
    if (DiffQty > mSkip) {
        THROW_AND_PRINT_RUNTIME_ERROR("The number of elements in the block cannot be larger than " << mSkip);
    }

    unsigned ByteQty = 1;

    for (unsigned i = 0; i < DiffQty; ++i) {
        size_t Val = aDiffVals[i];

        if (Val <= 0xFF) {
            ;
        } else if (Val <= 0xFFFF) { // > 0xFF
            ByteQty = max(ByteQty, 2U);
        } else if (Val <= 0xFFFFFF) { // > 0xFFFF
            ByteQty = max(ByteQty, 3U);
        } else { // > 0xFFFFFF
            ByteQty = max(ByteQty, sizeof(size_t));
        }

        mCurrVal += aDiffVals[i];
    }
    unsigned BlockSize = ByteQty * mSkip;

    if (mUsedDataSize + BlockSize >= mvData.size()) {
        mvData.resize((mUsedDataSize + BlockSize) * GROW_KOEFF);
    }

    for (unsigned i = 0; i < mSkip; ++i) {
        size_t Val = (i < DiffQty) ? aDiffVals[i] : 0;

        switch (ByteQty) {
            case 3: 
                    mvData[mUsedDataSize++] = (Val & 0xFF0000) >> 16;
            case 2: 
                    mvData[mUsedDataSize++] = (Val & 0xFF00) >> 8;
            case 1: 
                    mvData[mUsedDataSize++] = (Val & 0xFF);
                    break;
           
            default: 
                    memcpy(&mvData[mUsedDataSize], &Val, sizeof Val);
                    mUsedDataSize += sizeof Val;
        }
    }

    assert(mUsedDataSize <= mvData.size());

    ++mCurrBlockId;
    mpDataOffset[mCurrBlockId] = mUsedDataSize;
    mpSamples[mCurrBlockId] = mCurrVal;
};

void 
CCompactTable::Dump(ofstream& out)
{
    CTableHeader    Header(mBlockQty, mvData.size());

    out.exceptions(ios::failbit | ios::badbit | ios::eofbit);

    out.write(reinterpret_cast<const char*>(&Header), sizeof Header);
    out.write(reinterpret_cast<const char*>(mpDataOffset), (mBlockQty + 1) * sizeof *mpDataOffset);
    out.write(reinterpret_cast<const char*>(mpSamples), (mBlockQty + 1) * sizeof *mpSamples);
    out.write(reinterpret_cast<const char*>(&mvData[0]), mvData.size());
}
