#ifndef COMPACT_TABLE_HPP
#define COMPACT_TABLE_HPP

#include <algorithm>
#include <vector>
#include <fstream>

#include <string.h>

#include "myexcept.h"

#define GROW_KOEFF 1.2

class CCompactTable {
public:
    CCompactTable(size_t MinTableSize, unsigned SkipLog) : 
                    mSkipLog(SkipLog), mSkip(1<<mSkipLog), mSkipMask(mSkip - 1),
                    mUsedDataSize(0), mCurrVal(0), mCurrBlockId(0) {
        mBlockQty = (MinTableSize + mSkip - 1) >> mSkipLog;

        mpDataOffset = new size_t[mBlockQty + 1];
        mpSamples    = new size_t[mBlockQty + 1];

        std::fill(mpDataOffset, mpDataOffset + mBlockQty + 1, 0);
        std::fill(mpSamples, mpSamples + mBlockQty + 1, 0);

        mvData.resize(MinTableSize * GROW_KOEFF);
    }
    unsigned GetBlockQty() const { return mBlockQty; }
    unsigned GetSkip() const { return mSkip; }
    // If DiffQty < mSkip, the remaining values are filled with zeros.
    void AddBlock(size_t aDiffVals[], unsigned DiffQty);
    size_t GetVal(size_t Index) const {
        size_t BlockId = Index >> mSkipLog;
        unsigned BlockOffset = Index & mSkipMask;
        if (BlockId >= mBlockQty) {
            THROW_RUNTIME_ERROR("Index: " << Index << " is too big. " << 
                                "The largest possible value is: " << (mBlockQty * mSkip - 1)); 
        }
        unsigned ByteQty = (mpDataOffset[BlockId + 1] - mpDataOffset[BlockId]) >> mSkipLog;

        size_t  Start = mpDataOffset[BlockId];
        size_t  Val = mpSamples[BlockId];

        for (unsigned i = 0; i <= BlockOffset; ++i) {
            size_t  Diff = 0;

            switch (ByteQty) {
             case 3: Diff = mvData[Start++]; 
             case 2: Diff = (Diff << 8) + mvData[Start++];
             case 1: Diff = (Diff << 8) + mvData[Start++];
                     break;
             default:
                     Diff = *((size_t*)&mvData[Start]);
                     Start += sizeof(size_t);
            }
    
            Val += Diff;
        }

        return Val;
    }
    size_t GetDataSize() const { return mvData.size() + 2 * sizeof(size_t) * (mBlockQty + 1); }
    void Dump(std::ofstream& out);
private:
    unsigned                    mSkipLog;
    unsigned                    mSkip;
    unsigned                    mSkipMask;
    size_t*                     mpDataOffset;
    size_t*                     mpSamples;
    size_t                      mUsedDataSize;
    std::vector<unsigned char>  mvData;
    size_t                      mBlockQty;
    size_t                      mCurrVal;
    size_t                      mCurrBlockId;

    struct CTableHeader {
        size_t  mBlockQty;
        size_t  mDataSize;
        CTableHeader(size_t BlockQty, size_t DataSize) : mBlockQty(BlockQty), mDataSize(DataSize) {}
    };
};

#endif
