#ifndef FAST_STRLIST_HPP
#define FAST_STRLIST_HPP
#include <iostream>
#include <vector>
#include <string>
#include <cstring>
#include "myexcept.h"

using namespace std;

// You can fill this list quickly,
// with just a few allocations
class CFastStringList {
private:
    struct CBlockHead {
        size_t      mUsedSpace;
        unsigned    mBlockQty;
        size_t      maStringSize[0];
    };
public:
    CFastStringList(unsigned ReserveQty, unsigned BlockMaxQty, unsigned AvgStringLen, float ReserveKoeff = 1.2) : mBlockMaxQty(BlockMaxQty) {
        if (!AvgStringLen) { AvgStringLen = 10; }
        if (ReserveKoeff < 1) { ReserveKoeff = 1; }
        mBlockSize = sizeof(CBlockHead) + 
                     (sizeof(size_t) +
                     (AvgStringLen * ReserveKoeff + 1 /* zero char */)) * mBlockMaxQty;
        AddNewBlock();
        mvBlocks.reserve((ReserveQty + BlockMaxQty - 1) / ReserveQty);
    }
    ~CFastStringList() {
        for (unsigned i = 0; i < mvBlocks.size(); ++i) {
            delete [] mvBlocks[i];
        }
    }
    // pStr does not have to be 0-terminated
    void AddString(const char *pStr, unsigned len) {
        if (mpCurrBlockHead->mUsedSpace + len + 1 > mBlockSize ||
            mpCurrBlockHead->mBlockQty >= mBlockMaxQty) {
            AddNewBlock();
            if (mpCurrBlockHead->mUsedSpace + len + 1 > mBlockSize) {
                THROW_AND_PRINT_RUNTIME_ERROR("Cannot add string, because it's larger than a single block!");
            }
        }
        memcpy(mpCurrBlockData, pStr, len);
        mpCurrBlockHead->mUsedSpace += len;
        mpCurrBlockData += len;
        *mpCurrBlockData++ = 0;
        *mpStrLenghts++ = len;
        mpCurrBlockHead->mBlockQty++;
        mpCurrBlockHead->mUsedSpace++;
    }

    void ToStringList(vector<string>& vRes) {
        vRes.clear();

        for (iterator it(*this); !it.Finished(); it.Next()) {
            vRes.push_back(string(it.Data(), it.Size()));
        }
    }

    class iterator {
    public:
        iterator(CFastStringList& StrList) : 
                        mStrList(StrList), mvBlocks(mStrList.mvBlocks), mpCurrBlockHead(NULL) {
            First();
        }
        bool Next() {
            if (mpCurrBlockData < mpCurrBlockDataEnd) {
                mpCurrBlockData += *mpStrLenghts + 1;
                if (mpCurrBlockData < mpCurrBlockDataEnd) {
                    ++mpStrLenghts;
                    return true;
                }
            }
            ++mBlockId;
            return GoToNonEmpty();
        }
        bool First() {
            mBlockId = 0;
            return GoToNonEmpty();
        }
        bool Finished() const {
            return mpCurrBlockHead == NULL;
        }
        char*   Data() const { return mpCurrBlockData; }
        size_t  Size() const { return *mpStrLenghts; }
    private:
        CFastStringList&                    mStrList;
        vector<char*>&                      mvBlocks;
        const CFastStringList::CBlockHead*  mpCurrBlockHead;
        char*                         mpCurrBlockData;  
        char*                         mpCurrBlockDataEnd;  
        const size_t*                 mpStrLenghts;
        unsigned                      mBlockId;

        void InitBlock() {
            if (mBlockId >= mStrList.mvBlocks.size()) {
                THROW_AND_PRINT_RUNTIME_ERROR("Internal error: mBlockId is too large");
            }
            mStrList.InitHeaderExisting(mStrList.mvBlocks[mBlockId], mpCurrBlockHead, 
                                        mpCurrBlockData, mpCurrBlockDataEnd, mpStrLenghts);
        }

        bool GoToNonEmpty() {
            for (; mBlockId < mvBlocks.size(); ++mBlockId) {
                InitBlock();
                if (!CurrBlockEmpty()) {
                    return true;
                }
            }
            // If we reach this point, no non-empty block is found

            mpCurrBlockHead = NULL;
            mpCurrBlockDataEnd = mpCurrBlockData = NULL;
            return false;
        }

        bool CurrBlockEmpty() const {
            return mpCurrBlockData >= mpCurrBlockDataEnd;
        }
    };
private:

    unsigned        mBlockSize;
    unsigned        mBlockMaxQty;
    CBlockHead*     mpCurrBlockHead;
    char*           mpCurrBlockData;       
    size_t*         mpStrLenghts;
    
    vector<char*>   mvBlocks;

    void AddNewBlock() {
        mvBlocks.push_back(new char[mBlockSize]);
        InitHeaderNew(mvBlocks.back(), mpCurrBlockHead, mpCurrBlockData, mpStrLenghts);
    }

    size_t HeaderSize(size_t BlockMaxQty) const {
        return sizeof(CBlockHead) + BlockMaxQty * sizeof(size_t);
    }

    void InitHeaderNew(char *pData, CBlockHead*& pHeader, char*& pBlockData, size_t*& pStrLenghts) const {
        pHeader = reinterpret_cast<CBlockHead*>(pData);
        pHeader->mUsedSpace = HeaderSize(mBlockMaxQty);
        pHeader->mBlockQty = 0;
        pStrLenghts = reinterpret_cast<size_t*>(pData + sizeof(CBlockHead));
        pBlockData =  pData + mpCurrBlockHead->mUsedSpace;
    }

    void InitHeaderExisting(char *pData, const CBlockHead*& pHeader, 
                            char*& pBlockData, char*& pBlockDataEnd, const size_t*& pStrLenghts) const { 
        pHeader = reinterpret_cast<const CBlockHead*>(pData);
        pBlockData =  pData +  HeaderSize(mBlockMaxQty);
        pBlockDataEnd = pData + pHeader->mUsedSpace;
        pStrLenghts = reinterpret_cast<const size_t*>(pData + sizeof(CBlockHead));
    }

    friend class iterator;
};

#endif

