#include <fstream>
#include <algorithm>
#include <vector>

#include "perf_hash.h"
#include "myexcept.h"
#include "utils.h"

using namespace std;

bool
GeneratePerfectHash(const string& sInputFile, const string& sOutputFile, CMPH_ALGO eHashAlgo)
{

    FILE* fdHash = fopen(sOutputFile.c_str(), "w");

    if (!fdHash) {
        Perror("Cannot open file: '" + sOutputFile + "'");
        return false;
    }

    ifstream    inp(sInputFile.c_str());

    if (!inp) {
        Perror("Cannot open file: '" +sInputFile + "'");
        return false;
    }

    inp.exceptions(ios::badbit);

    string w;

    vector<string>        vTmpDict;
    vector<const char*>   vpStrPtr;

    cout << "Reading strings from the file: '" << sInputFile << "'" << endl;

    while (getline(inp, w)) {
        vTmpDict.push_back(w);
    }

    cout << "Sorting strings from the file: '" << sInputFile << "'" << endl;

    sort(vTmpDict.begin(), vTmpDict.end());
    vector<string>::const_iterator itEnd = unique(vTmpDict.begin(), vTmpDict.end());

    size_t Qty = itEnd - vTmpDict.begin();

    cout << Qty << " unique strings." << endl;

    vpStrPtr.resize(Qty);

    for (size_t i = 0; i < Qty; ++i) {
        vpStrPtr[i] = vTmpDict[i].c_str();
    }

    cmph_io_adapter_t*  pSource = cmph_io_vector_adapter((char **)&vpStrPtr[0], Qty);
    cmph_config_t*      pConfig = cmph_config_new(pSource);

    if (!pConfig) {
        cerr << "Cannot create a minimal perfect hash config!" << endl;
        return false;
    }

    cmph_config_set_algo(pConfig, eHashAlgo);
    cmph_config_set_mphf_fd(pConfig, fdHash);

    cout << "Creating a perfect hash for strings from the file: '" << sInputFile << "'" << endl;

    cmph_t* pHash = cmph_new(pConfig);

    if (!pHash) {
        cerr << "Cannot create a minimal perfect hash object!" << endl;
        return false;
    }

    cout << "Hash created!" << endl;

    if (!cmph_dump(pHash, fdHash)) {
        cerr << "Failed to dump the hash!";
        return false;
    }

    cmph_destroy(pHash);       
    cmph_config_destroy(pConfig);
    cmph_io_vector_adapter_destroy(pSource);
    fclose(fdHash);

    return true;
}
