#ifndef _INT_CODER
#define _INT_CODER
//
//  Copyright 2000, Leonid Boitsov
//



//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License as
//  published by the Free Software Foundation; either version 2 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.



//
// All procedures (except Golomb codes) are for the positive integers only
//

//
// For the encoding I use macros because the inline function
// that I wrote
// was not actually inlined, so that the coding worked almost
// three times slower!!
//

//
// Described in the article of Hugh E. Williams and Justin Zobel
// "Compressing Integers for Fast File Access" (1998)
//

// The coder writes directly to the stream
#include "search.h"
#include "bit_macro.h"

// Bellow goes Elias Gamma coding/decoding
// Note that GammaEncode codes only non-zero integers!
// Name is the name of the output buffer
// GammaEncode after writing [log_2(v)] 
// zeroes following by 1 
// writes to the bitstream digits of v in reveresed order
// less the most significant digit,
// that is starting from the the least significant bit
#define  GammaEncode(v,Name)					\
{register uint32 vn   = v>>1;					\
register int   logv =0;						\
for(;vn;logv++,vn>>=1) {PutBit(0,Name);}			\
 PutBit(1,Name);						\
vn = (v);							\
for(register int i=logv;i;i--,vn>>=1)				\
{PutBit(vn & 0x1,Name);}}		

#define GammaDecode(v,Name)					\
{register    int   b=0;						\
register uint32 n=0;						\
for(;!b;)							\
{ GetBit(b,Name);n+=(b) ? 0:1;}					\
register uint32 vn=1;						\
for((v)=0;n;n--,vn<<=1)						\
{								\
 GetBit(b,Name);						\
 if (b) (v)|=vn;						\
};								\
(v)|=vn;} /* The most significant bit is not explicitly coded! */


// Below goes the Elias Delta Coding/Decoding
// First codes [log_2(v)]+1 using GammaEncode
// Then stores digits of v in reversed order
// less the most significant digit starting from the
// least significant bit
#define DeltaEncode(v,Name)					\
{register uint32 vn_   = v>>1;					\
register int   logv_ =1;					\
for(;vn_;logv_++,vn_>>=1);		 			\
/* Got [log_2(v)]+1 !             */				\
/* Now store it using GammaEncode */                            \
{GammaEncode(logv_,Name);}					\
/* Now do as with GammaEncode method */				\
vn_ = (v);							\
for(register int i=logv_-1;i;i--,vn_>>=1)			\
{PutBit(vn_ & 0x1,Name);}}

#define DeltaDecode(v,Name)					\
{register     int   b_=0;					\
register uint32 n_=0;						\
{GammaDecode(n_,Name);}						\
n_--;								\
register uint32 vn_=1;						\
for((v)=0;n_;n_--,vn_<<=1)					\
{								\
 GetBit(b_,Name);						\
 if (b_) (v)|=vn_;						\
};								\
(v)|=vn_;}/* The most significant bit is not implicitly coded! */

#define DeclareGolomb(GName)					\
uint32 GName##k  ;						\
uint32 GName##ln ;						\
uint32 GName##deg ;						\
uint32 GName##d;

#define GolombKoeff(_mx,_num)  int(0.6 * (_mx + _num -1)/ double(_num) )

#define InitGolomb(param,GName)					\
GName##k  =param;						\
GName##ln =0;							\
for(register int i=((GName##k)>>1);i;i>>=1,GName##ln++);	\
GName##deg = (1<< (GName##ln+1));				\
GName##d = (GName##k) - GName##deg;

#define GolombEncode(v,Name,GName)				\
{uint32 q = ((v)/GName##k);					\
uint32 r = ((v)%GName##k);					\
for(register int32 vn=q;vn;vn--) {PutBit(0,Name);}		\
 PutBit(1,Name);						\
for(register int i=0;i< GName##ln;i++,r>>=1)			\
{PutBit(r & 0x01,Name);}					\
/* Choosing whether use ln or ln+1 bits to encode the remainder r */ \
/* If d == 0 then ln bits are choosen automatically		  */ \
if ( r % GName##deg < GName##d)					\
 PutBit(r & 0x01,Name);}

#define GolombDecode(v,Name,GName)				\
{register     int   b=0;						\
register uint32 n=0;						\
for(;!b;)							\
{ GetBit(b,Name);n+=(b) ? 0:1;}					\
register uint32 r=0;						\
register uint32 vn=1;						\
for(register int i=1;i<=GName##ln;i++,vn<<=1)			\
{								\
 GetBit(b,Name);						\
 if (b) r|=vn;							\
}								\
if (r<GName##d) {GetBit(b,Name);if (b) r|=vn;}			\
(v) = r + GName##k * n;}


// This is a 3-bit coding method, used by Andrew Aksyonoff
#define bit3Encode(v,Name)					\
{								\
 for(register int t=v;t;)					\
 {								\
  PutBit(t&1,Name);PutBit(t&2,Name);PutBit(t&4,Name);		\
  t>>=3;							\
  PutBit(t,Name);						\
 }								\
}					

#define varByteEncode(v,Name)					\
{								\
 for(register int t=v;t;)					\
 {								\
  register int b=t&127; 					\
  t>>=7;							\
  if (t) b|=128;PutByte(b,Name);				\
 }								\
}								

#endif
