#include <iostream>
#include <fstream>
#include <iterator>
#include <algorithm>
#include <stdlib.h>

#include "interface.h"
#include "gensupercond.h"
#include "hashport.h"

using namespace std;
void usage(const char *prog)
{
    cerr << "Usage: " << prog << " substr? transp? kerr file\n";
}

int main(int argc, char *argv[])
{
    if (argc < 5) {
        usage(argv[0]);
        return 1;
    }

    string              fn = argv[4];
    int                 kerr = atol(argv[3]);
    bool                hasTransp = atol(argv[2]);
    bool                substrCheck = atol(argv[1]);
    ifstream            inpFile(fn.c_str());

    string              s;
    GetEnv(ENV_ABC, s);

    EAlphabet           AlphaType = GetAlphabetType(s);

    if (AlphaType == eAlphaUnknown) {
        cerr << "Unknown alphabet: '" << s << "' check env var " << ENV_ABC << endl;
        return 1;
    }

    if (!inpFile) {
        string err = "Cannot open '" + fn + "'";
        perror(err.c_str());
        return -1;
    }



    //
    // Avg
    // S(n) = (x1 + x2 + ... + xn) / n  = (s(n-1)*(n-1) + xn)/n
    //
    double avgRatio = 0;
    string str;
    int    n = 0;

    while (inpFile >> str) {
        const char          *pStr = str.c_str();
        int                 len   = str.size();

        cout << "Generating neighborhood for '" << pStr << "' k = " << kerr << " String # " << n << endl;
        StringList      lst;
        StringHashSet   dict;

        size_t FullNgSize = GenCondNG(substrCheck ? eSuperCondensed: eCondensed, GetAlphabet(AlphaType), pStr, kerr, hasTransp, dict);

        cout << "Sizes: " << FullNgSize << " " << dict.size() << endl;
        n++;
        double avg = double(dict.size()) / FullNgSize;
        avgRatio = (avgRatio * (n - 1) + avg) / n;
    }

    cout << "Substr check? " << substrCheck << endl;
    cout << "kerr " << kerr << " has transp? " << hasTransp << endl;
    cout << "Avg ratio: " << avgRatio << "\n";
    return 0;
}
