#include "index.h"
#include "base/index.h"

int * search_intern(Index S, Obj, Tdist r, bool show, 
		int limit, int * result, int * result_count, int * sum_dist);

vector<char*>  vpDictList;

char* getDB (int nr) {
    if (nr <= 0) {
        THROW_AND_PRINT_RUNTIME_ERROR("nr should be > 0");
    }
    return vpDictList.at(nr - 1);
}

int openDB(char *) {
    return vpDictList.size();
};

void printobj (int obj) {
    THROW_AND_PRINT_RUNTIME_ERROR("printobj Not supported!");
}


CFastSSWrapper::CFastSSWrapper(bool hasTransp, EAlphabet AlphaType) : 
                                                        CInterface(hasTransp, AlphaType),
                                                        mpIndex(NULL)
{
}

CFastSSWrapper::~CFastSSWrapper() 
{
    // For some reason it crashes
    //if (mpIndex) { freeIndex(mpIndex, false); free(mpIndex); } 
}


bool 
CFastSSWrapper::CreateInMemIndex(const char *inputFile,
                         int insertCount, int *dcnt, 
                         bool silent)
{
    vpDictList.clear();

    ifstream f (inputFile);
    if (!f) {
        string err = string("Cannot open '") + inputFile + string("'");
        Perror(err.c_str());
        return false;
    }

    string s;

    while (getline(f, s)) {
        toUpper (s);

        vpDictList.push_back(strdup(s.c_str())); // Will be freed by a hash table
    }

    stringstream err;

    err << MaxTestKerr<CFastSSWrapper>();

    string sErrVal = err.str();
    char   aErrVal[128];

    strncpy(aErrVal, sErrVal.c_str(), sizeof(aErrVal));

    char*   pArgv  = aErrVal;
    char**  ppArgv = &pArgv;

    int argc = 1;
    mpIndex = build("" /* it's ignored */, vpDictList.size(), &argc, &ppArgv);

    if (dcnt) *dcnt = vpDictList.size();

    return true;
}

bool
CFastSSWrapper::Insert(const char *str) {
    cerr << "Not implemented!" << endl;
    return false;
}

bool
CFastSSWrapper::Search(const char *str, int kerr, int *numComp,
             StringList * resList, bool silent)
{

	int* pResult = reinterpret_cast<int*>(malloc(sizeof(int)));
	int result_count = 0;
	int sum_dist = 0;

	pResult = search_intern(mpIndex, (Obj)str, kerr, false, -1, pResult, &result_count, &sum_dist);

    if (resList) {
        for (unsigned i = 0; i < result_count; ++i) {
            resList->push_back(vpDictList[pResult[i] - 1]);    
        }
    }

	free(pResult);

    return true;
}

bool
CFastSSWrapper::DumpDB (const char *base_name)
{
    unsigned    len = strlen(base_name);
    char        aBaseName[len + 1];

    strcpy(aBaseName, base_name);
    saveIndex(mpIndex, aBaseName);
    return true;
}

bool
CFastSSWrapper::OpenDB (const char *base_name, bool create)
{
    return true;
}

bool
CFastSSWrapper::CloseDB ()
{
    return true;
}
