#include "storeobj.h"

void
 CFreqVectorList::Search (CFreqVectorMap & map, const std::string s,
                          int MaxErr,
                          const myagrep &a, 
                          int *numComp, StringList * resList,
                          bool hasTransp, bool silent )
{
    CFreqVector v (&map, 0, map.Size () - 1);
    v.InitFromString (s.c_str ());
    Search (v, 0, 0, s, MaxErr, a, numComp, resList, hasTransp, silent);
}

int CFreqVectorList::SearchInFile (int fd, CFreqVectorMap & map,
                                   const std::string s, int MaxErr,
                                   const myagrep &a,
                                   int *numComp, StringList * resList,
                                   bool hasTransp, bool silent)
{
    CFreqVector v (&map, 0, map.Size () - 1);
    v.InitFromString (s.c_str ());
    return SearchInFile (fd, 0, map, v, 0, 0, s, MaxErr, a, numComp,
                         resList, hasTransp, silent);
}


int CFreqVectorList::SearchInFile (int fd, off_t offset,
                                   CFreqVectorMap & map,
                                   const CFreqVector & v, int UpperPosDist,
                                   int UpperNegDist, const std::string s,
                                   int MaxErr, const myagrep & a, int *numComp,
                                   StringList * resList, bool hasTransp,
                                   bool silent)
{
    int buflen;
    if (lseek (fd, offset, SEEK_SET) < 0
        || read (fd, &buflen, sizeof (buflen)) < sizeof (buflen))
        return -1;
    int delta = 0;
    if (offset)
        delta = (offset % getpagesize ());
    void *buf =
        mmap (NULL, buflen + delta, PROT_READ, MAP_SHARED, fd,
              offset - delta);
    if (!buf || buf == MAP_FAILED)
        return -1;
    char *sbuf = (char *) buf + sizeof (buflen) + delta;
    int num = *((int *) (void *) sbuf);
    sbuf += sizeof (num);

    int pd = 0, nd = 0, plen = 1;
    // Now read node vectors
    for (int i = 0; i < num; i++) {
        // Initializing vector from sbuf
        int StartIndex = *((u_int8_t *) (void *) sbuf);
        sbuf++;
        int EndIndex = *((u_int8_t *) (void *) sbuf);
        sbuf++;
        off_t sub_off = *((off_t *) (void *) sbuf);
        sbuf += sizeof (off_t);
        CFreqVector vn (&map, StartIndex, EndIndex,
                        (u_int8_t *) (void *) sbuf);
        sbuf += vn.Size ();
        plen = Diff (vn, v, &pd, &nd);
        int _UpperPosDist = UpperPosDist + pd;
        int _UpperNegDist = UpperNegDist + nd;
        if (_UpperPosDist <= MaxErr && _UpperNegDist <= MaxErr) {
            // If EndIndex equals to v.EndIndex() then subtree is of type CStringHeap
            int res;
            if (v.EndIndex () > vn.EndIndex ()) {
                res =
                    SearchInFile (fd, sub_off, map,
                                  v.SubVector (v.StartIndex () +
                                               vn.Size (), v.EndIndex ()),
                                  _UpperPosDist, _UpperNegDist, s, MaxErr,
                                  a, numComp, resList, hasTransp, silent);
            } else
                res =
                    CStringHeap::SearchInFile (fd, sub_off, s, MaxErr, a,
                                               numComp, resList, hasTransp, silent);
            if (res < 0) {
                munmap (buf, buflen + delta);
                return -1;
            }
        }
    }
    if (munmap (buf, buflen + delta) < 0)
        return -1;
    return 1;
}

int CFreqVectorList::GetVectorCount() 
{
    int res = 0;

    for (int i = 0; i < num; ++i)
        res += plist[i]->GetVectorCount();

    return res;
}

void CFreqVectorList::Search (const CFreqVector & v, int UpperPosDist,
                              int UpperNegDist, const std::string s,
                              int MaxErr, 
                              const myagrep & a, int *numComp,
                              StringList * resList, bool hasTransp, 
                              bool silent)
{
// cout<<"Comparing with the vector:";v.Print();
    if (UpperPosDist < MaxErr || UpperNegDist < MaxErr) {
        // In this case inexact subvector matches are still possible
        int pd = 0, nd = 0, plen = 1;
        for (int i = 0; i < num; i++) {
            const CFreqVector & vn = vlist[i];
            plen = Diff (vn, v, &pd, &nd);
            int _UpperPosDist = UpperPosDist + pd;
            int _UpperNegDist = UpperNegDist + nd;
            if (_UpperPosDist <= MaxErr && _UpperNegDist <= MaxErr)
                plist[i]->Search (v.
                                  SubVector (v.StartIndex () + vn.Size (),
                                             v.EndIndex ()), _UpperPosDist,
                                  _UpperNegDist, s, MaxErr, a, numComp,
                                  resList, hasTransp, silent);
        }
    } else {
        //cout<<"Exact search!\n";v.Print();
        // Here the prefix of v must match some of the vectors from vlist
        int b1 = 0, b2 = num - 1, b;
        int plen = 0;
        for (; b2 >= b1;) {
            b = (b2 + b1) / 2;
            int nd, pd;
            plen = Diff (vlist[b], v, &nd, &pd);
            if (plen == -1 || !(nd + pd)) {
                b1 = b2 = b;
                break;
            } else if (CompFunc (vlist[b], v) > 0) {
                b1 = b + 1;
            } else {
                b2 = b - 1;
            }
        }
        if (b2 == b1) {
            if (plen >= 0)
                plist[b1]->Search (v.
                                   SubVector (v.StartIndex () + plen,
                                              v.EndIndex ()), UpperPosDist,
                                   UpperNegDist, s, MaxErr, a, numComp,
                                   resList, hasTransp, silent);
            // if plen == -1 the vectors are equal
            else
                plist[b1]->Search (v, UpperPosDist, UpperNegDist, s,
                                   MaxErr, a, numComp, resList, hasTransp, silent);
        }
    }



}

void CFreqVectorList::SearchExact (CFreqVectorMap & map,
                                   const std::string & s, 
                                   const myagrep &a, 
                                   int *numComp,
                                   StringList * resList, bool silent)
{
    CFreqVector v (&map, 0, map.Size () - 1);
    v.InitFromString (s.c_str ());
    SearchExact (v, s, a, numComp, resList, silent);
}

void CFreqVectorList::SearchExact (const CFreqVector & v,
                                   const std::string & s, 
                                   const myagrep &a, 
                                   int *numComp,
                                   StringList * resList, bool silent)
{
    int b1 = 0, b2 = num - 1, b;
    int plen = 0;
    for (; b2 >= b1;) {
        b = (b2 + b1) / 2;
        int nd, pd;
        plen = Diff (vlist[b], v, &nd, &pd);
        if (plen == -1 || !(nd + pd)) {
            b1 = b2 = b;
            break;
        } else if (CompFunc (vlist[b], v) > 0) {
            b1 = b + 1;
        } else {
            b2 = b - 1;
        }
    }
    if (b2 == b1) {
        if (plen >= 0)
            plist[b1]->SearchExact (v.
                                    SubVector (v.StartIndex () + plen,
                                               v.EndIndex ()), s, 
                                    a, numComp,
                                    resList, silent);
        // If plen==-1, vectors are equal
        else
            plist[b1]->SearchExact (v, s, a, numComp, resList, silent);
    }

}

void CFreqVectorList::Insert (CFreqVectorMap & map, const std::string & s)
{
    CFreqVector v (&map, 0, map.Size () - 1);
    v.InitFromString (s.c_str ());
    Insert (v, s);
}

void CFreqVectorList::Insert (const CFreqVector & vi,
                              const std::string & s)
{
// The case of empty node
    if (!num) {
// if (ptr==NULL) we initialize it 
        CStringHeap *ptr = new CStringHeap;
        ptr->Insert (s);
        vlist.insert (vlist.begin (), vi);
        plist.insert (plist.begin (), ptr);
        num++;
        return;
    }
// If not empty we find a frequency vector
// having a common prefix of at least MinPrefix long.
// If such a prefix does not exist, then v is inserted 
// in the list. The list is kept in a sorted order

// 1. Seeking the most closest vector
    int i, iclosest = -1, plen = 0, md = 256;
    for (i = 0; i < num; i++) {
        int nd, pd, d;
        int l = Diff (vi, vlist[i], &nd, &pd);


        d = (nd > pd) ? pd : nd;
        if (l == -1 || (nd + pd == 0))  // Equal vectors of equal size
        {
            if (nd + pd == 0 && l >= 0) {
                // We found a common prefix which equals to the 
                // node label (node frequency fector)
                // Must insert a string into this node
                CFreqVectorList *ptr =
                    dynamic_cast < CFreqVectorList * >(plist[i]);
                assert (ptr);
                ptr->Insert (vi.
                             SubVector (vi.StartIndex () + l,
                                        vi.EndIndex ()), s);
                return;
            }
            // Both node label and vector vi are equal
            assert (plist[i]);
            CStringHeap *ptr = dynamic_cast < CStringHeap * >(plist[i]);
            assert (ptr);
            ptr->Insert (s);
            return;
        }
        if ((l > plen) || (l == plen && d < md)) {
            plen = l;
            md = d;
            iclosest = i;
        }

    }
// 2. Prefix is long enough
    if (plen >= MinPrefix) {
        CFreqVector & vc = vlist[iclosest];
        // Prefix should be shorter than vector length
        assert (plen <= vi.EndIndex ());
        //assert(plen<=vc.EndIndex());
        CFreqVector vcsub (vc.
                           SubVector (vc.StartIndex (),
                                      vc.StartIndex () + plen - 1));
        // Creating subnode 
        CFreqVectorList *vlst = new CFreqVectorList ();
        // Filling subnode with splitted vectors
        CFreqVector vs1 =
            vc.SubVector (vc.StartIndex () + plen, vc.EndIndex ());
        CFreqVector vs2 =
            vi.SubVector (vi.StartIndex () + plen, vi.EndIndex ());
        CStoreObj *ptr = plist[iclosest];
        CStringHeap *sh = new CStringHeap;
        sh->Insert (s);
        // We must keep the sub-node in a sorted order!
        if (CompFunc (vs1, vs2) > 0) {
            vlst->vlist.insert (vlst->vlist.end (), vs1);
            vlst->plist.insert (vlst->plist.end (), ptr);
            vlst->vlist.insert (vlst->vlist.end (), vs2);
            vlst->plist.insert (vlst->plist.end (), sh);
        } else {
            vlst->vlist.insert (vlst->vlist.end (), vs2);
            vlst->plist.insert (vlst->plist.end (), sh);
            vlst->vlist.insert (vlst->vlist.end (), vs1);
            vlst->plist.insert (vlst->plist.end (), ptr);
        }
        // Substituting the data at the current level
        vlst->num = 2;
        plist[iclosest] = vlst;
        vlist[iclosest] = vcsub;    // vcsub is used in equation operator
        // Here vcsub must be destroyed
        return;
    }
// 3. Prefix is too short -- just filling array in sorted (ascending) order
    TVectList::iterator vlast = vlist.begin ();
    TPtrList::iterator plast = plist.begin ();
    for (i = 0; i < num; i++, vlast++, plast++) {
        int d = CompFunc (vlist[i], vi);
        if (!i && d < 0) {
            break;
        }
        if (d > 0)
            if (i < num - 1)
                if (CompFunc (vlist[i + 1], vi) < 0) {
                    vlast++;
                    plast++;
                    break;
                }
    }
    vlist.insert (vlast, vi);
    CStringHeap *ptr = new CStringHeap;
    ptr->Insert (s);
    plist.insert (plast, ptr);
    num++;

}

#ifdef _DEBUG
void CFreqVectorList::Print (int level)
{
    TVectList::iterator v = vlist.begin ();
    TPtrList::iterator p = plist.begin ();
    for (; v != vlist.end (); v++, p++) {
        cout << "L=" << level << "\n";
        for (int i = 0; i < level; i++)
            cout << " ";
        (*v).Print ();
        if (*p)
            (*p)->Print (level + 1);
    }
}
#endif

int CFreqVectorList::DumpToFile (int fd, off_t offset)
{
// We store the data in the following order
// 1. the buffer size, the number of sub-trees
// 2. the subvectors themselves with zero offsets
// 3. the sub-trees calculating and storing sub-tree offsets
    int cnt = 0;
    int buflen = 0;
    int sub_sz = 0;
    buflen += sizeof (cnt) + sizeof (buflen);

    TVectList::iterator v = vlist.begin ();
    for (; v != vlist.end (); v++, cnt++) {
        // Vector size + sub-tree offset size + vector indices 
        buflen += sizeof (offset) + (*v).Size () + 2;
    }
    int delta = 0;
    if (offset)
        delta = (offset % getpagesize ());
    if (lseek (fd, offset + buflen - 1, SEEK_SET) < 0
        || write (fd, "", 1) < 1) {
        return -1;
    }
    void *buf =
        mmap (NULL, buflen + delta, PROT_WRITE | PROT_READ, MAP_SHARED, fd,
              offset - delta);
    if (!buf || buf == MAP_FAILED)
        return -1;
    char *sbuf = (char *) buf + delta;
    // Step 1.
    *(int *) (void *) sbuf = buflen;
    sbuf += sizeof (buflen);
    *(int *) (void *) sbuf = cnt;
    sbuf += sizeof (cnt);
    // Step 2.
    off_t *ptrs[cnt];
    v = vlist.begin ();
    for (int i = 0; v != vlist.end (); v++, i++) {
        assert (i < cnt);
        // Indexes
        *((u_int8_t *) (void *) sbuf) = (*v).StartIndex ();
        sbuf++;
        *((u_int8_t *) (void *) sbuf) = (*v).EndIndex ();
        sbuf++;
        // Zero offset
        ptrs[i] = (off_t *) (void *) sbuf;
        *ptrs[i] = 0;
        sbuf += sizeof (off_t);
        // Sub-vector
        (*v).CopyToMem ((void *) sbuf);
        sbuf += (*v).Size ();
    }
    // Now storing subtrees starting from offset = offset +sz 
    TPtrList::iterator p = plist.begin ();
    sub_sz = 0;
    for (int j = 0; p != plist.end (); p++, j++) {
        assert (j < cnt);
        *ptrs[j] = offset + buflen + sub_sz;
        // Diving into subtree
        int s_sz = (*p)->DumpToFile (fd, offset + buflen + sub_sz);
        if (s_sz < 0) {
            munmap (buf, buflen + delta);
            return -1;
        }
        sub_sz += s_sz;
    }

    assert (sbuf - (char *) buf == delta + buflen);

    if (munmap (buf, buflen + delta) < 0)
        return -1;
    // Return the total number of bytes written
    return buflen + sub_sz;
}

#ifdef _DEBUG
void CStringHeap::Print (int level)
{
    TStringList::iterator si = sl.begin ();
    for (; si != sl.end (); si++) {
        for (int i = 0; i < level; i++)
            cout << " ";
        cout << (*si) << "\n";
    }

}
#endif

const char CStringHeap::termChar = 0;

void CStringHeap::Insert (const std::string & s)
{
    sl.append(s);
    sl.append(&termChar, 1);
    ++slCnt;
}

void CStringHeap::SearchExact(const CFreqVector & v,
                               const std::string & s, 
                               const myagrep &a, int *numComp,
                               StringList * resList, bool silent)
{
    if (numComp)
        (*numComp) += slCnt;

    a.findPattern(CInterface::filterType, termChar,
                 sl.data(), sl.data() + sl.size(),
                 silent, resList);

}

int CStringHeap::GetVectorCount() 
{
    return 1;
}

void CStringHeap::Search (const CFreqVector & v, int UpperPosDist,
                          int UpperNegDist, const std::string s,
                          int MaxErr, 
                          const myagrep & a, int *numComp,
                          StringList * resList, bool hasTransp, bool silent)
{
    if (numComp)
        (*numComp) += slCnt;

    a.findPattern(CInterface::filterType, termChar,
                 sl.data(), sl.data() + sl.size(),
                 silent, resList);

}




int CStringHeap::SearchInFile (int fd, off_t offset, const std::string s,
                               int MaxErr, 
                               const myagrep & a, int *numComp,
                               StringList * resList, bool hasTransp, bool silent)
{
    int delta = 0;
    if (offset)
        delta = (offset % getpagesize ());
    // First we must read the record size
    int cnt, buflen;
    if (lseek (fd, offset, SEEK_SET) < 0
        || read (fd, &buflen, sizeof (buflen)) < sizeof (buflen))
        return -1;
    // Now map to the file
    void *buf =
        (void *) mmap (NULL, buflen + delta, PROT_READ, MAP_SHARED, fd,
                       offset - delta);
    if (!buf || buf == MAP_FAILED)
        return -1;

    char *sbuf = (char *) buf + sizeof (buflen) + delta;

    cnt = *(int *) (void *) sbuf;
    sbuf += sizeof (cnt);

    // Now we have access to the string buffer
    if (numComp)
        (*numComp) += cnt;
    // Calculate strings size
    int sbufSize = buflen - sizeof buflen - sizeof cnt;

    a.findPattern(CInterface::filterType, termChar,
                 sbuf, sbuf + sbufSize,
                 silent, resList);

    if (munmap (buf, buflen + delta) < 0)
        return -1;

    return 1;
}

int CStringHeap::DumpToFile (int fd, off_t offset)
{
    int sz = sl.size();
         
    int delta = 0;
    if (offset)
        delta = (offset % getpagesize ());
    int buflen = sizeof (slCnt) + sizeof (sz) + sz;
    if (lseek (fd, offset + buflen - 1, SEEK_SET) < 0
        || write (fd, "", 1) < 1) {
        return -1;
    }
    void *buf =
        (void *) mmap (NULL, buflen + delta, PROT_WRITE | PROT_READ,
                       MAP_SHARED, fd, offset - delta);
    if (!buf || buf == MAP_FAILED)
        return -1;
    // First of all we write down the number of the strings
    // and record size
    char *sbuf = (char *) buf + delta;
    *(int *) sbuf = buflen;
    sbuf += sizeof (buflen);
    *(int *) sbuf = slCnt;
    sbuf += sizeof (slCnt);

    memcpy(sbuf, sl.data(), sl.size());

    assert (sbuf - (char *) buf == delta + buflen);

    if (munmap (buf, buflen + delta) < 0)
        return -1;
    // Not return the number of used bytes
    return buflen;
}
