#ifndef _STOREOBJ_H
#define _STOREOBJ_H

#include <unistd.h>
#include <sys/mman.h>

#include <vector>
#include <string>
#include "freqvector.h"
#include "myagrep.h"
#include "interface.h"



class CStoreObj {
  protected:
  public:
    virtual void SearchExact (const CFreqVector & v, const std::string & s,
                              const myagrep & a,
                              int *numComp, StringList * resList,
                              bool silent = true) = 0;
    virtual void Search (const CFreqVector & v, int UpperPosDist,
                         int UpperNegDist, const std::string s, int MaxErr,
                         const myagrep & a, 
                         int *numComp, StringList * resList, bool hasTransp,
                         bool silent) = 0;
    // returns the size of the dumped structure if successful or -1 otherwise
    virtual int DumpToFile (int fd, off_t offset) = 0;
    virtual int GetVectorCount() = 0;
#ifdef _DEBUG
    virtual void Print (int) = 0;
#endif
};

typedef CStoreObj *PCStoreObj;
typedef std::vector < CFreqVector > TVectList;
typedef std::vector < PCStoreObj > TPtrList;

using namespace std;

class CFreqVectorList:public CStoreObj {
  protected:
    static const int MinPrefix = 4;
    int num;
    TVectList vlist;
    TPtrList plist;
// === Searching/inserting private
    void Insert (const CFreqVector & v, const std::string & s);
    void SearchExact (const CFreqVector & v, const std::string & s,
                      const myagrep &a, 
                      int *numComp, StringList * resList, bool silent);
    void Search (const CFreqVector & v, int UpperPosDist, int UpperNegDist,
                 const std::string s, int MaxErr, 
                 const myagrep &a, int *numComp,
                 StringList * resList, bool hasTransp, bool silent);
  public:
    // === Searching/inserting public
    void SearchExact (CFreqVectorMap &, const std::string & s,
                      const myagrep &a, 
                      int *numComp, StringList * resList, bool silent);
    void Search (CFreqVectorMap &, const std::string s, int MaxErr,
                    const myagrep &a, int *numComp, 
                    StringList * resList, bool hasTransp, bool silent);
    CFreqVectorList ():num (0) {;
    } void Insert (CFreqVectorMap &, const std::string & s);
    // === Working directly with file on disk
    int DumpToFile (int fd, off_t offset);
    static int SearchInFile (int fd, CFreqVectorMap & map,
                             const std::string s, int MaxErr, 
                             const myagrep &a, int *numComp,
                             StringList * resList, bool hasTransp, bool silent);
    static int SearchInFile (int fd, off_t offset, CFreqVectorMap & map,
                             const CFreqVector & v, int UpperPosDist,
                             int UpperNegDist, const std::string s,
                             int MaxErr, const myagrep & a, int *numComp,
                             StringList * resList, bool hasTransp, bool silent);
    int GetVectorCount();
#ifdef _DEBUG
    void Print (int level = 0);
#endif
};

typedef std::vector < std::string > TStringList;

class CStringHeap:public CStoreObj {
    void SearchExact (const CFreqVector & v, const std::string & s,
                      const myagrep &a, 
                      int *numComp, StringList * resList, bool silent);
  protected:
    void Search (const CFreqVector & v, int UpperPosDist, int UpperNegDist,
                 const std::string s, int MaxErr, 
                 const myagrep & a, int *numComp,
                 StringList * resList, bool hasTransp, bool silent);
    const static char termChar;
  public:
    CStringHeap () : slCnt(0) {}
    int    slCnt;
    string sl;
    void Insert (const std::string &);
    // === Working directly with file on disk
    int DumpToFile (int fd, off_t offset);
    static int SearchInFile (int fd, off_t offset, const std::string s,
                             int MaxErr, 
                             const myagrep & a, int *numComp,
                             StringList * resList, bool hasTransp, bool silent);
    int GetVectorCount();
#ifdef _DEBUG
    virtual void Print (int);
#endif
};


#endif
