#ifndef _LEN1INDEX_NO_BDB_H
#define _LEN1INDEX_NO_BDB_H

#include "len1index.h"

class CLen1IndexNoBDB : public CLen1Index 
{
  protected:
    bool DumpDBNoCompress(const char *index_name, mOffset &);
    bool DumpDBCompress(const char *index_name, mOffset &);
    bool ReadListDBCompress(const char *str, int StrLen, int minpos, int maxpos,
                     OffsetList * ptr);
    bool ReadListDBNoCompress(const char *str, int StrLen, int minpos, int maxpos,
                     OffsetList * ptr);
  public:
    CLen1IndexNoBDB(int NgramLen, bool hasTransp, bool hasCompress) : 
            CLen1Index(NgramLen, hasTransp, hasCompress, false), fdGram(0), gramMapSize(0),
             gramFileStart(NULL), gramArray(NULL), gramCnt(0) {}

    bool OpenDB (const char *name, bool create = false);
    bool CloseDB ();
    bool DumpDB (const char *index_name, mOffset &);

    bool ReadListDB (const char *str, int StrLen, int minpos, int maxpos,
                     OffsetList * ptr);
    void ResetCounters();
    void PrintAddStat(int testNum, int kerr, ostream &);
#ifdef _DEBUG
    void Print ();
#endif
   protected:
    struct EntryData {
        size_t        listIndex;
        size_t        cnt;
        EntryData(size_t listIndex, size_t cnt) : listIndex(listIndex), cnt(cnt) {}
    };
    struct EntryDataCompress {
        size_t         offset;
        size_t         cnt;
        size_t         size;
        EntryDataCompress(size_t offset, size_t cnt, size_t size) :
                offset(offset), cnt(cnt), size(size) {}
    };
    int       fdGram;
    off_t     gramMapSize;
    size_t    *gramFileStart;
    char      *gramArray; 
    int       gramCnt;
    inline char      *END() {
        if (!opened) {
            cerr << "Using on closed index in " << __func__ << endl;
            return NULL;
        }
        return gramArray + gramCnt * EntrySize();
    }
    // Find the lower position for the n-gram represented by mins, mins is produced by ngram::toStr()
    // lower_bound analog
    char      *gram_lower_bound(char * mins);
    inline size_t EntrySize() {return (mHasCompress ? sizeof(EntryDataCompress):sizeof(EntryData)) + NgramLen + 2;}
    DECLARE_STATIC_COUNTER(load_list);
};

#endif
