#include "interface.h"
#include "langcase.h"
#include "myexcept.h"

#define MAX_RAND_ITER   100
//#define TRACE_RANDOM_GEN

bool CInterface::InsertList (const StringList & lst)
{
    for (StringList::const_iterator i = lst.begin (); i != lst.end (); i++) {
        if (!Insert (i->c_str ()))
            return false;
    }
    return true;
};


bool CInterface::CreateInMemIndex(const char *inputFile, int insertCount, int *dcnt, bool silent)
{
    *dcnt = 0;
    string s;

    ifstream f (inputFile);
    if (!f) {
        string err = string("Cannot open '") + inputFile + string("'");
        Perror(err.c_str());
        return false;
    }
    StringList lst;
    int lstCnt = 0;
    while (getline(f, s)) {
        toUpper (s);

        if (insertCount == 1) {
            Insert (s.c_str ());
        } else {
            lst.push_back (s);
            lstCnt++;
            if (lstCnt >= insertCount) {
                this->InsertList (lst);
                lstCnt = 0;
                lst.clear ();
            }
        }
        if (dcnt)
            (*dcnt)++;
        if (!silent && !((*dcnt) % 10000))
            cout << (*dcnt) << "\n";
    }
    if (lstCnt)
        this->InsertList (lst);

#if 0
    Print();
#endif
    return true;
}

string ApplyRandomEdits(string str,
                        bool   hasTransp,
                        int kerr,
                        const char *pABC,
                        unsigned    ABCLen,
                        int start,
                        const CProbDesc& ProbDesc)
{
    unsigned size = str.size ();
    if (!kerr || start > size) return str;
#ifdef TRACE_RANDOM_GEN
    string OrigStr = str;
#endif

    for (unsigned nIter = 0; nIter < MAX_RAND_ITER; ++nIter) {
        unsigned val = rand() % RANDOM_EDIT_PRECISION_VAR;

        float p = ProbDesc.mProbSubst;

        if (val < p * RANDOM_EDIT_PRECISION_VAR) {
            // Substitution
            unsigned i = start;

             if (i >= size) { // Cannot generate a substitution 
                continue;
            }

            i += rand() % (size - i);

            const char *ptr = pABC + rand() % ABCLen;

            str[i] = *ptr;
#ifdef TRACE_RANDOM_GEN
            cout << OrigStr << " -> " << str << " [sub i = " << i << "]" << endl;
#endif
            return ApplyRandomEdits(str, hasTransp, kerr - 1, pABC, ABCLen, i + 1, ProbDesc);
        }

        p += ProbDesc.mProbIns;

        if (val < p * RANDOM_EDIT_PRECISION_VAR) {
            unsigned i = start;

            i += rand() % (size - i + 1);

            const char *ptr = pABC + rand() % ABCLen;

            str.insert (str.begin () + i, *ptr);
#ifdef TRACE_RANDOM_GEN
            cout << OrigStr << " -> " << str << " [ins i = " << i << "]" << endl;
#endif
            return    ApplyRandomEdits(str, hasTransp, kerr - 1, pABC, ABCLen, i + 1, ProbDesc);
        }

        p += ProbDesc.mProbDel;

        if (val < p * RANDOM_EDIT_PRECISION_VAR) {
            unsigned i = start;

            if (i >= size) { // Cannot generate a deletion 
                continue;
            }

            i += rand() % (size - i);
            str.erase (str.begin () + i);
#ifdef TRACE_RANDOM_GEN
            cout << OrigStr << " -> " << str << " [del i = " << i << "]" << endl;
#endif
            return ApplyRandomEdits(str, hasTransp, kerr - 1,  pABC, ABCLen, i, ProbDesc);
        }


        if (hasTransp) {
            unsigned i = start;

            if (i + 1 >= size) { // Cannot generate a transposition
                continue;
            }

            i += rand() % (size - i - 1);
            swap(str[i], str[i+1]);
#ifdef TRACE_RANDOM_GEN
            cout << OrigStr << " -> " << str << " [transp i = " << i << "]" << endl;
#endif
            return ApplyRandomEdits(str, hasTransp, kerr - 1, pABC, ABCLen, i + 2, ProbDesc);
        }

        THROW_AND_PRINT_RUNTIME_ERROR("Internal error: Should not reach this point, if transpositions are not generated!");
    
    }

    THROW_AND_PRINT_RUNTIME_ERROR("Internal error: failed to produce a random edit after " << MAX_RAND_ITER << " iterations!");

    return "";
}

void FillExpansionList (string str, bool hasTransp, int kerr, StringList * lst,
                        const char *ABC, int minKerrLimited, bool limited,
                        int start, bool showVariants)
{
#if 0
    if (lst->find (str) == lst->end ()) {
        cout << str << endl;
    }
#endif

    if (1) {
        if (showVariants) {
            std::cout << "Variant: " << str << std::endl;
        }
        lst->push_back (str);
    }

    int size = str.size ();
    if (!kerr || start > size) return; 

    int i = 0;
    // Substitution
    char strCopy[size + 1];
    
    memcpy(strCopy, str.c_str(), size + 1);

    for (i = start; i < size; i++) {
        int NGen = 0;
        for (const char *ptr = ABC; *ptr; ptr++) {
            if (*ptr != str[i]) {
                ++NGen;
            }
            str[i] = *ptr;
            FillExpansionList (str, hasTransp, kerr - 1, lst, ABC, minKerrLimited,
                               limited, i + 1, showVariants);
            str[i] = strCopy[i];
            if (limited && kerr <= minKerrLimited && NGen > 1)
                break;
        }
        assert (str == strCopy);
    }

    // Insertion

    for (i = start; i <= size; i++) {
        int NGen = 0;
        for (const char *ptr = ABC; *ptr; ptr++) {
            str.insert (str.begin () + i, *ptr);
            FillExpansionList (str, hasTransp, kerr - 1, lst, ABC, minKerrLimited,
                               limited, i + 1, showVariants);
            str.erase (str.begin () + i);
                ++NGen;
            if (limited && kerr <= minKerrLimited && NGen > 1)
                break;
        }
        assert (str == strCopy);
    }

    // Deletion
    for (i = start; i < size; i++) {
        str.erase (str.begin () + i);
        FillExpansionList (str, hasTransp, kerr - 1, lst, ABC, minKerrLimited, limited,
                           i, showVariants);
        str.insert (str.begin () + i, strCopy[i]);
        assert (str == strCopy);
    }

    if (hasTransp) {
    // Transposition
        for (i = start; i < size - 1; i++) {
            str[i] = strCopy[i + 1];
            str[i + 1] = strCopy[i];
            FillExpansionList (str, hasTransp, kerr - 1, lst, ABC, minKerrLimited, limited, i + 2, showVariants);
            str[i] = strCopy[i];
            str[i + 1] = strCopy[i + 1];
        }
        assert (str == strCopy);
    }
}


void toUpper (string & s)
{
    for (int i = 0; i < s.size (); i++)
        s[i] = toupper_eng_rus (s[i]);
}
