#ifndef ALPHABET_H
#define ALPHABET_H

#include <memory.h>
#include <ctype.h>

#include <string>

#include "myexcept.h"

enum EAlphabet {
    eEnglish = 0,
    eRussian,
    eDNA,
    eEnglishNum,
    eEnglishShort,
    eAlphaUnknown
};

extern const char* EnglishABC;
extern const char* RussianABC;
extern const char* EnglishNumABC;
extern const char* DNA_ABC;
extern const char* EnglishShortABC;

extern const char*  paABCList[];

inline const char *GetAlphabet(EAlphabet type) {
    switch (type) {
        case eEnglish: return EnglishABC;
        case eRussian: return RussianABC;
        case eDNA:     return DNA_ABC;
        case eEnglishNum: return EnglishNumABC;
        case eEnglishShort: return EnglishShortABC;
    }
    return NULL;
}

inline EAlphabet GetAlphabetType(const std::string& Name) {
    if (!strcasecmp(Name.c_str(), "eng")) return eEnglish;
    if (!strcasecmp(Name.c_str(), "engnum")) return eEnglishNum;
    if (!strcasecmp(Name.c_str(), "rus")) return eRussian;
    if (!strcasecmp(Name.c_str(), "dna")) return eDNA;
    if (!strcasecmp(Name.c_str(), "engshort")) return eEnglishShort;

    return eAlphaUnknown;
}

inline unsigned ConvertChar(EAlphabet type, const char c) {
    if (type >= eAlphaUnknown) {
        THROW_RUNTIME_ERROR("Invalid alphabet type: " << type);
    }
    const char* pAlpha = paABCList[type];
    const char* p      = strchr(pAlpha, c);

    if (!p) {
        THROW_RUNTIME_ERROR("There is no symbol '" << c << "' for alphabet type: " << type);
    }

    return p - pAlpha;
};

#endif
