#include <string>
#include <sstream>
#include <iostream>
#include <fstream>

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>

#include "levendfa.h"

using namespace std;

DECLARE_STATIC_COUNTER(cr_dfa)


void Usage(const char* pProg)
{
    cerr << "Usage: " << pProg << " <kerr> <HasTransp> <pattern> <InputFileName> <StatFileName (to output automaton construction time)> " << endl;
}

void ScanSegment(const char* pBeg, const char* pEnd, LevenDFA& dfa, char TermChar)
{
    const char* pCurr   = pBeg;

    for (const char *p = pBeg; pBeg < pEnd; pBeg = p + 1) {
        p = const_cast<const char *>(strchr(pBeg, TermChar));
        if (!p) p = pEnd;

        dfa.Start();
        bool bOk    = true;
        bool bFinal = false;

        for (const char* p1 = pBeg; p1 < p; ++p1) {
            if (!dfa.Next(*p1, bFinal)) {
                bOk = false;
                break;
            }
        }
        
        if (bOk) {
            if (dfa.Next(LevenDFA::EOS, bFinal) && bFinal) {
                write(1, pBeg, p - pBeg);
                char c = '\n';
                write(1, &c, 1);
            }
        }
    }
}

int FindPattern(const char *pPattern, const char *pInputFileName, const char *pStatFileName, unsigned int kerr, bool bHasTransp, char TermChar)
{

    START_COUNTER(cr_dfa)
    LevenDFA*   pDFA = new LevenDFA(EnglishABC, pPattern, kerr, bHasTransp);
    STOP_COUNTER(cr_dfa)

    int fd = open(pInputFileName, O_RDONLY);

    if (fd < 0) {
        cerr << "Cannot open '" << pInputFileName << "' for reading" << endl;
        return 1;
    }

    ofstream Stat(pStatFileName, ios_base::app);

    int  nr = 0;
    int  maxReadLen = 1024 * 8;
    char buf[maxReadLen * 2 + 1];

    nr = read(fd, buf, maxReadLen);
    int start = 0;

    while (nr > 0)
    {
        buf[start + nr] = TermChar; // There should be always a following TermChar

        const char *pBeg = buf, *pEnd = pBeg + start + nr;

        ScanSegment(pBeg, pEnd, *pDFA, TermChar);

        int pos;
        for (pos = nr - 1; pos > 0 && pBeg[pos] != TermChar; --pos);
        start = nr - 1 - pos;
        if (start) {
            start = nr - 1 - pos;
            memmove(buf, pBeg + pos + 1, start);
        }
        nr = read(fd, &buf[start], maxReadLen);
    }

    START_COUNTER(cr_dfa)
    delete pDFA;
    STOP_COUNTER(cr_dfa)

    Stat << GET_COUNTER(cr_dfa) << endl;

    return 0;
}

int main(int argc, char* pArgv[])
{


    if (argc != 6) {
        Usage(pArgv[0]);
        return 1;
    }

    unsigned int        kerr        = atol(pArgv[1]);
    bool                bHasTransp  = atol(pArgv[2]);

    return FindPattern(pArgv[3], pArgv[4], pArgv[5], kerr, bHasTransp, '\n');
}
