#ifndef _FQNODE_H
#define _FQNODE_H

#include <vector>
#include <string>
#include <sys/types.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>
#include <stdio.h>
#include <fstream>
#include <filesize.h>
#include <myagrep.h>
#ifdef MYERS_DIST
    #include "myers.hpp"
#else
    #include "edist.h"
#endif

class CFQnode;


using namespace std;


typedef vector <string >     StringList;
typedef vector <CFQnode *>   FQNodeList;
typedef vector <StringList>  VStringList;


class CFQnode {
  protected:
    std::string   root;
    int           treeCnt;
    u_int16_t     max_dist;
    u_int16_t     bulkCnt; // 1 if no bulk

    FQNodeList nodes;
    void SearchAtLevel (const std::string & s, int nerr, const myagrep & a,
#ifdef MYERS_DIST
                        CMyersEdist& MyersEdist,
#endif
                        int *numComp, StringList * resList, bool hasTransp, 
                        bool silent);
    static bool SearchInFileAtLevel (int fd, off_t offset,
                                     const std::string & s, const myagrep & a,
                                     int nerr, int *numComp,
                                     StringList * resList, 
                                     bool hasTransp, bool silent);
    const static char termChar;

    void fillTreeCnt();
    void appendAllStrings(string & str);
  public:
   CFQnode (const std::string & _root = ""):root (_root), max_dist (0), bulkCnt(1), treeCnt(0) {;} 
   ~CFQnode();
    int DumpToFile (off_t offset, int fd);
    void Insert(const StringList & lst, bool hasTransp, bool silent, int level);
    void Search(const std::string & s, unsigned int nerr, int *numComp, 
                 StringList * resList, bool hasTransp, bool silent) {
        myagrep a(s.c_str(), nerr, hasTransp);
#ifdef MYERS_DIST
        CMyersEdist MyersEdist(s.c_str(), hasTransp);
#endif
        SearchAtLevel (s, nerr, a, 
#ifdef MYERS_DIST
            MyersEdist,
#endif
        numComp, resList, hasTransp, silent);
    }
    static bool SearchInFile (int fd, const std::string & s,
                              unsigned int nerr, int *numComp,
                              StringList * resList, bool hasTransp, bool silent) {
        myagrep a(s.c_str(), nerr, hasTransp);

        return SearchInFileAtLevel (fd, 0, s, a, nerr, numComp, resList, hasTransp,
                                    silent);
    }
    void Print (int level = 0);
    void CreateBuckets(int buckSize, int level = 0);
};

#endif
