#include "index.h"
#include "myagrep.h"

CFQnode::~CFQnode () {
    for (int i = 1; i <= max_dist; i++)
        if (nodes[i]) {
            delete nodes[i];
        }
}

CMtreeIndex::CMtreeIndex (bool hasTransp, EAlphabet AlphaType): CInterface(hasTransp, AlphaType), opened (false)
{
}

CMtreeIndex::~CMtreeIndex ()
{
    if (opened)
        CloseDB ();
}

bool CMtreeIndex::Insert(const char *str)
{
    StringList lst;
    lst.push_back (string (str));
    return InsertList(lst);
}

bool CMtreeIndex::InsertList(const StringList & lst)
{
    node.Insert(lst, mHasTransp, true, 0);
    return true;
}

bool CMtreeIndex::Search(const char *str, int kerr, int *numComp,
                          StringList * resList, bool silent)
{
    string s (str);
    node.Search (s, kerr, numComp, resList, mHasTransp, silent);
    return true;
}

bool CMtreeIndex::SearchInDB(const char *str, int kerr, int *numComp,
                              StringList * resList, bool silent)
{
    if (!opened) {
        cerr << "Database is not opened!" << endl;
    }
    return node.SearchInFile (fd, string (str), kerr, numComp, resList, mHasTransp,
                              silent);
}

bool CMtreeIndex::OpenDB (const char *file_name, bool create)
{
    if (opened)
        CloseDB ();
    if (!create) {
        fd = open (file_name, O_RDONLY, 0);
    } else {
        fd = open (file_name, O_RDWR | O_TRUNC | O_CREAT, defaultAccess);
    }
    if (fd < 0) {
        Perror ("Error opening file");
        return false;
    }
    return (opened = true);
}

bool CMtreeIndex::CloseDB ()
{
    if (!opened)
        return true;
    if (close (fd) != 0)
        return false;
    opened = false;
    return true;
}


bool CMtreeIndex::DumpDB (const char *file_name)
{
    bool stat;
    if (opened)
        CloseDB ();
    if (!OpenDB (file_name, true)) {
        return false;
    }

    stat = node.DumpToFile (0, fd);
    return stat && CloseDB ();
}

bool CMtreeIndex::CreateInMemIndex (const char *inputFile,
                                     int insertCount, int *dcnt, bool silent)
{
    *dcnt = 0;
    string s;

    ifstream f (inputFile);
    if (!f) {
        string err = string("Cannot open '") + inputFile + string("'");
        Perror(err.c_str());
        return false;
    }
    StringList lst;
    int lstCnt = 0;
    while (f >> s) {
        toUpper (s);

        if (insertCount == 1) {
            Insert (s.c_str ());
        } else {
            lst.push_back (s);
            lstCnt++;
            if (lstCnt >= insertCount) {
                this->InsertList (lst);
                lstCnt = 0;
                lst.clear ();
            }
        }
        if (dcnt)
            (*dcnt)++;
        if (!silent && !((*dcnt) % 10000))
            cout << (*dcnt) << "\n";
    }
    if (lstCnt)
        this->InsertList (lst);
    int num = BUCKET_SIZE;
    const char *pBuck = getenv("BUCKET_SIZE");
    if (pBuck && atol(pBuck)) {
        num = atol(pBuck);
    }
    cout << "Bucket size: "<< num << endl;
    node.CreateBuckets(num);

#if 0
    Print();
#endif
    return true;
}


void CMtreeIndex::Print ()
{
    node.Print ();
}
