#include "index.h"

#include <fstream>

extern const char* baseName;

CFlamingoMtreeIndex::CFlamingoMtreeIndex (bool hasTransp, EAlphabet AlphaType): CInterface(hasTransp, AlphaType), mpStrMap(NULL)
{
}

CFlamingoMtreeIndex::~CFlamingoMtreeIndex () {
    delete mpStrMap;
}

bool CFlamingoMtreeIndex::Insert (const char *str) 
{
    return false;
}

bool CFlamingoMtreeIndex::Search (const char *str, int kerr, int *numComp,
                          StringList * resList, bool silent)
{
    mpStrMap->Search(str, kerr, maThreshold[max(1, kerr)], true /* include duplicates */, resList);
    return true;
}

bool CFlamingoMtreeIndex::SearchInDB (const char *str, int kerr, int *numComp,
                              StringList * resList, bool silent)
{
    cerr << "SearchInDB is not supported.\n";
    return false;
}

bool CFlamingoMtreeIndex::OpenDB (const char *name, bool create)
{
    return true;
}

bool CFlamingoMtreeIndex::CloseDB ()
{
    return true;
}


bool CFlamingoMtreeIndex::DumpDB (const char *name)
{
    ofstream out(name);

    if (!out) {
        string err = string("Cannot create file '") + name + "'";
        Perror(err);
        return false;
    }
    return true;
}

bool CFlamingoMtreeIndex::CreateInMemIndex (const char *inputFile, int insertCount, int *dcnt, bool silent)
{
    int             cnt = 0;
    string          s;

    ifstream inp(inputFile);

    while (getline(inp, s)) {
        mvList.push_back(s);
        ++cnt;
    }

    if (inp.bad()) {
        Perror("Cannot read from the file: " + string(inputFile));
        return false;
    }

    if (dcnt) *dcnt = cnt;

    mpStrMap = new StringMap(mvList);

    for (unsigned k = 1; k <= MAX_KERR; ++k) {
        cout << "Sampling for k = " << k << endl;
        maThreshold[k] = mpStrMap->getNewThreshold(mvList, SAMPLE_PCT, k);
    }

    mpStrMap->ConstructSearchRtree(mvList, baseName);

    return true;
}

