#ifndef GEN_FULLNG_H
#define GEN_FULLNG_H

#include <string>
#include <algorithm>

#include "../Interface/interface.h"
#include "../StringTriesNew/afuzzycontext.h"
#include "../afuzzy/myers.hpp"


template <class TNGReceiver>
void GenFullNGInternal(TNGReceiver& recv, unsigned start, string& Str, const char *pPat, unsigned sz, unsigned kerr, const char* pABC, Uint aR[], unsigned StartState, Uint aSMapArr[])
{
    Uint        aR1[kerr + 1];

    assert(sz);

    Uint        MaskOk = (Uint(1) << (sz - 1));

    if (Str.size() < start + 1) {
        Str.resize(start + 1);
    }

    for (const char* pLetter = pABC; *pLetter; ++pLetter) {
        Str[start] = *pLetter;
        Uint SMap = aSMapArr[(unsigned char) *pLetter];

        // Recalculate automaton states
        unsigned    StartState1 = StartState;

        aR1[StartState] = (((aR[StartState] << 1) | Uint(start <= StartState)) & SMap);


        bool bCanContinue = false;
        bool bFinal = false;

        if (start > StartState && !aR1[StartState]) {
            ++StartState1;
        }

        if (aR1[StartState] != 0) { bCanContinue = true; }
        if ((aR1[StartState] & MaskOk) != 0) { bFinal = true; }

        for (unsigned d = StartState + 1; d <= kerr; d++) {
            aR1[d] = (((aR[d] << 1) | Uint(start <= d)) & SMap);
            aR1[d] |= ((aR[d - 1] | aR1[d - 1]) << 1) | aR[d - 1] | Uint(start <= d - 1);
           
            if (d == StartState1 && start >= d && !aR1[d]) {
                ++StartState1;
            } 

            if (aR1[d] != 0) { bCanContinue = true; }
            if ((aR1[d] & MaskOk) != 0) { bFinal = true; }
        }

        if (bFinal) { // If final, we ignore bCanContinue, because we are not interested in matches that have proper prefixes in a  neighborhood
            Str.resize(start + 1);
            recv(Str);
        } 
        if (bCanContinue && start + 1 < sz + kerr) {
            GenFullNGInternal(recv, start + 1, Str, pPat, sz, kerr, pABC, aR1, StartState1, aSMapArr);
        } 
    }
}

template <class TNGReceiver>
void GenFullNG(TNGReceiver& recv, const string& sPat, const char* pABC, unsigned kerr) {
    unsigned        sz = sPat.size(); 
    const char*     pPat = sPat.c_str();
    string          Str;

    Str.resize(sz + kerr);

    if (!sz) {
        cerr << "Empty patterns are not supported!" << endl;
        return;
    }

    unsigned        aMap[256];
    Uint            aSMapArr[256];

    memset (aMap, 0, sizeof (aMap));
    memset (aSMapArr, 0, sizeof (aSMapArr));
    
    Uint             aTmpS[sz + 1];

    aTmpS[0] = 0;

    for (unsigned i = 0, cnt = 0; i < sz; i++) {
        unsigned l = aMap[(unsigned char) pPat[i]];

        if (!l) {
            l = aMap[(unsigned char) pPat[i]] = ++cnt;
            aTmpS[l] = 0;
        }
        aTmpS[l] |= Uint(1) << i;
    }
    for (unsigned c = 0; c <= 255; ++c) {
        aSMapArr[c] = aTmpS[aMap[c]];
    }

    Uint    aR[kerr + 1];

    for (unsigned i = 0; i <= kerr; ++i) {
        aR[i] = (Uint(1) << i) - 1;
    }
    
    if (sz <= kerr) {
        recv("");
    }

    GenFullNGInternal(recv, 0, Str, pPat, sz, kerr, pABC, aR, 0, aSMapArr);
}

#endif
