#ifndef GEN_SUPERCOND_H
#define GEN_SUPERCOND_H

#include <string>
#include <algorithm>

#include "../Interface/interface.h"

enum ENGType {
   eFull,
   eCondensed,
   eSuperCondensed
};

unsigned GenCondNG(ENGType Type, const char* pABC, const string& s, unsigned kerr, bool bHasTransp, StringHashSet& dict);

// TODO/TBD no transpositions are supported yet

template <class TSuperCondReceiver>
void GenSuperCondNGInternal(TSuperCondReceiver& recv, unsigned start, unsigned LastActive, char aStr[], const char *pPat, unsigned sz, const char* pABC, unsigned kerr, unsigned* ppDP[], unsigned* ppStartSubstrTrace[]) {
    ppDP              [start + 1][0] = 0;
    ppStartSubstrTrace[start + 1][0] = start + 1;

    for (const char* pLetter = pABC; *pLetter; ++pLetter) {
        unsigned CurrentActive = 0;
        unsigned LastRowToCheck = min(LastActive + 1, sz);

        aStr[start] = *pLetter;

        unsigned row = 1, prow = 0;
    

        for (; row <= LastRowToCheck; ++row, ++prow) {
            unsigned StartSubstrTrace = ppStartSubstrTrace[start][row];
            unsigned d = ppDP[start][row] + 1;
            unsigned d1 = ppDP[start][prow] + static_cast<unsigned>(pPat[row - 1] != aStr[start]);

            if (d >= d1) {
                d = d1;
                StartSubstrTrace = ppStartSubstrTrace[start][prow];
            }

            unsigned d2 = ppDP[start + 1][prow] + 1;

            if (d >= d2) {
                d = d2;
                StartSubstrTrace = ppStartSubstrTrace[start + 1][prow];
            }

            ppDP[start + 1][row] = d;
            ppStartSubstrTrace[start + 1][row] = StartSubstrTrace;

            // If a substring trace starts from the current column, that means an empty substring.
            // Don't take an empty substring into account.
            if (d <= kerr && (!StartSubstrTrace || StartSubstrTrace == start + 1)) {
                CurrentActive = row;
            }
        }
        if (CurrentActive + 1 <= sz) {
            ppDP[start + 1][CurrentActive + 1] = kerr + 1;
            ppStartSubstrTrace[start + 1][CurrentActive + 1] = 0;
        }
        unsigned LastStartSubstrTrace = ppStartSubstrTrace[start + 1][sz];
        if (CurrentActive == sz && (!LastStartSubstrTrace || LastStartSubstrTrace == start + 1)) {
            aStr[start + 1] = 0;
            recv(aStr, start +1);
        } else if (CurrentActive + kerr >= start + 1 && start + 1 < sz + kerr) {
            GenSuperCondNGInternal(recv, start + 1, CurrentActive, aStr, pPat, sz, pABC, kerr, ppDP, ppStartSubstrTrace);
        }
    }
}

template <class TSuperCondReceiver>
void GenSuperCondNG(TSuperCondReceiver& recv, const string& sPat, const char* pABC, unsigned kerr) {
    unsigned        sz = sPat.size(); 
    const char*     pPat = sPat.c_str();

    unsigned        MaxLen = sz + kerr;
    unsigned        aDP                 [(MaxLen + 1) * (sz + 1)];
    unsigned        aStartSubstrTrace   [(MaxLen + 1) * (sz + 1)];

    char            aStr[MaxLen + 1];
    unsigned*       ppDP[MaxLen + 1];
    unsigned*       ppStartSubstrTrace[MaxLen + 1];

    unsigned*   p1 = aDP;
    unsigned*   p2 = aStartSubstrTrace;

    for (unsigned col = 0, *p = aDP; col <= MaxLen; 
                            ++col, 
                            p1 += sz + 1,
                            p2 += sz + 1) {
        ppDP[col]               = p1;
        ppStartSubstrTrace[col]   = p2;
    }
    for (unsigned row = 0; row <= min(kerr + 1, sz); ++row) {
        ppDP[0][row]                = row;
        ppStartSubstrTrace[0][row]  = 0;
    }

    GenSuperCondNGInternal(recv, 0, kerr, aStr, pPat, sz, pABC, kerr, ppDP, ppStartSubstrTrace);
}

#endif
