#ifndef GEN_FULLNG_H
#define GEN_FULLNG_H

#include <sys/types.h>

#include <string>
#include <algorithm>

#include "../Interface/interface.h"

typedef u_int32_t Uint;
template <class TNGReceiver>
void GenSuperCondNG_NFAInternal(TNGReceiver& recv, unsigned start, char aStr[], const char *pPat, unsigned sz, unsigned kerr, const char* pABC,
                                Uint aR[], unsigned StartState, Uint aS[], Uint aSMapArr[])
{
    aStr[start + 1] = 0;

    Uint        aS1[kerr + 1];
    Uint        aR1[kerr + 1];

    assert(sz);

    Uint        MaskOk = (Uint(1) << (sz - 1));

    for (const char* pLetter = pABC; *pLetter; ++pLetter) {
        Uint SMap = aSMapArr[(unsigned char) *pLetter];

        // Recalculate automaton states
        if (start) {
            // aS simulates an automaton that recognizes only a proper prefix.
            // That is why we skipped the first letter, i.e., start == 0
            aS1[0] = ((aS[0] << 1) | 1) & SMap;

            for (unsigned d = 1; d <= kerr; d++) {
                aS1[d]  = ((aS[d] << 1) | 1) & SMap;
                aS1[d] |= ((aS[d - 1] | aS1[d - 1]) << 1) | aS[d - 1] | 1;
            }
        } else {
            memcpy(aS1, aS, (kerr + 1) * sizeof *aS);
        }

        unsigned    StartState1 = StartState;

        aR1[StartState] = (((aR[StartState] << 1) | Uint(start <= StartState)) & SMap);


        bool bCanContinue = false;
        bool bFinal = false;

        if (start > StartState && !aR1[StartState]) {
            ++StartState1;
        }
        Uint CompleteNotSubstr = aR1[StartState] & (~aS1[StartState]);

        if (CompleteNotSubstr != 0) { bCanContinue = true; }
        if ((CompleteNotSubstr & MaskOk) != 0) { bFinal = true; }

        for (unsigned d = StartState + 1; d <= kerr; d++) {
            aR1[d] = (((aR[d] << 1) | Uint(start <= d)) & SMap);
            aR1[d] |= ((aR[d - 1] | aR1[d - 1]) << 1) | aR[d - 1] | (start <= d - 1);
           
            if (d == StartState1 && start >= d && !aR1[d]) {
                ++StartState1;
            } 
            CompleteNotSubstr = aR1[d] & (~aS1[d]);

            if (CompleteNotSubstr != 0) { bCanContinue = true; }
            if ((CompleteNotSubstr & MaskOk) != 0) { bFinal = true; }
        }

        if (bFinal) { // If final, we ignore bCanContinue, because we are not interested in matches that have proper prefixes in a  neighborhood
            aStr[start] = *pLetter;
            aStr[start + 1] = 0;
            recv(aStr, start + 1);
        } else if (bCanContinue && start + 1 < sz + kerr) {
            aStr[start] = *pLetter;

            GenSuperCondNG_NFAInternal(recv, start + 1, aStr, pPat, sz, kerr, pABC,
                                        aR1, StartState1, aS1, aSMapArr);
        }
    }
}

template <class TNGReceiver>
void GenSuperCondNG_NFA(TNGReceiver& recv, const string& sPat, const char* pABC, unsigned kerr) {
    unsigned        sz = sPat.size(); 
    const char*     pPat = sPat.c_str();
    char            aStr[sz + kerr + 1];

    if (!sz) {
        cerr << "Empty patterns are not supported!" << endl;
        return;
    }

    unsigned        aMap[256];
    Uint            aSMapArr[256];

    memset (aMap, 0, sizeof (aMap));
    memset (aSMapArr, 0, sizeof (aSMapArr));
    
    Uint             aTmpS[sz + 1];

    aTmpS[0] = 0;

    for (unsigned i = 0, cnt = 0; i < sz; i++) {
        unsigned l = aMap[(unsigned char) pPat[i]];

        if (!l) {
            l = aMap[(unsigned char) pPat[i]] = ++cnt;
            aTmpS[l] = 0;
        }
        aTmpS[l] |= Uint(1) << i;
    }
    for (unsigned c = 0; c <= 255; ++c) {
        aSMapArr[c] = aTmpS[aMap[c]];
    }

    Uint    aR[kerr + 1];
    Uint    aS[kerr + 1];

    for (unsigned i = 0; i <= kerr; ++i) {
        aS[i] = aR[i] = (Uint(1) << i) - 1;
    }

    GenSuperCondNG_NFAInternal(recv, 0, aStr, pPat, sz, kerr, pABC, aR, 0, aS,aSMapArr);
}

#endif
