#include "gensupercond.h"
#include "genfullng.h"
#include "hashport.h"
#include <iostream>
#include <string>

#define REP_QTY 10

using namespace std;

static void ExpandAndSearch(int start,
                        int &numComp, const StringHashSet & dict,
                        string & str, 
                        bool hasTransp, int kerr, 
                        StringList *resList,
                        const char *ABC)
{
    if (1) {
        ++numComp;
#if 0
        if (GetEnvSwitch("SHOW_VARIANTS")) {
            std::cout << "Variant: " << str << std::endl;
        }
#endif
        if (dict.find(str) != dict.end()) {
            if (resList) { resList->push_back (str); }
        }
    }

    int size = str.size ();
    if (!kerr || start > size) return;

    int i = 0;
    // Substitution
    for (i = start; i < size; i++) {
        char origChar = str[i];
        for (const char *ptr = ABC; *ptr; ptr++) {
            if (*ptr != str[i]) {
                str[i] = *ptr;
                ExpandAndSearch(i + 1, numComp, dict, str, hasTransp, kerr - 1, resList, ABC);
            }
        }
        str[i] = origChar;
    }
    // We well try to make as few insertions and erases as possible
    // It is enough to make 1 insertion, 1 deletion per insertion case
    // and 1 insertion, 1 deletion per deletion case
    // Insertion
    str.insert(str.begin() + start, ' ');
    for (i = start; i <= size; i++) {
        for (const char *ptr = ABC; *ptr; ptr++) {
            str[i] = *ptr; 
            ExpandAndSearch(i + 1, numComp, dict, str, hasTransp, kerr - 1, resList, ABC);
        }
        if (i < size) {
            swap(str[i], str[i + 1]);
        }
    }
    str.erase(str.end() - 1);
    if (hasTransp) {
    // Transposition
        for (i = start; i < size - 1; i++) {
            swap(str[i], str[i+1]);
            ExpandAndSearch(i + 2, numComp, dict, str, hasTransp, kerr - 1, resList, ABC);
            swap(str[i], str[i+1]);
        }
    }
    // Deletion
    if (start >= size) return;
    char origChar = str[start];
    str.erase(str.begin() + start);
    for (i = start; i < size; i++) {
        ExpandAndSearch(i, numComp, dict, str, hasTransp, kerr - 1, resList, ABC);
        if (i < size - 1) {
           swap(str[i], origChar);
        }
    }
    str.insert (str.end(), origChar);

}

struct NGPrint {
    void operator()(const string& s) {
        cout << s << endl;
    }
};

struct NGCollect {
    void operator()(const string& s) {
        mLst.push_back(s);
    }
    StringList   mLst;
};

void Usage(const char* pProg) {
    cerr << pProg << " <string> <kerr> <print Ng?>" << endl;
}

int main(int argc, char* pArgv[]) {
    if (argc != 4) {
        Usage(pArgv[0]);
        return -1;
    }
    string   sPat       = pArgv[1];
    unsigned kerr       = atol(pArgv[2]);
    bool     bPrintNG   = atol(pArgv[3]);
    bool     bHasTransp = false;

    cout << "String: '" << sPat << "' kerr = " << kerr << " HasTransp: " << bHasTransp << endl;


    string  sAlphaName;

    GetEnv(ENV_ABC, sAlphaName);

    EAlphabet           AlphaType = GetAlphabetType(sAlphaName);

    if (AlphaType == eAlphaUnknown) {
        cerr << "Unknown alphabet: '" << sAlphaName << "' check env var " << ENV_ABC << endl;
        return 1;
    }
    const char* pABC = GetAlphabet(AlphaType);

    NGCollect ngc;

    DECLARE_STATIC_COUNTER(GenTime);

    START_COUNTER(GenTime);

    GenFullNG(ngc, sPat, pABC, kerr);

    STOP_COUNTER(GenTime);

    cout << "Time spent: " << GET_COUNTER(GenTime) << " microsecs " << endl;

    StringHashSet   dict;


    RESET_COUNTER(GenTime);
    START_COUNTER(GenTime);

    GenCondNG(eFull, pABC, sPat, kerr, bHasTransp, dict);

    STOP_COUNTER(GenTime);

    cout << "Time spent (old algo): " << GET_COUNTER(GenTime) << " microsecs " << endl;


    RESET_COUNTER(GenTime);
    START_COUNTER(GenTime);

    StringList  ResList;
    int         numComp = 0;
        //ngc.mLst.clear();

    for (unsigned i = 0; i < REP_QTY; ++i) {


        ResList.clear();

        ExpandAndSearch(0,
                    numComp, dict,
                    sPat,
                    false, kerr, 
                    &ResList, pABC);
    }

    STOP_COUNTER(GenTime);

    cout << "Time spent (ExpandAndSearch ): " << GET_COUNTER(GenTime) / REP_QTY<< " microsecs, numComp " << numComp <<  " RepQty " << REP_QTY << endl;

    unsigned nUnfound1 = 0, nUnfound2 = 0;

    cout << "======================================" << endl;
    if (bPrintNG) {
    cout << ngc.mLst.size() << " strings in NG " << endl;
        for (unsigned i = 0; i < ngc.mLst.size(); ++i) {
            cout << ngc.mLst[i] << endl;
        }
    }

    cout << "======================================" << endl;

    for (unsigned i = 0; i < ngc.mLst.size(); ++i) {
        const string& w = ngc.mLst[i];

        if (dict.find(w) == dict.end()) {
            cout << "Unfound (1) string: " << w << endl;
            nUnfound1++;
        }

    }

    StringHashSet ListHash(ngc.mLst.begin(), ngc.mLst.end());

    for (StringHashSet::const_iterator it = dict.begin(); it != dict.end(); ++it) {
        const string& w = *it;

        if (ListHash.find(w) == ListHash.end()) {
         cout << "Unfound (2) string: " << w << endl;
         nUnfound2++;
        }
    }

    cout << ngc.mLst.size() << " -> " << dict.size() << endl;

    if (ngc.mLst.size() == dict.size() && !nUnfound1 && !nUnfound2) {
        cout << "FULL MATCH!" << endl;
    } else {
        cout << "Incomplete MATCH, have unfound1 = " << nUnfound1 << " unfound2 = " << nUnfound2 << endl;
    }
}
