
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

#define LBFIX "Leonid Boytsov's changes"
#define ulong unsigned long

ulong HSIZE; /* 21433 */

#define free(p) { if ((p) != 0) free(p); }
#define malloc(n)    ((n) == 0 ? 0 : malloc(n))
#define realloc(p,n) ((p) == NULL ? malloc(n) : realloc(p,n))

#define CTRL(p) { if ((p)==NULL) { fprintf (stderr,"out of memory\n"); exit(1); }}

typedef unsigned char byte;
/* typedef unsigned long ulong; */

static void step (register byte *c, int m, int k, 
		  register byte *p, register byte cc)

    { register int pc,nc,j,lcol;
      pc = 0; nc = 1; lcol = c[0];
      for (j=1;j<=lcol;j++)
         { if (c[j] < nc) nc = c[j];
           pc += (cc != p[j]);
           if (pc <= nc) nc = pc; else nc++;
           pc = c[j]; c[j] = nc;
         }
      while (c[lcol] > k) lcol--;
      if (lcol < m) c[++lcol] = k+1;
      c[0] = lcol;
    }

ulong hash (register byte *c)

   { register int i,lcol;
     register ulong h=0;
     lcol = c[0];
     for (i=0;i<=lcol;i++) h = 1013*h+c[i];
     return h % HSIZE;
   }

typedef struct slist
   { int pos;
     struct slist *next;
   } *list;

typedef list* htable;

htable hcreate (void)
   { htable h;
     CTRL(h = (htable)malloc(HSIZE*sizeof(list))); 
     memset (h,0,HSIZE*sizeof(list));
     return h;
   }

int igual (register byte *c1, register byte *c2)
   { register int i = c1[0];
     for (i=c1[0];i>=0;i--) if (c1[i] != c2[i]) return 0;
     return 1;
   }

typedef struct
   { byte *c;
     int *vec;
     int f;
   } nodo;

typedef struct
   { nodo *nodos;
     ulong size;
     byte S[256];
     int psigma;
     byte *sigma;
   } grafo;

int hadd (htable h, byte **c, int k, grafo *G)
   { ulong hpos = hash (*c);
     int j;
     list lo = NULL,l = h[hpos];
     nodo *N;
     while (l != NULL) 
	{ if (igual (G->nodos[l->pos].c,*c)) 
	     { if (lo != NULL)
		  { lo->next = l->next;
		    l->next = h[hpos];
		    h[hpos] = l;
		  }
	       return l->pos;
	     }
	  lo = l; l = l->next;
	}
     CTRL(l = (list) malloc (sizeof(struct slist))); 
     l->pos = G->size;
     l->next = h[hpos];
     h[hpos] = l;
     if ((G->size % 50) == 0)
        CTRL(G->nodos = (nodo*) realloc (G->nodos, 
					 (G->size+50) * sizeof(nodo))); 
     G->size++;
     N = &G->nodos[l->pos];
     N->c = *c; *c = NULL;
     CTRL(N->vec = (int*)malloc (G->psigma*sizeof(int)));
     for (j=0;j<G->psigma;j++) N->vec[j] = -1;
     N->f = (N->c[N->c[0]] <= k);
     return l->pos;
   }

int hfree (htable h)

   { int i;
     list l;
     for (i=0;i<HSIZE;i++)
	{ l = h[i];
	  while (l != NULL)
	    { list ol = l;
	      l = l->next;
	      free (ol);
	    }
	}
     free (h);
   }

htable H;
grafo G;
int m,k;
byte *pat;
byte *C;

void mkautom (void)

   { int i,j;

     for (i=0;i<256;i++) G.S[i] = 0;
     for (i=0;i<=m;i++) G.S[pat[i]]=1;
     G.psigma = 0; for (i=0;i<256;i++) G.psigma += G.S[i];
     CTRL(G.sigma = (byte*) malloc (G.psigma));
     G.psigma = 0; 
     for (i=0;i<256;i++) 
	if (G.S[i]) { G.S[i] = G.psigma; G.sigma[G.psigma++] = i; }

     CTRL(C = (byte*) malloc (m+1));
     C[0] = k+1;
     for (i=1;i<=k+1;i++) C[i] = i;
     H = hcreate();
     G.nodos = NULL; G.size = 0;
     hadd (H,&C,k,&G);
     CTRL(C = (byte*) malloc (m+1)); 
   }

int expand (int state, int pos)

   { int nstate;
     nodo *N = &G.nodos[state];
     memcpy (C,N->c,N->c[0]+1);
     step (C,m,k,pat,G.sigma[pos]);
     nstate = N->vec[pos] = hadd (H,&C,k,&G);
     if (C == NULL) CTRL(C = (byte*) malloc (m+1)); 
     return nstate;
   }

int main (int argc, char **argv)

   { int i,n,pos;
     byte *text;
     register byte *ptr;
     struct stat st;
     int matches;
     FILE *f;
     register int state,nstate;
     if (argc != 4)
	{ fprintf (stderr,"usage: %s <pat> k <file>\n",argv[0]);
	  exit(1);
	}
     pat = argv[1];
     m = strlen(pat); k = atoi(argv[2]);

#ifdef LBFIX
     if (m > 20) HSIZE = ~0;
     else { HSIZE = 1;
            for (i=1;i<m;i++) HSIZE *= 3;
            }
#else
     if (k > 20) HSIZE = ~0;
     else { HSIZE = 1;
            for (i=1;i<k;i++) HSIZE *= 3;
	  }
#endif
          

     if (stat (argv[3],&st) != 0) { fprintf (stderr,"%s not found\n",argv[3]); exit(1); }
     n = st.st_size;
     printf("File size: %d bytes\n", n);
     if (HSIZE > n) HSIZE = n;
     CTRL (text = (byte*)malloc (n+1));
     f = fopen (argv[3],"r");
     fread (text,n,1,f);
     fclose (f);
     text[n] = 0;

     HSIZE /= 10; if (HSIZE > 1024*1024) HSIZE = 1024*1024;
     HSIZE--;
#ifdef LBFIX
     if (HSIZE < 1024) {
        HSIZE = 1024;
     }
#endif
printf ("hsize = %i\n",HSIZE);
     mkautom ();
     pat--;
     state=0; matches = 0;
     
     for (ptr=text;*ptr;ptr++)
	{ pos = G.S[*ptr];
	  nstate = G.nodos[state].vec[pos];
	  state = (nstate != -1) ? nstate : expand (state,pos);
	  if (G.nodos[state].f) matches++;
	   if (1)  { byte *p;
	       printf ("%i: ...",i);
	       p = ptr-10; if (p<text) p=text;
	       while (p<ptr) putchar(*p++);
	       printf ("\n");
	     }
	}
     printf ("%i estados\n",G.size);
     printf ("%i matches\n",matches);
     exit(0);
   }
