
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <sys/stat.h>

#define ulong unsigned long
#define OUTPUT_WHOLE_LINE /* Output the whole line where a match is found */
#define PAT_MAX_LEN 20

ulong HSIZE; /* 21433 */

#define free(p) { if ((p) != 0) free(p); }
#define malloc(n)    ((n) == 0 ? 0 : malloc(n))
#define realloc(p,n) ((p) == NULL ? malloc(n) : realloc(p,n))

#define CTRL(p) { if ((p)==NULL) { fprintf (stderr,"out of memory\n"); exit(1); }}
#define IS_CR(c) ((c) == '\n' || (c) == '\r')

typedef unsigned char byte;
/* typedef unsigned long ulong; */

/* 
 * Calculates the next column of a dynamic programming matrix:
 * it considers only essential values, i.e. values <= k.
 */
static void step (register byte *c, int m, int k, 
		  register byte *p, register byte cc)

    { register int pc,nc,j,lcol;
      pc = c[1]; nc = pc + 1; lcol = c[0];
      for (j=2;j<=lcol+1;j++)
         { if (c[j] < nc) nc = c[j];
           pc += (cc != p[j- 1]);
           if (pc <= nc) nc = pc; else ++nc;
           pc = c[j]; c[j] = nc;
         }
      /* Watch out for k, it can be zero! */
      while (lcol > 1 && c[lcol + 1] > k) lcol--;
      if (lcol < m) {
        c[lcol + 2] = k+1;
        ++lcol;
      }
      c[0] = lcol;
      c[1]++;
    }

ulong hash (register byte *c)

   { register int i,lcol;
     register ulong h=0;
     lcol = c[0];
     for (i=0;i<=lcol + 1;i++) h = 1013*h+c[i];
     return h % HSIZE;
   }

/* Hash table overflow list */
typedef struct slist
   { int pos;
     struct slist *next;
   } *list;

typedef list* htable;

htable hcreate (void)
   { htable h;
     CTRL(h = (htable)malloc(HSIZE*sizeof(list))); 
     memset (h,0,HSIZE*sizeof(list));
     return h;
   }

/* Compares automaton states */
int equal (register byte *c1, register byte *c2)
   { register int i = c1[0] + 1;
     for (i=c1[0] + 1;i>=0;i--) if (c1[i] != c2[i]) return 0;
     return 1;
   }

/* Automaton state */
typedef struct
   { byte *c;    /* c[0] - the number of distance values (see next comment line for i > 0 
                    c[1] - the number of text character from the previous end of line.
                    i > 1, c[i] = EditDist(c[1] characters of the text from the previous end of line, pat[0..i - 2] */
     int *vec;   /* Transitions */
     int f;      /* Final state? */
   } nodo;

typedef struct
   { nodo *nodos; /* Automaton states */
     ulong size;  /* Number of states */
     byte S[256]; /* Maps from the ASCII character to reduced alphabet (consisting only of pattern characters
                   * plus a single character that represents any character not in pattern 
                   */
     int psigma;  /* The number of unique pattern characters        */
     byte *sigma; /* Maps from reduced alphabet to ASCI characters  */
   } grafo;

int hadd (htable h, byte **c, int k, grafo *G)
   { ulong hpos = hash (*c);
     int j;
     list lo = NULL,l = h[hpos];
     nodo *N;
     while (l != NULL) 
	{ if (equal (G->nodos[l->pos].c,*c)) 
    /* This moves list element l to the beginning of the list */
	     { if (lo != NULL)
		  { lo->next = l->next;
		    l->next = h[hpos];
		    h[hpos] = l;
		  }
	       return l->pos;
	     }
	  lo = l; l = l->next;
	}
    /* If not found, insert in the beginning of the list */
     CTRL(l = (list) malloc (sizeof(struct slist))); 
     l->pos = G->size;
     l->next = h[hpos];
     h[hpos] = l;

    /* Reallocate after every 50 insertions */
     if ((G->size % 50) == 0)
        CTRL(G->nodos = (nodo*) realloc (G->nodos, 
					 (G->size+50) * sizeof(nodo))); 

     G->size++;
     N = &G->nodos[l->pos];
     N->c = *c; *c = NULL;
     CTRL(N->vec = (int*)malloc (G->psigma*sizeof(int)));
     for (j=0;j<G->psigma;j++) N->vec[j] = -1;
     N->f = (N->c[N->c[0] + 1] <= k);
     return l->pos;
   }

htable H;   /* Hash table for quick search for automaton states */
grafo G;    /* Automaton states                 */
int m,      /* Pattern size                     */
    k;      /* Maximal allowed number of errors */
byte *pat;  /* Pattern                          */
byte *C;

void mkautom (void)

   { int i,j;

     for (i=0;i<256;i++) G.S[i] = 0;
     for (i=0;i<=m;i++) G.S[pat[i]]=1;
     G.psigma = 0; for (i=0;i<256;i++) G.psigma += G.S[i];
     /* 
      * The alphabet consists of pattern characters, 
      * plus one character that represents any character, 
      * which is not in the pattern.
      * Extra alphabet symbol is represented by a zero-code character.
      */
     CTRL(G.sigma = (byte*) malloc (G.psigma + 1));
     G.psigma = 0; 
     for (i=0;i<256;i++) 
	if (G.S[i]) { G.S[i] = G.psigma; G.sigma[G.psigma++] = i; }

     CTRL(C = (byte*) malloc (m+2));
     assert(k < m);
     C[0] = k+1;
     C[1] = 0;
     for (i=1;i<=k+1;i++) C[i + 1] = i;
     H = hcreate();
     G.nodos = NULL; G.size = 0;

     hadd (H,&C,k,&G);
     CTRL(C = (byte*) malloc (m+2)); 
   }

int expand (int state, int pos)

   { int nstate;
     nodo *N = &G.nodos[state];
     memcpy (C,N->c,N->c[0]+2);
     step (C,m,k,pat,G.sigma[pos]);
     nstate = N->vec[pos] = hadd (H,&C,k,&G);
     if (C == NULL) CTRL(C = (byte*) malloc (m+2)); 
     return nstate;
   }

int main (int argc, char **argv)

   { int i,n,pos;
     byte *text;
     register byte *ptr, *beg;
     struct stat st;
     int matches;
     FILE *f;
     register int state,nstate;
     if (argc != 4)
	{ fprintf (stderr,"usage: %s <pat> k <file>\n",argv[0]);
	  exit(1);
	}
     pat = argv[1];
     m = strlen(pat); k = atoi(argv[2]);

     if (k >= m) {
        fprintf(stderr, "Pattern length should be > than k");
        exit(1);
     }

     if (m > PAT_MAX_LEN) {
        fprintf(stderr, "Warning: pattern is longer than %d chars, truncating pattern\n", PAT_MAX_LEN);
        pat[PAT_MAX_LEN] = 0;
     } 

     HSIZE = 1;
     for (i=1;i<m;i++) HSIZE *= 3;

     if (stat (argv[3],&st) != 0) { fprintf (stderr,"%s not found\n",argv[3]); exit(1); }
     n = st.st_size;
     if (HSIZE > n) HSIZE = n;
     CTRL (text = (byte*)malloc (n+1));
     f = fopen (argv[3],"r");
     fread (text,n,1,f);
     fclose (f);
     text[n] = 0;

     HSIZE /= 10; if (HSIZE > 1024*1024) HSIZE = 1024*1024;
     if (HSIZE < 21433) {
        HSIZE = 21433;
     }
     printf ("Initial hash table size (hashize) = %i\n",HSIZE);
     mkautom ();
     pat--;
     state=0; matches = 0;
     
     for (beg = ptr=text;*ptr;ptr++)
	{ 
          if (IS_CR(*ptr))
          {
              beg = ptr + 1;
              state = 0;
              continue;
          }
          pos = G.S[*ptr];
	  nstate = G.nodos[state].vec[pos];
	  state = (nstate != -1) ? nstate : expand (state,pos);
	  if (G.nodos[state].f && (!ptr[1] || IS_CR(ptr[1]))) {
         matches++;
	     if (1) 
         { 
               byte *p = beg;
               byte *e = ptr;
               while (*e && !IS_CR(*e)) { ++e; }
	       while (p<e) putchar(*p++);
	       printf ("\n");
	     } 
      }
	}
     printf ("%i automaton states\n",G.size);
     printf ("%i matches\n",matches);
     exit(0);
   }
