
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <sys/stat.h>

#define ulong unsigned long
#define OUTPUT_WHOLE_LINE /* Output the whole line where a match is found */

ulong HSIZE; /* 21433 */

#define free(p) { if ((p) != 0) free(p); }
#define malloc(n)    ((n) == 0 ? 0 : malloc(n))
#define realloc(p,n) ((p) == NULL ? malloc(n) : realloc(p,n))

#define CTRL(p) { if ((p)==NULL) { fprintf (stderr,"out of memory\n"); exit(1); }}

typedef unsigned char byte;
/* typedef unsigned long ulong; */

/* 
 * Calculates the next column of a dynamic programming matrix:
 * it considers only essential values, i.e. values <= k.
 */
static void step (register byte *c, int m, int k, 
		  register byte *p, register byte cc)

    { register int pc,nc,j,lcol;
      pc = 0; nc = 1; lcol = c[0];
      for (j=1;j<=lcol;j++)
         { if (c[j] < nc) nc = c[j];
           pc += (cc != p[j]);
           if (pc <= nc) nc = pc; else nc++;
           pc = c[j]; c[j] = nc;
         }
      /* Watch out for k, it can be zero! */
      while (lcol && c[lcol] > k) lcol--;
      if (lcol < m) c[++lcol] = k+1;
      c[0] = lcol;
    }

ulong hash (register byte *c)

   { register int i,lcol;
     register ulong h=0;
     lcol = c[0];
     for (i=0;i<=lcol;i++) h = 1013*h+c[i];
     return h % HSIZE;
   }

/* Hash table overflow list */
typedef struct slist
   { int pos;
     struct slist *next;
   } *list;

typedef list* htable;

htable hcreate (void)
   { htable h;
     CTRL(h = (htable)malloc(HSIZE*sizeof(list))); 
     memset (h,0,HSIZE*sizeof(list));
     return h;
   }

/* Compares automaton states */
int equal (register byte *c1, register byte *c2)
   { register int i = c1[0];
     for (i=c1[0];i>=0;i--) if (c1[i] != c2[i]) return 0;
     return 1;
   }

/* Automaton state */
typedef struct
   { byte *c;    /* c[0] - the number of distance values (see next comment line for i > 0 
                    i > 0, c[i] = BestEditDist(suffix of the text, pat[0..i - 1] */
     int *vec;   /* Transitions */
     int f;      /* Final state? */
   } nodo;

typedef struct
   { nodo *nodos; /* Automaton states */
     ulong size;  /* Number of states */
     byte S[256]; /* Maps from the ASCII character to reduced alphabet (consisting only of pattern characters
                   * plus a single character that represents any character not in pattern 
                   */
     int psigma;  /* The number of unique pattern characters        */
     byte *sigma; /* Maps from reduced alphabet to ASCI characters  */
   } grafo;

int hadd (htable h, byte **c, int k, grafo *G)
   { ulong hpos = hash (*c);
     int j;
     list lo = NULL,l = h[hpos];
     nodo *N;
     while (l != NULL) 
	{ if (equal (G->nodos[l->pos].c,*c)) 
    /* This moves list element l to the beginning of the list */
	     { if (lo != NULL)
		  { lo->next = l->next;
		    l->next = h[hpos];
		    h[hpos] = l;
		  }
	       return l->pos;
	     }
	  lo = l; l = l->next;
	}
    /* If not found, insert in the beginning of the list */
     CTRL(l = (list) malloc (sizeof(struct slist))); 
     l->pos = G->size;
     l->next = h[hpos];
     h[hpos] = l;

    /* Reallocate after every 50 insertions */
     if ((G->size % 50) == 0)
        CTRL(G->nodos = (nodo*) realloc (G->nodos, 
					 (G->size+50) * sizeof(nodo))); 

     G->size++;
     N = &G->nodos[l->pos];
     N->c = *c; *c = NULL;
     CTRL(N->vec = (int*)malloc (G->psigma*sizeof(int)));
     for (j=0;j<G->psigma;j++) N->vec[j] = -1;
     N->f = (N->c[N->c[0]] <= k);
     return l->pos;
   }

/*
 * Isn't used. Anyway, memory is freed, when the program terminates.
 */
/*
int hfree (htable h)

   { int i;
     list l;
     for (i=0;i<HSIZE;i++)
	{ l = h[i];
	  while (l != NULL)
	    { list ol = l;
	      l = l->next;
	      free (ol);
	    }
	}
     free (h);
   }
*/

htable H;   /* Hash table for quick search for automaton states */
grafo G;    /* Automaton states                 */
int m,      /* Pattern size                     */
    k;      /* Maximal allowed number of errors */
byte *pat;  /* Pattern                          */
byte *C;

void mkautom (void)

   { int i,j;

     for (i=0;i<256;i++) G.S[i] = 0;
     for (i=0;i<=m;i++) G.S[pat[i]]=1;
     G.psigma = 0; for (i=0;i<256;i++) G.psigma += G.S[i];
     /* 
      * The alphabet consists of pattern characters, 
      * plus one character that represents any character, 
      * which is not in the pattern.
      * Extra alphabet symbol is represented by a zero-code character.
      */
     CTRL(G.sigma = (byte*) malloc (G.psigma + 1));
     G.psigma = 0; 
     for (i=0;i<256;i++) 
	if (G.S[i]) { G.S[i] = G.psigma; G.sigma[G.psigma++] = i; }

     CTRL(C = (byte*) malloc (m+1));
     assert(k < m);
     C[0] = k+1;
     for (i=1;i<=k+1;i++) C[i] = i;
     H = hcreate();
     G.nodos = NULL; G.size = 0;

     hadd (H,&C,k,&G);
     CTRL(C = (byte*) malloc (m+1)); 
   }

int expand (int state, int pos)

   { int nstate;
     nodo *N = &G.nodos[state];
     memcpy (C,N->c,N->c[0]+1);
     step (C,m,k,pat,G.sigma[pos]);
     nstate = N->vec[pos] = hadd (H,&C,k,&G);
     if (C == NULL) CTRL(C = (byte*) malloc (m+1)); 
     return nstate;
   }

int main (int argc, char **argv)

   { int i,n,pos;
     byte *text;
     register byte *ptr;
     struct stat st;
     int matches;
     FILE *f;
     register int state,nstate;
     if (argc != 4)
	{ fprintf (stderr,"usage: %s <pat> k <file>\n",argv[0]);
	  exit(1);
	}
     pat = argv[1];
     m = strlen(pat); k = atoi(argv[2]);

     if (k >= m) {
        fprintf(stderr, "Pattern length should be > than k");
        exit(1);
     }

     if (m > 20) {
        fprintf(stderr, "Pattern is longer than 20 chars\n");
        exit(1);
     } else { HSIZE = 1;
            for (i=1;i<m;i++) HSIZE *= 3;
	  }

     if (stat (argv[3],&st) != 0) { fprintf (stderr,"%s not found\n",argv[3]); exit(1); }
     n = st.st_size;
     if (HSIZE > n) HSIZE = n;
     CTRL (text = (byte*)malloc (n+1));
     f = fopen (argv[3],"r");
     fread (text,n,1,f);
     fclose (f);
     text[n] = 0;

     HSIZE /= 10; if (HSIZE > 1024*1024) HSIZE = 1024*1024;
     if (HSIZE < 21433) {
        HSIZE = 21433;
     }
     printf ("Initial hash table size (hashize) = %i\n",HSIZE);
     mkautom ();
     pat--;
     state=0; matches = 0;
     
     for (ptr=text;*ptr;ptr++)
	{ pos = G.S[*ptr];
	  nstate = G.nodos[state].vec[pos];
	  state = (nstate != -1) ? nstate : expand (state,pos);
	  if (G.nodos[state].f) {
         matches++;
	     if (1) 
         { 
           byte *p;
	       /* printf ("%i: ...",i); */
	       p = ptr-10; if (p<text) p=text;
	       while (p<ptr) putchar(*p++);
	       printf ("\n");
	     } 
      }
	}
     printf ("%i automaton states\n",G.size);
     printf ("%i matches\n",matches);
     exit(0);
   }
