#include <iostream>
#include <sstream>
#include <fstream>
#include "index.h"

using namespace std;

#include "calchash.h"
#include "rstat.h"

typedef vector<string> StringList;

const HMap &getHmap(const StringList & dictList)
{
    static bool   init;
    static HMap       hsignOpt;

    if (GetEnvSwitch("OPTIMIZE_BY_STAT")) {
        cout << "Using optimized\n";
        if (init) return hsignOpt;
        CollectStat st;

        st.CollectStringStat(dictList);
    	if (!CalcHMap (GetMaxDigit(), st.stat (), hsignOpt, GetEnvSwitch("PRINT_ADD_STAT"))) {
            std::cerr << "Hash generation failed!" << std::endl;
        }
        init = true;
        return hsignOpt;
    } else {
        cout << "Using standard\n";
        return getStandHmap();
    }
    init = true;
    return hsignOpt;
}


int test(unsigned int gramSize, unsigned int maxIntersectCount, const StringList & TestList, 
         bool hasTransp, const StringList &dictList, unsigned int testCutoff = 10)
{


    CNgramFiltIndex   indx(gramSize, maxIntersectCount, getHmap(dictList), hasTransp);

    for (unsigned int i = 0; i < TestList.size(); ++i)    
        indx.insertWord(TestList[i]);

    unsigned int     tcnt = 0;
    unsigned int     succ1 = 0, succ2 = 0;

    for (unsigned int flag = 0; flag <= 4; flag = (flag) ? flag<<1: 1) {
    cout << endl << "Flag = " << flag << ": " << endl;
    for (unsigned int kerr = 0; kerr < 3; ++kerr) {
        for (unsigned int i = 0; i < TestList.size (); ++i) {
            string s0 = TestList[i];
            StringList res;

            if (kerr > 1 && i > MAX_FUNC_TEST_SIZE)
                break;

            StringList lst;
            //cout<<" Filling expansion list for "<<s0<<" kerr = "<<kerr<<" ...";
            cout << ".";
            cout.flush ();
            FillExpansionList (s0, hasTransp, kerr, &lst, EnglishABC, kerr,
                               kerr == 2, 0, false);
            //cout<<" done, size "<<lst.size ()<<endl;

            if (kerr > 1 && lst.size () > testCutoff) {
                lst.erase (lst.begin () + testCutoff, lst.end ());
            }
            //cout<<" final size: "<<lst.size ()<<endl;

            for (StringList::iterator it = lst.begin ();
                 it != lst.end (); it++) {
                const string &    s = *it;
                unsigned int      cnt;


                tcnt++;

                indx.search(s, kerr, res, cnt, flag);
                int f = 0;
                for (int j = 0; j < res.size (); j++) {
                    if (res[j] == s0) {
                        f = 1;
                        break;
                    }
                }
                succ1 += f;
                if (f == 0) {
                    cerr << endl << " flag = " << flag << " '" << s0 << "' -> '" << s <<
                        "' kerr = " << kerr << " unfound" << endl;
                }
            }
        }
    }
    }
    bool OkFlag = (succ1 == tcnt);
    cout << endl;
    cout << "Functional test result for " << tcnt << " queries :" <<
        ((OkFlag) ? "Ok" : "with errors") << endl;
    if (!OkFlag) {
        cout << "search succeeded " << succ1 << " of " << tcnt
            << endl;
    }
    cout.flush ();
    return 0;
}

int main(int argc, char * argv[])
{
    if (argc != 6) {
       std::cerr << "Usage: " << argv[0] << " [n-gram size] [max intersect count] [test set file] [test dict file] [transposition flags]\n";
       return -1;
    }
    unsigned int gramSize = atol(argv[1]);
    unsigned int maxIntersectCount = atol(argv[2]);


    string            s;
    ifstream          ftest(argv[3]);
    bool              hasTransp(atol(argv[5]));


    StringList  testStrings;

    while (ftest >> s) {
        testStrings.push_back(s);
    }

    const char *pDictFile = argv[4];

    StringList dictList;

    ifstream  f(pDictFile);
    if (!f) {
       cerr << "Cannot open '" << pDictFile << "'\n";
       return 1;
    }
    while (f >> s) {
        dictList.push_back(s);
    }

    CNgramFiltIndex   indx(gramSize, maxIntersectCount, getHmap(dictList), hasTransp);

    if (!GetEnvSwitch("SKIP_FUNC_TEST") && test(gramSize, maxIntersectCount, testStrings, hasTransp, dictList)) {
        return -1;
    }
    for (int i = 0; i < dictList.size(); ++i) {
        indx.insertWord(dictList[i]);
    }
    cout << "Index is created\n";
    char               buf[1024];
    unsigned int       kerr;
    unsigned int       flags;
    unsigned int       cnt;
#if 0
    cin.clear();
    indx.printAll();
    cout << "Input word, kerr and flags: " << endl;
    while(cin.getline(buf, sizeof buf)) {
        StringList        res;
        
        stringstream str(buf);

        str >> s >> kerr >> flags;

        indx.search(s, kerr, res, cnt, flags);
        cout << "Result, " << " cnt to check " << cnt << " : " << endl;
        for (unsigned int i = 0; i < res.size(); ++i)
            cout << res[i] << endl;
    }
#else
    long long cntall[3][8];

    cout << "Signature size: " << GetMaxDigit() << endl;

    for (unsigned int kerr = 0; kerr < 3; ++kerr)
    for (unsigned int flags = 0; flags < 8; ++flags) {
        cntall[kerr][flags] = 0;
        cout << "Kerr: " << kerr << " flags: " << flags << endl;
        for (unsigned int i = 0; i < testStrings.size(); ++i) {
            StringList        res;
            unsigned int      cnt;

            indx.search(testStrings[i], kerr, res, cnt, flags);
            cntall[kerr][flags] += cnt;
        }
    }
    cout << "Index size: " << indx.size() << endl;
    for (unsigned int kerr = 0; kerr < 3; ++kerr) {
        cout << "Kerr = " << kerr << endl;
        for (unsigned int flags = 0; flags < 8; ++flags) {
            cout << setw(7) << flags << " ";
        }
        cout << endl << "================================================================================" << endl;
        for (unsigned int flags = 0; flags < 8; ++flags) {
            cout << setw(7) << setprecision(4) << (100.0 * cntall[kerr][flags] / (indx.size() * testStrings.size())) << " ";
        }
        cout << endl << endl;
    }
#endif
    
}
