#include "index.h"

#include <fstream>

CFlamingoNgramIndex::CFlamingoNgramIndex (bool hasTransp, EAlphabet AlphaType): CInterface(hasTransp, AlphaType), 
    mNgramLen(4), 
    mFanout(10), 
    mpIndexer(NULL),
    mpSimMetric(NULL), 
    mStrContVect(true) // true indicates statistics gathering, e.g. for auto part filtering

{
    GetEnv("NGRAM_LEN", mNgramLen);
    cout << "Qgram size: " << mNgramLen << endl;
    GetEnv("FANOUT", mFanout);
    cout << "Fanout: " << mFanout << endl;

}

CFlamingoNgramIndex::~CFlamingoNgramIndex () {
    delete mpGramGen;
    delete mpSimMetric;
    delete mpIndexer;
}

bool CFlamingoNgramIndex::Insert (const char *str) 
{
    return false;
}

bool CFlamingoNgramIndex::SearchInDB (const char *str, int kerr, int *numComp,
                              StringList * resList, bool silent)
{
    cerr << "SearchInDB is not supported.\n";
    return false;
}

bool CFlamingoNgramIndex::OpenDB (const char *name, bool create)
{
    return true;
}

bool CFlamingoNgramIndex::CloseDB ()
{
    return true;
}


bool CFlamingoNgramIndex::DumpDB (const char *name)
{
    ofstream out(name);

    if (!out) {
        string err = string("Cannot create file '") + name + "'";
        Perror(err);
        return false;
    }
    return true;
}

bool CFlamingoNgramIndex::CreateInMemIndex (const char *inputFile, int insertCount, int *dcnt, bool silent)
{
    int    cnt = 0;
    string s;

    ifstream inp(inputFile);

    while (getline(inp, s)) {
        ++cnt;
    }

    if (inp.bad()) {
        Perror("Cannot read from the file: " + string(inputFile));
        return false;
    }

    if (dcnt) *dcnt = cnt;


    int MaxStrLength = 32;
    
    mpGramGen = new GramGenFixedLen(mNgramLen);

    mpIndexer = new FtIndexerMem<>(&mStrContVect, mpGramGen, MaxStrLength, mFanout);

    bool bLenFilt = GetEnvSwitch("LEN_FILT");
    bool bCharSumFilt = GetEnvSwitch("CHARSUM_FILT");
    if (bLenFilt) {
        cout << "Using a length filter" << endl;
        mpIndexer->addPartFilter(new LengthFilter(MaxStrLength));
    }
    if (bCharSumFilt) {
        cout << "Using a char-sum filter" << endl;
        mpIndexer->addPartFilter(new CharsumFilter(MaxStrLength));
    }

    mStrContVect.initStatsCollector(mpGramGen);
    mStrContVect.fillContainer(inputFile, cnt);

    mpSimMetric = new SimMetricEd(*mpGramGen);

    mpIndexer->buildIndex();

    for (unsigned i = 0; i <= MAX_KERR; ++i) {
        mapSearcher[i] = new FtSearcherMem<>(&mMerger, mpIndexer);
    }

    return true;
}

bool CFlamingoNgramIndex::Search (const char *str, int kerr, int *numComp,
                          StringList * resList, bool silent)
{
    vector<unsigned> vStrId;

    Query query(str, *mpSimMetric, (float)kerr); // query string, similarity metric, similarity threshold, must be float!
    mapSearcher[kerr]->search(query, vStrId);

    if (resList) {
        string tmp;

        for (unsigned i = 0; i < vStrId.size(); ++i) {
            mStrContVect.retrieveString(tmp, vStrId[i]);
            resList->push_back(tmp);
        }
    }
    
    return true;
}

