#include "freqgramvector.h"

void CFreqGramVector::InitMem ()
{
    m_Vector = NULL;
    if (Size () > 0) {
        m_Vector = new u_int8_t[Size ()];
    }
}

void CFreqGramVector::FreeMem ()
{
    delete [] m_Vector;
    m_Vector = NULL;
}

CFreqGramVector::CFreqGramVector (const CFreqGramVector & v) : m_StartIndex (v.m_StartIndex), 
                                              m_EndIndex (v.m_EndIndex)
{
    InitMem ();
    if (Size () > 0) {
        memcpy (m_Vector, v.m_Vector, Size ());
    }
}

CFreqGramVector::~CFreqGramVector ()
{
    FreeMem ();
}

CFreqGramVector CFreqGramVector::SubVector (int si, int ei) const
{
    if (si < m_StartIndex)
        si = m_StartIndex;
    if (ei > m_EndIndex)
        ei = m_EndIndex;
    if (si > ei)
        si = ei;
    // Now m_StartIndex<=si<=ei<=m_EndIndex

    return CFreqGramVector (si, ei, &m_Vector[si - m_StartIndex]);
}


CFreqGramVector::CFreqGramVector (int StartIndex, int EndIndex): m_StartIndex (StartIndex),
                                                                 m_EndIndex (EndIndex)
{
    InitMem ();
    if (Size () > 0)
        memset (m_Vector, 0, Size ());
}

CFreqGramVector::CFreqGramVector (int StartIndex, int EndIndex, const VectorType * p) 
                                  : m_StartIndex (StartIndex), m_EndIndex (EndIndex)
{
    InitMem ();
    if (Size () > 0) {
        memcpy (m_Vector, p, Size ());
    }
}


const CFreqGramVector & CFreqGramVector::
operator= (const CFreqGramVector & v)
{
    if (this != &v) {
        FreeMem ();

        m_StartIndex = v.m_StartIndex;
        m_EndIndex = v.m_EndIndex;

        InitMem ();
        if (Size () > 0) {
            memcpy (m_Vector, v.m_Vector, Size ());
        }
    }
    return (*this);
}

u_int8_t CFreqGramVector::StartIndex () const
{
    return m_StartIndex;
}

u_int8_t CFreqGramVector::EndIndex () const
{
    return m_EndIndex;
}

void CFreqGramVector::InitFromString (const char *str)
{
    StringToNgramVec(str, NgramSize, Size (), m_Vector);
}

int CFreqGramVector::CopyToMem (void *mem) const
{
    int s = Size ();
    if (!s)
        return 0;
    memcpy (mem, (void *) m_Vector, s);
    return s;
}

u_int8_t CFreqGramVector::Size () const
{
    return m_EndIndex - m_StartIndex + 1;
}



void CFreqGramVector::Print () const
{
    for (int i = 0; i < Size (); i++)
        std::cout << int (m_Vector[i]) << " ";
    std::cout << "\n";
}



int
Diff (const CFreqGramVector & v1, const CFreqGramVector & v2, int *posdist,
      int *negdist)
{
    int ToIndex = ((v1.Size () < v2.Size ())? v1.Size () : v2.Size ()) - 1;
    *posdist = *negdist = 0;
    int iFirstDiff = -1;
    for (int i = 0; i <= ToIndex; i++) {
        if (v1.m_Vector[i] != v2.m_Vector[i]) {
            if (iFirstDiff < 0)
                iFirstDiff = i;
            int diff = v1.m_Vector[i] - v2.m_Vector[i];
            if (diff > 0)
                *posdist += diff;
            else
                *negdist -= diff;
        }
    }
    if (iFirstDiff < 0)
        if (v1.Size () > v2.Size ())
            iFirstDiff = v2.Size ();
        else if (v2.Size () > v1.Size ())
            iFirstDiff = v1.Size ();

    return iFirstDiff;

}

int CompFunc (const CFreqGramVector & v1, const CFreqGramVector & v2)
{
    int i;
    int ToIndex = ((v1.Size () < v2.Size ())? v1.Size () : v2.Size ()) - 1;
    for (i = 0; i <= ToIndex && v1.m_Vector[i] == v2.m_Vector[i]; i++);
    if (i <= ToIndex)
        return (int (v2.m_Vector[i]) - int (v1.m_Vector[i]));
    return v2.Size () - v1.Size ();

}
