#include "index.h"
#include "cgram.h"

CKdtreeIndex::CKdtreeIndex (bool hasTransp, EAlphabet AlphaType,
                            int _GramSize)
: CInterface(hasTransp, AlphaType), GramSize (_GramSize), opened (false)
{
    MaxFuzzy = _VECTOR_LENGTH;

    GetEnv("VECTOR_LENGTH", MaxFuzzy);
    GetEnv("NGRAM_LEN", GramSize);
   cout << "Ngram size: " << GramSize << endl;

    kdAlloc = new CFastAllocator(sizeof *kd);
    vecAlloc = new CFastAllocator(MaxFuzzy * sizeof(VectorType));
    void * p = kdAlloc->allocChunk();
    kd = new (p) kdtree (*kdAlloc, *vecAlloc, MaxFuzzy, NULL, NULL);
}

CKdtreeIndex::~CKdtreeIndex ()
{
    if (opened)
        CloseDB ();
    //delete kd;
    kd->~kdtree();
    delete kdAlloc;
    delete vecAlloc;
}

bool CKdtreeIndex::Insert (const char *str)
{
    VectorType    vec[MaxFuzzy];

    dict.push_back(string(str));

    StringToNgramVec (str, GramSize, MaxFuzzy, vec);
    kd->insertNodeIntoTree (vec, dict.back().c_str(), NULL);

    return true;
}

bool CKdtreeIndex::Search (const char *str, int kerr, int *numComp,
                           StringList * resList, bool silent)
{
    myagrep        a(str, kerr, mHasTransp);
    VectorType     vec[MaxFuzzy];
    StringList     res;
    unsigned int   gramDist = NgramMaxErrNum (kerr, GramSize, MaxFuzzy, mHasTransp);

    StringList *res1 = (resList == NULL) ? &res : resList;

    StringToNgramVec (str, GramSize, MaxFuzzy, vec);

    kd->Search (vec, gramDist, a, res1, numComp);
    if (!silent) {
        for (StringList::iterator i = res1->begin (); i != res1->end ();
             i++) {
            cout << *i << endl;
        }
    }
    return true;
}

void CKdtreeIndex::MakeBalanced ()
{
    kdtree * oldTree = kd;

    //std::cerr << "Data size: " << kd->dataSize() << std::endl;;
    CFastAllocator * newKdAlloc = new CFastAllocator(sizeof *kd);
    CFastAllocator * newVecAlloc = new CFastAllocator(MaxFuzzy * sizeof(VectorType));

#if 0
    std::cout << "Before balanced, consumed: " << kdAlloc->getAllocatedMem() << ", " << vecAlloc->getAllocatedMem() << std::endl;
#endif
    kd = oldTree->getBalancedCopy(*newKdAlloc, *newVecAlloc);
#if 0
    std::cerr << "After balanced, new consumed: " << newKdAlloc->getAllocatedMem() << ", " << newVecAlloc->getAllocatedMem() << std::endl;
#endif
#if 0
    oldTree->printTree();
    cout << " ==================== \n";
    kd->printTree();
#endif

    //delete oldTree;
    oldTree->~kdtree();
    delete kdAlloc;
    delete vecAlloc;

    kdAlloc = newKdAlloc;
    vecAlloc = newVecAlloc;
}

bool CKdtreeIndex::SearchInDB (const char *str, int kerr, int *numComp,
                               StringList * resList, bool silent)
{
    unsigned int gramDist = NgramMaxErrNum (kerr, GramSize, MaxFuzzy, mHasTransp);

    if (!opened) {
        cerr << "Database is not opened!" << endl;
    }
    myagrep       a(str, kerr, mHasTransp);
    VectorType    vec[MaxFuzzy];
    StringList    res;
    StringList *  res1 = (resList == NULL) ? &res : resList;

    StringToNgramVec (str, GramSize, MaxFuzzy, vec);
    //cout<<str<<" :";PrintNgramVec (MaxFuzzy, vec);

    if (!kdtree::SearchInFile (fd, vec, gramDist, a, res1, numComp)) {
        return false;
    }
    if (!silent) {
        for (StringList::iterator i = res1->begin (); i != res1->end ();
             i++) {
            cout << *i << endl;
        }
    }
    return true;
}

bool CKdtreeIndex::OpenDB (const char *file_name, bool create)
{
    if (opened)
        CloseDB ();
    if (!create) {
        fd = open (file_name, O_RDONLY, 0);
    } else {
        fd = open (file_name, O_RDWR | O_TRUNC | O_CREAT, defaultAccess);
    }
    if (fd < 0) {
        Perror ("Error opening file");
        return false;
    }
    return (opened = true);
}

bool CKdtreeIndex::CloseDB ()
{
    if (!opened)
        return true;
    if (close (fd) != 0)
        return false;
    opened = false;
    return true;
}

bool CKdtreeIndex::DumpDB (const char *file_name)
{
    bool stat;
    if (opened)
        CloseDB ();
    if (!OpenDB (file_name, true)) {
        return false;
    }
    stat = kd->DumpToFile (fd, 0);
    return stat && CloseDB ();
}

/*
** This function substitutes the function from a base class.
*/
bool CKdtreeIndex::CreateInMemIndex(const char *inputFile,
                                       int insertCount, int *dcnt,
                                       bool silent)
{
    *dcnt = 0;
    string s;

    ifstream f (inputFile);
    StringList lst;
    int lstCnt = 0;
    while (f >> s) {
        toUpper (s);
        if (insertCount == 1) {
            Insert (s.c_str ());
        } else {
            lst.push_back (s);
            lstCnt++;
            if (lstCnt >= insertCount) {
                this->InsertList (lst);
                lstCnt = 0;
                lst.clear ();
            }
        }
        if (dcnt)
            (*dcnt)++;
        if (!silent && !((*dcnt) % 10000))
            cout << (*dcnt) << "\n";
    }
    if (lstCnt)
        this->InsertList (lst);
    MakeBalanced ();
    
    return true;
}
