#ifndef _INDEX
#define _INDEX

#include <stdexcept>

#include "ngram.h"
#include "db_init.h"
#include "myagrep.h"
#include "interface.h"
#include "len1index.h"
#include "utils.h"
#include "myexcept.h"
#include "../SeqAgrep/index.h"

class CNgramIndex:public CInterface {
  public:
    int NgramLen;
    static const int MaxLen = 31;
    //static const int SeqIndexBound = 3;
    static string seqIndexName (const char *name, int k) {
        stringstream str; 
        str << name << ".seq" << k;
        return str.str();
    };
  protected:
    CAgrepIndex* pSeqIndex[MAX_KERR];
    CLen1Index *db[MaxLen + 1];
    PtrListVector dict;         // The list of all strings
    bool SearchExact (const char *str, bool silent = true);
    bool opened;

    int fd;
    char *mapDict;
    char *pCompDict; // dictionary strings conactenated
    char *pDictBuff;
    size_t DictBuffSize;
    size_t DictBuffPos;
    DB_ENV *dbenv;
    string baseName;
    int mapSize;

    char* AllocFromBuff(size_t len) {
        if (len + DictBuffPos > DictBuffSize) {
            THROW_AND_PRINT_RUNTIME_ERROR("Not enough free space in the buffer!");
        }
        DictBuffPos += len;
        return pDictBuff + DictBuffPos - len;
    }
  public:
    CNgramIndex (bool hasTransp, EAlphabet AlphaType);
    virtual ~ CNgramIndex ();
    //=========================================================================
    virtual bool Insert (const char *str);
    virtual bool Search (const char *str, int kerr, int *numComp,
                         StringList * resList, bool silent = true);
    //=========================================================================
    virtual bool SearchInDB (const char *str, int kerr, int *numComp,
                             StringList * resList, bool silent = true);
    //=========================================================================
    bool CreateInMemIndex (const char * inputFile, int insertCount, int * dcnt, bool silent = true);
    virtual bool DumpDB (const char *base_name);
    virtual bool OpenDB (const char *base_name, bool create = false);
    bool CloseDB ();
    void DisplayStat () {
			CLen1Index::DisplayStat ();
    }
    void ResetCounters();
    void PrintAddStat(SearchType type, int testNum, int kerr, ostream & ostr);
    DECLARE_STATIC_COUNTER(aseek)
#ifdef _DEBUG
    void Print ();
#endif
}

;

#endif
