#!/usr/bin/perl -w
use strict;

my $fname = $ARGV[0] or die("Please, specify the input file!");

open FN, "<$fname" or die("Cannot open '$fname'");

my %time1;
my %comp1;
my %filt1;
my %time2;
my %comp2;
my %filt2;

my $BucketSize;
my $MethName;

for (my $line = 1;<FN>; ++$line) {
    if (/Bucket size ([0-9]+)/) {
        $BucketSize = $1;
    } elsif (/^([^ ]+) hasTransp/) {
        $MethName = $1;
    } else {
        my ($N, $IndexSize, $tmp1, $TimeMem1, $TimeMem2, $tmp2, $TimeDisk1, $TimeDisk2, $tmp3,$TimeFlash1, $TimeFlash2, $tmp4, $NumComp1, $NumComp2) = split(/ /);
        if (!defined($N) || !defined($TimeMem1) || !defined($TimeMem2) || !defined($NumComp1) || !defined($NumComp2)) {
            die("Incorrect format, line $line");
        }
        if (!defined($BucketSize)) {
            die("Incorrect format, bucket size is expected to be defined (prior to line $line)");
        }
        if (!defined($MethName)) {
            die("Incorrect format, bucket size is expected to be defined (prior to line $line)");
        }
        if ($MethName eq "Mtree1") {
            $time1{$BucketSize}  =  $TimeMem1;
            $comp1{$BucketSize} =  $NumComp1;
            $filt1{$BucketSize} =  FormatFixFracDigit($NumComp1/$N, 4);
            $time2{$BucketSize}  =  $TimeMem2;
            $comp2{$BucketSize} =  $NumComp2;
            $filt2{$BucketSize} =  FormatFixFracDigit($NumComp2/$N, 4);
        }
    }
}

my @SortedKeys = sort {$::a <=> $::b} keys(%time1);
# Bucket numbers
my $Start = "<- c(" ;
my $End   = ")\n";

my $f = 0;

my $bvar;
my $time1var;
my $time2var;
my $filt1var;
my $filt2var;

for my $bucket (@SortedKeys) {
    if ($f) {
        $bvar .= ",";
        $time1var .= ",";
        $time2var .= ",";
        $filt1var .= ",";
        $filt2var .= ",";
    }
    $bvar .= $bucket;
    $time1var .= $time1{$bucket};
    $time2var .= $time2{$bucket};
    $filt1var .= $filt1{$bucket};
    $filt2var .= $filt2{$bucket};
    $f = 1;
}

print "bucket ${Start}${bvar}$End";
print "time1 ${Start}${time1var}$End";
print "time2 ${Start}${time2var}$End";
print "filt1 ${Start}${filt1var}$End";
print "filt2 ${Start}${filt2var}$End";

sub FormatFixFracDigit {
    my ($val, $FracDigits) = @_;

    if (!defined($val)) {return "";}

    # FracDigits after the point
    my $ret = sprintf("%.${FracDigits}f", $val);
    # If the point is in the end, remove it.
    $ret =~ s/\.[0]*$//;
    return $ret;
}
