#!/usr/bin/perl -w
require "./result.pm";

my $UseEverySecond = 1; # Outpus data sets 0.2, 0.8, 3.2 instead of the full dictionary set. Doesn't affect graphs
my $file = $ARGV[0];

$file or die("Should specify a config file");

require("$file");

defined(@InputFileList)     or die('Config should specify variable @InputFileList');
defined($MainFileSuffix)    or die('Config should specify variable $MainFileSuffix');
defined(@Headers)           or die('Config should specify varaible @Headers');
defined(@TranspFlag)        or die('Config should specify variable @TranspFlag');
defined(@RGraphSuffix)      or die('Config should specify variable @RGraphSuffix');
defined(@GenIndexSizeFlag)  or die('Config should specify variable @GenIndexSizeFlag');
defined(@GenTranspSlowDown) or die('Config should specify variable @GenTranspSlowDown');
defined(%ReportList)        or die('Config should specify variable @ReportList');

my $NFiles = scalar(@InputFileList);

if (scalar(@Headers) != $NFiles) {
    die('The number of @Headers elements in the config file does not match the number of elements in @InputFiles');
}

my $MtreeNameSyn1 = "Mtree1";
my $MtreeNameSyn2 = "Mtree1Myers";

my %MethNames = (
    AlternGenIndexExpansion     => ['Hybrid neighborhood generation-1', 'blue'],
    AlternGenIndexExpansion2    => ['Hybrid neighborhood generation-2', 'darkorchid4'], # Dark violet
    AlternGenMorFrank           => ['\mbox{Mor-Fraenkel} method', 'firebrick'],
    AlternGenLinHash            => ['Full neighborhood generation', 'chocolate'],
    FreqTries                   => ['Unigram frequency vector trie', 'darkgreen'],
# In DNA data sets, we use a different BKT modification, whose name is encoded as MtreeNameSyn2
    $MtreeNameSyn1              => ['BKT', 'brown' ],
    NgramFreqTriesChunk         => ['\mbox{$q$-gram} frequency vector trie', 'darkmagenta'],
    NgramMultCompress           => ['\mbox{Length-divided} inverted \mbox{$q$-gram} file', 'red2'],
    SeqAgrepOneChunk            => ['Sequential search', 'red'],
    SignHashOptimizedChunk      => ['Signature hashing', 'black'],
    StringTriesDouble           => ['FB-trie', 'darkorange'],
    StringTriesNew              => ['String trie', 'seagreen'],
    Cole_et_al                  => ['\mbox{$k$-errata} tree', 'tomato4'],
    AlternGenReducedAlpha       => ['Reduced alphabet neighborhood generation', 'violetred2'],
    AlternGenSuperCond          => ['Super-condensed neighborhood generation', 'tan3']  
);

my @FullMethodList      = (sort(SortMethNames keys(%MethNames)));
#my @FullMethodList      = (AlternGenMorFrank, NgramFreqTriesChunk);
#my @FullMethodList      = (Cole_et_al);
my @FilteringMethodList = sort(SortMethNames ("FreqTries", $MtreeNameSyn1, "NgramFreqTriesChunk",  "NgramMultCompress", "SignHashOptimizedChunk", "AlternGenReducedAlpha", "AlternGenSuperCond", "AlternGenIndexExpansion", "AlternGenIndexExpansion2", "AlternGenMorFrank"));
my %LegendColors;
my %LegendSyms;

my $i = 1;
for my $MethKey (@FullMethodList) {
   $LegendColors{$MethKey} = "'$MethNames{$MethKey}->[1]'";
   $LegendSyms{$MethKey}   = $i;
   ++$i;
}

my @dicts = (2, 4, 8, 16, 32);

my @res;

for (my $i = 0; $i < $NFiles; ++$i) {
    my $fd;

    open($fd, "<$InputFileList[$i]") or die("Cannot open '$InputFileList[$i]'");

    $res[$i] = ReadFile($fd);
}

if ($ReportList{"TableSearch"}) {
    DoTable("TableSearch${MainFileSuffix}.tex",         \&GenTableSearch            , \@res, \@TranspFlag, \@InputFileList);
}
if ($ReportList{"TableIndexSize"}) {
    DoTable("TableIndexSize${MainFileSuffix}.tex",      \&GenTableIndexSize         , \@res, \@TranspFlag, \@InputFileList);
}
if ($ReportList{"TableFiltEfficiency"}) {
    DoTable("TableFiltEfficiency${MainFileSuffix}.tex", \&GenTableFiltEfficiency    , \@res, \@TranspFlag, \@InputFileList);
}
if ($ReportList{"TableTranspSlowdown"}) {
    DoTable("TableTranspSlowdown${MainFileSuffix}.tex", \&GenTableTranspSlowdown    , \@res, \@TranspFlag, \@InputFileList);
}

# Special tables to build graphs
if ($ReportList{"TableSpaceTimeGraph"}) {
    for (my $i = 0; $i < $NFiles; ++$i) {
        my $AddSuffix = $RGraphSuffix[$i];
        if ($AddSuffix) {
            DoTable("TableSpaceTimeGraph${MainFileSuffix}_${AddSuffix}.R",    \&GenTableSpaceTimeGraph, [$res[$i]], [$TranspFlag[$i]], [$InputFileList[$i]]);
        }
    }
}

if ($ReportList{"TableLenTimeGraph"}) {
    defined($MinLen) or die('Config should specify variable $MinLen (if you want to generate data for TableLenTimeGraph)');
    defined($MaxLen) or die('Config should specify variable $MaxLen (if you want to generate data for TableLenTimeGraph)');
    defined($NDict)  or die('Config should specify variable $NDict (if you want to generate data for TableLenTimeGraph)');
    defined($XLabelName) or die('Config should specify variable $XLabelName');
    DoTable("TableLenTimeGraph${MainFileSuffix}.R", \&GenTableLengthTimeGraph, \@res, \@TranspFlag, \@InputFileList);
}

if ($ReportList{"FullLegend"}) {
    my $txt = GenColSym(@FullMethodList);
    my $Names;

    for my $MethKey (@FullMethodList) {
        my $nm = $MethNames{$MethKey}->[0];
        if ($Names) { $Names .= ", "; }
        $Names .= "'".RemoveTexSpecials($nm)."'";
    }
    open (FN, ">TableFullLegend.R") or die("Cannot open TableFullLegend.R");

    print FN "$txt\nNames <- c($Names)";
}

sub DoTable {
    my ($File, $Func, $ResRef, $TranspFlagRef, $InputFileListRef) = @_;

    defined($ResRef)         or die('$ResRef should be passed to the function');
    defined($TranspFlagRef)  or die('$TranspFlagRef should be passed to the function');

    if (scalar(@$ResRef) != scalar(@$TranspFlagRef)) {
        die('Non-matching result and transpflag qtys: '.scalar(@$ResRef).' != '.scalar(@$TranspFlagRef));
    }
    if (scalar(@$ResRef) != scalar(@$InputFileListRef)) {
        die('Non-matching result and inputfile qtys: '.scalar(@$ResRef).' != '.scalar(@$InputFileListRef));
    }

    for (my $i = 0; $i < scalar(@$TranspFlagRef); ++$i) {
        defined($TranspFlagRef->[$i]) or die("Undefined TranspFlag, File=$File, SetId = $i, Qty=" .scalar(@$TranspFlagRef));
    }

    my $fn = "$File";
    open FN, ">$fn" or die("Cannot open: $fn");
   
    print FN &$Func($ResRef, $TranspFlagRef, $InputFileListRef);
    close FN;
}



sub GenTableFiltEfficiency {
   my ($ResRef, $TranspFlagRef, $InputFileListRef) = @_; 
   my @res = @$ResRef;

   my $txt = "";
   # Generate table header
   my $AddColumn = $UseEverySecond ? '' : '@{\extracolsep{1pt}}r';
   $txt .= '\hspace{-0.15in}\begin{minipage}{\textwidth}
\scriptsize
\renewcommand{\captionfont}{\normalsize}
\setlength\minrowclearance{0.75pt}
\begin{longtable}{@{\extracolsep{-5pt}}p{2.15in}|@{\extracolsep{2pt}}r'.$AddColumn.'@{\extracolsep{1pt}}r'.$AddColumn.'@{\extracolsep{1pt}}r|@{\extracolsep{2pt}}r'.$AddColumn.'@{\extracolsep{1pt}}r'.$AddColumn.'@{\extracolsep{1pt}}r|@{\extracolsep{2pt}}r'.$AddColumn.'@{\extracolsep{1pt}}r'.$AddColumn.'@{\extracolsep{1pt}}r}
\caption{\label{TableFiltEfficiency'.$MainFileSuffix.'} {Filtering efficiency}} \\\\
\endhead \hline\hline
';
    my $DiffColumn = $UseEverySecond ? 2 : 0;

    for (my $i = 0; $i < scalar(@res); ++$i) {
        my $infoRef = $res[$i];

        $txt .=  '\multicolumn{'.(16 - $DiffColumn * 3).'}{c}{\textbf{\tiny{'.$Headers[$i].'}}} \\\\ \hline\hline';

        my $Add4 = $UseEverySecond ?  '' : ' \textbf{\tiny{0.4M}} &';
        my $Add16 = $UseEverySecond ? '' : ' \textbf{\tiny{1.6M}} &';

        if (!$i) {
            $txt .= '& \multicolumn{'.(5 - $DiffColumn).'}{c|}{\textbf{\tiny{$k=1$}}} &  \multicolumn{'.(5 - $DiffColumn).'}{c|}{\textbf{\tiny{$k=2$}}} &  \multicolumn{'.(5 - $DiffColumn).'}{c}{\textbf{\tiny{$k=3$}}} \\\\
\textbf{\tiny{Number of dictionary strings}} & \textbf{\tiny {0.2M}} & '.$Add4.' \textbf{\tiny{0.8M}} & '.$Add16.' \textbf{\tiny{3.2M}} & \textbf{\tiny{0.2M}} & '.$Add4.' \textbf{\tiny{0.8M}} & '.$Add16.' \textbf{\tiny{3.2M}} & \textbf{\tiny{0.2M}} & '.$Add4.' \textbf{\tiny{0.8M}} & '.$Add16.' \textbf{\tiny{3.2M}} \\\\ \hline
';
        }

        for my $MethKey (@FilteringMethodList) {
            my $HasTransp = $TranspFlagRef->[$i]; 
            my $key = RecKey($MethKey, $HasTransp);
            my $row = "";
            my $info = $infoRef->{$key};
            my $nm = $MethNames{$MethKey}->[0];

            if (!$info) {next;} # No data for a given method

            #$row .= '\raggedright \tiny{'.$nm.'}';
            $row .= '\raggedright '.$nm.'';

            my $part1 = "";
            my $part2 = "";
            my $part3 = "";
            my $Mult = $UseEverySecond ? 4 : 2;

            for (my $N = 200000; $N <= 3200000; $N *= $Mult) {
                my ($IndexSize, $TimeMem1, $TimeMem2, $TimeMem3, $TimeDisk1, $TimeDisk2, $TimeDisk3, $TimeFlash1, $TimeFlash2, $TimeFlash3, $NumComp1, $NumComp2, $NumComp3) = $info->GetValues($N);
                $part1 .= "& ";
                $part2 .= "& ";
                $part3 .= "& ";
                if ($NumComp1) {$part1 .= '{\\tiny{'.FormatValFixed(($N - $NumComp1) / $N, 5).'}}';}
                if ($NumComp2) {$part2 .= '{\\tiny{'.FormatValFixed(($N - $NumComp2) / $N, 5).'}}';}
                if ($NumComp3) {$part3 .= '{\\tiny{'.FormatValFixed(($N - $NumComp3) / $N, 4).'}}';}
            }
            $row .= $part1;
            $row .= $part2;
            $row .= $part3;
            $txt .= $row."\\\\\n";
        }
        $txt .= '\hline \hline';
    }
    $txt .= '\multicolumn{'.(16 - $DiffColumn * 3).'}{l}{\tiny\textbf{Notes:} Filtering efficiency is calculated as $1 - N_{check}/N$, where $N_{check}$ is a number of verifications performed} \\\\';
    $txt .= '\multicolumn{'.(16 - $DiffColumn * 3).'}{l}{\tiny during the checking step. M stands for million entries.} \\\\';

    # Generate table footer
    $txt .= '
    \end{longtable}
    \end{minipage}
    \normalsize
    ';

    return $txt;
}

sub GenTableTranspSlowdown {
    my ($ResRef, $TranspFlagRef, $InputFileListRef) = @_; 
    my @res = @$ResRef;

    my $txt = "";
    # Generate table header

    my $KerrQty        =  3;
    my $OneSetHeadDesc = '|@{\extracolsep{1pt}}p{0.28in}@{\extracolsep{0pt}}|@{\extracolsep{2pt}}p{0.28in}@{\extracolsep{0pt}}|@{\extracolsep{2pt}}p{0.28in}@{\extracolsep{0pt}}';
    my $OneSetKerrDesc = '& $k = 1$ & $k = 2$ & $k = 3$ ';

    my $HeadDesc = "";
    my $HeadDataSetNames = "";
    my $DataSetNames = "";
    my $HeadDictKerr = "";

    my $SetQty = scalar(@res);
    my $SetId;
    my $ColumnQty = 1;
    #my $ClineText = "";

    for ($SetId = 0; $SetId < $SetQty; ++$SetId) {
        if ($GenTranspSlowDown[$SetId]) {
            my $IsLast = 1;
            for ($i = $SetId + 1; $i < $SetQty; ++$i) {
                if ($GenTranspSlowDown[$i]) {
                    $IsLast = 0;
                    last;
                }
            }
            $HeadDesc       .= $OneSetHeadDesc;

            $DictHeadKerr   .= $OneSetKerrDesc; 
            $DataSetNames   .= '&\multicolumn{'.$KerrQty.'}{c'.($IsLast ? '':'|').'}{\tiny{\textbf{'. $Headers[$SetId] . '}}}';
            #$ClineText      .= '\\cline{'.($ColumnQty+1).'-'.($ColumnQty+$KerrQty).'}';
            $ColumnQty      += $KerrQty;
        }
    }
    $txt .= '\begin{minipage}{\textwidth}
    \tiny
    \renewcommand{\captionfont}{\normalsize}
    \setlength\minrowclearance{1pt}
\begin{longtable}{@{\extracolsep{2pt}}p{1.62in}@{\extracolsep{2pt}}'.$HeadDesc.'}
\caption{  \label{TableTranspSlowdown'.$MainFileSuffix.'} {Ratio of retrieval time
for transposition-aware modification to retrieval time for  transposition-unaware modification.} }
\\\\ \hline \hline \endhead
'. 
$DataSetNames .  '\\\\ \hline\hline
' .
$DictHeadKerr .  '\\\\ \hline \hline
';
    my $N = 3200000;


    for my $MethKey (@FullMethodList) {
        my $key1 = RecKey($MethKey, 0);
        my $key2 = RecKey($MethKey, 1);
        my $row = "";
         

        my $nm = $MethNames{$MethKey}->[0];
        my $bold = "";

        if ($nm =~ /sequential search/i) {
            $bold = "\\textbf";
        }


        my $RowVals = "";
        my $HasVals = 0;

        for ($SetId = 0; $SetId < $SetQty; ++$SetId) {
            if ($GenTranspSlowDown[$SetId]) {
                my $infoRef = $res[$SetId];

                my $info1 = $infoRef->{$key1};
                my $info2 = $infoRef->{$key2};

                if (!$info1) {
                   print "Info1 undefined for $MethKey SetId = $SetId, File=$InputFileListRef->[$SetId]\n";
                }
                if (!$info2) {
                   print "Info2 undefined for $MethKey SetId = $SetId, File=$InputFileListRef->[$SetId]\n";
                }

                my $val1 = "";
                my $val2 = "";
                my $val3 = "";

                if ($info1 && $info2) {

                    my ($IndexSize1, $TimeMem1_1, $TimeMem1_2, $TimeMem1_3); 
                    my ($IndexSize2, $TimeMem2_1, $TimeMem2_2, $TimeMem2_3); 
    
                    # Find the first row with the data
                    for (my $N1 = $N; $N1 >= 200000; $N1 /= 2) {
                        ($IndexSize1, $TimeMem1_1, $TimeMem1_2, $TimeMem1_3) = $info1->GetValues($N1);
                        ($IndexSize2, $TimeMem2_1, $TimeMem2_2, $TimeMem2_3) = $info2->GetValues($N1);
                        if ($TimeMem1_1 && $TimeMem2_1) {last;}
                    }

                    if ($TimeMem1_1) {
                        $val1 = FormatFixFracDigit($TimeMem2_1 / $TimeMem1_1, 2);
                        $HasVals = 1;
                    }

                    if ($TimeMem1_2) {
                        $val2 = FormatFixFracDigit($TimeMem2_2 / $TimeMem1_2, 2);
                        $HasVals = 1;
                    }

                    if ($TimeMem1_3) {
                        $val3 = FormatFixFracDigit($TimeMem2_3 / $TimeMem1_3, 2);
                        $HasVals = 1;
                    }
                }

                $RowVals .= " & {\\tiny ${bold}{".$val1. " }}";
                $RowVals .= " & {\\tiny ${bold}{".$val2. " }}";
                $RowVals .= " & {\\tiny ${bold}{".$val3. " }}";
            }
        }

        if ($HasVals) {
            $row .= "\\raggedright \\tiny{${bold}{$nm}}" . $RowVals;
            $txt .= $row."\\\\\n";
        }
    }
    # Generate table footer
    $txt .= '\hline \hline
'.
#'\multicolumn{'.$ColumnQty.'}{l}{\hspace{-0.5em}\tiny\textbf{Notes: } '.($N/1000000).' million entries.} \\\\
#'.
'\end{longtable}
\vspace{0.1in}
\end{minipage}';
    return $txt;
}

sub GenTableSearch {
    my ($ResRef, $TranspFlagRef, $InputFileListRef) = @_; 
    my @res = @$ResRef;

    my $AddColumn = $UseEverySecond ? '' : '@{\extracolsep{3pt}}r';
    my $DiffColumn = $UseEverySecond ? 2 : 0;

    my $txt = "";
    # Generate table header
    $txt .= '
    \begin{minipage}{\textwidth}
    \tiny
    \renewcommand{\captionfont}{\normalsize}
    \setlength\minrowclearance{0.75pt}
    \begin{longtable}{@{\extracolsep{2pt}}p{1.82in}@{\extracolsep{3pt}}|@{\extracolsep{3pt}}r'.$AddColumn.'@{\extracolsep{3pt}}r'.$AddColumn.'@{\extracolsep{3pt}}r|@{\extracolsep{3pt}}r'.$AddColumn.'@{\extracolsep{3pt}}r'.$AddColumn.'@{\extracolsep{3pt}}r|@{\extracolsep{3pt}}r'.$AddColumn.'@{\extracolsep{3pt}}r'.$AddColumn.'@{\extracolsep{3pt}}r}
    \caption{\label{TableSearch'.$MainFileSuffix.'} In-memory search time (ms) } \\\\
     \hline \hline\endhead
    ';

    my $Add4 = $UseEverySecond ?  '' : ' \textbf{\tiny{0.4M}} &';
    my $Add16 = $UseEverySecond ? '' : ' \textbf{\tiny{1.6M}} &';


    for (my $i = 0; $i < scalar(@res); ++$i) {
        my $infoRef = $res[$i];
        $txt .=  '\multicolumn{'.(16 - 3 * $DiffColumn).'}{c}{\textbf{\tiny '.$Headers[$i].'}} \\\\ \hline\hline';
        if (!$i) { 
            $txt .= '& \multicolumn{'.(5 - $DiffColumn).'}{c|}{\textbf{\tiny{$k=1$}}} &  \multicolumn{'.(5 - $DiffColumn).'}{c|}{\textbf{\tiny{$k=2$}}} &  \multicolumn{'.(5 - $DiffColumn).'}{c}{\textbf{\tiny{$k=3$}}} \\\\
\textbf{\tiny{Number of dictionary strings}} & \textbf{\tiny {0.2M}} & '.$Add4.' \textbf{\tiny{0.8M}} & '.$Add16.' \textbf{\tiny{3.2M}} & \textbf{\tiny{0.2M}} & '.$Add4.' \textbf{\tiny{0.8M}} & '.$Add16.' \textbf{\tiny{3.2M}} & \textbf{\tiny{0.2M}} & '.$Add4.' \textbf{\tiny{0.8M}} & '.$Add16.' \textbf{\tiny{3.2M}} \\\\ \hline
     ';
        }
        for my $MethKey (@FullMethodList) {
            my $HasTransp = $TranspFlagRef->[$i]; 
            my $key = RecKey($MethKey, $HasTransp);
            my $row = "";
            my $info = $infoRef->{$key};
            my $nm = $MethNames{$MethKey}->[0];

            #print "$MethKey -> $key\n";

            if (!$info) {next;} # No data for a given method

            my $bold = "";
            if ($nm =~ /sequential search/i) {
                $bold = "\\textbf";
            }
            $row .= "\\raggedright ${bold}{\\tiny{$nm}}";

            my $part1 = "";
            my $part2 = "";
            my $part3 = "";

            my $Mult = $UseEverySecond ? 4 : 2;

            for (my $N = 200000; $N <= 3200000; $N *= $Mult) {
                my ($IndexSize, $TimeMem1, $TimeMem2, $TimeMem3, $TimeDisk1, $TimeDisk2, $TimeDisk3, $TimeFlash1, $TimeFlash2, $TimeFlash3, $NumComp1, $NumComp2, $NumComp3) = $info->GetValues($N);
                $part1 .= "& {\\tiny ${bold}{ " . FormatValFixed($TimeMem1, 4)."}}";
                $part2 .= "& {\\tiny ${bold}{ " . FormatValFixed($TimeMem2, 4)."}}";
                $part3 .= "& {\\tiny ${bold}{ " . FormatValFixed($TimeMem3, 4)."}}";
            }
            $row .= $part1;
            $row .= $part2;
            $row .= $part3;
            $txt .= $row."\\\\\n";
        }
        $txt .= '\hline \hline';
    }
    $txt .= '\multicolumn{'.(16 - 3 * $DiffColumn).'}{l}{\tiny \textbf{Notes: } M stands for million entries.} \\\\';

    # Generate table footer
    $txt .= '
    \end{longtable}
    \end{minipage}'.
    #'\vspace{-25pt}'.
    '\normalsize
    ';

    return $txt;
}

sub GenTableIndexSize {
    my ($ResRef, $TranspFlagRef, $InputFileListRef) = @_; 
    my @res = @$ResRef;

    my $txt = "";
    my $FirstColWidth = "1.6in";
    my $ColumnQty  = 1;

    # Generate table header
    #my $DataSetQty = 5;
    #my $OneSetHeadDesc = '|l@{\extracolsep{2pt}}l@{\extracolsep{2pt}}l@{\extracolsep{2pt}}l@{\extracolsep{2pt}}l@{\extracolsep{3pt}}';
    #my $OneSetQtyDesc  = '& \textbf{\tiny{0.2M}} & \textbf{\tiny{0.4M}} & \textbf{\tiny{0.8M}} & \textbf{\tiny{1.6M}} & \textbf{\tiny{3.2M}} ';
    
    my $DataSetQty = 3;
    my $OneSetHeadDesc = '|p{0.3in}@{\extracolsep{2pt}}p{0.3in}@{\extracolsep{2pt}}p{0.3in}@{\extracolsep{3pt}}';
    my $OneSetQtyDesc  = '& \textbf{\tiny{0.2M}} & \textbf{\tiny{0.8M}} & \textbf{\tiny{3.2M}} ';

    my $HeadDesc;
    my $HeadDataSetNames;
    my $HeadDictQtys;

    my $SetQty = scalar(@res);
    my $SetId;

    for ($SetId = 0; $SetId < $SetQty; ++$SetId) {
        if ($GenIndexSizeFlag[$SetId]) {
            my $IsLast = 1;
            for ($i = $SetId + 1; $i < $SetQty; ++$i) {
                if ($GenIndexSizeFlag[$i]) {
                    $IsLast = 0;
                    last;
                }
            }
            $HeadDesc       .= $OneSetHeadDesc;
            $DictHeadQtys   .= $OneSetQtyDesc; 
            $DataSetNames   .= '&\multicolumn{'.$DataSetQty.'}{c'.($IsLast ? '':'|').'}{\tiny{\textbf{'. $Headers[$SetId] . '}}}';
            $ColumnQty      += $DataSetQty;
        }
    }

    $txt .= '
\hspace{-0.25in}
\begin{minipage}{\textwidth}
\renewcommand{\captionfont}{\normalsize}
\scriptsize
\setlength\minrowclearance{0.75pt}
\begin{longtable}{p{'.$FirstColWidth.'}@{\extracolsep{1pt}}'. $HeadDesc.'}'.
'\caption{\label{TableIndexSize'.$MainFileSuffix.'} Index Size}
\endhead
\hline\hline
'.$DataSetNames.' \\\\ '.
'\textbf{\raggedright\tiny{Number of dictionary strings}} '.$DictHeadQtys .'  \\\\ \hline\hline'.
#'\textbf{} '.$DictHeadQtys .'  \\\\ \hline\hline'.
'';
    my @infoSeq;
    my %MethHasData;
    for ($SetId = 0; $SetId < $SetQty; ++$SetId) {
        if ($GenIndexSizeFlag[$SetId]) {
            for my $MethKey (@FullMethodList) {
                #next if ($MethKey eq "AlternGenSuperCond"); # Super. cond. NG are not implemented efficiently with respect to index size. Thus, we don't report index size for them.
                my $HasTransp = $TranspFlagRef->[$SetId]; 
                defined($HasTransp) or die("TranspFlag is undefined, SetId=$SetId, File=$InputFileListRef->[$SetId]"); 
                my $key = RecKey($MethKey, $HasTransp);
                my $info = $res[$SetId]->{$key};
                my $nm = $MethNames{$MethKey}->[0];

                if (!$info) {next;} # No data for a given method

                $MethHasData{$MethKey} = 1;

                if ($nm =~ /sequential search/i) {
                    $infoSeq[$SetId] = $info;
                }
            }
            $infoSeq[$SetId] or die("Cannot locate info on sequential searching for SetId = $SetId, File=$InputFileListRef->[$SetId]");
        }
    }



    for my $MethKey (@FullMethodList) {
        my $row = "";
        my $row1 = "";
        my $nm = $MethNames{$MethKey}->[0];

        if (!defined($MethHasData{$MethKey})) {
            next; # No data for a given method
        }


        if ($nm =~ /sequential search/i) {
            next;
        }
        $row .= "\\multirow{2}{$FirstColWidth}{\\raggedright \\tiny{$nm}}";

        for ($SetId = 0; $SetId < $SetQty; ++$SetId) {
            if ($GenIndexSizeFlag[$SetId]) {
                my $HasTransp = $TranspFlagRef->[$SetId];
                defined($HasTransp) or die("TranspFlag is undefined, SetId=$SetId, File=$InputFileListRef->[$SetId]"); 
                my $key = RecKey($MethKey, $HasTransp);
                my $info = $res[$SetId]->{$key};

                for (my $N = 200000; $N <= 3200000; $N *= 4) {
                    my $IndexSize = ($info) ? ($info->GetValues($N))[0] : undef;
                    my $DataSize  = ($infoSeq[$SetId]->GetValues($N))[0];

                    $row  .= "& ";
                    $row1 .= "& ";
                    if (defined($IndexSize) && $IndexSize ne "") {
                        $row1 .=  "{\\tiny ".FormatFixFracDigit(100.0 * $IndexSize / $DataSize, 0)."\\% }"; 
                        $row .=  "{\\tiny ".FormatFixFracDigit($IndexSize / 1024, 0)."MB }"; 
                    } else {
                        $row1 .=  "{\\tiny }"; 
                        $row .=  "{\\tiny }"; 
                    }
                }
            }
        }
        $txt .= $row."\\\\\n". $row1."\\\\\\hline\n";
    }
    $txt .= '\hline \hline';

    # Generate table footer
    $txt .= '\multicolumn{'.$ColumnQty.'}{l}{\tiny \textbf{Notes: } Percentage represents the ratio of the index size to the lexicon size.} \\\\
\multicolumn{'.$ColumnQty.'}{l}{\tiny M stands for million entries.} \\\\
\end{longtable}
\end{minipage}
';

    return $txt;
}

sub GenTableSpaceTimeGraph {
    my ($ResRef, $TranspFlagRef, $InputFileListRef) = @_; 
    my @res = @$ResRef;

    my $infoRef = $res[0];
    my $HasTransp = $TranspFlagRef->[0];

    my $txt = "";

    my $dt1 = "";
    my $dt2 = "";
    my $dt3 = "";
    my $IndexSize = "";

    my $tmp = "";


    my $Names = "";


    my @MethList;

    for my $MethKey (@FullMethodList) {
        my $key = RecKey($MethKey, $HasTransp);
        my $row = "";
        my $info = $infoRef->{$key};
        my $nm = $MethNames{$MethKey}->[0];

        if (!$info) {next;} # No data for a given method

        my ($part1, $part2, $part3, $partIndex);

        my $HasData = 0;

        for (my $N = 200000; $N <= 3200000; $N *= 2) {
            my ($IndexSize, $TimeMem1, $TimeMem2, $TimeMem3, $TimeDisk1, $TimeDisk2, $TimeDisk3, $TimeFlash1, $TimeFlash2, $TimeFlash3, $NumComp1, $NumComp2, $NumComp3) = $info->GetValues($N);
            if ($part1) { $part1 .= ", "; }
            $part1 .= FormatValFixedNA($TimeMem1, 5);
            if ($part2) { $part2 .= ", "; }
            $part2 .= FormatValFixedNA($TimeMem2, 5);
            if ($part3) { $part3 .= ", "; }
            $part3 .= FormatValFixedNA($TimeMem3, 5);
            if ($partIndex) { $partIndex .= ", "; }
            if (defined($IndexSize) && $IndexSize > 0) {
                $partIndex .= FormatValFixedNA($IndexSize / 1024.0, 1);
            } else {
                $partIndex .= "NA";
            }
            if (defined($TimeMem1) || defined($TimeMem2) || defined($TimeMem3)) {
                $HasData = 1;
            }
        }
        if ($HasData) {
            push(@MethList, $MethKey);
        } else {
            next;
        }
        if ($Names) { $Names .= ", "; }
        $Names .= "'".RemoveTexSpecials($nm)."'";

        if ($dt1 ) {
            $dt1 .= "#$nm\nTime[[1]] <- rbind(Time[[1]], c($part1))\n";
            $dt2 .= "#$nm\nTime[[2]] <- rbind(Time[[2]], c($part2))\n";
            $dt3 .= "#$nm\nTime[[3]] <- rbind(Time[[3]], c($part3))\n";
            $IndexSize .= "#$nm\nIndexSize <- rbind( IndexSize, c($partIndex))\n";
        } else {
            $dt1 .= "#$nm\nTime[[1]] <- c($part1)\n";
            $dt2 .= "#$nm\nTime[[2]] <- c($part2)\n";
            $dt3 .= "#$nm\nTime[[3]] <- c($part3)\n";
            $IndexSize .= "#$nm\nIndexSize <- c($partIndex)\n";
        }
    }

    $txt .= GenColSym(@MethList);
    $txt .= "Time <- list()\n";
    $txt .= $dt1;
    $txt .= $dt2;
    $txt .= $dt3;
    $txt .= $IndexSize;
    $txt .= "ymaxMult <- ". ($ymaxMult ? $ymaxMult : 1)."\n";
    $txt .= "Names <- c($Names)\n";

    return $txt;
}

sub GenTableLengthTimeGraph {
    my ($ResRef, $TranspFlagRef, $InputFileListRef) = @_; 
    my @res = @$ResRef;

    my $txt = "";

    scalar(@res) == ($MaxLen - $MinLen + 1) or die('The number of data sets does not match $MinLen and $MaxLen');

    my $dt1 = "";
    my $dt2 = "";
    my $dt3 = "";
    my $LenData = "";

    my $tmp = "";


    my $Names = "";
    my @MethList;

    for my $MethKey (@FullMethodList) {
        my $HasData = 0;
        my $row = "";
        my $nm = $MethNames{$MethKey}->[0];

        my ($part1, $part2, $part3, $partLen);


        for (my $Len = $MinLen, my $SetId = 0; $Len <= $MaxLen; ++$Len, ++$SetId) {
            my $HasTransp = $TranspFlagRef->[$SetId]; 
            my $key = RecKey($MethKey, $HasTransp);

            my $infoRef = $res[$SetId];
            my $info = $infoRef->{$key};

            my ($IndexSize, $TimeMem1, $TimeMem2, $TimeMem3, $TimeDisk1, $TimeDisk2, $TimeDisk3, $TimeFlash1, $TimeFlash2, $TimeFlash3, $NumComp1, $NumComp2, $NumComp3);

            if ($info) {
                ($IndexSize, $TimeMem1, $TimeMem2, $TimeMem3, $TimeDisk1, $TimeDisk2, $TimeDisk3, $TimeFlash1, $TimeFlash2, $TimeFlash3, $NumComp1, $NumComp2, $NumComp3) = $info->GetValues($NDict);
            }

            if ($part1) { $part1 .= ", "; }
            $part1 .= ((!defined($TimeMem1)) ? "NA" : FormatValFixedNA($TimeMem1, 5));
            if ($part2) { $part2 .= ", "; }
            $part2 .= ((!defined($TimeMem2)) ? "NA" : FormatValFixedNA($TimeMem2, 5));
            if ($part3) { $part3 .= ", "; }
            $part3 .= ((!defined($TimeMem3)) ? "NA" : FormatValFixedNA($TimeMem3, 5));
            if ($partLen) { $partLen .= ", "; }
            $partLen .= $Len;

            if (defined($TimeMem1) || defined($TimeMem2) || defined($TimeMem3)) {
                $HasData = 1;
            }
        }

        if ($HasData) {
            push(@MethList, $MethKey);
        } else {
            next;
        }

        if ($Names) { $Names .= ", "; }
        $Names .= "'".RemoveTexSpecials($nm)."'";

        if ($dt1 ) {
            $dt1 .= "#$nm\nTime[[1]] <- rbind(Time[[1]], c($part1))\n";
            $dt2 .= "#$nm\nTime[[2]] <- rbind(Time[[2]], c($part2))\n";
            $dt3 .= "#$nm\nTime[[3]] <- rbind(Time[[3]], c($part3))\n";
            $LenData .= "#$nm\nLenData <- rbind( LenData, c($partLen))\n";
        } else {
            $dt1 .= "#$nm\nTime[[1]] <- c($part1)\n";
            $dt2 .= "#$nm\nTime[[2]] <- c($part2)\n";
            $dt3 .= "#$nm\nTime[[3]] <- c($part3)\n";
            $LenData .= "#$nm\nLenData <- c($partLen)\n";
        }
    }
    $txt .= GenColSym(@MethList);
    $txt .= "Time <- list()\n";
    $txt .= $dt1;
    $txt .= $dt2;
    $txt .= $dt3;
    $txt .= $LenData;
    $txt .= "Names <- c($Names)\n";
    $txt .= "XNumVals  <- ".($MaxLen - $MinLen)."\n";
    $txt .= "ymaxMult <- ". ($ymaxMult ? $ymaxMult : 1)."\n";
    $txt .= "yminMult <- ". ($yminMult ? $yminMult : 1)."\n";
    $txt .= "XLabelName <- \"$XLabelName\"\n";

    return $txt;
}

#== Auxilliary functions ==

sub RecKey {
    my ($MethName, $HasTransp) = @_;

    return "$MethName|$HasTransp";
}

sub ReadFile {
    my $fd = shift;
    my %info = ();

    my $MethName;
    my $HasTransp;
    my $ResRef;

    while(<$fd>) {
        chomp;
        if (/^([^ ]+) hasTransp = ([01])/) {
            $MethName  = $1;
            $MethName  = $MtreeNameSyn1 if ($MethName eq $MtreeNameSyn2);
            $HasTransp = $2;
            my $k = RecKey($MethName, $HasTransp);
            $ResRef = $info{$k};
            if (!$ResRef) {
                $ResRef =  new result($MethName, $HasTransp);
                $info{$k} = $ResRef;
            }
        } else {
            my ($N, $IndexSize, $tmp1, $TimeMem1, $TimeMem2, $TimeMem3, $tmp2, $TimeDisk1, $TimeDisk2, $TimeDisk3, $tmp3, $TimeFlash1, $TimeFlash2, $TimeFlash3, $tmp4, $NumComp1, $NumComp2, $NumComp3) = split(/\s+/);
            if ($IndexSize <= 0) {$IndexSize = "";}
            $ResRef->AddValues($N, $IndexSize, $TimeMem1, $TimeMem2, $TimeMem3, $TimeDisk1, $TimeDisk2, $TimeDisk3, $TimeFlash1, $TimeFlash2, $TimeFlash3, $NumComp1, $NumComp2, $NumComp3);
        }
    }

    return \%info;
}

sub GenColSym {
    my @MethList = @_;
    my $txt = "";

    my $txtCol = "legendColors <- c(";
    my $txtSym = "legendSyms   <- c(";
    my $f = 1;

    for my $MethKey (@MethList) {
        if (!$f) {
            $txtCol .= ",";
            $txtSym .= ",";
        }
        $f = 0;
        $txtCol .= $LegendColors{$MethKey};
        $txtSym .= $LegendSyms{$MethKey};
    }
    $txtCol .= ")";
    $txtSym .= ")";

    $txt .= $txtCol."\n";
    $txt .= $txtSym."\n";

    return $txt;
}

sub FormatFixFracDigit {
    my ($val, $FracDigits) = @_;

    if (!defined($val)) {return "";}

    # FracDigits after the point
    my $ret = sprintf("%.${FracDigits}f", $val);
    # If the point is in the end, remove it.
    $ret =~ s/\.[0]*$//;
    return $ret;
}

sub FormatValFixedNA {
    my ($val, $TotDigits) = @_;

    my $res = FormatValFixed($val, $TotDigits);
 
    return ($res ne "") ? $res : "NA";
}

sub FormatValFixed {
    my ($val, $TotDigits) = @_;

    if (!defined($val)) {return "";}
    # 1. If the total number of digits before point <= $TotDigits,
    #    then return only first $TotDigits (not counting point).
    # 2. If the point is in the end, remove it.
    my $ret = sprintf("%f", $val);
    $ret =~ s/\.$//;

    if ($ret =~ /(\-?)([0-9]+)\.([0-9]+)/) {
        my $v0 = $1;
        my $v1 = $2;
        my $v2 = $3;
        if (length($v1) >= $TotDigits) { $ret = $v1; }
        else { 
            $ret = substr("$v0$v1.$v2", 0, $TotDigits + 1); 
        }
    }

    return $ret;
}

sub RemoveTexSpecials {
    my $s = $_[0];

    $s =~ s/\\mbox//g;
    $s =~ s/[\${}]//g;

    return $s;
}

sub SortMethNames {
    my $s1 = RemoveTexSpecials(lc($MethNames{$main::b}->[0]));
    my $s2 = RemoveTexSpecials(lc($MethNames{$main::a}->[0]));

    $s2 cmp $s1;
}

sub PrintArray {
    for (my $i = 0; $i <= $#_; ++$i) {
        print "$_[$i] ";
    }
    print "\n";
}
