#!/usr/bin/perl -w
my $DirEng = $ARGV[0];
my $DirRus = $ARGV[1];
my $DirClue = $ARGV[2];

if (!$DirEng || !$DirRus || !$DirClue) {
    Usage()
}

if (! -d $DirEng) {
    print STDERR "$DirEng is not a directory\n";
    Usage();
}

if (! -d $DirRus) {
    print STDERR "$DirRus is not a directory\n";
    Usage();
}

if (! -d $DirClue) {
    print STDERR "$DirClue is not a directory\n";
    Usage();
}

$| = 1;
my @Dirs =  ($DirEng, $DirRus, $DirClue);
my @Names = ("English synthetic", "Russian synthetic", "ClueWeb09");
my @Suffix = ("eng", "rus", "engnum");

for (my $i = 0; $i <= $#Dirs; ++$i) {
    my  $Dir = $Dirs[$i];

    print "$Names[$i]\t";

    for my $d (2, 4, 8, 16, 32) {
    #for my $d (2) {
        my $DictName = "$Dir/dict${d}_$Suffix[$i]";
        my @res = ProcessOneDict($DictName);
        my ($AlphaSize, $MinLen, $MaxLen, $AvgLen, $LineQty) = @res;

        #print "$DictName -> $AlphaSize $MinLen $MaxLen $AvgLen $LineQty\n";
        if ($d == 2) {
            print "&\t $AlphaSize\t";
        }
        print "&\t $MinLen--$MaxLen/".sprintf("%.1f",$AvgLen)."\t";
    }
    print "\\\\ \\hline\n";
}

sub Usage {
    die("Usage: $0 <DirEng> <DirRus> <DirClue>");
}

sub ProcessOneDict {
    my $DictName = shift;

    open(FN, "<$DictName") or die("Cannot open: '$DictName'");

    my $CharQty = 0;
    my $LineQty = 0;
    my $MinLen  = 1e10;
    my $MaxLen  = 0;
    my %Chars;

    while (<FN>) {
        chomp;
        s/\s*//g;
        my $s = $_;

        ++$LineQty;
        my $len = length($s);
        my @str = unpack(rep("A", $len), $s);
        for (my $i = 0; $i < $#str; ++$i) {
            $c = lc($str[$i]);
            $Chars{$c} = 1;
        }
        $MaxLen = $len if ($len > $MaxLen);
        $MinLen = $len if ($len < $MinLen);
        $CharQty += $len;
    }

    close(FN);

    #delete($Chars{""});
    #print "#".join('',sort(keys(%Chars)))."#",scalar(keys(%Chars)),"\n";

    return (scalar(keys(%Chars)), $MinLen, $MaxLen, $CharQty/$LineQty, $LineQty);
}

sub rep {
    my ($str, $qty) = @_;

    my $res = "";
    for (;$qty >= 0; --$qty) {
        $res .= $str;
    }
    return $res;
}
