#!/usr/bin/perl -w
use POSIX qw(ceil floor);

if (!$ARGV[0] || !$ARGV[1] || !$ARGV[2]) {
    die("Usage: <input file> <output graph for inclusion in TEX> <output data on B(h, i)>");
}
open(FN, "<$ARGV[0]") or die("cannot open $ARGV[0]");
open(OF1, ">$ARGV[1]") or die("cannot open $ARGV[1] for writing");
open(OF2, ">$ARGV[2]") or die("cannot open $ARGV[2] for writing");

my @branch;

for (my $l = 1; <FN>; ++$l) {
    if ($l == 2) { @branch = split(/\s+/, $_); }
}

my $dx = 15;
my $xCnt = min(20, scalar(@branch));
my $N    = $xCnt;
if ($xCnt > 18) { $xCnt = 18; }
my $xWidth = $dx * $xCnt;
my $maxBranch = $branch[0];
my $height = 4 * $maxBranch;


# Output graph
print OF1 '\begin{wrapfigure}{l}{0.44\textwidth}
\vspace{-25pt}
\begin{center}
\caption{\label{Figure5} The average number of children per trie level (English data, 3.2M strings)}
%%\begin{picture}(380,160)\thinlines
\begin{picture}(190,90)\thinlines
\footnotesize
\setlength{\unitlength}{0.5pt}
\put(30,40){\framebox('.$xWidth.','.$height.'){}} %Frame
\put(10, -2){\makebox('.(30 + $xWidth).',10){Node depth (distance to the root)}}
';

my $yStep = 5;
my $dh = 4;
my $lastI = $maxBranch;
if ($lastI % $yStep) { $lastI -= $yStep; }
for (my $i = 0; $i <= $lastI; $i += $yStep) {
 print OF1 '\put(10,'.(35 + $dh*$i).'){\makebox(10,10){'.$i.'}}
\put(30,'.($dh * $i  + 40).'){\line(-1,0){5}}% Y-axis hatches
';
}
if ($maxBranch % $yStep) {
 print OF1 '\put(10,'.(35 + $dh * $maxBranch).'){\makebox(10,10){'.$maxBranch.'}}
\put(30,'.($dh * $maxBranch  + 40).'){\line(-1,0){5}}% Y-axis hatches
';
}
print OF1 '\multiput(40,40)('.$dx.',0){' . $xCnt. '}{\line(0,-1){5}} % X-axis hatches
';

for (my $i = 0; $i < $xCnt; ++$i) {
  my $start = 35 + $i * $dx;
  if (!($i % 2)) {
     print OF1 '\put('.$start.',20){\makebox(10,10){'.($i + 1).'}}'."\n";
  }
  print OF1 '\multiput('.$start.',40)(0.25,0){40}{\line(0,1){'.(($branch[$i]/$maxBranch)*$height).'}}'."\n";
} 

print OF1 '\end{picture}
\end{center}
\vspace{-15pt}
\end{wrapfigure}
';

# Calculate geometric mean

# Init array
my @B;
my @BMax;

for (my $i = 0; $i < $N; ++$i) {
   $B[$i] = [];
}

for (my $h = 0; $h < $N; ++$h) {
    for (my $m = 1; $m <= $N - $h; ++$m) {
        my $res = 1;

        for (my $i = $h; $i < $h + $m; ++$i) {
            $res *= ceil($branch[$i]);
        }
        $res = exp(log($res) / $m);
        $B[$h]->[$m - 1] = $res;
        $BMax[$h ] = max($BMax[$h], $res);
    }
}

print OF2 "Max values\n";
for (my $h = 0; $h < $N; ++$h) {
    print OF2 "$h\t";
}

print OF2 "\n";
for (my $h = 0; $h < $N; ++$h) {
    print OF2 sprintf("%.1f", $BMax[$h])."\t";
}
print OF2 "\n";

print OF2 "Detailed values\n";
for (my $h = 0; $h < $N; ++$h) {
    print OF2 "$h\t";
}
print OF2 "\n";
for (my $m = 1; $m < $N; ++$m) {
    for (my $h = 0; $h <= $N - $m - 1; ++$h) {
        print OF2 sprintf("%.1f", $B[$h]->[$m - 1])."\t";
    }
    print OF2 "\n";
}

print OF2 "Simulation of B(floor(n/2) - ceil(k/2), ceil(n/2) + ceil(k/2))^(ceil(n/2) + ceil(k/2))\n";
for (my $k = 1; $k <= 4; ++$k) {
    for (my $n = 2; $n <= 15; ++$n) {
        my $depth = ceil($n/2) + ceil($k/2);
        my $res = $B[max(0, floor($n/2) - ceil($k/2))]->[min($N - 1, $depth)];
        
        print OF2 sprintf("%.1f", exp(log($res) *$depth) )."\t";
    }
    print OF2 "\n";
}


sub max {
    my ($a, $b) = @_;

    if (!defined($a)) { return $b; }
    if (!defined($b)) { return $a; }

    return ($a < $b) ? $b:$a;
}


sub min {
    my ($a, $b) = @_;

    if (!defined($a)) { return $b; }
    if (!defined($b)) { return $a; }

    return ($a > $b) ? $b:$a;
}
