library(graphics)
library(stats)
library(Hmisc)

source("figureDrawFig.R")

Usage <- function(err) {
    print(paste("Error:", err))
    stop("Usage: <OutputFile> <InputFile1> <transp1> <InputFile2> <transp2> ...")
}

args <- commandArgs(TRUE)

if ((length(args) %% 2) == 0) {
   Usage("Even number of arguments!")
}

OutputFile      <- args[1]

InputFiles <- c()
HasTransp  <- c()
iArg <- 2
NGraph <- (length(args) - 1)/2;

while (iArg < length(args)) {
    InputFiles <- c(InputFiles, args[iArg])
    HasTransp  <- c(HasTransp , args[iArg + 1])
    iArg <- iArg + 2
}

postscript(OutputFile, family = "Times", encoding = "TeXtext.enc", height= 4.5, width= 6 * NGraph, paper = "special")

par(mfrow = c(1, NGraph) )
par(cex = 0.9)

par(font.main = 1) # Main title font is plan instead of bold
par(font.sub = 1) # Sub title font is plan instead of bold

XLabel <- "k"
YLabel <- "Time (ms)"

leg.txt <- c("NR-grep","magrep1","magrep2","DFA-estimated", "Myers-Hyyro", "Lazy DFA", "agrep")

Color_agrep     <- "darkmagenta"
Color_NRgrep    <- "darkviolet"
Color_magrep1   <- "green4"
Color_magrep2   <- "darkorange4"
Color_DFA       <- "red"
Color_LazyDFA   <- "black"
Color_Myers     <- "blue"

ymaxMult <- 1.2

for (iGraph in 1:NGraph) {
    source(InputFiles[iGraph])

    vals <- NRgrep
    vals <- rbind(vals, magrep1) 
    vals <- rbind(vals, magrep2) 
    vals <- rbind(vals, DFA)
    vals <- rbind(vals, MyersHyyro)

    Colors <- c(Color_NRgrep, Color_magrep1, Color_magrep2, Color_DFA, Color_Myers)

    if (HasTransp[iGraph] == 0) {
        vals <- rbind(vals, LazyDFA)
        vals <- rbind(vals, agrep)

        Colors <- c(Colors, Color_LazyDFA, Color_agrep)
    }

    par(mai=c(1.5,1.1,0.25,0.1)) 
    if (iGraph > 1) {
        par(new=F)
    }



    xmin <- 1
    xmax <- 3
    ymin <- min(Inf, MatMinNoNA(vals))
    ymax <- max(0,   MatMaxNoNA(vals))

    MethQty <- length(Colors)
    KData <- matrix(1:3, nrow = MethQty, ncol = 3, byrow = T)
    Names <- leg.txt[1:MethQty]
    legendSyms <- 1:MethQty

    DrawFigure(xmin, xmax, ymin, ymax * ymaxMult, Names, Colors, legendSyms, KData, vals, 
               "", XLabel, YLabel, bLogY = 1, bLogX = 0, bDrawLegend = T, bAdjustYMax = 1, lasY = 1, DivX = 2);
}
dev.off()

