#!/usr/bin/perl -w
use POSIX qw(ceil);
# The scripts several result file, merges them.
# For identical items, the smallest time is chosen. 
my %results;

# 1. Process files
for my $file (@ARGV) {
    ProcessFile(\%results, $file);
}
# 2. Output merged data
my $phead = "";
for my $key (sort(keys(%results))) {
    my ($MethName, $HasTransp, $DictCnt, $BucketSize) = SplitKey($key);
    # Print headers
    my $head = "$MethName|$HasTransp|$BucketSize";

    if ($head ne $phead) {
        if (defined($BucketSize) && $BucketSize ne "") {
            print "Bucket size $BucketSize\n";
        }
        print "$MethName hasTransp = $HasTransp\n";
        $phead = $head;
    }
    my $ref = $results{$key};
    # Print data
    for (my $i = 0; $i <= $#$ref; ++$i) {
        if ($i) {print " ";}
        print $ref->[$i];
    }
    print "\n";
}

sub ProcessFile {
    my ($ResRef, $file) = @_;

    my $log_file = $file.".log";
    open(F, "<$file") or die ("Cannot open '$file' for reading");
    open(LOG, ">$log_file") or die ("Cannot open log for '$log_file' ");

    my ($MethName, $HasTransp, $BucketSize);
    my $res;
    for(my $line = 1;<F>; ++$line) {
        chomp($_);
        
        if (/^([a-z0-9_\-]+) hasTransp = ([0-1]) *$/i) {
            $HasTransp = $2;
            $MethName  = $1;
        } elsif (/^Bucket size ([0-9]+) *$/i) {
            $BucketSize = $1;
        } else {
            my @fields = split(/ /);

            my $n = scalar(@fields);
            my $m = 0;
            if ($n == 2 + 9 * 2) {
                $m = 2;
            } elsif ($n == 2 + 9 * 3) {
                $m = 3;
            } elsif ($n == 2 + 9 * 4) {
                $m = 4;
            } else {
                die("Wrong format '$file' line: $line, field number: $n");
            }
            my $DictCnt = $fields[0];
            my $key = MakeKey($MethName, $DictCnt, $HasTransp, $BucketSize);

            if (!exists($ResRef->{$key})) {
                $ResRef->{$key} = \@fields;
            } else {
            # Calculate mins
                my $oldRef = $ResRef->{$key};

                if ($#$oldRef != $#fields) {
                   die("The number of fields differs from previous '$file' line: $line, current field number: $n, previous: ".($#$oldRef + 1));
                }
                for (my $i = 2; $i < 2 + $m * 3; ++$i) {
                   if (($fields[$i] < 0 && $oldRef->[$i] >= 0) || 
                       ($fields[$i] > 0 && $oldRef->[$i] < 0)) {
                       die("Signs differs from previous '$file' line: $line, current field number: $n, previous: ".($#$oldRef + 1));
                   }
                   if ($oldRef->[$i] > $fields[$i]) {
                       print LOG "Choosing the smaller: $fields[$i] vs $oldRef->[$i] file $file, line: $line ($key)\n";
                       $oldRef->[$i] = $fields[$i];
                   }
                }
                # Workaround for wrong number of comparisions. At some point the number of comparisions
                # was calculated incorrectly for hybrid NG and Mor-Fraenkel.
                # The real figure is larger.
                for (my $i = 2 + $m * 3; $i < 2 + $m * 6; ++$i) {
                   POSIX::ceil($fields[$i]) == $fields[$i] or die("Non-integer number of comp. '$file', line: $line");
                   ($fields[$i] >= -1) or die("Negative number of comp. '$file', line: $line");
                   if (($fields[$i] < 0 && $oldRef->[$i] >= 0) || 
                       ($fields[$i] > 0 && $oldRef->[$i] < 0)) {
                       die("Signs differs from previous '$file' line: $line, current field number: $n, previous: ".($#$oldRef + 1));
                   }
                   if ($oldRef->[$i] < $fields[$i]) {
                       print LOG "Choosing the bigger: $fields[$i] vs $oldRef->[$i] file $file, line: $line ($key)\n";
                       $oldRef->[$i] = $fields[$i];
                   }
                }
            }
        }
    }
    close(F) or die("Cannot close '$file'");
    close(LOG) or die("Cannot close '$log_file'");
}

sub MakeKey {
    my ($MethName, $DictCnt, $HasTransp, $BucketSize) = @_;

    $BucketSize = "" if (!defined($BucketSize));

    return "$MethName|$HasTransp|".sprintf("%09d",$DictCnt)."|$BucketSize";
}

sub SplitKey {
    my ($key) = @_;

    my @res = split(/\|/, $key);
    $res[2] *= 1;
    $res[3] = "" if (!defined($res[3]));
    return @res;
}
