#!/bin/bash
. common_var.sh

base=dict

. mainlist.sh

methlist="$methlist AlternGenIndexExpansion AlternGenIndexExpansion2 AlternGenMorFrank AlternGenReducedAlpha AlternGenSuperCond"

export RES_FILE_BASE=results

if [ "$1" = "--help" -o "$1" = "-h" -o -z "$1" ] ; then
   echo "Usage:"
   echo "$0 <dataset directory> <alphabet code> <test set size>  <method list> "
   echo "=================================================================================="
   echo  "NOTE1: alphabet code and dataset directory (one of the TestVoc* dirs) are mandatory"
   echo "=================================================================================="
   echo  "NOTE2: that only the following combinations of dataset directories and language codes are valid:"
   echo  "<dataset directory>=TestVoc <alphabet code>=eng"
   echo  "<dataset directory>=TestVoc <alphabet code>=rus"
   echo  "<dataset directory>=TestVocCLUEWEB09 <alphabet code>=engnum"
   echo  "<dataset directory>=TestVocDNA <alphabet code>=dna"
   echo "=================================================================================="
   echo "NOTE3: Default parameter values:"
   echo  "<test set size> = 1000"
   echo  "<method list> = \"$methlist\""
   exit 0
fi

TestVocDir=$1

if [ -z "$TestVocDir" ] ; then
    echo "Directory that contains test/dictionary files is empty!"
    exit 1
fi

if [ ! -d "$TestVocDir" ] ; then
    echo "'$TextVocDir' is not a directory that contains test/dictionary files!"
    exit 1
fi

N=$3
if [ -z "$3" ] ; then
    N=1000
fi
if [ ! -z "$4" ] ; then
    methlist=$4
fi

METHLIST=$methlist ./check_debug || exit

echo "Testing only main methods:  $methlist"

. common_param_lang.sh

TestFile1=../$TestVocDir/test${N}_$dict
TestFile2=../$TestVocDir/test${N}_$dict
DictFile=../$TestVocDir/dict1_$dict
echo "Testfile: N = $N"
export SKIP_FUNC_TEST=1
export SKIP_DISK=1
unset DISK_DIR
unset FLASH_DIR
export NO_CLEAR_CACHE=1
for dir in $methlist ; do
 for M in ORIG CMP ; do
  echo "M = $M"
  TestFileBase=../$TestVocDir/test${N}_${M}
  DictFile=../$TestVocDir/dict${M}_$dict
  dictList="${TestFileBase}_0_$dict ${TestFileBase}_1_$dict ${TestFileBase}_2_$dict ${TestFileBase}_2_$dict  $DictFile"
  . ./run_test.sh $dir "1" || exit 1
 done
done
mv $RES_FILE ${RES_FILE}_CMP_$dict
exit 0
