#!/bin/bash

TestVocDir=$1

if [ "$1" = "--help" -o "$1" = "-h" -o -z "$1" ] ; then
   echo "Usage (alphabet list and dataset directory are mandatory):"
   echo "$0 <dataset directory> <alphabet list>  <dict list> <method list> <bucket size list> <transp list>"
fi

if [ -z "$TestVocDir" ] ; then
    echo "You must specify TestVocDir as the first argument!"
    exit 1
fi

if [ ! -d "$TestVocDir" ] ; then
    echo "$TextVocDir is not a directory!"
    exit 1
fi

LangList=$2

if [ -z "$LangList" ] ; then
    echo "You must specify an alphabet as the second argument!"
    exit 1
fi

DictList=$3

if [ -z "$DictList" ] ; then
    DictList=16
fi

MethList=$4

if [ -z "$MethList" ] ; then
    MethList="Mtree1Myers AESA Mtree1 MtreeMultiPivotMyers MtreeMultiPivot"
fi

BucketSizeList=$5

if [ -z "$BucketSizeList" ] ; then
    BucketSizeList="1 10 20 30 40 50 75 100 150 200 300 400 500"
fi

TranspList=$6

if [ -z "$TranspList" ] ; then
    TranspList=1
fi

export SKIP_DISK=1

echo "TestVocDir:       $TestVocDir"
echo "BucketSizeList:   $BucketSizeList"
echo "LangList:         $LangList"
echo "DictList:         $DictList"
echo "MethList:         $MethList"
echo "TranspList:       $TranspList"

for lang in $LangList ; do

    RES_FILE_BASE="results_Mtree"
    RES_FILE=$RES_FILE_BASE.$lang
    echo > $RES_FILE

    for b in $BucketSizeList ; do
        export BUCKET_SIZE=$b
        echo "Bucket size $BUCKET_SIZE " | tee -a $RES_FILE

        for dir in $MethList ; do 
            if [ "$dir" = "MtreeMultiPivot" -o "$dir" = "MtreeMultiPivotMyers" ] ; then
                for pc in 2 3 ; do
                    export PIVOT_CNT=$pc 
                    echo "Pivot cnt $PIVOT_CNT " | tee -a $RES_FILE

                    ./testSPEED.sh "$TestVocDir" "$lang" 1000 "$dir" "$DictList" "$TranspList" "$RES_FILE_BASE"
                done
            else
                ./testSPEED.sh "$TestVocDir" "$lang" 1000 "$dir" "$DictList" "$TranspList" "$RES_FILE_BASE"
            fi
        done

    done

done
