#!/usr/bin/perl -w
require "./result.pm";


my $MAX_KERR = 4;
my $UseEverySecond = 1; # Outpus data sets 0.2, 0.8, 3.2 instead of the full dictionary set. Doesn't affect graphs
my $file = $ARGV[0];

$file or die("Should specify a config file");

require("$file");

defined(@InputFileList)     or die('Config should specify variable @InputFileList');
defined($MainFileSuffix)    or die('Config should specify variable $MainFileSuffix');
defined(@Headers)           or die('Config should specify varaible @Headers');
defined(@TranspFlag)        or die('Config should specify variable @TranspFlag');
defined(@RGraphSuffix)      or die('Config should specify variable @RGraphSuffix');
defined(%ReportList)        or die('Config should specify variable @ReportList');

my $NFiles = scalar(@InputFileList);

if (scalar(@Headers) != $NFiles) {
    die('The number of @Headers elements in the config file does not match the number of elements in @InputFiles');
}

my %MethNames = (
    AlternGenMorFrankPerfHashCompOth                => ['\mbox{Mor-Fraenkel} method (perfect hash)',    'gray0',   0],
    AlternGenReducedAlphaMorFrankPerfHashCompact2   => ['Hybrid of Mor-Fraenkel method and reduced alphabet neighborhood generation', 'gray12', 1],
    FastSS                                          => ['FastSS',                                       'gray24',   2],
    NgramBest                                       => ['\mbox{$q$-gram} index (best time)',            'gray36',   3],
    StringTriesDouble                               => ['FB-trie',                                      'gray48',   4],
    AlternGenReducedAlpha                           => ['Reduced alphabet neighborhood generation',     'gray60',   5],
    AlternGenLinHash                                => ['Full neighborhood generation',                 'gray72',   6],
);

# Order is given by the third array element
my @FullMethodList      = (sort(SortMethOrder keys(%MethNames)));
my %LegendColors;
my %LegendSyms;

my $i = 1;
for my $MethKey (@FullMethodList) {
   $LegendColors{$MethKey} = "'$MethNames{$MethKey}->[1]'";
   $LegendSyms{$MethKey}   = $i;
   ++$i;
}

my @dicts = (2, 4, 8, 16, 32);

my @res;

for (my $i = 0; $i < $NFiles; ++$i) {
    my $fd;

    open($fd, "<$InputFileList[$i]") or die("Cannot open '$InputFileList[$i]'");

    $res[$i] = result::ReadFile($fd);
}

#if ($ReportList{"TableIndexSize"}) {
    #DoTable("TableIndexSize${MainFileSuffix}.tex",      \&GenTableIndexSize         , \@res, \@TranspFlag, \@InputFileList);
#}

# Special tables to build graphs
if ($ReportList{"TableSpaceTimeGraph"}) {
    for (my $i = 0; $i < $NFiles; ++$i) {
        my $AddSuffix = $RGraphSuffix[$i];
        if ($AddSuffix) {
            DoTable("TableSpaceTimeGraph${MainFileSuffix}_${AddSuffix}.R",    \&GenTableSpaceTimeGraph, [$res[$i]], [$TranspFlag[$i]], [$InputFileList[$i]]);
        }
    }
}

if ($ReportList{"TableLenTimeGraph"}) {
    defined($MinLen) or die('Config should specify variable $MinLen (if you want to generate data for TableLenTimeGraph)');
    defined($MaxLen) or die('Config should specify variable $MaxLen (if you want to generate data for TableLenTimeGraph)');
    defined($NDict)  or die('Config should specify variable $NDict (if you want to generate data for TableLenTimeGraph)');
    defined($XLabelName) or die('Config should specify variable $XLabelName');
    DoTable("TableLenTimeGraph${MainFileSuffix}.R", \&GenTableLengthTimeGraph, \@res, \@TranspFlag, \@InputFileList);
}

if ($ReportList{"FullLegend"}) {
    my $txt = GenColSym(@FullMethodList);
    my $Names;

    for my $MethKey (@FullMethodList) {
        my $nm = $MethNames{$MethKey}->[0];
        if ($Names) { $Names .= ", "; }
        $Names .= "'".RemoveTexSpecials($nm)."'";
    }
    open (FN, ">TableFullLegend.R") or die("Cannot open TableFullLegend.R");

    print FN "$txt\nNames <- c($Names)";
}

sub DoTable {
    my ($File, $Func, $ResRef, $TranspFlagRef, $InputFileListRef) = @_;

    defined($ResRef)         or die('$ResRef should be passed to the function');
    defined($TranspFlagRef)  or die('$TranspFlagRef should be passed to the function');

    if (scalar(@$ResRef) != scalar(@$TranspFlagRef)) {
        die('Non-matching result and transpflag qtys: '.scalar(@$ResRef).' != '.scalar(@$TranspFlagRef));
    }
    if (scalar(@$ResRef) != scalar(@$InputFileListRef)) {
        die('Non-matching result and inputfile qtys: '.scalar(@$ResRef).' != '.scalar(@$InputFileListRef));
    }

    for (my $i = 0; $i < scalar(@$TranspFlagRef); ++$i) {
        defined($TranspFlagRef->[$i]) or die("Undefined TranspFlag, File=$File, SetId = $i, Qty=" .scalar(@$TranspFlagRef));
    }

    my $fn = "$File";
    open FN, ">$fn" or die("Cannot open: $fn");
    print "Opened $fn for writing\n";
   
    print FN &$Func($ResRef, $TranspFlagRef, $InputFileListRef);
    close FN;
}


sub GenTableSpaceTimeGraph {
    my ($ResRef, $TranspFlagRef, $InputFileListRef) = @_; 
    my @res = @$ResRef;

    my $infoRef = $res[0];
    my $HasTransp = $TranspFlagRef->[0];

    my $txt = "";


    my $Names = "";


    my @MethList;
    my $bFirst = 1;
    my @DataTime;
    my @DataIndexSize;

    for (my $i = 0; $i < $MAX_KERR; ++$i) {
        $DataTime[$i] = "";
        $DataIndexSize[$i] = "";
    }

    for my $MethKey (@FullMethodList) {
        my $key = result::RecKey($MethKey, $HasTransp);
        my $row = "";
        my $info = $infoRef->{$key};
        my $nm = $MethNames{$MethKey}->[0];

        if (!$info) {next;} # No data for a given method

        my @PartTime;
        my @PartIndexSize;

        for (my $i = 0; $i < $MAX_KERR; ++$i) {
            $PartTime[$i] = "";
            $PartIndexSize[$i] = "";
        }

        my $HasData = 0;

        my @dt;

        for my $N ($info->GetRangeN()) {
            my @IndexSize;
            my @TimeMem;

            my @RangeK = $info->GetRangeK($N);

            # Two cases are possible.
            # 1. Index with sizes not depending on k
            # They have only one entry.
            # 2. Index with sizes growing with k
            if (scalar(@RangeK) == 1) {
                my $kerr = $RangeK[0];

                my ($TmpIndx, @tmp) = $info->GetValues($N, $kerr);

                for (my $i = 0; $i < $kerr; ++$i) {
                    $IndexSize[$i] = $TmpIndx;
                    $TimeMem[$i] = $tmp[$i];
                }
            } else {
                my $PrevK = 0;
                for (my $i = 0; $i <= $#RangeK; ++$i) {
                    my $kerr= $RangeK[$i];
                    $kerr == $PrevK + 1 or die("Invalid format, non contiguous data for $MethKey, N=$N, PrevK=$PrevK, kerr=$kerr\n"); 
                    my ($TmpIndx, @tmp) = $info->GetValues($N, $kerr);
                    $IndexSize[$i] = $TmpIndx;
                    $TimeMem[$i] = $tmp[$i];
                    $PrevK = $kerr;
                }
            }

            for (my $i = 0; $i < $MAX_KERR; ++$i) {
                if ($PartTime[$i]) { $PartTime[$i] .= ", "; }
                if ($PartIndexSize[$i]) { $PartIndexSize[$i] .= ", "; }

                $PartTime[$i] .= FormatValFixedNA($TimeMem[$i], 5);
                $PartIndexSize[$i] .= defined($IndexSize[$i]) ? FormatValFixedNA($IndexSize[$i] / 1024.0, 1) : "NA";

                $HasData = 1 if (defined($TimeMem[$i]));
            }
        }

        if ($HasData) {
            push(@MethList, $MethKey);
        } else {
            next;
        }

        if ($Names) { $Names .= ", "; }
        $Names .= "'".RemoveTexSpecials($nm)."'";

        for ($i = 0; $i < $MAX_KERR; ++$i) {
            my $iPlusOne = $i + 1;
            if ($bFirst) {
                $DataTime[$i]       .= "#$nm\nTime[[$iPlusOne]] <- c($PartTime[$i])\n";
                $DataIndexSize[$i]  .= "#$nm\nIndexSize[[$iPlusOne]] <- c($PartIndexSize[$i])\n";
            } else {
                $DataTime[$i]       .= "#$nm\nTime[[$iPlusOne]] <- rbind(Time[[$iPlusOne]], c($PartTime[$i]))\n";
                $DataIndexSize[$i]   .= "#$nm\nIndexSize[[$iPlusOne]] <- rbind(IndexSize[[$iPlusOne]], c($PartIndexSize[$i]))\n";
            }
        }

        $bFirst = 0;
    }

    $txt .= GenColSym(@MethList);
    $txt .= "Time <- list()\n";
    $txt .= "IndexSize <- list()\n";

    for ($i = 0; $i < $MAX_KERR; ++$i) {
        $txt .= $DataTime[$i];
        $txt .= $DataIndexSize[$i];
    }

    $txt .= "ymaxMult <- ". ($ymaxMult ? $ymaxMult : 1)."\n";
    $txt .= "Names <- c($Names)\n";

    return $txt;
}

sub GenTableLengthTimeGraph {
    my ($ResRef, $TranspFlagRef, $InputFileListRef) = @_; 
    my @res = @$ResRef;

    my $txt = "";

    scalar(@res) == ($MaxLen - $MinLen + 1) or die('The number of data sets does not match $MinLen and $MaxLen');

    my $LenData = "";
    my $tmp = "";


    my $Names = "";
    my @MethList;

    my @DataTime;
    my $bFirst = 1;

    for (my $i = 0; $i < $MAX_KERR; ++$i) {
        $DataTime[$i] = "";
    }

    for my $MethKey (@FullMethodList) {
        my $row = "";
        my $nm = $MethNames{$MethKey}->[0];

        my @PartIndexSize;
        my $bHasData = 0;

        my @PartTime;

        for (my $i = 0; $i < $MAX_KERR; ++$i) {
            $PartTime[$i] = "";
        }

        my $partLen = "";

        for (my $Len = $MinLen, my $SetId = 0; $Len <= $MaxLen; ++$Len, ++$SetId) {
            my $HasTransp = $TranspFlagRef->[$SetId]; 
            my $key = result::RecKey($MethKey, $HasTransp);

            my $infoRef = $res[$SetId];
            my $info = $infoRef->{$key};

            if (!$info) {next;} # No data for a given method

            my @RangeK = $info->GetRangeK($NDict) or die("Cannot obtain values of kerr for N=$NDict");
            my $kerr = $RangeK[$#RangeK];

            my ($TmpIndx, @TimeMem) = $info->GetValues($NDict, $kerr);

            if ($partLen) { $partLen .= ", "; }
            $partLen .= $Len;

            for (my $i = 0; $i < $MAX_KERR; ++$i) {
                if ($PartTime[$i]) { $PartTime[$i] .= ", "; }
                $PartTime[$i] .= FormatValFixedNA($TimeMem[$i], 5);

                $bHasData = 1 if (defined($TimeMem[$i]));
            }

        }

        #print "@@@ $partLen\n";

        if ($bHasData) {
            push(@MethList, $MethKey);
        } else {
            next;
        }

        if ($Names) { $Names .= ", "; }
        $Names .= "'".RemoveTexSpecials($nm)."'";

        for ($i = 0; $i < $MAX_KERR; ++$i) {
            my $iPlusOne = $i + 1;
            if ($bFirst) {
                $DataTime[$i]  .= "#$nm\nTime[[$iPlusOne]] <- c($PartTime[$i])\n";
            } else {
                $DataTime[$i]   .= "#$nm\nTime[[$iPlusOne]] <- rbind(Time[[$iPlusOne]], c($PartTime[$i]))\n";
            }
        }
        if ($LenData) {
            $LenData        .= "#$nm\nLenData <- rbind( LenData, c($partLen))\n";
        } else {
            $LenData       .= "#$nm\nLenData <- c($partLen)\n";
        }
        $bFirst = 0;
    }
    $txt .= GenColSym(@MethList);
    $txt .= "Time <- list()\n";
    for (my $i = 0; $i < $MAX_KERR; ++$i) {
        $txt .= $DataTime[$i];
    }
    $txt .= $LenData;
    $txt .= "Names <- c($Names)\n";
    $txt .= "XNumVals  <- ".($MaxLen - $MinLen)."\n";
    $txt .= "ymaxMult <- ". ($ymaxMult ? $ymaxMult : 1)."\n";
    $txt .= "yminMult <- ". ($yminMult ? $yminMult : 1)."\n";
    $txt .= "XLabelName <- \"$XLabelName\"\n";

    return $txt;
}

#== Auxilliary functions ==

sub GenColSym {
    my @MethList = @_;
    my $txt = "";

    my $txtCol = "legendColors <- c(";
    my $txtSym = "legendSyms   <- c(";
    my $f = 1;

    for my $MethKey (@MethList) {
        if (!$f) {
            $txtCol .= ",";
            $txtSym .= ",";
        }
        $f = 0;
        $txtCol .= $LegendColors{$MethKey};
        $txtSym .= $LegendSyms{$MethKey};
    }
    $txtCol .= ")";
    $txtSym .= ")";

    $txt .= $txtCol."\n";
    $txt .= $txtSym."\n";

    return $txt;
}

sub FormatFixFracDigit {
    my ($val, $FracDigits) = @_;

    if (!defined($val)) {return "";}

    # FracDigits after the point
    my $ret = sprintf("%.${FracDigits}f", $val);
    # If the point is in the end, remove it.
    $ret =~ s/\.[0]*$//;
    return $ret;
}

sub FormatValFixedNA {
    my ($val, $TotDigits) = @_;

    my $res = FormatValFixed($val, $TotDigits);
 
    return ($res ne "") ? $res : "NA";
}

sub FormatValFixed {
    my ($val, $TotDigits) = @_;

    if (!defined($val)) {return "";}
    # 1. If the total number of digits before point <= $TotDigits,
    #    then return only first $TotDigits (not counting point).
    # 2. If the point is in the end, remove it.
    my $ret = sprintf("%f", $val);
    $ret =~ s/\.$//;

    if ($ret =~ /(\-?)([0-9]+)\.([0-9]+)/) {
        my $v0 = $1;
        my $v1 = $2;
        my $v2 = $3;
        if (length($v1) >= $TotDigits) { $ret = $v1; }
        else { 
            $ret = substr("$v0$v1.$v2", 0, $TotDigits + 1); 
        }
    }

    return $ret;
}

sub RemoveTexSpecials {
    my $s = $_[0];

    $s =~ s/\\mbox//g;
    $s =~ s/[\${}]//g;

    return $s;
}

sub SortMethOrder {
    my $s1 = $MethNames{$main::b}->[2];
    my $s2 = $MethNames{$main::a}->[2];

    $s2 <=> $s1;
}

sub SortMethNames {
    my $s1 = RemoveTexSpecials(lc($MethNames{$main::b}->[0]));
    my $s2 = RemoveTexSpecials(lc($MethNames{$main::a}->[0]));

    $s2 cmp $s1;
}

sub PrintArray {
    for (my $i = 0; $i <= $#_; ++$i) {
        print "$_[$i] ";
    }
    print "\n";
}
