#!/usr/bin/perl -w
# Merge results for two n-gram implementations: Flamingo and NgramMultCompress
use result;

my $InpFile = $ARGV[0] or die("Specify input file");
my $fd;

open $fd, "<$InpFile" or die("Cannot open '$InpFile' for reading");

my $InfoHashRef = result::ReadFile($fd, {});

my $MorFrankNameSyn1 = "AlternGenMorFrankPerfHashCompOth";
my $MorFrankNameSyn2 = "AlternGenMorFrankPerfHashCompDNA";

for my $key (keys(%$InfoHashRef)) {
    my $RecRef = $InfoHashRef->{$key};
    my @RangeN = $RecRef->GetRangeN();

    my $hRef = result::SplitKey($key);

    for (my $i=0; $i <= $#RangeN; ++$i) {
        my $N = $RangeN[$i];
        my @RangeK = $RecRef->GetRangeK($N);
        for (my $j = 0; $j <= $#RangeK; ++$j) {
            my $MethName = $hRef->{MethName};
            my $kerr = $RangeK[$j];
            next if ($MethName eq "NgramFlamingo");

            my ($IndexSize, $TimeMem1, $TimeMem2, $TimeMem3, $TimeMem4, $NumComp1, $NumComp2, $NumComp3, $NumComp4) = $RecRef->GetValues($N, $kerr);

            # Find numbers for NgramFlamingo to output best results
            if ($MethName eq "NgramMultCompress") {
                my $RecRefFlam = $InfoHashRef->{result::RecKey("NgramFlamingo", 0)} or die("Cannot retrieve information for Flamingo!");
                my @vals = $RecRefFlam->GetValues($N, $kerr) or die("Cannot obtain retrieval times for Flamingo, N=$N, kerr=$kerr");
                my ($tmp, $TimeMem1Flam, $TimeMem2Flam, $TimeMem3Flam, $TimeMem4Flam) = @vals;

                $TimeMem1 = $TimeMem1Flam if ($TimeMem1Flam < $TimeMem1);
                $TimeMem2 = $TimeMem2Flam if ($TimeMem2Flam < $TimeMem2);
                $TimeMem3 = $TimeMem3Flam if ($TimeMem3Flam < $TimeMem3);
                $TimeMem4 = $TimeMem4Flam if ($TimeMem4Flam < $TimeMem4);

                $MethName = "NgramBest";
            } else {
                $MethName = $MorFrankNameSyn1 if($MethName eq $MorFrankNameSyn2);
            }
            $IndexSize = 0 if (!defined($IndexSize));
            $TimeMem1 = -1 if (!defined($TimeMem1));
            $TimeMem2 = -1 if (!defined($TimeMem2));
            $TimeMem3 = -1 if (!defined($TimeMem3));
            $TimeMem4 = -1 if (!defined($TimeMem4));
            $NumComp1 = -1 if (!defined($NumComp1));
            $NumComp2 = -1 if (!defined($NumComp2));
            $NumComp3 = -1 if (!defined($NumComp3));
            $NumComp4 = -1 if (!defined($NumComp4));

            print "$MethName hasTransp=$hRef->{hasTransp} Sigma=0 kerr=$kerr\n"; 
            print "$N $IndexSize 0 $TimeMem1 $TimeMem2 $TimeMem3 $TimeMem4 ", 
                  "-1 -1 -1 -1 -1 ", 
                  "-1 -1 -1 -1 -1 ", 
                  "0 $NumComp1 $NumComp2 $NumComp3 $NumComp4 ",
                  "-1 -1 -1 -1 -1 ", 
                  "-1 -1 -1 -1 -1 ", 
                  "\n";
        }
    }
}
