LogCeil  <- function(x) { return (exp(log(10)* ceiling(log(x)/log(10)))); }
LogFloor <- function(x) { return (exp(log(10)* floor(log(x)/log(10)))); }

MatMinNoNA <- function (dt) {
    res <- 1/0
    for (i in 1:length(dt[,1])) {
        for (j in 1:length(dt[1,])) {
            if (!is.na(dt[i,j])) {
                res <- min(res, dt[i,j])
            }
        }
    }
    return (res);
}

MatMaxNoNA <- function (dt) {
    res <- -1/0
    for (i in 1:length(dt[,1])) {
        for (j in 1:length(dt[1,])) {
            if (!is.na(dt[i,j])) {
                res <- max(res, dt[i,j])
            }
        }
    }
    return (res);
}


DrawFigure <- function (xmin, xmax, ymin, ymax, names, color, sym, x, val, title, XLabel, YLabel, bLogY = 0, bLogX = 0, bDrawLegend = 0, bAdjustYMax = 1, lasY = 0, DivY = 10, DivX = 5) {

  DigitNum <- 2
  MultY <- 8
  MultX <- 4

  par(cex.lab=1.5, cex=1.2, mar=c(5,2.7,1,0.5))

  if (bLogY) {
      LogAxis <- "y"

      yAxis <- c()

      MultSpaceLegend <- 1

      if (bDrawLegend) {
          MultSpaceLegend <- Mult ^ 8
      }

      ymin <- LogFloor(ymin)

      yTickMax <- ymax

      #if (bAdjustYMax) {
          #ymax <- LogCeil(ymax * MultSpaceLegend)
          #yTickMax <- ymax / MultSpaceLegend
      #}


      y1   <- ymin

      while (y1 < yTickMax) {
          yAxis <- c(yAxis, y1)
          y1 <- y1 * MultY
      }
  } else {
    LogAxis  <- ""

    yAxis    <- ymax /DivY * (1:DivY)

    Delta <- ymax - ymin

    ymax <- ymax + Delta/ DivY
    if (bDrawLegend) {
        ymax <- ymax + 3 * Delta/ DivY
    }
    ymin <- ymin - Delta / DivY
  }

  if (bLogX) {
    xmax <- LogCeil(xmax)
    xmin <- LogFloor(xmin)

    xAxis <- c()

    x1   <- xmin

    while (x1 <= xmax) {
          xAxis <- c(xAxis, x1)
          x1 <- x1 * MultX
      }

    LogAxis <- paste("x", LogAxis, sep = "")
  } else {
    xAxis    <- rep(xmin, DivX) + ((xmax - xmin) /DivX) * (1:DivX)
  }

  yAxis   <- signif(yAxis, digits = 3)
  yLabels <- c()
  xLabels <- c()

  for (k in 1:length(yAxis)) {
      yLabels[k] <- print(yAxis[k], digits = DigitNum)
  }

  for (k in 1:length(xAxis)) {
      xLabels[k] <- print(xAxis[k], digits = DigitNum)
  }

  for (i in 1:length(color)) {
    if (i == 1) {
       par(new = FALSE)
    } else {
       par(new = TRUE)
    }


    res <- try(plot(x[i, ], (val[i,]), "b", col = color[i], pch = sym[i], axes = FALSE, 
         xlim=c(xmin, xmax), ylim=c(ymin,ymax), xlab= XLabel, ylab = YLabel, 
         main = title, LogAxis 
         ), silent = FALSE)

    if (is.object(res)) {
        if (class(res) == "try-error") {
            print("Failed for parameters:")
            print(paste("i=", i))
            print(names[i])
            print(x[i,])
            print(val[i,])
        }
    }
  }


  axis(1, at = xAxis, labels = xLabels, las = 2, cex.axis=1)
  axis(2, at = yAxis, labels = yLabels, las = lasY, cex.axis=1)

  box()

  par(new = FALSE)

  if (bDrawLegend) {
    legend(x = xmin, y = ymax + 50, names, col = color, pch = sym, lty = 1, ncol = 1, bty = "o", cex = 0.8)
  }

}

