library(graphics)
library(stats)
library(Hmisc)

Usage <- function(err) {
    print(paste("Error:", err))
    stop("Usage: <InputFile> <OutputFilePrefix>")
}

args <- commandArgs(TRUE)

if (length(args) != 2) {
   Usage("Wrong number of arguments!")
}

source(args[1])
OutputFilePrefix <- args[2]

XLabel = XLabelName
YLabel = "Time (ms)"

postscript(paste(OutputFilePrefix, "eps", sep = "."), family = "Times", encoding = "TeXtext.enc", height=5, width=15, paper="special")

par(mfrow = c(1, 4)) 
par(cex = 0.8) 
par(cex.axis = 0.8) 
par(font.main = 1) # Main title font is plain instead of bold
par(font.sub = 1) # Sub title font is plain instead of bold
par(font.axis = 1) # Axis font is plain instead of bold
par(font = 1) 


source('figureDrawFig.R')

ymin <-  Inf
ymax <- -Inf

for (k in 1:4) {
    ymin <- min(ymin, MatMinNoNA(Time[[k]]))
    ymax <- max(ymax, MatMaxNoNA(Time[[k]]))
}

xmax <- MatMaxNoNA(LenData)
xmin <- MatMinNoNA(LenData)

for (k in 1:4) {
    YLabelCurr <- ""

    if (k == 1) {
        par(mai=c(0.65,0.65,0.25,0.25))
        YLabelCurr <- YLabel
    } else if (k == 2) {
        par(mai=c(0.65,0.25,0.25,0.25))
    }

    #DrawFigure(xmin, xmax, ymin, ymax * ymaxMult, Names, legendColors, legendSyms, LenData, Time[[k]], paste("k = ",k,"", sep=""), XLabel, YLabelCurr, bLogY = 1, bLogX = 0, bDrawLegend = as.integer(k == 1), bAdjustYMax = 0, lasY = 1, DivX = XNumVals);
    DrawFigure(xmin, xmax, ymin * yminMult, ymax * ymaxMult, Names, legendColors, legendSyms, LenData, Time[[k]], paste("k = ",k,"", sep=""), XLabel, YLabelCurr, bLogY = 1, bLogX = 0, bDrawLegend = 0, bAdjustYMax = 0, lasY = 1, DivX = XNumVals);

}


dev.off()
