library(graphics)
library(stats)
library(Hmisc)

Usage <- function(err) {
    print(paste("Error:", err))
    stop("Usage: <InputFile> <OutputFilePrefix>")
}

args <- commandArgs(TRUE)

if (length(args) != 2) {
   Usage("Wrong number of arguments!")
}

OutputFilePrefix <- args[2]

source(args[1])

XLabel = "Index size (MBs)"
YLabel = "Time (ms)"

postscript(paste(OutputFilePrefix, "eps", sep = "."), family = "Times", encoding = "TeXtext.enc", height=7, width=15, paper="special")

par(mfrow = c(1, 4)) 
par(cex = 0.8) 
par(cex.axis = 0.8) 
par(font.main = 1) # Main title font is plain instead of bold
par(font.sub = 1) # Sub title font is plain instead of bold
par(font.axis = 1) # Axis font is plain instead of bold
par(font = 1) 


source('figureDrawFig.R')

ymin <-  Inf
ymax <- -Inf
xmin <-  Inf
xmax <- -Inf

for (k in 1:4) {
    ymin <- min(ymin, MatMinNoNA(Time[[k]]))
    ymax <- max(ymax, MatMaxNoNA(Time[[k]]))
    xmin <- min(xmin, MatMinNoNA(IndexSize[[k]]))
    xmax <- max(xmax, MatMaxNoNA(IndexSize[[k]]))
}


for (k in 1:4) {
    YLabelCurr <- ""
    if (k == 1) {
        par(mai=c(0.65,0.65,0.25,0.25))
        YLabelCurr <- YLabel
    } else if (k == 2) {
        par(mai=c(0.65,0.25,0.25,0.25))
    }

    DrawFigure(xmin, xmax, ymin, ymax * ymaxMult, Names, legendColors, legendSyms, IndexSize[[k]], Time[[k]], paste("k = ",k,"", sep=""), XLabel, YLabelCurr, bLogY = 1, bLogX = 1, 0, lasY = 1);

}


dev.off()
