#include "index.h"
#include "myagrep.h"

// here you can change this
const char CAgrepIndex::termChar = 0;

const AlgoType type = FILTER_AGREP;

CAgrepIndex::CAgrepIndex (bool hasTransp, EAlphabet AlphaType):CInterface(hasTransp, AlphaType), opened (false), fd (0), recNum(0)
{
}

CAgrepIndex::~CAgrepIndex ()
{
    if (opened) {
        CloseDB ();
    }
}


bool CAgrepIndex::Insert (const char *str)
{
    wordChunk.append(str, strlen(str));
    wordChunk.append(&termChar, 1);
    ++recNum;
    return true;
}

bool CAgrepIndex::Search(const char *str, int kerr, int *numComp,
                          StringList * resList, bool silent)
{
    myagrep a(str, kerr, mHasTransp);

    a.findPattern(type,
                  termChar, wordChunk.data(), wordChunk.data() + wordChunk.size(),
                  true, resList);

    if (resList) {
        cntFnd = resList->size();
    }
    if (numComp) (*numComp) = recNum;
    
    return true;
}

bool CAgrepIndex::SearchInDB(const char *str, int kerr, int *numComp,
                              StringList * resList, bool silent)
{
    if (!opened) {
        cerr << "Database is not opened!" << endl;
    }
    off_t fsz = FileSize (Name.c_str ());
	if (fsz < 0) {
		return false;
	}
	if (fsz == 0) {
		return true;
	}

    myagrep a(str, kerr, mHasTransp);

    if (lseek (fd, 0, SEEK_SET)) {
        Perror ("Cannot lseek");
        return false;
    }
    void *buf = mmap (NULL, fsz, PROT_READ, MAP_SHARED, fd, 0);
    if (buf == MAP_FAILED) {
        Perror ("Cannot mmap!");
        return false;
    }
    int len = 0;
    char *buf1 = ((char *) buf) + sizeof recNum;
    recNum = *((int *) buf);

    a.findPattern(type,
                  termChar, buf1, buf1 + fsz - sizeof recNum, true, resList); 
    if (resList) cntFnd = resList->size();
    if (numComp) *numComp = recNum;

    if (munmap (buf, fsz)) {
        Perror ("Cannot unmap");
    }

    return true;
}


bool CAgrepIndex::OpenDB (const char *file_name, bool create)
{
    if (opened)
        CloseDB ();
    if (!create) {
//#define O_DIRECT 040000
        fd = open (file_name, O_RDONLY);
    } else {
        fd = open (file_name, O_RDWR | O_TRUNC | O_CREAT, defaultAccess);
    }
    if (fd < 0) {
        Perror ("Error opening file");
        return false;
    }
    Name = file_name;
    return (opened = true);
}

bool CAgrepIndex::CloseDB ()
{
    if (!opened)
        return true;
    if (fd) {
        if (close (fd) != 0)
            return false;
    }
    fd = 0;
    opened = false;
    return true;
}


bool CAgrepIndex::DumpDB (const char *file_name)
{
    if (opened)
        CloseDB ();
    if (!OpenDB (file_name, true)) {
        return false;
    }
    if (write (fd, &recNum, sizeof recNum) != sizeof recNum) {
        return false;
    }
    int len = wordChunk.size();
    if (write (fd, wordChunk.data(), len) != len) {
        return false;
    }
    bool f = CloseDB ();
    return f;
}
