#include "sign.h"

using namespace std;

Sign::Sign (const HMap & _hsign, int _MaxDigit,
               const char *str):hsign (_hsign), MaxDigit (_MaxDigit) 
{
    val = (str) ? FromString (str) : 0;
} 

Sign::Sign (const HMap & _hsign, int _MaxDigit,
                int32 _val):hsign (_hsign), MaxDigit (_MaxDigit),
val (_val) 
{
} 

int32  Sign::abs (int32 val) 
{
    int32 cnt = 0;
    for (int32 i = 0; i < MaxDigit; i++, val >>= 1)
        if (val & 1)
            cnt++;
    return cnt;
};

int32  Sign::FromString (const char *str) const 
{
    int32 i, c, mask;
    int32 val = 0;
    for (i = 0; str[i]; i++)
         {
        register int t = hsign[(unsigned char) str[i]] - 1;
        if (t >= 0)
            val |= 1 << t;
        }
    return val;
};

Sign::operator  int32 () const 
{
    return val;
};

int32  Sign::operator - (const Sign & op) 
{
    int32 sum (0);
    int32 res (val ^ op.val);
    for (; res; res >>= 1)
        sum += res % 2;        // res is not zero according to the cycle condition
    return sum;
}


//  |A\B| 
int32  Sign::Diff (const Sign & op) 
{
    int32 sum (0);
    int32 res (val & ~op.val);
    for (; res; res >>= 1)
        sum += res % 2;        // res is not zero according to the cycle condition
    return sum;
};

ostream & operator << (ostream & out, Sign & s) 
{
    out << setw(4) << setfill('0') << hex << s.getVal() << dec << endl;
    return out;
};


