#include "index.h"
#include "alphabet.h"
#include "langcase.h"

CSignHashIndex::CSignHashIndex (bool hasTransp, EAlphabet AlphaType):CInterface(hasTransp, AlphaType), opened (false), MaxDigit(GetMaxDigit()), ind (MaxDigit, hsign, hasTransp), dind(hasTransp) 
{
}

CSignHashIndex::~CSignHashIndex ()
{
    if (opened)
        CloseDB ();
}

int CSignHashIndex::GetMaxDigit()
{
    int res = HASH_SIZE, tmp;
    if (getenv("VECTOR_LENGTH") && (tmp = atol(getenv("VECTOR_LENGTH")))) {
        res = tmp;            
    }
    return res;
}

bool CSignHashIndex::Insert (const char *str)
{
    cerr << "Insert unsupported" << endl;
    return false;
}

bool CSignHashIndex::InsertList (const StringList & lst)
{
    cerr << "InsertList unsupported" << endl;
    return false;
}

bool CSignHashIndex::Search (const char *str, int kerr, int *numComp,
                             StringList * resList, bool silent)
{
    ind.Search (str, kerr, numComp, resList, silent);
    return true;
}

bool CSignHashIndex::SearchInDB (const char *str, int kerr, int *numComp,
                                 StringList * resList, bool silent)
{
    if (!opened) {
        cerr << "Database is not opened!" << endl;
    }
    return dind.Search (str, kerr, numComp, resList, silent);
}

bool CSignHashIndex::OpenDB (const char *file_name, bool create)
{
    if (opened)
        CloseDB ();
    if (!create) {
        return (opened = dind.Open (file_name));
    } else {
        cerr << "OpenDB with create flag set to true is unsupported!" <<
            endl;
        return false;
    }
}

bool CSignHashIndex::CloseDB ()
{
    if (!opened)
        return true;
    if (!dind.Close ())
        return false;
    opened = false;
    return true;
}


bool CSignHashIndex::DumpDB (const char *baseName)
{

    if (!dind.FastCreate (mInputFile.c_str(), baseName, hsign, MaxDigit, true)) {
        Perror ("Cannot create an index");
        return false;
    }
    string tmpIndName = string (baseName) + ".tmp";
    if (!dind.Defragment (baseName, tmpIndName.c_str ())) {
        Perror ("Cannot defragment an index");
        return false;
    }

    return true;
}

/*
** This function substitutes the function from a base class.
*/
bool CSignHashIndex::CreateInMemIndex(const char *inputFile,
                                         int insertCount, int *dcnt,
                                         bool silent)
{
    mInputFile = inputFile;

#ifdef OPTIMIZE_BY_STAT
    CollectStat st;

    st.CollectFileStat (inputFile);
    if (!CalcHMap (MaxDigit, st.stat (), hsign, getenv("PRINT_ADD_STAT"))) {
        std::cerr << "Hash generation failed!" << std::endl;
        return false;
    }
#else
    const char* pABC = GetAlphabet(mAlphaType);

    for (const char *pSym = pABC; char c = *pSym; ++pSym) {
        int v = (pSym - pABC) % MaxDigit;

        hsign.setValue(tolower_eng_rus(c), v);
        hsign.setValue(toupper_eng_rus(c), v);
    }
#endif

    ifstream f (inputFile);
    string s;
    for ((*dcnt) = 0; f >> s; (*dcnt)++) {
        toUpper (s);
        ind.Insert (s.c_str ());
    }
    return true;
}
