#include "hashindex.h"

// you may change it here
const char CHashIndex::termChar = 0;


void CHashIndex::Insert (const char *str)
{
    int j = 0;
    // Let's calc. the signature value
    int SignVal = (int) Sign (hsign, MaxDigit, str);
#ifndef USE_CHUNK
    StrList::iterator it = HashList[SignVal].begin (),
        ite = HashList[SignVal].end ();
    for (; it != ite; it++)
        // The string already exists
        if (!strcmp (*it, str))
            return;
    char *new_str = new char[strlen (str) + 1];
    strcpy (new_str, str);
    HashList[SignVal].insert (ite, new_str);
#else
    HashList[SignVal].append(str, strlen(str));
    HashList[SignVal].append(&termChar, 1);
    HashListCnt[SignVal]++;
#endif
}



CHashIndex::CHashIndex(int _MaxDigit, const HMap & _hsign, bool hasTransp):
                        MaxDigit (_MaxDigit), hsign (_hsign), mHasTransp(hasTransp)
{
    int N = 1 << MaxDigit;
#ifdef USE_CHUNK
    HashListCnt.resize(N);
    HashList = new string[N];
#else
    HashList = new StrList[N];
#endif
}


void CHashIndex::Search (const char *str, int kerr, int *numComp,
                     StringList * resList, bool silent)
{
    myagrep a(str, kerr, mHasTransp);

    int imin = 0, imax = (1 << MaxDigit) - 1;
    Sign sgn (hsign, MaxDigit, str);
    if (!kerr)
        imin = imax = int32 (sgn);
    int bucknum = 0;
    for (int i = imin; i <= imax; i++) {
        bool WorthScanFlag = false;

        if (kerr) {
            Sign SignI (hsign, MaxDigit, i);
            int dst = (sgn - SignI);

            int nerr_e = kerr;
#if 0
            if (kerr >= 2)
                nerr_e--;       // All mismatches of type "change" are unlikely
#endif
            // so that we can speed up the process considerably
            // without really losing search quality


            for (int err_change = 0; err_change <= nerr_e; err_change++) {
                if (dst <= (kerr - err_change) + 2 * err_change &&
                    abs (sgn.abs () - SignI.abs ()) <= (kerr - err_change)
                    ) {
                    WorthScanFlag = true;
                    break;
                }
            }
        } else
            WorthScanFlag = true;

        // Now we'll check whether the signature could be ok
        if (WorthScanFlag) {
            bucknum++;
#ifndef USE_CHUNK
            StrList::iterator it = HashList[i].begin (),
                ite = HashList[i].end ();
            for (; it != ite; it++) {
                if (numComp)
                    (*numComp)++;
                if (a.checkRTRFilter(*it)) {
                    if (!silent)
                        std::cout << *it << "\n";
                    if (resList != NULL)
                        resList->push_back (string (*it));
                }
            }
#else
            if (numComp)
              (*numComp) += HashListCnt[i];
            a.findPattern(CInterface::filterType, 
                  termChar, HashList[i].data(), HashList[i].data() + HashList[i].size(),
                  true, resList);
#endif
        }
    }
    //cout<<" ## "<<bucknum<<"\n";

}

CHashIndex::~CHashIndex ()
{

#ifndef USE_CHUNK
    for (int i = 0; i < (1 << MaxDigit); i++) {
        for (int j = 0; j < HashList[i].size (); j++) {
            delete[] HashList[i][j];
        }
        HashList[i].clear ();
    }
#endif

    delete[] HashList;
}
