#include "edistk.h"
#include <string.h>

unsigned int distCol::m_dummy;

unsigned int distCol::size ()
{
    return sizeof (*m_pCol) * (m_hi - m_lo + 1);
}

void distCol::allocMem ()
{
    if (m_lo <= m_hi) {
        m_pCol = new unsigned int[size ()];
    }
}

void distCol::freeMem ()
{
    delete[]m_pCol;
    m_pCol = NULL;
}

distCol::~distCol ()
{
    freeMem ();
}

distCol::distCol (int pos, unsigned int kerr):m_pos (pos), m_kerr (kerr),
m_lo (0), m_hi (-1), m_pCol (NULL)
{
    if (m_pos >= 0) {
        m_lo = m_pos - kerr;
        m_hi = m_pos + kerr;
        allocMem ();
        for (int i = m_lo; i <= m_hi; i++) {
            m_pCol[i - m_lo] = kerr + 1;
        }
    }
}

int
 distCol::realIndex (unsigned int i)
{
    // if m_pos < 0 then the below if will fail
    if (int (i) <= m_hi && int (i) >= m_lo) {
        return int (i) - m_lo;
    }
    return -1;
}

unsigned int & distCol::getElem (unsigned int i)
{
    int ri = realIndex (i);
    return (ri < 0) ? (m_dummy = m_kerr + 1) : m_pCol[ri];
}

void distCol::copyContent (const distCol & src)
{
    m_kerr = src.m_kerr;
    m_lo = src.m_lo;
    m_hi = src.m_hi;
    m_pos = src.m_pos;
    allocMem ();
    if (m_lo <= m_hi)
        memcpy (m_pCol, src.m_pCol, size ());
}

distCol::distCol (const distCol & src)
{
    copyContent (src);
}

const distCol & distCol::operator= (const distCol & src)
{
    if (this != &src) {
        freeMem ();
        copyContent (src);
    }
    return *this;
}


checkDistContext::checkDistContext (unsigned int kerr, const char *str):
m_kerr (kerr), m_pos (0), m_currCol (0), m_patSize (strlen (str)),
m_pStr (str)
{
    m_apCol[0] = new distCol (0, kerr);
    m_apCol[1] = NULL;
    for (unsigned int i = 0; i <= std::min (m_kerr, m_patSize); i++) {
        m_apCol[0]->getElem (i) = i;
    }
}

void
 checkDistContext::freeMem ()
{
    delete m_apCol[0];
    delete m_apCol[1];
    m_apCol[0] = m_apCol[1] = NULL;
}

checkDistContext::~checkDistContext ()
{
    freeMem ();
}

void checkDistContext::copyContent (const checkDistContext & src)
{
    m_kerr = src.m_kerr;
    m_pos = src.m_pos;
    m_patSize = src.m_patSize;
    m_pStr = src.m_pStr;
    m_currCol = src.m_currCol;
    m_apCol[0] =
        (src.m_apCol[0] != NULL) ? (new distCol (*src.m_apCol[0])) : NULL;
    m_apCol[1] =
        (src.m_apCol[1] != NULL) ? (new distCol (*src.m_apCol[1])) : NULL;
}

checkDistContext::checkDistContext (const checkDistContext & src)
{
    copyContent (src);
}

const checkDistContext &
    checkDistContext::operator= (const checkDistContext & src)
{
    if (this != &src) {
        freeMem ();
        copyContent (src);
    }
    return *this;
}

void
 checkDistContext::Print (void)
{
    for (int i = 0; i <= m_patSize; i++) {
	    std::cout << m_apCol[m_currCol]->getElem (i) << " ";
    }
    std::cout << std::endl;
}

bool distCol::noCutoff (int upperIndex, bool hasTransp)
{
    int KERR = (hasTransp) ? m_kerr - 1: m_kerr;
    for (int i = std::max (m_lo, 0); i <= std::min (m_hi, upperIndex); i++) {
        if (m_pCol[i - m_lo] <= KERR) {
            return true;
            break;
        }
    }
    return false;
}

bool checkDistContext::checkDist (const char *str, bool complete, bool hasTransp)
{
    unsigned int len = strlen (str);
    return checkDist(str, len, complete, hasTransp, NULL);
}

bool checkDistContext::checkDist (const char *str, unsigned int len, bool complete, bool hasTransp, int *pDist)
{
    if (!m_patSize)
        return len <= m_kerr;

    m_pos++;
    for (; m_pos <= len; m_pos++) {
        distCol newCol (m_pos, m_kerr);

        distCol & prevCol = *(m_apCol[m_currCol]);
        for (int i = std::max (m_pos - int (m_kerr), 0);
             i <= std::min (m_pos + m_kerr, m_patSize); i++) {
            unsigned int c = prevCol.getElem (i) + 1;   // DpMatrix[row,col] = DpMatrix[row,col - 1] + 1 deletion of character in str
            if (i > 0) {
                c = std::min (c, newCol.getElem (i - 1) + 1);    // DpMatrix [row,col] = DpMatrix [row-1,col] + 1 delete of character in m_str
                c = std::min (c, prevCol.getElem (i - 1) + int (str[m_pos - 1] != m_pStr[i - 1]));   // substitution
            }
            if (hasTransp) {
                if (m_pos > 1 && i > 1 && str[m_pos - 1] == m_pStr[i - 2]
                    && str[m_pos - 2] == m_pStr[i - 1]) {
                    c = std::min (c, 1 + m_apCol[1 - m_currCol]->getElem (i - 2));
                }
            }
            newCol.getElem (i) = c;
        }
        // switching columns
        m_currCol = 1 - m_currCol;
        if (m_pos == 1) {
            m_apCol[m_currCol] = new distCol (newCol);  // copy constructor
        } else {
            *m_apCol[m_currCol] = newCol;   // operator=
        }
    }
    if (m_pos > len)
        m_pos = len;
    if (complete) {
        int val = m_apCol[m_currCol]->getElem(m_patSize);
        if (pDist) *pDist = val;
        return val <= m_kerr;
    }
    if (!hasTransp) {
        if (m_apCol[m_currCol]->noCutoff(m_patSize, false))
            return true;
    } else {
        if (m_apCol[0] != NULL && m_apCol[0]->noCutoff (m_patSize, m_currCol != 0))
            return true;
        if (m_apCol[1] != NULL && m_apCol[1]->noCutoff (m_patSize, m_currCol != 1))
            return true;
    }
    return false;
}
