#include "index.h"

CTrieIndex::CTrieIndex (bool hasTransp, EAlphabet AlphaType)
            : CInterface(hasTransp, AlphaType), opened (false)
{
}

CTrieIndex::~CTrieIndex ()
{
    if (opened)
        CloseDB ();
}

bool CTrieIndex::Insert (const char *str)
{
    return node.Insert (string (str) + EOS) != NULL;
}

bool CTrieIndex::Search (const char *str, int kerr, int *numComp,
                         StringList * resList, bool silent)
{
    string s (str);

    return (kerr != 0) ? node.Search (s, kerr, numComp, resList,
                                      mHasTransp, silent) : node.SearchExact (s,
                                                                  numComp,
                                                                  resList,
                                                                  silent);
}

bool CTrieIndex::SearchInDB (const char *str, int kerr, int *numComp,
                             StringList * resList, 
                             bool silent)
{
    if (!opened) {
        cerr << "Database is not opened!" << endl;
    }
    return node.SearchInFile (fd, string (str), kerr, numComp, resList,
                              mHasTransp, silent);
}

bool CTrieIndex::OpenDB (const char *file_name, bool create)
{
    if (opened)
        CloseDB ();
    if (!create) {
        fd = open (file_name, O_RDONLY, 0);
    } else {
        fd = open (file_name, O_RDWR | O_TRUNC | O_CREAT, defaultAccess);
    }
    if (fd < 0) {
        Perror ("Error opening file");
        return false;
    }
    return (opened = true);
}

bool CTrieIndex::CloseDB ()
{
    if (!opened)
        return true;
    if (close (fd) != 0)
        return false;
    opened = false;
    return true;
}


bool CTrieIndex::DumpDB (const char *file_name)
{
    bool stat;
    if (opened)
        CloseDB ();
    if (!OpenDB (file_name, true)) {
        return false;
    }
    stat = node.DumpToFile (fd, 0);
    return stat && CloseDB ();
}

#ifdef _DEBUG
void CTrieIndex::Print ()
{
    node.Print ();
}
#endif
