#include "leaf.h"

TrieLeaf::TrieLeaf (std::string Key, TrieLeaf * _UpperLink, TrieNode * _SubTree):SubTree (_SubTree), NextLeaf (NULL), KeyPart (Key),
UpperLink
(_UpperLink)
{
};

TrieLeaf::~TrieLeaf ()
{
    if (SubTree)
        delete SubTree;
}

TrieLeaf *TrieLeaf::Append (std::string Key, TrieLeaf * _UpperLeaf,
                            TrieNode * _SubTree)
{
    if (NextLeaf)               // go recursively
        return NextLeaf->Append (Key, _UpperLeaf, _SubTree);
    else {
        NextLeaf = new TrieLeaf (Key, _UpperLeaf, _SubTree);
        return NextLeaf;
    }
};

TrieLeaf *TrieLeaf::Insert (std::string Key, TrieNode * _SubTree)
{
    std::string CommKey, Key1, Key2;
    int l1 = KeyPart.length (), l2 = Key.length ();
    int i = 0;
    for (; i < min (Key.length (), KeyPart.length ());)
        if (Key[i] == KeyPart[i])
            i++;
        else
            break;

    Key2 = std::string (Key, i, l2 - 1);

    if (l1 == i)                /* Just insert to the lower level */
        if (SubTree) {
            return SubTree->Insert (Key2, this, _SubTree);
        } else {
            return NULL;        /* One string is a proper suffix of the other this 
                                   is not possible if every string is unique and end at EOS */
        }

    CommKey = std::string (KeyPart, 0, i);
    Key1 = std::string (KeyPart, i, l1 - 1);
    TrieNode *NewSubTree = new TrieNode (), *OldSubTree = SubTree;
    KeyPart = CommKey;          // shorten the key in the leaf
    SubTree = NewSubTree;
    if (!NewSubTree->Insert (Key2, this, NULL))
        return NULL;
    if (!OldSubTree)
        return SubTree->Insert (Key1, this, NULL);
    TrieLeaf *uplnk = NewSubTree->Insert (Key1, this, OldSubTree);
    OldSubTree->UpdateUpperLink (uplnk);
    return uplnk;
};

int
 TrieLeaf::CheckCommon (const std::string &Key)
{
    int i = 0;
    for (; i < min (Key.length (), KeyPart.length ());)
        if (Key[i] == KeyPart[i])
            i++;
        else
            break;
    return i;
};


#ifdef _DEBUG
void TrieLeaf::Print (std::string s, int level)
{
    for (int i = 0; i < level; i++)
        cout << " ";
    cout << "#" << s +
        KeyPart << " Stored key=" << KeyPart << " Sub=" << SubTree;
    cout << '\n';
    if (SubTree) {
        std::string st = s + KeyPart;
        SubTree->Print (st, level);
    }
}
#endif
