#include "index.h"
#include <numeric>

CTrieIndexDFA::CTrieIndexDFA (bool hasTransp, EAlphabet AlphaType)
            : CInterface(hasTransp, AlphaType), opened (false)
{
}

CTrieIndexDFA::~CTrieIndexDFA ()
{
    if (opened)
        CloseDB ();
}

bool CTrieIndexDFA::Insert (const char *str)
{
    try {
        node.Insert(str);
    } catch (const string & e) {
        cerr << "Error inserting of '" << (string(str)) << "': " << e << endl;
        throw;
    }
    return true;
}

bool CTrieIndexDFA::Search (const char *str, int kerr, int *numComp,
                         StringList * resList, bool silent)
{
    string s (str);

    if (kerr != 0) {
        node.Search (GetAlphabet(mAlphaType), s, kerr, numComp, resList, mHasTransp, silent);
    } else {
        node.SearchExact (s, numComp, resList, silent);
    }
    return true;
}

bool CTrieIndexDFA::SearchInDB (const char *str, int kerr, int *numComp,
                             StringList * resList, 
                             bool silent)
{
    if (!opened) {
        cerr << "Database is not opened!" << endl;
    }
    return node.SearchInFile(GetAlphabet(mAlphaType), fd, string (str), kerr, numComp, resList,
                              mHasTransp, silent);
}

bool CTrieIndexDFA::OpenDB (const char *file_name, bool create)
{
    if (opened)
        CloseDB ();
    if (!create) {
        fd = open (file_name, O_RDONLY, 0);
    } else {
        fd = open (file_name, O_RDWR | O_TRUNC | O_CREAT, defaultAccess);
    }
    if (fd < 0) {
        Perror ("Error opening file");
        return false;
    }
    return (opened = true);
}

bool CTrieIndexDFA::CloseDB ()
{
    if (!opened)
        return true;
    if (close (fd) != 0)
        return false;
    opened = false;
    return true;
}


bool CTrieIndexDFA::DumpDB (const char *file_name)
{
    bool stat;
    if (opened)
        CloseDB ();
    if (!OpenDB (file_name, true)) {
        return false;
    }
    //node.Print();
    stat = node.DumpToFile (fd, 0);
    return stat && CloseDB ();
}

void CTrieIndexDFA::PrintAddStat(SearchType type, int testNum, int kerr, ostream & ostr)
{
     if (type != SearchTypeMem) return;
     if (!kerr) {
          vector<int>   cnt;
          vector<int>   cntBranch; 

          node.countBranch(cnt, cntBranch);
          int sz = cnt.size();
          for (int i = 0; i < sz; ++i) {
              cout << i << "\t";
          }
          cout << endl;
          for (int i = 0; i < sz; ++i) {
              if (cnt[i]) {
                  cout << double(cntBranch[i])/cnt[i] << "\t";
              } else {
                  cout << "N/A\t" ;
              }
          }
          cout << endl;
     } else {
         node.PrintAddStat(testNum, kerr, ostr);
     }
}

#ifdef _DEBUG
void CTrieIndexDFA::Print ()
{
    node.Print ();
}
#endif
