#ifndef AFUZZY_CONT_H
#define AFUZZY_CONT_H

#include <iostream>
#include <iomanip>

/*
  Leonid Boitsov
 */
#include "interface.h"

typedef u_int32_t Uint;

class context {
    friend class afuzzycontext;
    Uint R[MAX_KERR + 1];
    Uint RP[MAX_KERR + 1];
    int  k;
    int  startpos;
    int  cnt;
    int  cntp;
    bool hasTransp;

    void copy(const context &c) {
        /*
        if (k != c.k) {
            delete R;
            R = new Uint[c.k + 1];
        }
        if (k != c.k || !c.hasTransp) {
            delete RP;
            RP = NULL;
        }
        */
        k = c.k;
        cnt = c.cnt;
        cntp = c.cntp;
        startpos = c.startpos;
        hasTransp = c.hasTransp;

        memcpy((void *) R, (void *) c.R, (c.k + 1) * sizeof (Uint));
        if (hasTransp) {
            /*
            if (!RP) {
                RP = new Uint[c.k + 1];
            }
            */
            memcpy ((void *) RP, (void *) c.RP, (c.k + 1) * sizeof (Uint));
        }
    }
  public:
     context ();
    ~context ();
    inline context (const context & c);
    inline const context & operator= (const context & c);
    void init (int k, bool hasTransp);
    //void setStartPos(int pos){startpos = pos;}
    int getStartPos() const {return startpos;}
};

context::context (const context & c): /*R(NULL), RP(NULL),*/ k(0), hasTransp(0)
{
    copy(c);
}

const context & context::operator=(const context & c)
{
    if (this == &c)
        return (*this);

    copy(c);

    return (*this);
}

class afuzzycontext {
    Uint  *S;
    Uint  RI[MAX_KERR + 1];
    int   Map[256];
    int   SMapArr[256];
    int   k;
    int   m;
    Uint  mask_ok;
    Uint  filter_ok;
    Uint  filter_shift;
    int   r_size;
    bool  hasTransp;
  public:
    afuzzycontext ();
    void init (const std::string & p, int k, bool hasTransp, context * cntx);
    inline bool checkRTR (const char *t, int tLen, context *, bool final, int *pDist) const;
    inline void checkRTRStep(const char *t, context *) const;
    inline bool checkRTRState(context *, bool &finalState, int *pDist) const;
    ~afuzzycontext ();
    static int MaxPatSize () {return sizeof (Uint) * 8;}
    int getKerr() const {return k;}
};

/**  */
bool afuzzycontext::checkRTR (const char *t, int tLen, context * cntx, bool final, int *pDist) const
{
    Uint R1[MAX_KERR + 1];

    register int cnt1;

    register int j = cntx->startpos;
    Uint SMapP = (j > 0) ? (SMapArr[(unsigned char) t[j - 1]]) : 0;
    for (; j < tLen; j++) {
        cnt1 = cntx->cnt;
        register Uint SMap = SMapArr[(unsigned char) t[j]];
        for (register int d = cntx->cnt; d <= k; d++) {
            R1[d] = (((cntx->R[d] << 1) | (j <= d)) & SMap);
            if (d > cntx->cnt) {
                R1[d] |= ((cntx->R[d - 1] | R1[d - 1]) << 1) | 
                           cntx->R[d - 1] | (j <= d - 1);
            }
        }
        /* Now transpositions */
        if (hasTransp && j > 0) {
            int d0 = max(cntx->cntp + 1, 1);
            for (int d = d0; d <= k; ++d) {
                R1[d] |= (cntx->RP[d - 1] << 2 | 
                     (Uint(j <= d) << 1)) & SMapP & (SMap << 1);
            }
        }
        for (register int d = cntx->cnt; d <= k; d++) {
            if (d == cnt1 && j >= d && !(R1[d])) {
               cnt1 = d + 1;
            }
        }

        if (hasTransp) { 
            for (register int i = cntx->cnt; i <= k; i++) {
                cntx->RP[i] = cntx->R[i];
                cntx->R[i] = R1[i];
            }
        } else {
            for (register int i = cntx->cnt; i <= k; i++) {
                cntx->R[i] = R1[i];
            }
        }

        cntx->cntp = cntx->cnt;
        cntx->cnt = cnt1;
        SMapP = SMap;

    }
    cntx->startpos = j;   // Moving start position

    if (final) {
        if (cntx->R[k] & mask_ok) {
            if (pDist) {
                int &dist = *pDist;
                for (dist = k - 1; dist > 0 && (cntx->R[dist] & mask_ok); --dist);

                ++dist;
            }
            return true;
        } else {
            return false;
        }
    } else {
        if (hasTransp) {
            return cntx->cnt <= k || cntx->cntp < k;
        } else {
            return cntx->cnt <= k;
        }
    }
}

void afuzzycontext::checkRTRStep (const char *t, context * cntx) const
{
    Uint R1[MAX_KERR + 1];

    register int cnt1;

    register int j = cntx->startpos;
    Uint SMapP = (j > 0) ? (SMapArr[(unsigned char) t[j - 1]]) : 0;

    if (1) {
        cnt1 = cntx->cnt;
        register Uint SMap = SMapArr[(unsigned char) t[j]];
        for (register int d = cntx->cnt; d <= k; d++) {
            R1[d] = (((cntx->R[d] << 1) | (j <= d)) & SMap);
            if (d > cntx->cnt) {
                R1[d] |= ((cntx->R[d - 1] | R1[d - 1]) << 1) | 
                           cntx->R[d - 1] | (j <= d - 1);
            }
        }
        /* Now transpositions */
        if (hasTransp && j > 0) {
            int d0 = max(cntx->cntp + 1, 1);
            for (int d = d0; d <= k; ++d) {
                R1[d] |= (cntx->RP[d - 1] << 2 | 
                     (Uint(j <= d) << 1)) & SMapP & (SMap << 1);
            }
        }
        for (register int d = cntx->cnt; d <= k; d++) {
            if (d == cnt1 && j >= d && !(R1[d])) {
               cnt1 = d + 1;
            }
        }

        if (hasTransp) { 
            for (register int i = cntx->cnt; i <= k; i++) {
                cntx->RP[i] = cntx->R[i];
                cntx->R[i] = R1[i];
            }
        } else {
            for (register int i = cntx->cnt; i <= k; i++) {
                cntx->R[i] = R1[i];
            }
        }

        cntx->cntp = cntx->cnt;
        cntx->cnt = cnt1;
        SMapP = SMap;

    }
    cntx->startpos = j + 1;   // Moving start position
}

bool afuzzycontext::checkRTRState(context * cntx, bool &finalState, int *pDist) const
{
    if (cntx->R[k] & mask_ok) {
        if (pDist) {
            int &dist = *pDist;
            for (dist = k - 1; dist > 0 && (cntx->R[dist] & mask_ok); --dist);
           ++dist;
        }
        finalState = true;
    }
    
    if (hasTransp) {
        if  (cntx->cnt <= k || cntx->cntp < k) return true;
    } else {
        if (cntx->cnt <= k) return true;
    }

    return false;
}

#endif
