#include "trie.h"
#include <algorithm>
#include "../Interface/triedeb.h"


void CTrie::SearchAtLevelExact(const std::string & s, int stLen,
                              int *numComp,
                             StringList * resList, bool silent) const
{
   if (stLen >= s.size()) {
       if (!silent)
           cout << s << '\n';
       if (resList != NULL)
           resList->push_back (s.substr (0, s.size () - 1));

       return;
   }
   int chSize = children.size();
   
   if (!chSize) 
       return;

   unsigned int minPos =  0, sz = chSize;

#if 1
   const char *pBeg = s.c_str() + stLen;
   int cmp;
   unsigned int half; 
   unsigned int middle;

   while (sz > 0) {
       half = sz >> 1;
       middle = minPos + half;

       cmp = strcmp(children[middle].label.c_str(), pBeg);
       if (cmp < 0) {
           minPos = middle + 1;
           sz = sz - half - 1;
       } else {
           sz = half;
       }
   }

   if (minPos == chSize)  {
        --minPos;
   }
   
   if (1) {
        const Child *pChld = &children[minPos];
        int len = pChld->checkCommon(s, stLen);
        // We stand after if we have an exact prefix
        CALC_NUM_COMP(numComp, pChld->ptr);
        if (len < pChld->label.size() && minPos) {
            pChld = &children[minPos - 1];
            len = pChld->checkCommon(s, stLen);
        }
        if (len >= pChld->label.size()) {
            int stLen1 = stLen + pChld->label.size();
            if (stLen1 == s.size()) {
                if (!silent)
                    cout << s << '\n';
                if (resList != NULL)
                    resList->push_back (s.substr (0, s.size () - 1));
            } else {
                if (pChld->ptr) {
                    pChld->ptr->SearchAtLevelExact(s, stLen1, numComp, resList,silent);
                }
            }
        }
   }
#else
   vector<Child>::const_iterator ite = children.end();
   for (vector<Child>::const_iterator it = children.begin(); it != ite ; ++it) {
        CALC_NUM_COMP(numComp, it->ptr);
        int len = it->checkCommon(s, stLen);
        if (len >= it->label.size()) {
            int stLen1 = stLen + it->label.size();
            if (stLen1 == s.size()) {
                if (!silent)
                    cout << s << '\n';
                if (resList != NULL)
                    resList->push_back (s.substr (0, s.size () - 1));
            } else {
                if (it->ptr) {
                    it->ptr->SearchAtLevelExact(s, stLen1, numComp, resList,silent);
                }
            }
        }
   }
#endif
}

void CTrie::SearchAtLevelPartExact(
                char *st, int stLen,
                const string &str,
                int switchLen,
                const afuzzycontext & a2, const context & tmpCon2,
                bool exDist2,
                int *numComp, 
                StringList * resList, bool hasTransp, bool silent) const 
{
    int sz = children.size();
    for (int i = 0; i < sz; ++i) {
        const string     &label = children[i].label;
        const char       *pLabelData = label.c_str();
        int              childSz = label.size();
        int              newSz = stLen + childSz;

        PrintStr(stLen, pLabelData);

        strcpy(st + stLen, pLabelData);

        const CTrie *SubTree = children[i].ptr;
        CALC_NUM_COMP(numComp, SubTree);
        if (newSz < switchLen) {
            if (SubTree && !strncmp(st + stLen, str.c_str() + stLen, label.size())) {
                SubTree->SearchAtLevelPartExact(st, newSz,
                                        str,
                                        switchLen,
                                        a2, tmpCon2, 
                                        exDist2,
                                        numComp, resList, hasTransp, silent);
            }
        } else {
            int dist;
            if (!strncmp(st + stLen, str.c_str() + stLen, switchLen - stLen)) {
                int newLen = newSz - switchLen; // The length of the pattern, which is going to be matched using a new context
                context c2(tmpCon2);
                //c2.setStartPos(switchLen);

                if (newLen) {

                    if (a2.checkRTR(st + switchLen, newLen, &c2, SubTree == NULL, &dist))
                     {
                        if (SubTree) {
                            SubTree->SearchAtLevel(
                                        switchLen,
                                        st, newSz,
                                        a2, c2, 
                                        exDist2,
                                        numComp, resList, hasTransp, silent);
                        } else if (!exDist2 || dist == a2.getKerr()
                                        ) {
                            if (!silent)
                            cout << st << '\n';
                            if (resList != NULL)
                            resList->push_back (string(st, newSz - 1));
                        }
                    }
                } 
                    else if (SubTree) {
                    SubTree->SearchAtLevel(
                                        switchLen,
                                        st, newSz,
                                        a2, c2,
                                        exDist2,
                                        numComp, resList, hasTransp, silent);

                }
            }
        }
    }
}

void CTrie::SearchAtLevel(
                char *st, int stLen,
                const afuzzycontext & a1, const context & tmpCon1, 
                bool exDist1,
                int switchLen,
                const afuzzycontext & a2, const context & tmpCon2,
                bool exDist2, int *numComp, 
                StringList * resList, bool hasTransp, bool silent) const 
{
    context tmpCon1New;
    int sz = children.size();
    for (int i = 0; i < sz; ++i) {
        tmpCon1New = tmpCon1; 
        const string     &label = children[i].label;
        int              childSz = label.size();
        int              newSz = stLen + childSz;
        const char       *pLabelData = label.c_str();

        PrintStr(stLen, pLabelData);
        
        strncpy(st + stLen, pLabelData, childSz + 1);

        const CTrie *SubTree = children[i].ptr;
        CALC_NUM_COMP(numComp, SubTree)

        int startPos = stLen;
        int gap = a1.getKerr();

        for (;startPos < newSz; ++startPos) {

            int startPos1 = startPos + 1;

            if (startPos1 > switchLen + gap) {
                break; 
            }
            a1.checkRTRStep(st, &tmpCon1New);
            bool finalState = false;
            if ( !a1.checkRTRState(&tmpCon1New, finalState, NULL)) {
                break;
            }
            if (startPos1 < switchLen - gap) {
                continue;
            }
            if (finalState) {

                context     tmpCon2New = tmpCon2;      
                int         dist1;

                bool        ok2 = true;
                bool        final2 = false;

                const char *pNewSt = st + startPos1;
                for (int k = startPos1; k < newSz; ++k) {
                    a2.checkRTRStep(pNewSt, &tmpCon2New);
                    if (!a2.checkRTRState(&tmpCon2New, final2, &dist1)) {
                         ok2 = false;
                         break;
                    }
                }

                if (startPos1 == newSz || ok2) {
                    if (SubTree) {
                        SubTree->SearchAtLevel(startPos1,
                                    st, newSz,
                                    a2, tmpCon2New, 
                                    exDist2,
                                    numComp, resList, hasTransp, silent);
                    } else if (final2) {
                       if (!exDist2 || dist1 == a2.getKerr()) {
                        if (!silent)
                            cout << st << '\n';
                            if (resList != NULL) {
                                resList->push_back (string(st, newSz - 1));
                        }
                    }
                    }
                }
            } 
        }
        // Additional case to check
        if (startPos == newSz && SubTree) { // it might be also
        // ok to continue with matching for the first part in the subtree
                SubTree->SearchAtLevel(
                st, newSz,
                a1, tmpCon1New, 
                exDist1,
                switchLen,
                a2, tmpCon2,
                exDist2, numComp, 
                resList, hasTransp, silent);
        }
    }
}

void CTrie::SearchAtLevel(
                int startCmpLen,
                char *st, int stLen,
                const afuzzycontext & a, const context & tmpCon, 
                bool exDist,
                int *numComp, 
                StringList * resList, bool hasTransp, bool silent) const 
{
    context c1;
    int sz = children.size();
    for (int i = 0; i < sz; ++i) {
        int dist;
        c1 = tmpCon; // copy is faster than every time construction 
        const string     &label = children[i].label;
        int              childSz = label.size();
        const char       *pLabelData = label.c_str();
        int              newSz = stLen + childSz;

        PrintStr(stLen, pLabelData);
        
        strncpy(st + stLen, pLabelData, childSz + 1);

        const CTrie *SubTree = children[i].ptr;
        CALC_NUM_COMP(numComp, SubTree)

        if (a.checkRTR (st + startCmpLen, newSz - startCmpLen, &c1, SubTree == NULL, &dist))    
        {
            if (SubTree) {
                SubTree->SearchAtLevel(startCmpLen,
                                        st, newSz,
                                        a, c1, 
                                        exDist,
                                        numComp, resList, hasTransp, silent);
            } else if (!exDist || dist == a.getKerr()) {
                if (!silent)
                    cout << st << '\n';
                if (resList != NULL) {
                    resList->push_back (string(st, newSz - 1));
                }
            }
        }
    }
}

bool CTrie::SearchInFile(int fd, 
                         const std::string & s, int kerr,
                         int *numComp, StringList * resList,
                         bool hasTransp, bool silent)
{
    string ss = s + EOS;
    context tmpCon;
    afuzzycontext a;

    static char st[8192]; // enough for us
    st[0] = 0;

    a.init(ss, kerr, hasTransp, &tmpCon);

    return (kerr) ? SearchInFileAtLevel(fd, 0, 
                        0,  // startCmpLen
                        st, 0,
                        a, tmpCon,
                        false,
                        numComp, resList, hasTransp, silent):
                    SearchInFileAtLevelExact (fd, 0, ss, "", numComp, resList, silent);
}

bool CTrie::SearchInFileSplit(int fd,
                 const std::string & s, 
                 int kerr1, bool exDist1,
                 int kerr2, bool exDist2,
                 int *numComp,
                 int switchLen,
                 StringList * resList, 
                 bool hasTransp, bool silent) 
{
    static char st[8192]; // enough for us
    st[0] = 0;

    if (!kerr1 && !kerr2) {
        return SearchInFileAtLevelExact (fd, 0, s.c_str(), st, numComp, resList, silent);
    }
    if (switchLen <= 0) {
        return SearchInFile(fd, s, kerr1 + kerr2, numComp, resList, hasTransp, silent);
    }
    string ss1 = s.substr(0, switchLen);
    string ss2 = s.substr(switchLen) + EOS;
    context       tmpCon1, tmpCon2;
    afuzzycontext a1, a2;

    a1.init(ss1, kerr1, hasTransp, &tmpCon1);
    a2.init(ss2, kerr2, hasTransp, &tmpCon2);

    if (kerr1) {
        // It is possible that ss1 matches an empty string.
        // This case cannot be processed by SearchAtLevel and
        // requires special attention.
        if (ss1.size() == kerr1 || (!exDist1 && ss1.size() < kerr1)) {
            if (!SearchInFileAtLevel (fd, 0, 0,
                    st,  0,
                    a2, tmpCon2, 
                    exDist2,
                    numComp, resList, hasTransp, silent)) {
                return false;
             }
        }
        return SearchInFileAtLevel (fd, 0,
                    st,  0,
                    a1, tmpCon1, 
                    exDist1,
                    switchLen,
                    a2, tmpCon2, 
                    exDist2,
                    numComp, resList, hasTransp, silent);
    } else {
        return SearchInFileAtLevelPartExact (fd, 0,
                    st,  0,
                    s,
                    switchLen,
                    a2, tmpCon2, 
                    exDist2,
                    numComp, resList, hasTransp, silent);
    
    }
    return true;
}

bool CTrie::SearchInFileAtLevelPartExact(
                int fd, off_t offset,
                char *st, int stLen,
                const string &str,
                int switchLen,
                const afuzzycontext & a2, const context & tmpCon2,
                bool exDist2,
                int *numComp, 
                StringList * resList, bool hasTransp, bool silent) 
{

    int buflen;
    if (lseek (fd, offset, SEEK_SET) < 0
        || read (fd, &buflen, sizeof (buflen)) < sizeof (buflen))
        return false;
    int delta = 0;
    if (offset)
        delta = (offset % getpagesize ());
    void *buf =
        mmap (NULL, buflen + delta, PROT_READ, MAP_SHARED, fd,
              offset - delta);
    if (!buf || buf == MAP_FAILED)
        return false;
    char *sbuf = (char *) buf + sizeof (buflen) + delta;
    int sz = *((int *) (void *) sbuf);
    sbuf += sizeof (sz);

    int pd = 0, nd = 0, plen = 1;
    // Now reading leafs 
    // The sequence of storing leafs data:
    // 1. string length +1 (the size of string buffer including trailing `\0'
    // 2. offset
    // 3. string buffer
    
    for (int i = 0; i < sz; ++i) {
        int slen = *(u_int8_t *) sbuf;
        sbuf++;
        off_t sub_off = *(off_t *) sbuf;
        sbuf += sizeof (off_t); // The part of the string stored in the leaf  
        std::string label(sbuf);

        int              childSz = slen - 1;
        int              newSz = stLen + childSz;

        strcpy(st + stLen, sbuf);

        CALC_NUM_COMP(numComp, sub_off)

        if (newSz < switchLen) {
            if (sub_off && !strncmp(st + stLen, str.c_str() + stLen, label.size())) {
                SearchInFileAtLevelPartExact(
                                        fd, sub_off,
                                        st, newSz,
                                        str,
                                        switchLen,
                                        a2, tmpCon2, 
                                        exDist2,
                                        numComp, resList, hasTransp, silent);
            }
        } else {
            int dist;
            if (!strncmp(st + stLen, str.c_str() + stLen, switchLen - stLen)) {
                int newLen = newSz - switchLen; // The length of the pattern, which is going to be matched using a new context
                context c2(tmpCon2);
                //c2.setStartPos(switchLen);

                if (newLen) {

                    if (a2.checkRTR(st + switchLen, newLen, &c2, sub_off == 0,  &dist)) {
                        if (sub_off) {
                            SearchInFileAtLevel(
                                        fd, sub_off,
                                        switchLen,
                                        st, newSz,
                                        a2, c2, 
                                        exDist2,
                                        numComp, resList, hasTransp, silent);
                        } else if (!exDist2 || dist == a2.getKerr()) {
                            if (!silent)
                            cout << st << '\n';
                            if (resList != NULL) {
                                resList->push_back (string(st, newSz - 1));
                            }
                        }
                    }
                } else if (sub_off) {
                    SearchInFileAtLevel(
                                        fd, sub_off,
                                        switchLen,
                                        st, newSz,
                                        a2, c2,
                                        exDist2,
                                        numComp, resList, hasTransp, silent);

                }
            }
        }

        sbuf += slen;
    }
    if (munmap (buf, buflen + delta) < 0)
        return false;
    
    return true;
}

bool CTrie::SearchInFileAtLevel(
                int fd, off_t offset,
                char *st, int stLen,
                const afuzzycontext & a1, const context & tmpCon1, 
                bool exDist1,
                int switchLen,
                const afuzzycontext & a2, const context & tmpCon2,
                bool exDist2, int *numComp, 
                StringList * resList, bool hasTransp, bool silent) 
{
    context tmpCon1New;
    int buflen;
    if (lseek (fd, offset, SEEK_SET) < 0
        || read (fd, &buflen, sizeof (buflen)) < sizeof (buflen))
        return false;
    int delta = 0;
    if (offset)
        delta = (offset % getpagesize ());
    void *buf =
        mmap (NULL, buflen + delta, PROT_READ, MAP_SHARED, fd,
              offset - delta);
    if (!buf || buf == MAP_FAILED)
        return false;
    char *sbuf = (char *) buf + sizeof (buflen) + delta;
    int sz = *((int *) (void *) sbuf);
    sbuf += sizeof (sz);

    int pd = 0, nd = 0, plen = 1;
    // Now reading leafs 
    // The sequence of storing leafs data:
    // 1. string length +1 (the size of string buffer including trailing `\0'
    // 2. offset
    // 3. string buffer
    
    for (int i = 0; i < sz; ++i) {
        tmpCon1New = tmpCon1; 

        int slen = *(u_int8_t *) sbuf;
        sbuf++;
        off_t sub_off = *(off_t *) sbuf;
        sbuf += sizeof (off_t); // The part of the string stored in the leaf  
        std::string label(sbuf);

        int              childSz = slen - 1;
        int              newSz = stLen + childSz;

        strcpy(st + stLen, sbuf);

        CALC_NUM_COMP(numComp, sub_off)

        int startPos = stLen;

        for (;startPos < newSz; ) {
            a1.checkRTRStep(st, &tmpCon1New);

            int startPos1 = startPos + 1;
            bool finalState;
            if (startPos1 > switchLen + a1.getKerr() ||
                !a1.checkRTRState(&tmpCon1New, finalState, NULL)) {
                break; // no matches follow
            }
            if (finalState) {
                bool final1 = !sub_off;

                context tmpCon2New = tmpCon2;
                int     dist1;

                if (startPos1 == newSz || 
                    a2.checkRTR(st + startPos1,
                            newSz - startPos1,
                            &tmpCon2New, final1, &dist1)) {
                    if (sub_off) {
                        if (!SearchInFileAtLevel(
                                        fd, sub_off,
                                        startPos1,
                                        st, newSz,
                                        a2, tmpCon2New, 
                                        exDist2,
                                        numComp, resList, 
                                        hasTransp, silent)) {
                            return false;
                        }
       
                    } else if (!exDist2 || dist1 == a2.getKerr()) {
                        if (!silent) {
                            cout << st << endl;
                        }
                        if (resList) {
                            resList->push_back(string(st, newSz - 1));
                        }
                    }
                }
            }
            startPos = startPos1;
        } /* for (;startPos < newSz;) */
            // Additional case to check
            if (startPos == newSz && sub_off) { // it might be also
            // ok to continue matching the first part in the subtree
                if (!SearchInFileAtLevel(
                                        fd, sub_off,
                                        st, newSz,
                                        a1, tmpCon1New, 
                                        exDist1,
                                        switchLen,
                                        a2, tmpCon2, 
                                        exDist2,
                                        numComp, resList, hasTransp, silent)) {
                   return false;
                }
            }

        sbuf += slen;
    }
    if (munmap (buf, buflen + delta) < 0) {
        return false;
    }
    return true;
}


bool CTrie::SearchInFileAtLevel(int fd, off_t offset,
                                    int startCmpLen,
                                    char * st, int stLen,
                                    const afuzzycontext & a, const context & tmpCon,
                                    bool exDist,
                                    int *numComp, StringList * resList,
                                    bool hasTransp, bool silent)
{
    int buflen;
    if (lseek (fd, offset, SEEK_SET) < 0
        || read (fd, &buflen, sizeof (buflen)) < sizeof (buflen))
        return false;
    int delta = 0;
    if (offset)
        delta = (offset % getpagesize ());
    void *buf =
        mmap (NULL, buflen + delta, PROT_READ, MAP_SHARED, fd,
              offset - delta);
    if (!buf || buf == MAP_FAILED)
        return false;
    char *sbuf = (char *) buf + sizeof (buflen) + delta;
    int num = *((int *) (void *) sbuf);
    sbuf += sizeof (num);

    int pd = 0, nd = 0, plen = 1;
    // Now reading leafs 
    // The sequence of storing leafs data:
    // 1. string length +1 (the size of string buffer including trailing `\0'
    // 2. offset
    // 3. string buffer
    for (int i = 0; i < num; i++) {
        context c1 = tmpCon;

        int slen = *(u_int8_t *) sbuf;
        sbuf++;
        off_t sub_off = *(off_t *) sbuf;
        sbuf += sizeof (off_t); // The part of the string stored in the leaf  

        int              childSz = slen - 1;
        int              newSz = stLen + childSz;

        strcpy(st + stLen, sbuf);
        
        CALC_NUM_COMP(numComp, sub_off)

        if (a.checkRTR (st + startCmpLen, newSz - startCmpLen, &c1, false, NULL))    
        {
            if (sub_off) {
                if (!SearchInFileAtLevel
                    (fd, sub_off, 
                     startCmpLen,
                     st, newSz,
                                        a,
                                        c1, 
                                        exDist,
                                        numComp, resList, hasTransp, silent))
                    return false;
            } else {
                if (!silent)
                    cout << st << '\n';
                if (resList != NULL) {
                    resList->push_back (string(st, newSz - 1));
                }
            }
        }
        sbuf += slen;
    }
    if (munmap (buf, buflen + delta) < 0)
        return false;
    return true;
}

bool CTrie::SearchInFileAtLevelExact(int fd, off_t offset,
                                    const std::string & s, std::string st,
                                    int *numComp, StringList * resList, bool silent)
{
    int buflen;
    if (lseek (fd, offset, SEEK_SET) < 0
        || read (fd, &buflen, sizeof (buflen)) < sizeof (buflen))
        return false;
    int delta = 0;
    if (offset)
        delta = (offset % getpagesize ());
    void *buf =
        mmap (NULL, buflen + delta, PROT_READ, MAP_SHARED, fd,
              offset - delta);
    if (!buf || buf == MAP_FAILED)
        return false;
    char *sbuf = (char *) buf + sizeof (buflen) + delta;
    int num = *((int *) (void *) sbuf);
    sbuf += sizeof (num);

    int pd = 0, nd = 0, plen = 1;
    // Now reading leafs 
    // The sequence of storing leafs data:
    // 1. string length +1 (the size of string buffer including trailing `\0'
    // 2. offset
    // 3. string buffer
    for (int i = 0; i < num; i++) {
        int slen = *(u_int8_t *) sbuf;
        sbuf++;
        off_t sub_off = *(off_t *) sbuf;
        sbuf += sizeof (off_t); // The part of the string stored in the leaf  
        std::string sk = st + std::string (sbuf);

        CALC_NUM_COMP(numComp, sub_off)
        
        if (sub_off) {
            if (sk.size() <= s.size() && !strncmp(s.c_str(), sk.c_str(), sk.size())) {
                if (!SearchInFileAtLevelExact(fd, sub_off, s, sk, numComp, resList, silent))
                    return false;
            }
        } else {
            if (sk == s) {
                if (!silent)
                    cout << sk << '\n';
                if (resList != NULL)
                    resList->push_back (sk.substr (0, sk.length () - 1));
            }
        }
        sbuf += slen;
    }
    if (munmap (buf, buflen + delta) < 0)
        return false;
    return true;
}

CTrie::CTrie()
{
}

CTrie::~CTrie()
{
    for (int i = 0; i < children.size(); ++i) {
        if (children[i].ptr) delete children[i].ptr;
    }
}

void CTrie::Search(const std::string & s, int kerr, int *numComp,
                 StringList * resList, bool hasTransp, bool silent) const
{
    PrintPat(s.c_str(), kerr);
    if (!kerr) {
        SearchExact(s, numComp, resList, silent);
        return;
    }
    static char st[8192]; // enough for us
    st[0] = 0;

    string ss = s + EOS;
    context tmpCon;
    afuzzycontext a;

    a.init(ss, kerr, hasTransp, &tmpCon);

    SearchAtLevel (0,
                    st,  0,
                    a, tmpCon, 
                    false,
                    numComp, resList, hasTransp, silent);
}

void CTrie::SearchSplit(const std::string & s, 
                 int kerr1, bool exDist1,
                 int kerr2, bool exDist2,
                 int *numComp,
                 int switchLen,
                 StringList * resList, bool hasTransp, bool silent) const
{
    PrintPat(s.c_str(), kerr1, kerr2);
    if (!kerr1 && !kerr2) {
        SearchExact(s, numComp, resList, silent);
        return;
    }
    if (switchLen <= 0) {
        Search(s, kerr1 + kerr2, numComp, resList, hasTransp, silent);
        return;
    }
    string ss1 = s.substr(0, switchLen);
    string ss2 = s.substr(switchLen) + EOS;
    context       tmpCon1, tmpCon2;
    afuzzycontext a1, a2;

    a1.init(ss1, kerr1, hasTransp, &tmpCon1);
    a2.init(ss2, kerr2, hasTransp, &tmpCon2);

    static char st[8192]; // enough for us
    st[0] = 0;

    if (kerr1) {
        // It is possible that ss1 matches an empty string.
        // This case cannot be processed by SearchAtLevel and
        // requires special attention.
        if (ss1.size() == kerr1 || (!exDist1 && ss1.size() < kerr1)) {
            SearchAtLevel (0, st,  0,
                    a2, tmpCon2, 
                    exDist2,
                    numComp, resList, hasTransp, silent);
        }
        SearchAtLevel (st,  0,
                    a1, tmpCon1, 
                    exDist1,
                    switchLen,
                    a2, tmpCon2, 
                    exDist2,
                    numComp, resList, hasTransp, silent);
    } else {
        SearchAtLevelPartExact (st,  0,
                    s,
                    switchLen,
                    a2, tmpCon2, 
                    exDist2,
                    numComp, resList, hasTransp, silent);
    
    }
}

void CTrie::SearchExact(const std::string & s, int *numComp,
                      StringList * resList, 
                      bool silent) const
{
    string ss = s + EOS;

    SearchAtLevelExact(ss, 0, numComp, resList, silent);
}

void CTrie::Insert(const std::string &s)
{
    int indx = -1;
    int maxLen = 0;
    int len;
    for (int i = 0; i < children.size(); ++i) 
    if ((len = children[i].checkCommon(s, 0)) > maxLen) {
        indx = i;
        maxLen = len;
    }
    if (!maxLen) {
        Child chld(s);

        vector<Child>::iterator it = lower_bound(children.begin(), children.end(), chld);
        children.insert(it, chld);
        return;
    }
    string rest(s.begin() + maxLen, s.end());
    if (rest.empty()) {
        throw string("String '") + s + string("' length: ") + Format("%d", s.size()) + string(" is a duplicate");
    }
    Child &maxChld = children[indx];
    // If the string to be inserted has label as its prefix
    // insert it into the corresponding subtree
    if (maxLen == maxChld.label.size()) {
        if (maxChld.ptr) {
            maxChld.ptr->Insert(rest);
        } else {
            throw string("String '") + s + string("' has a proper prefix in the dictionary!");
        }
        return;
    }
    //... otherwise the node must be splitted and subtree be created
    string restChld(maxChld.label.begin() + maxLen, maxChld.label.end());
    // Shorten the label
    maxChld.label.erase(maxChld.label.begin() + maxLen, maxChld.label.end());
    CTrie *oldPtr = maxChld.ptr;
    CTrie *pNewNode= new CTrie;
    maxChld.ptr = pNewNode;
    pNewNode->children.push_back(Child(restChld, oldPtr));
    pNewNode->children.push_back(Child(rest));
    if (pNewNode->children[0] > pNewNode->children[1]) {
        swap(pNewNode->children[0], pNewNode->children[1]);
    }
}

void CTrie::Print(int level)
{
    string indent(level, ' ');
    cout << indent << "Node address =" << this << " level=" << level << "\n";
    for (int i = 0; i < children.size(); ++i) {
        cout << indent << " Key " << children[i].label << " : " << children[i].ptr << endl;
        if (children[i].ptr) {
            children[i].ptr->Print(level + 1);
        }
    }
}

int CTrie::DumpToFile(int fd, off_t offset)
{
// We store the data in the following order
// 1. the buffer size, the number of leaves
// 2. the leaf keys themselvs, with zero offsets
// 3. leaf subtrees - updating offsets

    int buflen = 0;
    int sub_sz = 0;
    int cnt = children.size();
    buflen += sizeof (cnt) + sizeof (buflen);
    for (int i = 0; i < cnt; ++i) {
        buflen += sizeof (off_t) /* <- pointer size */ + 
                 + 1 /* string record size (must be<255) */  
                 + children[i].label.size() + 1;
    }
    int delta = 0;
    if (offset)
        delta = (offset % getpagesize ());
    if (lseek (fd, offset + buflen - 1, SEEK_SET) < 0
        || write (fd, "", 1) < 1) {
        return -1;
    }
    void *buf = mmap (NULL, buflen + delta, PROT_WRITE | PROT_READ, MAP_SHARED, fd, offset - delta);
    if (!buf || buf == MAP_FAILED)
        return -1;
    char *sbuf = (char *) buf + delta;
    // Step 1.
    *(int *) (void *) sbuf = buflen;
    sbuf += sizeof (buflen);
    *(int *) (void *) sbuf = cnt;
    sbuf += sizeof (cnt);
    // Step 2.
    off_t *ptrs[cnt];
    for (int i = 0; i < cnt; ++i) {
        // String length
        int len = children[i].label.size();
        *((u_int8_t *) (void *) sbuf) = len + 1;
        sbuf++;
        // Zero offset
        ptrs[i] = (off_t *) ((void *) sbuf);
        *ptrs[i] = 0;
        // Leaf key
        sbuf += sizeof (off_t);
        memcpy (sbuf, children[i].label.c_str (), len + 1);
        sbuf += len + 1;
    }
    // Step 3.
    for (int i = 0; i < cnt; ++i) {
        // Diving into subtree
        int s_sz = (children[i].ptr)? children[i].ptr->DumpToFile (fd, offset + buflen + sub_sz) : 0;
        if (s_sz > 0)
            *ptrs[i] = offset + buflen + sub_sz;
        if (s_sz < 0) {
            munmap (buf, buflen + delta);
            return -1;
        }
        sub_sz += s_sz;
    }

    assert (sbuf - (char *) buf == delta + buflen);

    if (munmap (buf, buflen + delta) < 0)
        return -1;
    // Returning the total number of bytes written
    return buflen + sub_sz;
}
