#ifndef  TRIE_H
#define  TRIE_H

#include <iostream>
#include <string>
#include <sys/types.h>
#include <sys/timeb.h>
#include <fstream>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>

const std::string EOS = "\1";

#include <vector>
#include <string>
#ifdef USE_AFUZZYCONTEXT
#   include "afuzzycontext.h"
#else
#   include "edistk.h"
#endif
#include "interface.h"

class CTrie {
    class Child {
        CTrie  *ptr;
        string label;
        friend class CTrie;
    public:

        Child(const string & label = "", CTrie *ptr = NULL) : label(label), ptr(ptr) {}
        ~Child(){} // do not delete here it will be done on the upper level
        CTrie *getPtr(){return ptr;}
        bool operator<(const Child &o) const {if (label != o.label) return label < o.label;return ptr < o.ptr;}
        bool operator>(const Child &o) const {if (label != o.label) return label > o.label;return ptr > o.ptr;}
        inline int checkCommon(const string &s, int start) const;
        inline int checkCommon(const char* pStr, int start) const;
    };
    vector<Child>  children;
  protected:
    /* 
       The following functions perform a search startings from the current node
       The function SearchAtLevelExact uses the results of prefix comparisons
       at upper levels. 
    */
    bool SearchAtLevelExact(const char *pStr, int StrLen, int StartLen, int *numComp,
                             StringList * resList, bool silent) const;
    void SearchAtLevel(const std::string & s, 
                        const char * st, int stLen,
#ifdef USE_AFUZZYCONTEXT
                        const afuzzycontext & a, 
                        const context & tmpCon,
#else
                        const checkDistContext & c, 
#endif
                        int *numComp,
                        StringList * resList, bool hasTransp, bool silent) const;
    static bool SearchInFileAtLevel(int fd, off_t offset,
                                    const std::string & s, std::string st,
#ifdef USE_AFUZZYCONTEXT
                                    const afuzzycontext & a, 
                                    const context & tmpCon,
#else
                                    const checkDistContext & c,
#endif
                                    int *numComp, StringList * resList,
                                    bool hasTransp, bool silent);
    static bool SearchInFileAtLevelExact(int fd, off_t offset,
                                    const std::string & s, std::string st,
                                    int *numComp, StringList * resList, bool silent);
  public:

     CTrie();
    ~CTrie();

    void Search(const std::string & s, int kerr, int *numComp,
                 StringList * resList, bool hasTransp, bool silent) const;
    static bool SearchInFile(int fd, const std::string & s, int kerr,
                             int *numComp, StringList * resList,
                             bool hasTransp, bool silent);
    bool SearchExact(const std::string & s, int *numComp,
                      StringList * resList, 
                      bool silent) const;
    void Insert(const std::string &s);

// Storing node content to the file
    int DumpToFile(int fd, off_t offset);
    void Print(int level = 0);
    void countBranch(vector<int> &cnt, vector<int> &cntBranch, vector<int> &cntLeaves, int level = 0);
    int  countLeaves();
};

int CTrie::Child::checkCommon(const std::string &Key, int start) const
{
    int i = 0, k = start;

    int kSize = Key.size() - start;
    int iMax = min(int(label.size()), kSize);
    for (; i < iMax;)
        if (Key[k] == label[i]) {
            i++; k++;
        }
        else
            break;
    return i;
}

int CTrie::Child::checkCommon(const char* pKey, int start) const
{
    int i = 0;

    int iMax = int(label.size());
    const char *p = pKey + start;

    for (; *p && i < iMax;) {
        if (*p++ == label[i]) {
            i++;
        }
        else break;
    }
    return i;
}

#endif
