#include <string>
#include <vector>
#include <iostream>
#include <iomanip>
#include <algorithm>

const int ChangeCost = 1, InsertCost = 1, DeleteCost = 1, TranspCost = 1;

using namespace std;

int EditDistanceOld (const string & ar1, const string & ar2, bool checkTransp, bool printMattr)
{
    const int m = ar1.size (), n = ar2.size ();
    const char *dt1 = ar1.data (), *dt2 = ar2.data ();
    int a, b, p1, p2, p3, p4;

    int graph[m + 1][n + 1];


    // Fill first row and column of graph
    graph[0][0] = 0;
    for (a = 1; a <= m; a++)
        graph[a][0] = graph[a - 1][0] + InsertCost;
    for (a = 1; a <= n; a++)
        graph[0][a] = graph[0][a - 1] + DeleteCost;


    for (a = 1; a <= m; a++) {
        for (b = 1; b <= n; b++) {

            p1 = graph[a - 1][b - 1] + ChangeCost * int (dt1[a - 1] !=
                                                         dt2[b - 1]);
            p2 = graph[a - 1][b] + InsertCost;
            p3 = graph[a][b - 1] + DeleteCost;

            int p = min(p1, min(p2, p3));
            p4 = p + 1;
            if (checkTransp && a > 1 && b > 1) { // checking for transpostions is done so as resulting function be a metric !!! 
                if (dt1[a - 1] == dt2[b - 2] && dt1[a - 2] == dt2[b - 1]) {
                    p4 = graph[a - 2][b - 2] + TranspCost;
                }
                if (dt1[a - 1] == dt2[b - 2]) {
                    int a1min = max(a - p, 0); // there is no need to go below
                    // the case of a1 == a - 2 is checked above
                    int icost = InsertCost;
                    for (int a1 = a - 3; a1 >= a1min; --a1, icost += InsertCost)
                    if (dt1[a1] == dt2[b - 1]) {
                        p4 = min(p4, graph[a1][b - 2] + TranspCost + icost); 
                    }
                }
                if (dt1[a - 2] == dt2[b - 1]) {
                    int b1min = max(b - p, 0);
                    // the case of b1 == b - 2 is checked above
                    int icost = InsertCost;
                    for (int b1 = b - 3; b1 >= b1min; --b1, icost += InsertCost)
                    if (dt1[a - 1] == dt2[b1]) {
                        p4 = min(p4, graph[a - 2][b1] + TranspCost + icost);
                    }
                }
            }
            p = min(p, p4);
            graph[a][b] = p;
            //cout<<ar1[a-2]<<ar1[a-1]<<" "<<ar2[b-2]<<ar2[b-1]<<" "<<p4<<" "<<p1<<" "<<a<<" "<<b<<"\n";

        }


    }
    if (printMattr) {
        for (int i = 0; i <= m; ++i) {
            for (int k = 0; k <= n; ++k) {
                cout << setw(2) << graph[i][k] << " ";
            }
            cout << endl;
        }
    }

    return graph[m][n];
}


int EditDistance(const string & ar1, const string & ar2, bool checkTransp, bool printMattr)
{
    //return EditDistanceOld(ar1, ar2, checkTransp, printMattr);

    const int m = ar1.size (), n = ar2.size ();
    const char *dt1 = ar1.data (), *dt2 = ar2.data ();
    int a, b, p1, p2, p3, p4;

    int graph[m + 1][n + 1];


    // Fill first row and column of graph
    graph[0][0] = 0;
    for (a = 1; a <= m; a++)
        graph[a][0] = graph[a - 1][0] + InsertCost;
    for (a = 1; a <= n; a++)
        graph[0][a] = graph[0][a - 1] + DeleteCost;


    if (!checkTransp) {
        for (a = 1; a <= m; a++) {
            for (b = 1; b <= n; b++) {

                p1 = graph[a - 1][b - 1] + ChangeCost * int (dt1[a - 1] !=
                                                             dt2[b - 1]);
                p2 = graph[a - 1][b] + InsertCost;
                p3 = graph[a][b - 1] + DeleteCost;

                graph[a][b] =  min(p1, min(p2, p3));
                //cout<<ar1[a-2]<<ar1[a-1]<<" "<<ar2[b-2]<<ar2[b-1]<<" "<<p4<<" "<<p1<<" "<<a<<" "<<b<<"\n";
            }

        }
    } else {
        static unsigned CP[255]; // NOTE: note thread-safe!

        for (a = 0; a < m; a++) {
            CP[(unsigned char)ar1[a]] = 0;
        }
        for (b = 0; b < n; b++) {
            CP[(unsigned char)ar2[b]] = 0;
        }

        for (a = 1; a <= m; a++) {
            unsigned CS = 0;
            for (b = 1; b <= n; b++) {
                bool NonEqFlag = dt1[a - 1] != dt2[b - 1];

                p1 = graph[a - 1][b - 1] + ChangeCost * int (NonEqFlag);
                p2 = graph[a - 1][b] + InsertCost;
                p3 = graph[a][b - 1] + DeleteCost;

                int p = min(p1, min(p2, p3));
                p4 = p + 1;

                if (unsigned bprime = CS) {
                    if (unsigned aprime = CP[(unsigned char)ar2[b - 1]]) {
                        p4 = graph[aprime - 1][bprime -1] + (a - aprime) + (b - bprime) - 1;
                    }
                }
                graph[a][b] = min(p, p4);
                //cout<<ar1[a-2]<<ar1[a-1]<<" "<<ar2[b-2]<<ar2[b-1]<<" "<<p4<<" "<<p1<<" "<<a<<" "<<b<<"\n";
                
                if (!NonEqFlag) {
                    CS = b;
                } 

            }
            CP[(unsigned char)ar1[a - 1]] = a;
        }
    }

    if (printMattr) {
        for (int i = 0; i <= m; ++i) {
            for (int k = 0; k <= n; ++k) {
                cout << setw(2) << graph[i][k] << " ";
            }
            cout << endl;
        }
    }

    return graph[m][n];
}
