#include <fcntl.h>
#include <sys/mman.h>

#include <iostream>
#include <string>

#include "myagrep.h"
#include "../Common/utils.h"
#include "../Common/filesize.h"

using namespace std;

void usage(const char *progName)
{
    cerr << "Usage: " << progName << " -fFilterType -nErrorNum -t<num> Pattern FileName \n";
    cerr << " Type: none, exact, exact_immediate, agrep, bitap\n";
    cerr << " <num>: 1 - count transpositions, 0 - do not count transpositions \n";
}

int main(int argc, char *argv[])
{
    AlgoType            type;
    unsigned int        kerr;
    string              pattern;
    string              fileName;
    bool                hasTransp = false;

    
    if (argc != 6) {
        cerr << "Wrong argument number" << endl;
        usage(argv[0]);
        return 1;
    }
    if (strncmp(argv[1], "-f", 2)) {
        cerr << "Wrong filter type argument\n";
        usage(argv[0]);
        return 1;
    }
    const char *pFilterType = argv[1] + 2;
    if (!strcmp(pFilterType, "none")) {
        type = FILTER_NONE;
    } else if (!strcmp(pFilterType, "exact")) {
        type = FILTER_EXACT;
    } else if (!strcmp(pFilterType, "exact_immediate")) {
        type = FILTER_EXACT_IMMEDIATE;
    } else if (!strcmp(pFilterType, "agrep")) {
        type = FILTER_AGREP;
    } else if (!strcmp(pFilterType, "combined")) {
        type = FILTER_COMBINED;
    } else {
        cerr << "Unknown filter type: " << pFilterType << endl;
        usage(argv[0]);
        return 1;
    }
    if (strncmp(argv[2], "-n", 2)) {
        cerr << "Wrong 'error number' argument\n";
        usage(argv[0]);
        return 1;
    }
    if (strncmp(argv[3], "-t", 2)) {
        cerr << "Wrong transp. argument\n";
        usage(argv[0]);
        return 1;
    }

    const char *pErrNum = argv[2] + 2;
    const char *pHasTransp = argv[3] + 2;
    kerr = atol(pErrNum);
    hasTransp = atol(pHasTransp);

    pattern = argv[4];
    fileName = argv[5];
    
    //cout << "Searching '" << pattern << "' within " << kerr << " errors, file " << fileName << " size " << filesize << endl;

    int fd = open(fileName.c_str(), O_RDONLY);
    if (fd < 0) {
        Perror("Cannot open file '" + fileName + "'");
        return -1;
    }

    myagrep agr(pattern.c_str(), kerr, hasTransp);
    agr.findPatternFile(fd,  type, '\n');

            
    return 0;
}
