#include "myagrep.h"

#include <iostream>
#include <algorithm>

using namespace std;

myagrep::myagrep(const char *p, unsigned int k, bool hasTransp):
 RI (NULL), mHasTransp(false)
{
    init(p, k, hasTransp);
}

void myagrep::init(const char *p, unsigned int maxErr, bool hasTransp)
{
    mPattern = p;
    mHasTransp = hasTransp;
    m = mPattern.size();
    kerr = maxErr;

    // Firstly init exact searching
    initBoyerMoore(p);
    // Secondly agrep filter, kerr and 
    initAgrepFilt();

    unsigned int FilterMap[256];
    unsigned int Map[256];
    Uint S[m + 1];
    Uint FilterS[m + 1];
    unsigned char PatFilter[sizeof (Uint) * 8 + 1];

    FilterSet = false;
    memset (Map, 0, sizeof (Map));
    if (RI)
        delete[]RI;
    
    RI = new Uint[kerr + 1];
    S[0] =  RI[0] = 0;

    for (unsigned int i = 0, cnt = 0; i < m; i++) {
        int l = Map[(unsigned char) p[i]];
        if (!l) {
            l = Map[(unsigned char) p[i]] = ++cnt;
            S[l] = 0;
        }
        S[l] |= Uint(1) << i;
    }


    for (unsigned int d = 0; d <= kerr; d++)
        RI[d] = (Uint(1) << d) - 1;

    mask_ok = (Uint(1) << (m - 1));
    r_size = sizeof (Uint) * (kerr + 1);
    unsigned int p_len = m;
    if (p_len > sizeof (Uint) * 8)
        p_len = sizeof (Uint) * 8;
    /* If kerr is zero then no filter is needed! */
    bool UseExactFilter = (p_len >= 2 * (kerr + 1));

    // Create faster map 
    for (unsigned int i1 = 0; i1 < 256; ++i1) {
            SMapArr[i1] = S[Map[i1]];
    }

    // For transpositions filter strings are one character shorter
    if (UseExactFilter && p_len >= (kerr + 1) * (1 + int(mHasTransp))) {

        filterInitValue = ~0;
        FilterSet = true;
        memset (FilterMap, 0, sizeof (FilterMap));
        
        FilterS[0] = filterInitValue;
        // Not let's fill the interleaved pattern
        // If this pattern is used for matches with transpositions
        // We are using shortened string pieces
        unsigned int dd = p_len / (kerr + 1);
        p_len = dd * (kerr + 1);

#if 0
        cout << "FILTER SET hasTransp = " << mHasTransp << " DD = " << (dd - mHasTransp) << " \n";
#endif

        for (unsigned int i = 0, cnt = 0; i < dd - int(mHasTransp); i++)
            for (unsigned int j = 0; j < (kerr + 1); j++, cnt++) {
                PatFilter[cnt] = (unsigned char) p[j * dd + i];
            }

        if (mHasTransp) {
            p_len -= (kerr + 1);
        }
        PatFilter[p_len] = 0;
#if 0
        cout << "PatFilter: " << PatFilter << endl;
#endif

        for (unsigned int i = 0, cnt = 0; i < p_len; i++) {
            int l = FilterMap[PatFilter[i]];
            if (!l) {
                l = FilterMap[PatFilter[i]] = ++cnt;
                FilterS[l] = filterInitValue;
            }
            FilterS[l] &= (~((Uint(1) << i)));
        }
        filter_ok = 0;
        for (unsigned int i = p_len - kerr - 1; i < p_len; i++) { // kerr+1 times
            filter_ok |= (Uint(1) << i);
        }
#if 0
        cout << hex << " p_len :" << p_len << " Filter ok: " << filter_ok << endl;
#endif

        for (unsigned int i1 = 0; i1 < 256; ++i1) {
            FilterSFilterMapArr[i1] = FilterS[FilterMap[i1]];
        }
    }
}

void myagrep::initAgrepFilt()
{
    int i, j, k, M, p, shift, b_size;
    unsigned hash;
    string &patLoc = mPattern;

    M = m / (kerr + 1);
    if (M < mHasTransp) {
        return;
    }
    int mLoc = m;

    M = mLoc/(kerr+1);
    p = mLoc - M*(kerr+1);

#if 0
    cout << "\n\nPreparing: " << patLoc << " Kerr = " << kerr << endl;
#endif

	for (i = 0; i < 256; i++) {
            mShift[i] = M - mHasTransp;
        }
	for (i = mLoc-1; i>=p ; i--) {
		shift = (mLoc - 1 - i) % M;
		hash = patLoc[i];
	        if(mShift[hash] > shift) mShift[hash] = shift;
	}
//#ifndef NDEBUG
#if 1
	for (i = 0; i < 256; i++) {
            if (mShift[i] > M - mHasTransp) abort();
        }
#endif
#if 0
	for(i=0; i<mLoc; i++) cout << " " << int(mShift[patLoc[i]]) << ",";
	cout << endl;
#endif

	shift_1 = M - mHasTransp;
	for(i=0; i< kerr +1; i++) {
		j = mLoc-1 - M*i;
		for(k = 1; k < M; k++) {
			for(p = 0; p < kerr+1; p++) 
				if(patLoc[j-k] == patLoc[mLoc-1-M*p]) 
					if(k < shift_1) shift_1 = k;
		}
	}
#if 0
	cout << "\nshift_1 = " << shift_1;
#endif
	if(shift_1 == 0) shift_1 = 1;
	
        memset(mMember, 0, sizeof mMember);

        b_size = 3;
	if (M - mHasTransp < b_size) b_size = M - mHasTransp;
	for(i=0; i<kerr+1; i++) {
		j = mLoc-1 - M*i;
		hash = 0;
		for(k=0; k<b_size; k++) {
			hash = (hash << 2) + patLoc[j-k];
		}
#if 0
	cout << " hash = " << hash << ",";
#endif
		mMember[hash] = 1;
	}
#if 0
    cout << "\nEnd of Preparing: " << patLoc << endl;
#endif
}

void myagrep::initBoyerMooreBadChar(const char *pat) 
{
    int i;
            
    for (i = 0; i < 256; ++i) {
        boyerMooreBadChar[i] = m;
    }
        
    for (i = 0; i < m - 1; ++i) {
        boyerMooreBadChar[(unsigned char)pat[i]] = m - i - 1;
    }
}
 
 
void myagrep::suffixes(const char *pat) 
{
    int f, g, i;
            
    suff[m - 1] = m;
    g = m - 1;
    for (i = int(m) - 2; i >= 0; --i) {
        if (i > g && suff[i + m - 1 - f] < i - g)
            suff[i] = suff[i + m - 1 - f];
        else {
            if (i < g) {
                g = i;
            }
            f = i;
            while (g >= 0 && pat[g] == pat[g + m - 1 - f]) {
                --g;
            }
            suff[i] = f - g;
        }
    }
}
 
void myagrep::initBoyerMooreGoodSuffix(const char *pat) 
{
    int i, j;
            
    suffixes(pat);
               
    for (i = 0; i < m; ++i) {
        boyerMooreGoodSuffix[i] = m;
    }

    j = 0;
    for (i = m - 1; i >= -1; --i) {
        if (i == -1 || suff[i] == i + 1) {
            for (; j < m - 1 - i; ++j) {
                if (boyerMooreGoodSuffix[j] == m)
                    boyerMooreGoodSuffix[j] = m - 1 - i;
            }
        }
    }

    for (i = 0; i <= int(m) - 2; ++i) {
        assert(i <= int(m) -2);
        assert(suff[i] <= m - 1);
        boyerMooreGoodSuffix[m - 1 - suff[i]] = m - 1 - i;
    }
}

void myagrep::initBoyerMoore(const char *pat)
{
    m = strlen(pat);
    FilterSet = true;
#if 0
    cout << "#" << pat << endl;
#endif

    suff.resize(m);
    if (!m) {
        return;
    }

    fill(suff.begin(), suff.end(), 0);
    boyerMooreGoodSuffix.resize(m);
    fill(boyerMooreGoodSuffix.begin(), boyerMooreGoodSuffix.end(), 0);

    /* Preprocessing */
    initBoyerMooreGoodSuffix(pat);
    initBoyerMooreBadChar(pat);
}

myagrep::~myagrep ()
{
    delete[]RI;
}

void myagrep::findPatternBoyerMoore(const char *pBeg, const char *pEnd, char termChar, bool silent, StringList *res) const
{
    int i, j;
    const char *pat = mPattern.c_str();

    if (mPattern.empty()) {
        return;
    }
    if (!m) {
        return;
    }
    if (pEnd - pBeg < m) {
        return;
    }
    const char *pEnd1 = pEnd - m + 1;
    for (const char *pCur = pBeg; pCur < pEnd1; ) {
        i = m - 1; 
        //for (; i >= 0 && pat[i] == pCur[i]; --i);
        switch (m % 8) { // here m > 0
#undef CHECK_CODE
#define  CHECK_CODE if (pat[i] != pCur[i]) goto cont;--i;
            case 0: do { CHECK_CODE
            case 7: CHECK_CODE
            case 6: CHECK_CODE
            case 5: CHECK_CODE 
            case 4: CHECK_CODE 
            case 3: CHECK_CODE
            case 2: CHECK_CODE
            case 1: CHECK_CODE 
            } while (i >= 0);
        }
cont:
        if (i < 0) {
            if ((pCur > pBeg && pCur[-1] != termChar) ||
                (pCur[m] != termChar)) { // this is safe, since pEnd1 < pEnd

                pCur += boyerMooreGoodSuffix[0]; // this is not a whole-line match
            } else {
                if (!silent) {
                    (void)write(1, pCur, m);
                    putchar('\n');
                }
                if (res) {
                    res->push_back(string(pCur, m));
                }
                pCur += m;
            }
                
        } else {
            pCur += std::max(boyerMooreGoodSuffix[i], 
                             int(boyerMooreBadChar[(unsigned char)pCur[i]] - m + 1 + i));
        }
    }

    return;
}

void myagrep::findPatternNone(const char *pBeg, const char *pEnd, char termChar, bool silent, StringList *res) const
{
    for (const char *p = pBeg; pBeg < pEnd; pBeg = p + 1) {
        p = const_cast<const char *>(strchr(pBeg, termChar));
        if (!p) p = pEnd;
        if (checkRTR(pBeg, p)) {
            if (!silent) {
                (void)write(1, pBeg, p - pBeg);
                putchar('\n');
            }
            if (res) {
                res->push_back(string(pBeg, p - pBeg));
            }
        }
    };
}

void myagrep::findPatternExact(const char *pBeg, const char *pEnd, char termChar, bool silent, StringList *res) const
{
    if (!kerr) {
        findPatternBoyerMoore(pBeg, pEnd, termChar, silent, res);
        return;
    }
    Uint FilterR = filterInitValue;
    static const char *apCand[1024*1];

    for (;pBeg < pEnd && *pBeg == termChar; ++pBeg) {
    }

    int         candCnt = 0;
    int         candCntMax = sizeof(apCand) / sizeof(*apCand);


    if (pEnd <= pBeg) {
       return;
    }
    const char *pCur = pBeg;
    unsigned int kerr1 = kerr + 1;

    switch ((pEnd - pBeg) & 7) {
        case 0: do {
#undef  CHECK_CODE
#define CHECK_CODE \
            FilterR <<= kerr + 1;  \
            FilterR |= FilterSFilterMapArr[(unsigned char) *pCur];\
            /*cout << "CUR: " << *pCur << " FilterR: " << hex << FilterR << " FilterR & filter_ok " << (FilterR & filter_ok) << endl; */\
            if ((FilterR & filter_ok) != filter_ok) {\
                apCand[candCnt++] = pCur;\
                if (candCnt >= candCntMax) {\
                    checkCand(pBeg, pEnd, apCand, candCnt, termChar, silent, res);\
                    candCnt = 0;\
                }\
                FilterR = filterInitValue;\
            }\
            ++pCur;
            CHECK_CODE
        case 7:
            CHECK_CODE
        case 6:
            CHECK_CODE
        case 5:
            CHECK_CODE
        case 4:
            CHECK_CODE
        case 3:
            CHECK_CODE
        case 2:
            CHECK_CODE
        case 1:
            CHECK_CODE
        } while (pCur < pEnd);
    }
    checkCand(pBeg, pEnd, apCand, candCnt, termChar, silent, res);
}

void myagrep::findPatternAgrep(const char *pBeg, const char *pEnd, char termChar, bool silent, StringList *res) const
{
    if (!kerr) {
        findPatternBoyerMoore(pBeg, pEnd, termChar, silent, res);
        return;
    }
    unsigned chunkSize = 1024 * 32 + m;
    int addMem = m;
    char text[chunkSize + addMem + 1];
    
    unsigned textSize = pEnd - pBeg;
    for (const char *p = pBeg; p < pEnd; p += chunkSize - m) {
        unsigned sz = std::min((unsigned int)(pEnd - p), chunkSize);

        memcpy(text, p, sz);
        text[sz] = termChar;
        // To make algorithm working correctly we must put a pattern
        // after the text
        for(int i = sz; i < sz + addMem; ++i) {
            text[i] = mPattern[m - 1];
        }
        text[sz + mPattern.size()] = termChar;
        agrepIntern(text, text + sz, termChar, silent, res);
    }
}

void myagrep::agrepIntern(const char *text, const char *textend, char termChar, bool silent, StringList * res) const
{
  int           d1 = shift_1;
  int           i, shift, j;
  unsigned int  r1;
  int           SHIFT, HASH;
  int           M = m / (kerr + 1);
  assert(M);
  const char    *textstart = text;

  static const  char *apCand[1024*1];
  int           candCnt = 0;
  int           candCntMax = sizeof(apCand) / sizeof(*apCand);

  if(M - mHasTransp < 3) r1 = M - mHasTransp;
  else r1 = 3;
  shift = M - 1;
  int  prevCand = 0;
  j = 0;

  while (text < textend) {
	shift = mShift[*(text += shift)];
        assert(shift <= M - mHasTransp);
        assert(text < textend + M);
	while(shift) {
		shift = mShift[*(text += shift)];
                assert(shift <= M - mHasTransp);
                assert(text < textend + M);
		shift = mShift[*(text += shift)];
                assert(shift <= M - mHasTransp);
                assert(text < textend + M);
	}
        j = 1; HASH = *text;
	while(j < r1) { 
                assert(text >= textstart + j);
                HASH = (HASH << 2) + *(text-j);
	        j++; 
        }
        assert(HASH < sizeof(mMember)/sizeof(*mMember));
        if (mMember[HASH] && text < textend) { 
                i = text - textstart;
                apCand[candCnt++] = text;
                if (candCnt >= candCntMax) {
                    checkCand(textstart, textend, apCand, candCnt, termChar, silent, res);
                    candCnt = 0;
                }
		shift = d1;
        } else shift = d1;
  }
  checkCand(textstart, textend, apCand, candCnt, termChar, silent, res);
}

void myagrep::findPatternExactImmediate(const char *pBeg, const char *pEnd, char termChar, bool silent, StringList *res) const
{
    if (!kerr) {
        findPatternBoyerMoore(pBeg, pEnd, termChar, silent, res);
        return;
    }
    Uint FilterR = filterInitValue;

    for (;pBeg < pEnd && *pBeg == termChar; ++pBeg) {
    }

    const char *pLast = NULL;

    if (pEnd <= pBeg) {
       return;
    }
    const char *pCur = pBeg;
    unsigned int kerr1 = kerr + 1;

    switch ((pEnd - pBeg) & 7) {
        case 0: do {
#undef CHECK_CODE
#define CHECK_CODE \
            FilterR <<= kerr + 1;  \
            FilterR |= FilterSFilterMapArr[(unsigned char) *pCur];\
            if ((FilterR & filter_ok) != filter_ok && (pCur > pLast || !pLast)) {\
                const char *p1 = pCur;\
                for (;p1 > pBeg; --p1) {\
                    if (*p1 == termChar) {++p1;break;}\
                }\
                pLast = strchr(pCur, termChar);\
                if (!pLast) {\
                     break;\
                }\
                if (checkRTR(p1, pLast)) {\
                    if (!silent) {(void)write(1, p1, pLast - p1); putchar('\n');}\
                    if (res) {res->push_back(string(p1, pLast - p1));}\
                }\
                FilterR = filterInitValue; \
            }\
            ++pCur;
            CHECK_CODE
        case 7:
            CHECK_CODE
        case 6:
            CHECK_CODE
        case 5:
            CHECK_CODE
        case 4:
            CHECK_CODE
        case 3:
            CHECK_CODE
        case 2:
            CHECK_CODE
        case 1:
            CHECK_CODE
        } while (pCur < pEnd);
    }
}

void myagrep::findPattern(AlgoType checkMeth, char termChar, 
                const char *pBeg, const char *pEnd,
                bool silent, StringList *res) const
{
    if (!kerr && checkMeth != FILTER_NONE) {
        findPatternBoyerMoore(pBeg, pEnd, termChar, silent, res);
        return;
    } else if (checkMeth == FILTER_EXACT && FilterSet) {
        findPatternExact(pBeg, pEnd, termChar, silent, res);
        return;
    } else if (checkMeth == FILTER_AGREP && FilterSet) {
        findPatternAgrep(pBeg, pEnd, termChar, silent, res);
        return;
    } else if (checkMeth == FILTER_EXACT_IMMEDIATE && FilterSet) {
        findPatternExactImmediate(pBeg, pEnd, termChar, silent, res);
        return;
    } else if (checkMeth == FILTER_COMBINED && FilterSet) {
        if (kerr < 2) {
	        findPatternAgrep(pBeg, pEnd, termChar, silent, res);
        } else {
            findPatternExact(pBeg, pEnd, termChar, silent, res);
        }
    } else {
        findPatternNone(pBeg, pEnd, termChar, silent, res);
    }
}

void myagrep::findPatternFile(int fd, AlgoType checkMeth, 
                char termChar, bool silent, StringList *res) const
{
    int  nr = 0;
    int  maxReadLen = 1024 * 8;
    char buf[maxReadLen * 2];

    nr = read(fd, buf, maxReadLen);
    int start = 0;
    while (nr > 0)
    {
        const char *pBeg = buf, *pEnd = pBeg + start + nr;

        findPattern(checkMeth, termChar, pBeg, pEnd, silent, res);
        
        int pos;
        for (pos = nr - 1; pos > 0 && pBeg[pos] != termChar; --pos);
        start = nr - 1 - pos;
        if (start) {
            start = nr - 1 - pos;
            memmove(buf, pBeg + pos + 1, start);
        }
        nr = read(fd, &buf[start], maxReadLen);
#if 0
        // this is debug check
        if (start) {
            const char *p = strchr(buf, termChar);
            if (!p || p > buf + nr + start) {
                p = buf + nr + start;
            }
            printf("Moved piece of size (%d) '%.*s' -> '%.*s' \n", start, start - 1, buf, p - pBeg, buf);
        }
#endif
    }
    if (errno) {
        perror("Cannot read file");
        return;
    }
}
