#include <iostream>
#include <vector>
#include <string>
#include <fstream>
#include <time.h>
#include <stdlib.h>

#include <edist.h>

using namespace std;

void usage(const char *pProg) 
{
    cerr << "Usage : " << pProg << " <file> <testcnt> " << endl;
}

int main(int argc, char *pArgv[])
{
    if (argc != 3) {
        usage(pArgv[0]);
        return -1;
    }


    ifstream f(pArgv[1]);

    if (!f) {
        perror("Cannot open file");
        usage(pArgv[0]);
        return -1;
    }

    int testCnt = atol(pArgv[2]);

    srand(time(NULL));

    vector<string> vs;
    vector<int>    dist;
    char           buf[1024];

    while(f.getline(buf, sizeof buf)) {
        vs.push_back(buf); 
    }

    if (!testCnt || vs.size() < 50) {
        cerr << "To few words to test" << endl;
        return -1;
    }

    double avg = 0.0;

    int n = 0;
    for (int i = 0; i < testCnt; ++i, ++n) {
        int k = rand() % vs.size();
        int l;
        do {
            l = rand() % vs.size();
        } while (l == k);
        int ed = EditDistance(vs[k], vs[l]);
        dist.push_back(ed);
        avg = (avg * n + ed) / (n + 1);
    }
    // now dispersion
    double d = 0.0;
    n = 0;
    for (int i = 0; i < testCnt; ++i, ++n) {
        double r = (dist[i] - avg) * (dist[i] - avg);
        d = (d * n + r) / (n + 1);
    }

    cout << "Average dist: " << avg << endl;
    cout << "Dist dispersion: " << d << endl;
    cout << "Intrinsic dimension: " << avg * avg / (2 * d) << endl;

    return 0;
}
