#include <iostream>
#include <vector>
#include <string>
#include <fstream>
#include <time.h>
#include <map>
#include "../Common/utils.h"

#include "interface.h"

#define REUSE_NG_QTY_KOEFF 2

using namespace std;

void usage(const char *pProg) 
{
    cerr << "Usage : " << pProg << " <file> <hasTransp> <kerr> <testcnt> <OutPrefix> <start len> <end len>" << endl;
}

int main(int argc, char *pArgv[])
{
    if (argc != 8) {
        usage(pArgv[0]);
        return -1;
    }


    ifstream f(pArgv[1]);

    if (!f) {
        perror("Cannot open file");
        usage(pArgv[0]);
        return -1;
    }

    string      AlphaName;

    GetEnv(ENV_ABC, AlphaName);

    EAlphabet           AlphaType = GetAlphabetType(AlphaName);
    if (AlphaType == eAlphaUnknown) {
        cerr << "Unknown alphabet: '" << AlphaName << "' check env var " << ENV_ABC << endl;
        return 1;
    }

    bool        hasTransp = atol(pArgv[2]);
    int         kerr = atol(pArgv[3]);
    int         cnt = atol(pArgv[4]);
    string      OutPrefix = pArgv[5];
    int         len1 = atol(pArgv[6]);
    int         len2 = atol(pArgv[7]);

    vector<ofstream*>   vpOutFile(len2 + 1);
    vector<unsigned>    vQty(len2 + 1);
    unsigned            FullQty = 0;

    int DictQty = len2 - len1 + 1;

    for (int i = len1; i <= len2; ++i) {
        stringstream FileName;

        FileName << OutPrefix << "." << i << "_" << kerr << "_" << AlphaName;
        vpOutFile[i] = new ofstream(FileName.str().c_str());

        if (!(*vpOutFile[i])) {
            cerr << "Cannot open: " << FileName.str() << " for writing" << endl;
            return 1;
        }
    }

    srand(time(NULL));

    vector<string> vs;
    char           buf[1024];

    while(f.getline(buf, sizeof buf - 1)) {
        vs.push_back(buf); 
        buf[sizeof buf - 1] = 0;
    }


    int pos = 0; 

    map<string, int> used;

    if (!vs.size()) {
        cerr << "Empty dictionary\n";
        return 1;
    }

    unsigned ReuseQty = 1;

    for (int i = 1; i <= kerr; ++i) {
        ReuseQty *= REUSE_NG_QTY_KOEFF;
    }


    while(FullQty < DictQty) {
        StringList lst;

        int pos = rand() % vs.size();
        string str = vs[pos];

        bool    Ok = false;


        // Check if the generated string might produce a string of a desired length
        for (int i = len1; i <= len2; ++i) 
        if (vQty[i] < cnt) {
            if (str.size() + kerr >= i && str.size() <= i + kerr) {
                Ok = true;
                break;
            }
        }


        if (!Ok) {
            continue; // To short or two long
        }

        vector<string> vLst;

        if (kerr) {
            FillExpansionList (str,
                            hasTransp,
                            kerr,
                            &lst,
                            GetAlphabet(AlphaType),
                            (kerr < 3) ? 0: kerr - 1, true, 
                            0, false);

            for (int i = 0; i < ReuseQty; ++i) {

                 int sz = lst.size();
                 assert(sz);
                 int indx = rand() % lst.size();

                 const string& str = lst[indx];
    
                 //cout << str << " -> " << lst[indx] << endl;
                 int len = str.size();
                 if (len < len1 || len > len2 || vQty[len] >= cnt || used[str]) {
                     continue;
                 }
                 used[str] = 1;
                 vLst.push_back(str);
            }
        } else {
            vLst.push_back(str);
        }

        for (int i = 0; i < vLst.size(); ++i) {
            const string& str = vLst[i];
            int len = str.size();
            //cout << str << " " << len << endl;
            if (len >= len1 && len <= len2 && vQty[len] < cnt) { 
                (*vpOutFile[len]) << str << endl;
                if (vQty[len]++ == cnt - 1) {
                    cout << "Len: " << len << " is ready." << endl;
                    ++FullQty;
                }
            }
        }
    }

    return 0;
}
