#include <iostream>
#include <vector>
#include <string>
#include <fstream>
#include <time.h>
#include <map>
#include "../Common/utils.h"

#include "interface.h"

using namespace std;

void usage(const char *pProg) 
{
    cerr << "Usage : " << pProg << " <file> <hasTransp> <kerr> <testcnt> " << endl;
}

CProbDesc   ProbDescNoTransp(0.33, 0.33, 0.34, 0);
CProbDesc   ProbDescTransp(0.25, 0.25, 0.25, 0.25);

int main(int argc, char *pArgv[])
{
    if (argc != 5) {
        usage(pArgv[0]);
        return -1;
    }


    ifstream f(pArgv[1]);

    if (!f) {
        perror(("Cannot open file: '" + string(pArgv[1]) +  "'").c_str());
        usage(pArgv[0]);
        return -1;
    }

    bool hasTransp = atol(pArgv[2]);
    int kerr = atol(pArgv[3]);
    int cnt = atol(pArgv[4]);

    srand(time(NULL));

    vector<string>  vs;
    string          tmp;

    while(getline(f, tmp)) {
        vs.push_back(tmp); 
    }

    int cntLeft = cnt;

    int pos = 0; 

    map<string, int> used;

    if (!vs.size()) {
        cerr << "Empty dictionary\n";
        return 1;
    }

    string              s;
    GetEnv(ENV_ABC, s);

    EAlphabet           AlphaType = GetAlphabetType(s);

    if (AlphaType == eAlphaUnknown) {
        cerr << "Unknown alphabet: '" << s << "' check env var " << ENV_ABC << endl;
        return 1;
    }

    const char *pABC = GetAlphabet(AlphaType);
    unsigned    ABCLen = strlen(pABC);


    while(cntLeft > 0) {
        StringList lst;

        int pos = rand() % vs.size();
        string str = vs[pos];

        if (kerr) {
             string sRandEdit = ApplyRandomEdits(str, hasTransp, kerr, pABC, ABCLen, 0,
                                                 hasTransp ? ProbDescTransp : ProbDescNoTransp);

             if (used[sRandEdit]) {
                 continue;
             }
             used[sRandEdit] = 1;
             cout << sRandEdit << endl;
        } else {
             cout << str << endl;
        }
        --cntLeft;
    }

    return 0;
}
